/**
 * Version 1.0 is to handle individual single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
// java dependencies
import java.awt.geom.Point2D;
import java.io.File;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.math.BigDecimal;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.utilities.ErrorHandler;
import uk.ac.leeds.ccg.andyt.grids.utilities.UnsignedLongPowersOf2;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;

/**
 * Grid2DSquareCellAbstract provides inner classes for CellID and ChunkID
 * referencing and general geometry methods for extended classes. It also
 * controls what methods extended classes must implement and so acts somewhat
 * like an interface.
 *
 * The basic geometries are ordered in set numbers of rows and columns and are
 * arranged sequentially as their base two-dimensional orthogonal coordinate
 * axes. The sequential arrangement goes along the x-axis row by row from the
 * y-axis, then up the y-axis taking each row in turn.
 *
 * TODO:
 * Add a constant int TYPE which can be used to switch between different types
 * of this.
 * Handling for NumberFormatExceptions and ArithmeticExceptions in calculations
 */
public abstract class Grid2DSquareCellAbstract extends ErrorHandler implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    /**
     * Local directory used for caching.
     */
    protected File directory;
    
    /**
     * The chunks that make up this instance. HashMap is prefered to an array
     * or an array of arrays due to garbage collection caching advantages.
     * TODO:
     * 1. Would it be better to use a; ReferenceMap, ReferenceIdentityMap,
     * IdentityHashMap, or Hashtable in place of this HashMap?
     * 2. Should this be transient?
     */
    protected HashMap grid2DSquareCellChunks;
    
    /**
     * A reference to the grid Statistics Object.
     */
    protected GridStatisticsAbstract gridStatistics;
    
    /**
     * For storing the number of chunk rows.
     */
    protected int nChunkRows;
    
    /**
     * For storing the number of chunk columns.
     */
    protected int nChunkCols;
    
    /**
     * For storing the (usual) number of cells in a chunk row.
     * This may be different for chunks in the final row of
     * chunks (final as in row major order)
     */
    protected int chunkNrows;
    
    /**
     * For storing the (usual) number of cells in a chunk column.
     * This may be different for chunks in the final row of
     * chunks (final as in row major order)
     */
    protected int chunkNcols;
    
    /**
     * For storing the number of rows in the grid.
     */
    protected long nrows;
    
    /**
     * For storing the number of columns in the grid.
     */
    protected long ncols;
    
    /**
     * For storing the name of the grid.
     */
    protected String name = "";
    
    /**
     * For storing cellsize, minx, miny, maxx, maxy
     * Although maxx and maxy could be easily calculated if required, for
     * convenience they are calculated by default during grid construction.
     */
    protected BigDecimal[] dimensions;
    
    /**
     * For storing the minimum number of decimal places needed to store the
     * dimensions.
     */
    protected int dimensionsScale;
    
    /**
     * For storing values mapped to a binary encoded long. The long is a key
     * which indicates if the value is that at a given location. Both keys and
     * values are unique. This is used in Grid2DSquareCellDoubleChunk64CellMap.
     */
    protected static UnsignedLongPowersOf2 unsignedLongPowersOf2;
    
    /**
     * Initialises all fields from an existing instance. Main use in
     * constructors:
     *   Grid2DSquareCellDouble(File,File,ObjectInputStream).
     *   Grid2DSquareCellInt(File,File,ObjectInputStream).
     * @param grid2DSquareCellAbstract the Grid2DSquareCellAbstract from which
     *   the field values of this are set.
     */
    protected abstract void initAll( Grid2DSquareCellAbstract grid2DSquareCellAbstract );
    
    /**
     * Initialises all fields from an existing instance including
     * grid2DSquareCellDoubleChunks.
     * For use with constructors:
     *   Grid2DSquareCellDouble(File,File,ObjectInputStream).
     *   Grid2DSquareCellInt(File,File,ObjectInputStream).
     * @param grid2DSquareCell the Grid2DSquareCellAbstract from which the field
     *   values of this are set.
     * @param flag This is to distinguish this method from initMain( Grid2DSquareCellAbstract )
     */
    protected void initAll( Grid2DSquareCellAbstract grid2DSquareCell, boolean flag ) {
        this.grid2DSquareCellChunks = grid2DSquareCell.grid2DSquareCellChunks;
        this.chunkNcols = grid2DSquareCell.chunkNcols;
        this.chunkNrows = grid2DSquareCell.chunkNrows;
        this.nChunkCols = grid2DSquareCell.nChunkCols;
        this.nChunkRows = grid2DSquareCell.nChunkRows;
        this.memoryReserve = grid2DSquareCell.memoryReserve;
        this.ncols = grid2DSquareCell.ncols;
        this.nrows = grid2DSquareCell.nrows;
        this.directory = grid2DSquareCell.directory;
        this.name = grid2DSquareCell.name;
        this.dimensions = grid2DSquareCell.dimensions;
        this.dimensionsScale = grid2DSquareCell.dimensionsScale;
        this.unsignedLongPowersOf2 = grid2DSquareCell.unsignedLongPowersOf2;
    }
    
    /**
     * Returns this.grid2DSquareCellChunks
     * TODO:
     * Should we clone, i.e. and return this.grid2DSquareCellChunks.clone()?
     * The methods for caching and clearing in core need not directly refer to this.grid2DSquareCellChunks.
     * The method is protected to avoid problems of freeing memory by caching and clearing which can be
     * caused if an external Object holds a reference to an Object in this.grid2DSquareCellChunks.
     */
    protected HashMap getGrid2DSquareCellChunks() {
        return this.grid2DSquareCellChunks;
    }
    
    /**
     * Returns this.gridStatistics
     * TODO:
     * For safety, this method should probably be removed and instead, this
     * class be made to implement GridStatisticsInterface. This done the
     * methods introduced would be made to call the relevant ones in
     * this.gridStatistics.
     */
    public GridStatisticsAbstract getGridStatistics() {
        return this.gridStatistics;
    }
    
    /**
     * Returns the Grid2DSquareCellChunkAbstract with the given chunkRowIndex and
     * chunkColIndex
     * @param chunkRowIndex
     * @param chunkColIndex
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected Grid2DSquareCellChunkAbstract getGrid2DSquareCellChunk( int chunkRowIndex, int chunkColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getGrid2DSquareCellChunk( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return getGrid2DSquareCellChunk( chunkRowIndex, chunkColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns grid2DSquareCellChunks.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected Grid2DSquareCellChunkAbstract getGrid2DSquareCellChunk( ChunkID chunkID, boolean handleOutOfMemoryError ) {
        try {
            if ( inGrid( chunkID, handleOutOfMemoryError ) ) {
                if ( ! ( getGrid2DSquareCellChunks().containsKey( chunkID ) ) ) {
                    //if ( this.grid2DSquareCellChunks.get( chunkID ) == null ) {
                    //if ( this.grid2DSquareCellChunks[ chunkRowIndex ][ chunkColIndex ] == null ) {
                    restoreChunk( chunkID, handleOutOfMemoryError );
                }
                return ( Grid2DSquareCellChunkAbstract ) this.grid2DSquareCellChunks.get( chunkID );
                //return this.grid2DSquareCellDoubleChunks[ chunkRowIndex ][ chunkColIndex ];
            }
            return null;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                cacheChunkExcept( chunkID , false);
                initMemoryReserve();
                return getGrid2DSquareCellChunk( chunkID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns an iterator over the cell values in this.
     */
    public abstract Iterator iterator( boolean handleOutOfMemoryError );
    
    /**
     * Returns a string description of the instance.
     */
    public abstract String toString( boolean handleOutOfMemoryError );
    
    
    /**
     * Returns a string description of the Abstract fields of this instance.
      * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param flag This is ignored. It is simply to distinguish this method from 
     *   the abstract method toString(boolean).
     */
    public String toString( boolean handleOutOfMemoryError, boolean flag ) {
        try {
            String result = "chunkNcols( " + this.chunkNcols + " ), " +
                    "chunkNrows( " + this.chunkNrows + " ), " +
                    "dimensionsScale( " + this.dimensionsScale + " ), " +
                    "nChunkCols( " + this.nChunkCols + " ), " +
                    "nChunkRows( " + this.nChunkRows + " ), " +
                    "ncols( " + this.ncols + " ), " +
                    "nrows( " + this.nrows + " ), " +
                    "directory( " + this.directory + " ), " +
                    "name( " + this.name + " ), " +
                    "dimensions( " + this.dimensions[ 0 ].toString();
            for ( int dimensionsID = 1; dimensionsID < this.dimensions.length; dimensionsID ++ ) {
                result += ", " + dimensions[ dimensionsID ].toString();
            }
            if ( this.memoryReserve == null ) {
                result += " ), memoryReserve( null";
            } else {
                result += " ), memoryReserve.length( " + this.memoryReserve.length;
            }
            result += " ), grid2DSquareCellChunks.size( " + this.grid2DSquareCellChunks.size();
            //result += " ), grid2DSquareCellIntProcessor( " + this.grid2DSquareCellIntProcessor.toString();
            if ( this.unsignedLongPowersOf2 == null ) {
                result += " ), unsignedLongPowersOf2( null ), ";
            } else {
                result += " ), unsignedLongPowersOf2( " + this.unsignedLongPowersOf2.toString() + " ), ";
            }
            result += " Statistics( " + this.gridStatistics.toString() + " )";
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                cacheChunk( false );
                initMemoryReserve();
                return toString( handleOutOfMemoryError, flag );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns directory.
     */
    public File getDirectory() {
        return this.directory;
    }
    
    /**
     * Sets name.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setName( String newName, boolean handleOutOfMemoryError ) {
        try {
            this.name = newName;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                cacheChunk(false);
                initMemoryReserve();
                setName( newName, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns name.
     */
    public String getName() {
        return this.name;
    }
    
    /**
     * Sets chunk with chunkID to grid2DSquareCellDoubleChunk.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setChunk( Grid2DSquareCellChunkAbstract grid2DSquareCellChunk, ChunkID chunkID, boolean handleOutOfMemoryError ) {
        try {
            this.grid2DSquareCellChunks.put( chunkID, grid2DSquareCellChunk );
            System.gc();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                cacheChunkExcept( chunkID , false);
                initMemoryReserve();
                setChunk( grid2DSquareCellChunk, chunkID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns ncols.
     */
    public final long getNcols() {
        return this.ncols;
    }
    
    /**
     * Initialises ncols.
     */
    protected final void initNcols( long ncols ) {
        this.ncols = ncols;
    }
    
    /**
     * Returns nrows.
     */
    public final long getNrows() {
        return this.nrows;
    }
    
    /**
     * Initialises nrows.
     */
    protected final void initNrows( long nrows ) {
        this.nrows = nrows;
    }
    
    /**
     * Returns nChunkRows.
     */
    public final int getNChunkRows(boolean handleOutOfMemoryError) {
        try {
            return this.nChunkRows;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getNChunkRows( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    ///**
    // * Initialises nChunkRows.
    // */
    //protected void initNChunkRows( int nChunkRows ) {
    //    this.nChunkRows = nChunkRows;
    //}
    
    /**
     * Initialises nChunkRows
     * @param nrows
     * @param chunkNrows
     */
    public final void initNChunkRows( long nrows, int chunkNrows ) {
        // Calculate the number of row chunks
        //this.nChunkRows = ( int ) Math.ceil( ( double ) nrows / ( double ) chunkNrows );
        if ( ( nrows % chunkNrows ) > 0 ) {
            this.nChunkRows = ( int ) ( nrows / chunkNrows ) + 1;
        } else {
            this.nChunkRows = ( int ) nrows / chunkNrows;
        }
    }
    
    /**
     * Returns nChunkCols.
     */
    public final int getNChunkCols( boolean handleOutOfMemoryError ) {
        try {
            return this.nChunkCols;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getNChunkCols( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    ///**
    // * Initialises nChunkCols.
    // */
    //protected void initNChunkCols( int nChunkCols ) {
    //    this.nChunkCols = nChunkCols;
    //}
    
    /**
     * Initialises nChunkCols
     * @param ncols
     * @param chunkNcols
     */
    public final void initNChunkCols( long ncols, int chunkNcols ) {
        // Calculate the number of column chunks
        //this.nChunkCols = ( int ) Math.ceil( ( double ) ncols / ( double ) chunkNcols );
        if ( ( ncols % chunkNcols ) > 0.0d ) {
            this.nChunkCols = ( int ) ( ncols / chunkNcols ) + 1;
        } else {
            this.nChunkCols = ( int ) ncols / chunkNcols;
        }
    }
    
    /**
     * Returns chunkNrows.
     */
    public final int getChunkNrows( boolean handleOutOfMemoryError ) {
        try {
            return this.chunkNrows;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkNrows( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises chunkNrows.
     */
    protected final void initChunkNrows( int chunkNrows ) {
        this.chunkNrows = chunkNrows;
    }
    
    /**
     * Returns chunkNcols.
     */
    public final int getChunkNcols( boolean handleOutOfMemoryError ) {
        try {
            return this.chunkNcols;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkNcols( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises chunkNcols.
     */
    protected final void initChunkNcols( int chunkNcols ) {
        this.chunkNcols = chunkNcols;
    }
    
    /**
     * Returns the number of rows in the final row Chunk
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNrowsFinalRowChunks( boolean handleOutOfMemoryError ) {
        try {
            return ( int ) ( getNrows() - ( ( long ) ( getNChunkRows(this.handleOutOfMemoryErrorFalse) - 1 ) * ( long ) getChunkNrows(this.handleOutOfMemoryErrorFalse) ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkNrowsFinalRowChunks( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the number of cols in the final col Chunk
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNcolsFinalColChunks( boolean handleOutOfMemoryError ) {
        try {
            return ( int ) ( getNcols() - ( ( long ) ( getNChunkCols(this.handleOutOfMemoryErrorFalse) - 1 ) * ( long ) getChunkNcols(this.handleOutOfMemoryErrorFalse) ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkNcolsFinalColChunks( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    
    /**
     * Returns the number of rows in chunkID as an int
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNrows( ChunkID chunkID, boolean handleOutOfMemoryError ) {
        try {
            if ( chunkID.chunkRowIndex < getNChunkRows(this.handleOutOfMemoryErrorFalse) - 1 ) {
                return getChunkNrows(this.handleOutOfMemoryErrorFalse);
            } else {
                return getChunkNrowsFinalRowChunks( handleOutOfMemoryError );
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkNrows( chunkID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the number of columns in chunkID as an int
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNcols( ChunkID chunkID, boolean handleOutOfMemoryError ) {
        try {
            if ( chunkID.chunkColIndex < getNChunkCols(this.handleOutOfMemoryErrorFalse) - 1 ) {
                return getChunkNcols(this.handleOutOfMemoryErrorFalse);
            } else {
                return getChunkNcolsFinalColChunks( handleOutOfMemoryError );
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkNcols( chunkID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns dimensions.
     */
    public final BigDecimal[] getDimensions() {
        //double dimensions0 = this.dimensions[0].doubleValue(); // For debugging
        //double dimensions1 = this.dimensions[1].doubleValue(); // For debugging
        //double dimensions2 = this.dimensions[2].doubleValue(); // For debugging
        //double dimensions3 = this.dimensions[3].doubleValue(); // For debugging
        //double dimensions4 = this.dimensions[4].doubleValue(); // For debugging
        return this.dimensions;
    }
    
    /**
     * Returns dimensions[0] converted to a double.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellsizeDouble( boolean handleOutOfMemoryError ) {
        return this.dimensions[ 0 ].doubleValue();
    }
    
    /**
     * Returns dimensionsScale.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getDimensionsScale( boolean handleOutOfMemoryError ) {
        return this.dimensionsScale;
    }
    
    /**
     * Initialises dimensions.
     */
    protected final void initDimensions( BigDecimal[] dimensions ) {
        this.dimensions = dimensions;
        this.dimensionsScale = Integer.MIN_VALUE;
        for ( int i = 0; i < dimensions.length; i ++ ) {
            this.dimensionsScale = Math.max( this.dimensionsScale, dimensions[ i ].scale() );
        }
    }
    
    /**
     * Returns the value of the cell at point as a double.
     * @param point The Point2D.Double for which the cell value is returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble( Point2D.Double point, boolean handleOutOfMemoryError ) {
        try {
            return getCellDouble( getChunkRowIndex( point.getY(), handleOutOfMemoryError ), getChunkColIndex( point.getX(), handleOutOfMemoryError ), getChunkCellRowIndex( point.getY(), handleOutOfMemoryError ), getChunkCellColIndex( point.getX(), handleOutOfMemoryError ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                cacheChunkExcept( getChunkRowIndex( point.getY(), handleOutOfMemoryError ), getChunkColIndex( point.getX(), handleOutOfMemoryError ) , false);
                initMemoryReserve();
                return getCellDouble( point, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex as a double
     * @param cellRowIndex The cell row index.
     * @param cellColIndex The cell column index.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getCellDouble( getChunkRowIndex( cellRowIndex, handleOutOfMemoryError ), getChunkColIndex( cellColIndex, handleOutOfMemoryError ), getChunkCellRowIndex( cellRowIndex, handleOutOfMemoryError ), getChunkCellColIndex( cellColIndex, handleOutOfMemoryError ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                cacheChunkExcept( getChunkRowIndex( cellRowIndex, handleOutOfMemoryError ), getChunkColIndex( cellColIndex, handleOutOfMemoryError ) , false);
                initMemoryReserve();
                return getCellDouble( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at chunkRowIndex, chunkColIndex, chunkCellRowIndex,
     * chunkCellColIndex as a double.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble( int chunkRowIndex, int chunkColIndex, int chunkCellRowIndex, int chunkCellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getCellDouble( getGrid2DSquareCellChunk( chunkRowIndex, chunkColIndex, handleOutOfMemoryError ), chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return getCellDouble( chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at chunkRowIndex, chunkColIndex, chunkCellRowIndex,
     * chunkCellColIndex as a double.
     * @param grid2DSquareCellChunk The Grid2DSquareCellChunkAbstract containing
     *   the cell.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble( Grid2DSquareCellChunkAbstract grid2DSquareCellChunk, int chunkRowIndex, int chunkColIndex, int chunkCellRowIndex, int chunkCellColIndex, boolean handleOutOfMemoryError ) {
        try {
            if ( inGrid( chunkRowIndex, chunkColIndex, handleOutOfMemoryError ) ) {
                if ( grid2DSquareCellChunk instanceof Grid2DSquareCellDoubleChunkAbstract ) {
                    double noDataValue = ( ( Grid2DSquareCellDoubleChunkAbstract ) grid2DSquareCellChunk ).getGrid2DSquareCellDouble().getNoDataValue();
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                        return ( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                        return ( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                        return ( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                        return ( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkRAF.class ) {
                        return ( ( Grid2DSquareCellDoubleChunkRAF ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                } else {
                    int noDataValue = ( ( Grid2DSquareCellIntChunkAbstract ) grid2DSquareCellChunk ).getGrid2DSquareCellInt().getNoDataValue( this.handleOutOfMemoryErrorFalse );
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunk64CellMap.class ) {
                        return ( double ) ( ( Grid2DSquareCellIntChunk64CellMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                        return ( double ) ( ( Grid2DSquareCellIntChunkArray ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                        return ( double ) ( ( Grid2DSquareCellIntChunkJAI ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                        return ( double ) ( ( Grid2DSquareCellIntChunkMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkRAF.class ) {
                        return ( double ) ( ( Grid2DSquareCellIntChunkRAF ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                }
                
            }
            return getNoDataValueDouble( handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return getCellInt( grid2DSquareCellChunk, chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    //    /**
    //     * @depricated
    //     * Returns the value at cellRowIndex, cellColIndex as a int.
    //     * @param cellRowIndex The cell row index.
    //     * @param cellColIndex The cell column index.
    //     */
    //    public int getCellInt( long cellRowIndex, long cellColIndex ) {
    //        return getCellInt( cellRowIndex, cellColIndex, true );
    //    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex as a int.
     * @param cellRowIndex The cell row index.
     * @param cellColIndex The cell column index.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getCellInt( getChunkRowIndex( cellRowIndex, handleOutOfMemoryError ), getChunkColIndex( cellColIndex, handleOutOfMemoryError ), getChunkCellRowIndex( cellRowIndex, handleOutOfMemoryError ), getChunkCellColIndex( cellColIndex, handleOutOfMemoryError ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                cacheChunkExcept( getChunkRowIndex( cellRowIndex, handleOutOfMemoryError ), getChunkColIndex( cellColIndex, handleOutOfMemoryError ) , false);
                initMemoryReserve();
                return getCellInt( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at chunkRowIndex, chunkColIndex, chunkCellRowIndex,
     * chunkCellColIndex as a int.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt( int chunkRowIndex, int chunkColIndex, int chunkCellRowIndex, int chunkCellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getCellInt( getGrid2DSquareCellChunk( chunkRowIndex, chunkColIndex, handleOutOfMemoryError ), chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return getCellInt( chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at chunkRowIndex, chunkColIndex, chunkCellRowIndex,
     * chunkCellColIndex as a int.
     * @param grid2DSquareCellChunk The Grid2DSquareCellChunkAbstract containing
     *   the cell.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt( Grid2DSquareCellChunkAbstract grid2DSquareCellChunk, int chunkRowIndex, int chunkColIndex, int chunkCellRowIndex, int chunkCellColIndex, boolean handleOutOfMemoryError ) {
        try {
            if ( inGrid( chunkRowIndex, chunkColIndex, handleOutOfMemoryError ) ) {
                if ( grid2DSquareCellChunk instanceof Grid2DSquareCellDoubleChunkAbstract ) {
                    double noDataValue = ( ( Grid2DSquareCellDoubleChunkAbstract ) grid2DSquareCellChunk ).getGrid2DSquareCellDouble().getNoDataValue();
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                        return ( int ) ( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                        return ( int ) ( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                        return ( int ) ( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                        return ( int ) ( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkRAF.class ) {
                        return ( int ) ( ( Grid2DSquareCellDoubleChunkRAF ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                } else {
                    int noDataValue = ( ( Grid2DSquareCellIntChunkAbstract ) grid2DSquareCellChunk ).getGrid2DSquareCellInt().getNoDataValue( this.handleOutOfMemoryErrorFalse );
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunk64CellMap.class ) {
                        return ( ( Grid2DSquareCellIntChunk64CellMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                        return ( ( Grid2DSquareCellIntChunkArray ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                        return ( ( Grid2DSquareCellIntChunkJAI ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                        return ( ( Grid2DSquareCellIntChunkMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkRAF.class ) {
                        return ( ( Grid2DSquareCellIntChunkRAF ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    }
                }
            }
            return Integer.MIN_VALUE;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return getCellInt( grid2DSquareCellChunk, chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, true );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk column index for the chunks intersecting the
     * x-coordinate x.
     * @param x
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkColIndex( double x, boolean handleOutOfMemoryError ) {
        try {
            return getChunkColIndex( getCellColIndex( x, handleOutOfMemoryError ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkColIndex( x, handleOutOfMemoryError);
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk column index for the chunk intersecting the cell at
     * cell column index cellColIndex.
     * @param cellColIndex
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkColIndex( long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return ( int ) cellColIndex / getChunkNcols(this.handleOutOfMemoryErrorFalse);
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkColIndex( cellColIndex, handleOutOfMemoryError);
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell column Index for the cells that intersect the
     * x-coordinate x.
     * @param x
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellColIndex( double x, boolean handleOutOfMemoryError ) {
        try {
            return getCellColIndex( BigDecimal.valueOf( x ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellColIndex( x, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell column Index for the cells that intersect the
     * x-coordinate x.
     * @param xBigDecimal
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellColIndex( BigDecimal xBigDecimal, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            //double xBigDecimalDoubleValue = xBigDecimal.doubleValue(); // For Debugging
            //int scale = Math.max( yBigDecimal.scale(), getDimensionsScale() ); // For Debugging
            //double subtract = ( yBigDecimal.subtract( dimensions[2] ) ).doubleValue(); // For Debugging
            //long divide = ( ( yBigDecimal.subtract( dimensions[2] ) ).divide( dimensions[0], scale, BigDecimal.ROUND_DOWN ) ).longValue(); // For Debugging
            //int scale = Math.max( xBigDecimal.scale(), getDimensionsScale() );
            return ( ( xBigDecimal.subtract( dimensions[ 1 ] ) ).divide( dimensions[ 0 ], Math.max( xBigDecimal.scale(), getDimensionsScale( handleOutOfMemoryError ) ), BigDecimal.ROUND_DOWN ) ).longValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellColIndex( xBigDecimal, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk cell column Index of the cells that intersect the
     * x-coordinate x.
     * @param x
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellColIndex( double x, boolean handleOutOfMemoryError ) {
        try {
            //long cellColIndex = getCellColIndex( x ); // For Debugging
            //return getChunkCellColIndex( cellColIndex );
            return getChunkCellColIndex( getCellColIndex( x, handleOutOfMemoryError ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkCellColIndex( x, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunks cell column index of the cell in the chunk thats
     * centoid is closest to the x-coordinate x.
     * @param cellColIndex
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellColIndex( long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            int chunkNcols = getChunkNcols(this.handleOutOfMemoryErrorFalse);
            return ( int ) ( cellColIndex - ( ( cellColIndex / chunkNcols ) * chunkNcols ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkCellColIndex( cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk row index for the chunks intersecting the
     * y-coordinate y.
     * @param y
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkRowIndex( double y, boolean handleOutOfMemoryError ) {
        try {
            return getChunkRowIndex( getCellRowIndex( y, handleOutOfMemoryError ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkRowIndex( y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk row index for the chunk intersecting the cell at
     * cell row index cellRowIndex.
     * @param cellRowIndex
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkRowIndex( long cellRowIndex, boolean handleOutOfMemoryError ) {
        try {
            return ( int ) cellRowIndex / getChunkNrows(this.handleOutOfMemoryErrorFalse);
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkRowIndex( cellRowIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell row Index for the cells that intersect the
     * y-coordinate y.
     * @param y
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellRowIndex( double y, boolean handleOutOfMemoryError ) {
        try {
            return getCellRowIndex( BigDecimal.valueOf( y ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellRowIndex( y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell row Index for the cells that intersect the
     * y-coordinate yBigDecimal.
     * @param yBigDecimal The y-coordinate for which the cell row index is returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellRowIndex( BigDecimal yBigDecimal, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            //double yBigDecimalDoubleValue = yBigDecimal.doubleValue(); // For Debugging
            //int scale = Math.max( yBigDecimal.scale(), getDimensionsScale() ); // For Debugging
            //double subtract = ( yBigDecimal.subtract( dimensions[2] ) ).doubleValue(); // For Debugging
            //long divide = ( ( yBigDecimal.subtract( dimensions[2] ) ).divide( dimensions[0], scale, BigDecimal.ROUND_DOWN ) ).longValue(); // For Debugging
            return ( ( yBigDecimal.subtract( dimensions[2] ) ).divide( dimensions[0], Math.max( yBigDecimal.scale(), getDimensionsScale( handleOutOfMemoryError ) ), BigDecimal.ROUND_DOWN ) ).longValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellRowIndex( yBigDecimal, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk cell row Index of the cells that intersects the
     * y-coordinate y.
     * @param y .
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellRowIndex( double y, boolean handleOutOfMemoryError ) {
        try {
            //long cellRowIndex = getCellRowIndex( y ); // For Debugging
            //return getChunkCellRowIndex( cellRowIndex ); // For Debugging
            return getChunkCellRowIndex( getCellRowIndex( y, handleOutOfMemoryError ), handleOutOfMemoryError ); // For Debugging
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkCellRowIndex( y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunks cell row Index of the cells in the chunk thats
     * centoid is closest to the y-coordinate y.
     * @param cellRowIndex .
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellRowIndex( long cellRowIndex, boolean handleOutOfMemoryError ) {
        try {
            int chunkNrows = getChunkNrows(this.handleOutOfMemoryErrorFalse);
            return ( int ) ( cellRowIndex - ( ( cellRowIndex / chunkNrows ) * chunkNrows ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getChunkCellRowIndex( cellRowIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID of the cell given by cellRowIndex, cellColIndex even if
     * that cell would not be in the grid.
     * @param cellRowIndex The cell row index.
     * @param cellColIndex The cell column index.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID getCellID( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ){
        try {
            return new CellID( getNcols(), cellRowIndex, cellColIndex );
            //return new CellID( cellRowIndex, cellColIndex );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellID( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID of the cell given by x-coordinate x, y-coordinate y even
     * if that cell would not be in the grid.
     * @param x The x-coordinate.
     * @param y The y-coordinate.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID getCellID( double x, double y, boolean handleOutOfMemoryError ) {
        try {
            return new CellID( getNcols(), getCellRowIndex( y, handleOutOfMemoryError ), getCellColIndex( x, handleOutOfMemoryError ) );
            //return new CellID( getCellRowIndex( y ), getCellColIndex( x ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellID( x, y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method to cache this instance by writing it out to an
     * ObjectOutputStream File called Cache to be located in the File
     * (directory) returned by getDirectory(). If cacheChunks is true then each
     * of this.grid2DSquareCellChunks is cached first.
     * @param cacheChunks If true then chunks are cached individually prior to caching this.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void cache( boolean cacheChunks , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            if ( cacheChunks ) {
                cacheChunks(false);
            }
            try {
                ObjectOutputStream oos = new ObjectOutputStream( new BufferedOutputStream( new FileOutputStream( new File( getDirectory(), "Cache" ) ) ) );
                oos.writeObject( this );
                oos.flush();
                oos.close();
            } catch ( java.io.IOException ioe0 ) {
                ioe0.printStackTrace();
            }
            initMemoryReserve();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cache( " + cacheChunks + " ), memoryReserve must have been insufficiently large or garbage collection is not working as expected!" );
        }
    }
    
    /**
     * Attempts to cache a seriailized version of a Chunk. Returns the ChunkID
     * of that which was cached. The first chunk that is attempted to be cached
     * is that with a chunk row index of 0, and a chunk column index of 0.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final ChunkID cacheChunk(boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            int chunkRowIndex;
            int chunkColIndex;
            int chunkNrows = getChunkNrows(this.handleOutOfMemoryErrorFalse);
            int chunkNcols = getChunkNcols(this.handleOutOfMemoryErrorFalse);
            ChunkID chunkID;
            for ( chunkRowIndex = 0; chunkRowIndex < chunkNrows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < chunkNcols; chunkColIndex ++ ) {
                    chunkID = new ChunkID( chunkNcols, chunkRowIndex, chunkColIndex );
                    if ( cacheChunk( chunkID , false) ) {
                        return chunkID;
                    }
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
            return null;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk(), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Attempts to cache a seriailized version of Chunk given by; chunk row
     * index chunkRowIndex, chunk column index chunkColIndex. Returns true if
     * chunk was cached and false otherwise.
     * (NB. Sufficient memory should be available. Encountering an
     * OutOfMemoryError is likely to be fatal.)
     * @param chunkRowIndex The chunk row index.
     * @param chunkColIndex The chunk column index.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final boolean cacheChunk( int chunkRowIndex, int chunkColIndex , boolean reinitialiseMemoryReserve ) {
        try {
            clearMemoryReserve();
            boolean result = cacheChunk( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ) , false);
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( chunkRowIndex( " + chunkRowIndex + " ), chunkColIndex( " + chunkColIndex + " ) ) memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    
    /**
     * Caches a seriailsed version of Chunk with given chunkID.
     * (NB. Sufficient memory should be available. Encountering an
     * OutOfMemoryError is likely to be fatal.)
     * @param chunkID The chunkID of the chunk to be cached to file.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final boolean cacheChunk( ChunkID chunkID, boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            if ( this.grid2DSquareCellChunks.get( chunkID ) != null ) {
                if ( ! isCached( chunkID , false) ) {
                    Grid2DSquareCellChunkAbstract grid2DSquareCellChunk = getGrid2DSquareCellChunk( chunkID, this.handleOutOfMemoryErrorFalse );
                    try {
                        ObjectOutputStream oos = new ObjectOutputStream( new BufferedOutputStream( new FileOutputStream( new File( getDirectory(), "" + chunkID.getChunkRowIndex() + "_" + chunkID.getChunkColIndex() ) ) ) );
                        oos.writeObject( grid2DSquareCellChunk );
                        oos.flush();
                        oos.close();
                    } catch ( java.io.IOException ioe0 ) {
                        ioe0.printStackTrace();
                    }
                    if ( reinitialiseMemoryReserve ) {
                        initMemoryReserve();
                    }
                    return true;
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
            return false;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( ChunkID( " + chunkID.toString() + " ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Caches a seriailsed version of all grid2DSquareCellChunks.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void cacheChunks( boolean reinitialiseMemoryReserve ) {
        try {
            clearMemoryReserve();
            Iterator iterator = getGrid2DSquareCellChunks().keySet().iterator();
            ChunkID chunkID;
            while ( iterator.hasNext() ) {
                chunkID = ( ChunkID ) iterator.next();
                if ( ! isCached( chunkID, false ) ) {
                    cacheChunk( chunkID, false );
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunks(), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
        //try {
        //    clearMemoryReserve();
        //    int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
        //    int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
        //    int chunkRowIndex;
        //    int chunkColIndex;
        //    ChunkID chunkID;
        //    for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
        //        for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
        //            chunkID = new ChunkID( chunkNcols, chunkRowIndex, chunkColIndex );
        //            if ( ! isCached( chunkID , false) ) {
        //                cacheChunk( chunkRowIndex, chunkColIndex , false);
        //            }
        //        }
        //    }
        //    if ( reinitialiseMemoryReserve ) {
        //        initMemoryReserve();
        //    }
        //} catch ( java.lang.OutOfMemoryError oome0 ) {
        //    throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunks(), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        //}
    }
    
    /**
     * Caches a seriailsed version of all grid2DSquareCellChunks in
     * grid2DSquareCellChunkIDs.
     * @param chunkIDs A HashSet containing the ChunkIDs of the
     * grid2DSquareCellChunks to be cached.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void cacheChunks( HashSet chunkIDs, boolean reinitialiseMemoryReserve ) {
        try {
            clearMemoryReserve();
            Iterator iterator = chunkIDs.iterator();
            ChunkID chunkIDToCache;
            while( iterator.hasNext() ) {
                chunkIDToCache = ( ChunkID ) iterator.next();
                if ( ! isCached( chunkIDToCache , false) ) {
                    cacheChunk( chunkIDToCache , false);
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( ChunkID( " + chunkIDs.toString() + " ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Caches all grid2DSquareCellChunks other that with ChunkID chunkID.
     * @param chunkID The ChunkID of the chunk not to be cached.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void cacheChunksExcept( ChunkID chunkID, boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
            int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
            int chunkRowIndex;
            int chunkColIndex;
            ChunkID chunkIDToCache;
            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
                    chunkIDToCache = new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex );
                    if ( ! isCached( chunkIDToCache , false) && ! chunkIDToCache.equals( chunkID ) ) {
                        cacheChunk( chunkRowIndex, chunkColIndex , false);
                    }
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( ChunkID( " + chunkID.toString() + " ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    
    /**
     * Caches a Grid2DSquareCellChunk of grid2DSquareCellChunks (other that with
     * ChunkID chunkID) then returns its ChunkID. The first chunk that is attempted
     * to be cached is that with a chunk row index of 0, and a chunk column index
     * of 0.
     * @param chunkID The ChunkID of the chunk not to be cached.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final ChunkID cacheChunkExcept( ChunkID chunkID , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            ChunkID result = cacheChunkExcept( chunkID.getChunkRowIndex(), chunkID.getChunkColIndex() , false);
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( ChunkID( " + chunkID.toString() + " ) ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    
    
    /**
     * Caches a Grid2DSquareCellChunk of grid2DSquareCellChunks (other that with;
     * chunk row index chunkRowIndex, chunk column index chunkColIndex) then
     * returns its ChunkID. The first chunk that is attempted to be cached is that
     * with a chunk row index of 0, and a chunk column index of 0.
     * @param chunkRowIndex The chunk row index.
     * @param chunkColIndex The chunk column index.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final ChunkID cacheChunkExcept( int chunkRowIndex, int chunkColIndex , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
            int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
            int cri;
            int cci;
            ChunkID result;
            for ( cri = 0; cri < nChunkRows; cri ++ ) {
                for ( cci = 0; cci< nChunkCols; cci ++ ) {
                    result = new ChunkID( nChunkCols, cri, cci );
                    if ( ! isCached( result , false) ) {
                        cacheChunk( cri, cci , false);
                        if ( reinitialiseMemoryReserve ) {
                            initMemoryReserve();
                        }
                        return result;
                    }
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
            return null;
            //throw new Error( "Failed to cacheChunkExceptReverseOrder( chunkRowIndex( " + chunkRowIndex + " ), chunkColIndex( " + chunkColIndex + " ) ) in " + this.getClass().toString() + "!" );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( chunkRowIndex( " + chunkRowIndex + " ), chunkColIndex( " + chunkColIndex + " ) ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Caches a Grid2DSquareCellChunk of grid2DSquareCellChunks (other that with
     * ChunkID chunkID) then returns its ChunkID. The first chunk that is attempted
     * to be cached is that with a chunk row index of chunkNrows - 1, and a chunk
     * column index of chunkNcols - 1. This is primarily used in constructors that
     * read a data format that presents values in reverse row order.
     * @param chunkRowIndex The chunk row index.
     * @param chunkColIndex The chunk column index.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final ChunkID cacheChunkExceptReverseOrder( int chunkRowIndex, int chunkColIndex , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
            int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
            ChunkID result;
            int cri;
            int cci;
            for ( cri = nChunkRows - 1; cri > -1; cri -- ) {
                for ( cci = nChunkCols - 1; cci > -1; cci -- ) {
                    if ( cri != chunkRowIndex && cci != chunkRowIndex ) {
                        result = new ChunkID( nChunkCols, cri, cci );
                        if ( ! isCached( result , false) ) {
                            cacheChunk( cri, cci , false);
                            if ( reinitialiseMemoryReserve ) {
                                initMemoryReserve();
                            }
                            return result;
                        }
                    }
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
            return null;
            //throw new Error( "Failed to cacheChunkExceptReverseOrder( chunkRowIndex( " + chunkRowIndex + " ), chunkColIndex( " + chunkColIndex + " ) ) in " + this.getClass().toString() + "!" );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunkExceptReverseOrder( chunkRowIndex( " + chunkRowIndex + " ), chunkColIndex( " + chunkColIndex + " ) ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Caches a Grid2DSquareCellChunk of grid2DSquareCellChunks (other that with
     * ChunkID chunkID) then returns its ChunkID. The first chunk that is attempted
     * to be cached is that with a chunk row index of chunkNrows - 1, and a chunk
     * column index of chunkNcols - 1. This is primarily used in constructors that
     * read a data format that presents values in reverse row order.
     * @param chunkID The ChunkID of the chunk not to be cached.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final ChunkID cacheChunkExceptReverseOrder( ChunkID chunkID , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            ChunkID result = cacheChunkExceptReverseOrder( chunkID.getChunkRowIndex(), chunkID.getChunkColIndex() , false);
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( ChunkID( " + chunkID.toString() + " ) ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Caches all grid2DSquareCellChunks other than those with ChunkIDs in
     * chunkIDs.
     * @param chunkIDs A hashSet containing ChunkIDs not to be cached.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void cacheChunksExcept( HashSet chunkIDs , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
            int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
            int chunkRowIndex;
            int chunkColIndex;
            ChunkID chunkIDToCache;
            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
                    chunkIDToCache = new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex );
                    if ( ! ( ! isCached( chunkIDToCache , false) && chunkIDs.contains( chunkIDToCache ) ) ) {
                        cacheChunk( chunkRowIndex, chunkColIndex , false);
                    }
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( ChunkID( " + chunkIDs.toString() + " ) ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Caches a seriailsed version of all grid2DSquareCellChunks and clears them.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void cacheAndClearChunks( boolean reinitialiseMemoryReserve ) {
        try {
            clearMemoryReserve();
            Iterator iterator = this.grid2DSquareCellChunks.entrySet().iterator();
            ChunkID chunkID;
            while ( iterator.hasNext() ) {
                chunkID = ( ChunkID ) ( ( Map.Entry ) iterator.next() ).getKey();
                cacheChunk( chunkID, false );
                iterator.remove();
                //clearChunk( chunkID );
            }
            System.gc();
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunks(), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
        //try {
        //    clearMemoryReserve();
        //    int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
        //    int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
        //    int chunkRowIndex;
        //    int chunkColIndex;
        //    ChunkID chunkID;
        //    for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
        //        for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
        //            chunkID = new ChunkID( chunkNcols, chunkRowIndex, chunkColIndex );
        //            if ( ! isCached( chunkID , false) ) {
        //                cacheChunk( chunkRowIndex, chunkColIndex , false);
        //            }
        //        }
        //    }
        //    if ( reinitialiseMemoryReserve ) {
        //        initMemoryReserve();
        //    }
        //} catch ( java.lang.OutOfMemoryError oome0 ) {
        //    throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunks(), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        //}
    }
    
    /**
     * Returns true iff grid2DSquareCellChunk given by chunkID is cached. This
     * must be an upToDate cache.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk tested to see if
     *   it is cached.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final boolean isCached( ChunkID chunkID , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            if ( this.grid2DSquareCellChunks.containsKey( chunkID ) ) {
                if ( reinitialiseMemoryReserve ) {
                    initMemoryReserve();
                }
                return false;
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
            return true;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( ChunkID( " + chunkID.toString() + " ) ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory. This is usually only
     * done after the equivallent of cacheChunk(chunkRowIndex,chunkColIndex) has
     * been called.
     * @param chunkRowIndex The chunk row index of the grid2DSquareCellChunk
     *   to be cleared.
     * @param chunkColIndex The chunk column index of the grid2DSquareCellChunk
     *   to be cleared.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void clearChunk( int chunkRowIndex, int chunkColIndex , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            clearChunk( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ) );
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".cacheChunk( chunkRowIndex( " + chunkRowIndex + " ), chunkColIndex( " + chunkColIndex + " ) ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory. This is usually only
     * done after the equivallent of cacheChunk(chunkID) has been called.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk to be cleared.
     */
    public final void clearChunk( ChunkID chunkID ) {
        try {
            //this.grid2DSquareCellDoubleChunks.remove( chunkID );
            //this.grid2DSquareCellDoubleChunks.purge();
            //this.grid2DSquareCellDoubleChunks[ chunkRowIndex ][ chunkColIndex ] = null;
            //if ( getGrid2DSquareCellChunks().remove( chunkID ) == null ) {
            //  System.gc();
            //}
            //getGrid2DSquareCellChunks().remove( chunkID );
            this.grid2DSquareCellChunks.remove( chunkID );
            System.gc();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".clearChunk( ChunkID( " + chunkID.toString() + " ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory. This is usually
     * only done after the cacheChunks() has been called.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void clearChunks(boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            Iterator iterator = this.grid2DSquareCellChunks.entrySet().iterator();
            ChunkID chunkID;
            while ( iterator.hasNext() ) {
                iterator.next();
                iterator.remove();
            }
//int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
//int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
//int chunkRowIndex;
//int chunkColIndex;
//for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
//    for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
//        clearChunk( chunkRowIndex, chunkColIndex , false);
//    }
//}
            System.gc();
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".clearChunks(), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Clears Grid2DSquareCellChunkAbstracts with ChunkIDs in chunkIDs.
     * @param chunkIDs A HashSet containing the ChunkIDs of the
     * Grid2DSquareCellChunkAbstracts to be cleared.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void clearChunks( HashSet chunkIDs , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            Iterator iterator = chunkIDs.iterator();
            ChunkID chunkIDToClear;
            while( iterator.hasNext() ) {
                chunkIDToClear = ( ChunkID ) iterator.next();
                clearChunk( chunkIDToClear );
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".clearChunks( HashSet( " + chunkIDs.toString() + " ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by chunkID.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void clearChunkExcept( ChunkID chunkID , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
            int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
            int chunkRowIndex;
            int chunkColIndex;
            ChunkID chunkIDToCache;
            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
                    chunkIDToCache = new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex );
                    if ( ! chunkIDToCache.equals( chunkID ) ) {
                        clearChunk( chunkRowIndex, chunkColIndex , false);
                        return;
                    }
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".clearChunkExcept( ChunkID( " + chunkID.toString() + " ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by chunkID.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void clearChunksExcept( ChunkID chunkID , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
            int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
            int chunkRowIndex;
            int chunkColIndex;
            ChunkID chunkIDToCache;
            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
                    chunkIDToCache = new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex );
                    if ( ! chunkIDToCache.equals( chunkID ) ) {
                        clearChunk( chunkRowIndex, chunkColIndex , false);
                    }
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".clearChunksExcept( ChunkID( " + chunkID.toString() + " ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except those with
     * ChunkIDs in chunkIDs.
     * @param chunkIDs The HashSet of ChunkIDs of the grid2DSquareCellChunks not
     *   to be cleared.
     * @param reinitialiseMemoryReserve If true then this.memoryReserve is initialised
     *   prior to returning from this method.
     */
    public final void clearChunksExcept( HashSet chunkIDs , boolean reinitialiseMemoryReserve) {
        try {
            clearMemoryReserve();
            int nChunkRows = getNChunkRows(this.handleOutOfMemoryErrorFalse);
            int nChunkCols = getNChunkCols(this.handleOutOfMemoryErrorFalse);
            int chunkRowIndex;
            int chunkColIndex;
            ChunkID chunkIDToCache;
            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
                    chunkIDToCache = new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex );
                    if ( ! chunkIDs.contains( chunkIDToCache ) ) {
                        clearChunk( chunkRowIndex, chunkColIndex , false);
                    }
                }
            }
            if ( reinitialiseMemoryReserve ) {
                initMemoryReserve();
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            throw new Error( "OutOfMemoryError in " + this.getClass().toString() + ".clearChunksExcept( HashSet( " + chunkIDs.toString() + " ), memoryReserve must have been insufficiently large or not garbage collected appropriately!" );
        }
    }
    
    /**
     * Restores a grid2DSquareCellChunk from a cache.
     * @param chunkRowIndex The chunk row index of the grid2DSquareCellChunk
     *   to be restored.
     * @param chunkColIndex The chunk column index of the grid2DSquareCellChunk
     *   to be restored.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void restoreChunk( int chunkRowIndex, int chunkColIndex, boolean handleOutOfMemoryError ) {
        try {
            restoreChunk( new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                restoreChunk( chunkRowIndex, chunkColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Restores a grid2DSquareCellChunk from a cache.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk to be restored.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void restoreChunk( ChunkID chunkID, boolean handleOutOfMemoryError ) {
        try {
            if ( isCached( chunkID , false) ) {
                File file = new File( getDirectory(), "" + chunkID.getChunkRowIndex() + "_" + chunkID.getChunkColIndex() );
                ObjectInputStream ois = null;
                try {
                    ois = new ObjectInputStream( new BufferedInputStream( new FileInputStream( file ) ) );
                } catch ( Exception e0 ) {
                    e0.printStackTrace();
                }
                if ( this.getClass() == Grid2DSquareCellInt.class ) {
                    //Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = ( Grid2DSquareCellDoubleChunkAbstract ) ois.readObject();
                    Grid2DSquareCellIntChunkAbstract grid2DSquareCellIntChunk = null;
                    try {
                        grid2DSquareCellIntChunk = ( Grid2DSquareCellIntChunkAbstract ) ois.readObject();
                    } catch ( Exception e1 ) {
                        e1.printStackTrace();
                    }
                    //grid2DSquareCellDoubleChunk.initGrid2DSquareCellDouble( this );
                    grid2DSquareCellIntChunk.initGrid2DSquareCell( this );
                    //grid2DSquareCellDoubleChunk.initChunkID( chunkID );
                    grid2DSquareCellIntChunk.initChunkID( chunkID );
                    if ( grid2DSquareCellIntChunk.getClass() == Grid2DSquareCellIntChunk64CellMap.class ) {
                        this.grid2DSquareCellChunks.put( chunkID, ( Grid2DSquareCellIntChunk64CellMap ) grid2DSquareCellIntChunk );
                    } else {
                        if ( grid2DSquareCellIntChunk.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                            this.grid2DSquareCellChunks.put( chunkID, ( Grid2DSquareCellIntChunkArray ) grid2DSquareCellIntChunk );
                        } else {
                            if ( grid2DSquareCellIntChunk.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                                this.grid2DSquareCellChunks.put( chunkID, ( Grid2DSquareCellIntChunkMap ) grid2DSquareCellIntChunk );
                            } else {
                                if ( grid2DSquareCellIntChunk.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                                    this.grid2DSquareCellChunks.put( chunkID, ( Grid2DSquareCellIntChunkJAI ) grid2DSquareCellIntChunk );
                                } else {
                                    System.out.println("Warning: Unrecognised type of Grid2DSquareCellIntChunkAbstract or null " + grid2DSquareCellIntChunk + " in Grid2DSquareCellIntAbstract.restoreChunk( ChunkID( " + chunkID.toString() + " ) )" );
                                    // log( "Warning: Unrecognised type of Grid2DSquareCellIntChunkAbstract or null " + grid2DSquareCellIntChunk + " in Grid2DSquareCellIntAbstract.restoreChunk( ChunkID( " + chunkID.toString() + " ) )" );
                                }
                            }
                        }
                    }
                } else {
                    //this.getClass() == Grid2DSquareCellDouble.class
                    //Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = ( Grid2DSquareCellDoubleChunkAbstract ) ois.readObject();
                    Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = null;
                    try {
                        grid2DSquareCellDoubleChunk = ( Grid2DSquareCellDoubleChunkAbstract ) ois.readObject();
                    } catch ( Exception e1 ) {
                        e1.printStackTrace();
                    }
                    //grid2DSquareCellDoubleChunk.initGrid2DSquareCellDouble( this );
                    grid2DSquareCellDoubleChunk.initGrid2DSquareCell( this );
                    //grid2DSquareCellDoubleChunk.initChunkID( chunkID );
                    grid2DSquareCellDoubleChunk.initChunkID( chunkID );
                    if ( grid2DSquareCellDoubleChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                        this.grid2DSquareCellChunks.put( chunkID, ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellDoubleChunk );
                    } else {
                        if ( grid2DSquareCellDoubleChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                            this.grid2DSquareCellChunks.put( chunkID, ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellDoubleChunk );
                        } else {
                            if ( grid2DSquareCellDoubleChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                                this.grid2DSquareCellChunks.put( chunkID, ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellDoubleChunk );
                            } else {
                                if ( grid2DSquareCellDoubleChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                                    this.grid2DSquareCellChunks.put( chunkID, ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellDoubleChunk );
                                } else {
                                    System.out.println("Warning: Unrecognised type of Grid2DSquareCellDoubleChunkAbstract or null " + grid2DSquareCellDoubleChunk + " in Grid2DSquareCellDoubleAbstract.restoreChunk( ChunkID( " + chunkID.toString() + " ) )" );
                                    // log( "Warning: Unrecognised type of Grid2DSquareCellDoubleChunkAbstract or null " + grid2DSquareCellDoubleChunk + " in Grid2DSquareCellDoubleAbstract.restoreChunk( ChunkID( " + chunkID.toString() + " ) )" );
                                }
                            }
                        }
                    }
                }
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkID , false) );
                initMemoryReserve();
                restoreChunk( chunkID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - the cell IDs for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     * returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     * returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID[] getCellIDs( double x, double y, double distance, boolean handleOutOfMemoryError ) {
        try {
            return getCellIDs( x, y, getCellRowIndex( y, handleOutOfMemoryError ), getCellColIndex( x, handleOutOfMemoryError ), distance, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellIDs( x, y, distance, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - the cell IDs for cells thats centroids would be
     * intersected by circle with centre at centroid of cell given by cell row
     * index cellRowIndex, cell column index cellColIndex, and radius distance.
     * @param cellRowIndex the row index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param cellColIndex the column index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID[] getCellIDs( long cellRowIndex, long cellColIndex, double distance, boolean handleOutOfMemoryError ) {
        try {
            return getCellIDs( getCellX( cellColIndex, handleOutOfMemoryError ), getCellY( cellRowIndex, handleOutOfMemoryError ), cellRowIndex, cellColIndex, distance, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellIDs( cellRowIndex, cellColIndex, distance, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns double[] cells - the values for cells thats centroids would be
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     *        returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     *        returned.
     * @param cellRowIndex the row index at y.
     * @param cellColIndex the col index at x.
     * @param distance the radius of the circle for which intersected cell
     *        values are returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * remove need for copy to new array.
     */
    protected CellID[] getCellIDs( double x, double y, long cellRowIndex, long cellColIndex, double distance, boolean handleOutOfMemoryError ) {
        try {
            CellID[] cellIDs0;
            BigDecimal[] dimensions = getDimensions();
            int cellDistance = ( int ) Math.ceil( distance / dimensions[ 0 ].doubleValue() );
            int limit = ( ( 2 * cellDistance ) + 1 ) * ( ( 2 * cellDistance ) + 1 );
            cellIDs0 = new CellID[ limit ];
            long row;
            long col;
            double thisX;
            double thisY;
            int count = 0;
            long ncols = getNcols();
            //if ( limit > 0 ) {
            for ( row = cellRowIndex - cellDistance; row <= cellRowIndex + cellDistance; row ++ ) {
                thisY = getCellY( row, handleOutOfMemoryError );
                for ( col = cellColIndex - cellDistance; col <= cellColIndex + cellDistance; col ++ ) {
                    thisX = getCellX( col, handleOutOfMemoryError );
                    if ( Utilities.distance( x, y, thisX, thisY ) <= distance ) {
                        cellIDs0[ count ] = new CellID( ncols, row, col );
                        //cellIDs0[ count ] = new CellID( row, col );
                        count ++;
                    }
                }
            }
            CellID[] cellIDs = new CellID[ count ];
            System.arraycopy( cellIDs0, 0, cellIDs, 0, count );
            return cellIDs;
            //}
            //return cellIDs0;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellIDs( x, y, cellRowIndex, cellColIndex, distance, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the the nearest cells cellRowIndex and cellColIndex as a long[]
     * from ID to point given by x-coordinate x, y-coordinate y.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID getNearestCellID( double x, double y, boolean handleOutOfMemoryError ) {
        try {
            return getNearestCellID( x, y, getCellRowIndex( y, handleOutOfMemoryError ), getCellColIndex( x, handleOutOfMemoryError ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getNearestCellID( x, y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the the nearest cells cellRowIndex and cellColIndex as a long[]
     * from ID to point given by cell row index cellRowIndex, cell column index
     * cellColIndex.
     * @param cellRowIndex the row index from which nearest cell ID is returned.
     * @param cellColIndex the column index from which nearest cell ID is returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * return CellID[] as could be more than one nearest CellID
     */
    public CellID getNearestCellID( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getNearestCellID( getCellX( cellColIndex, handleOutOfMemoryError ), getCellY( cellRowIndex, handleOutOfMemoryError ), cellRowIndex, cellColIndex, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getNearestCellID( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the the nearest CellID to point given by x-coordinate x,
     * y-coordinate y in position given by cellRowIndex, cellColIndex.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param cellRowIndex The cell row index of cell containing point.
     * @param cellColIndex The cell column index of cell containing point.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected CellID getNearestCellID( double x, double y, long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            CellID cellID;
            long ncols = getNcols();
            if ( ! inGrid( x, y, handleOutOfMemoryError ) ) {
                BigDecimal[] dimensions = getDimensions();
                long row;
                long col;
                if ( x >= dimensions[3].doubleValue() ) {
                    col = getNcols() - 1;
                    if ( y > dimensions[4].doubleValue() ) {
                        row = 0;
                    } else {
                        if ( y < dimensions[2].doubleValue() ) {
                            row = getNrows() - 1;
                        } else {
                            row = getCellRowIndex( y, handleOutOfMemoryError );
                        }
                    }
                } else {
                    if ( x < dimensions[1].doubleValue() ) {
                        col = 0;
                        if ( y >= dimensions[4].doubleValue() ) {
                            row = 0;
                        } else {
                            if ( y < dimensions[2].doubleValue() ) {
                                row = getNrows() -1;
                            } else {
                                row = getCellRowIndex( y, handleOutOfMemoryError );
                            }
                        }
                    } else {
                        col = getCellColIndex( x, handleOutOfMemoryError );
                        if ( y >= dimensions[4].doubleValue() ) {
                            row = 0;
                        } else {
                            row = getNrows() - 1;
                        }
                    }
                }
                cellID = new CellID( ncols, row, col );
                //cellID = new CellID( row, col );
            } else {
                cellID = new CellID( ncols, cellRowIndex, cellColIndex );
                //cellID = new CellID( cellRowIndex, cellColIndex );
            }
            return cellID;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getNearestCellID( x, y, cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the height of the grid.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getHeightDouble( boolean handleOutOfMemoryError ) {
        try {
            return getHeightBigDecimal( handleOutOfMemoryError ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getHeightDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the height of the grid.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getHeightBigDecimal( boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( dimensions[ 4 ].subtract( dimensions[ 2 ] ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getHeightBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the width of the grid as a double.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getWidthDouble( boolean handleOutOfMemoryError ) {
        try {
            return getWidthBigDecimal( handleOutOfMemoryError ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getWidthDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the width of the grid as a BigDecimal.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getWidthBigDecimal( boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( dimensions[ 3 ].subtract( dimensions[ 1 ] ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getWidthBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param xBigDecimal The x-coordinate of the point.
     * @param yBigDecimal The y-coordinate of the point.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final boolean inGrid( BigDecimal xBigDecimal, BigDecimal yBigDecimal, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( xBigDecimal.compareTo( dimensions[ 1 ] ) != -1 &&
                    yBigDecimal.compareTo( dimensions[ 2 ] ) != -1 &&
                    xBigDecimal.compareTo( dimensions[ 3 ] ) != 1 &&
                    yBigDecimal.compareTo( dimensions[ 4 ] ) != 1 );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return inGrid( xBigDecimal, yBigDecimal, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean inGrid( double x, double y, boolean handleOutOfMemoryError ) {
        try {
            return inGrid( BigDecimal.valueOf( x ), BigDecimal.valueOf( y ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return inGrid( x, y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff position given by cell row index cellRowIndex, cell
     * column index cellColIndex is in the Grid.
     * @param cellRowIndex The cell row index to test.
     * @param cellColIndex The cell column index to test.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean inGrid( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return cellRowIndex >= 0 && cellRowIndex < getNrows() &&
                    cellColIndex >= 0 && cellColIndex < getNcols() ;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return inGrid( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff position given by chunk row index chunkRowIndex, chunk
     * column index chunkColIndex is in the Grid.
     * @param chunkRowIndex The chunk row index to test.
     * @param chunkColIndex The chunk column index to test.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean inGrid( int chunkRowIndex, int chunkColIndex, boolean handleOutOfMemoryError ) {
        try {
            return chunkRowIndex >= 0 && chunkRowIndex < getChunkNrows(this.handleOutOfMemoryErrorFalse) &&
                    chunkColIndex >= 0 && chunkColIndex < getChunkNcols(this.handleOutOfMemoryErrorFalse) ;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return inGrid( chunkRowIndex, chunkColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff cell given by cellID is in the Grid.
     * @param cellID The CellID of a cell to test.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean inGrid( CellID cellID, boolean handleOutOfMemoryError ) {
        try {
            return inGrid( cellID.getCellRowIndex(), cellID.getCellColIndex(), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return inGrid( cellID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff ChunkID chunkID is in the Grid.
     * @param chunkID The ChunkID of a cell to test.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean inGrid( ChunkID chunkID, boolean handleOutOfMemoryError ) {
        try {
            return  chunkID.chunkRowIndex > -1 && chunkID.chunkRowIndex < getNChunkRows(this.handleOutOfMemoryErrorFalse) &&
                    chunkID.chunkColIndex > -1 && chunkID.chunkColIndex < getNChunkCols(this.handleOutOfMemoryErrorFalse);
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return inGrid( chunkID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff cell given by chunkRowIndex, chunkColIndex,
     * chunkCellRowIndex, chunkCellColIndex is in the Grid.
     * @param chunkRowIndex
     * @param chunkColIndex
     * @param chunkCellRowIndex
     * @param chunkCellColIndex
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean inGrid( int chunkRowIndex, int chunkColIndex, int chunkCellRowIndex, int chunkCellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return inGrid( ( ( long ) chunkRowIndex * ( long ) getChunkNcols(this.handleOutOfMemoryErrorFalse) ) + ( long ) chunkCellRowIndex, ( ( long ) chunkColIndex * ( long ) getChunkNrows(this.handleOutOfMemoryErrorFalse) ) + ( long ) chunkCellColIndex, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return inGrid( chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a BigDecimal.
     * @param cellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellXBigDecimal( long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( dimensions[1].add( dimensions[0].multiply( BigDecimal.valueOf( cellColIndex ) ) ) ).add( dimensions[0].divide( BigDecimal.valueOf( 2L ), getDimensionsScale( handleOutOfMemoryError ), BigDecimal.ROUND_HALF_EVEN ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellXBigDecimal( cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a double.
     * @param cellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellXDouble( long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getCellXBigDecimal( cellColIndex, handleOutOfMemoryError ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellXDouble( cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * @deprecated use getCellXDouble(long,boolean)
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a double.
     * @param cellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public @Deprecated  final double getCellX( long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getCellXBigDecimal( cellColIndex, handleOutOfMemoryError ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellX( cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cell with cell ID cellID as
     * a BigDecimal.
     * @param cellID
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellXBigDecimal( CellID cellID, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( dimensions[1].add( dimensions[0].multiply( BigDecimal.valueOf( cellID.cellColIndex ) ) ) ).add( dimensions[0].divide( BigDecimal.valueOf( 2L ), getDimensionsScale( handleOutOfMemoryError ), BigDecimal.ROUND_HALF_EVEN ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellXBigDecimal( cellID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cell with cell ID cellID as
     * a double
     * @param cellID
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellXDouble( CellID cellID, boolean handleOutOfMemoryError ) {
        try {
            return getCellXBigDecimal( cellID, handleOutOfMemoryError ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellXDouble( cellID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * @deprecated use getCellXDouble(CellID,boolean)
     * Returns the x-coordinate of the centroid for cell with cell ID cellID as
     * a double.
     * @param cellID
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public @Deprecated final double getCellX( CellID cellID, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( ( dimensions[1].add( dimensions[0].multiply( BigDecimal.valueOf( cellID.cellColIndex ) ) ) ).add( dimensions[0].divide( BigDecimal.valueOf( 2L ), getDimensionsScale( handleOutOfMemoryError ), BigDecimal.ROUND_HALF_EVEN ) ) ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellX( cellID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a BigDecimal.
     * @param cellRowIndex the cell column index thats centroid y-coordinate is
     *   returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellYBigDecimal( long cellRowIndex, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            //BigDecimal cellRowIndexBigDecimal = new BigDecimal( cellRowIndex );
            //BigDecimal part1 = ( dimensions[2].add( dimensions[0].multiply( cellRowIndexBigDecimal ) ) );
            //BigDecimal part2 = ( dimensions[0].divide( new BigDecimal( 2L ), getDimensionsScale(), BigDecimal.ROUND_HALF_EVEN ) );
            //double result = ( part1.add( part2 ) ).doubleValue();
            //return result;
            return ( dimensions[2].add( dimensions[0].multiply( BigDecimal.valueOf( cellRowIndex ) ) ) ).add( dimensions[0].divide( BigDecimal.valueOf( 2L ), getDimensionsScale( handleOutOfMemoryError ), BigDecimal.ROUND_HALF_EVEN ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellYBigDecimal( cellRowIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a double.
     * @param cellRowIndex the cell column index thats centroid y-coordinate is
     * returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellYDouble( long cellRowIndex, boolean handleOutOfMemoryError ) {
        try {
            return getCellYBigDecimal( cellRowIndex, handleOutOfMemoryError ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellYDouble( cellRowIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * @deprecated use getCellYDouble(long,boolean)
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a double.
     * @param cellRowIndex the cell column index thats centroid y-coordinate is
     *   returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public @Deprecated final double getCellY( long cellRowIndex, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( ( dimensions[2].add( dimensions[0].multiply( BigDecimal.valueOf( cellRowIndex ) ) ) ).add( dimensions[0].divide( BigDecimal.valueOf( 2L ), getDimensionsScale( handleOutOfMemoryError ), BigDecimal.ROUND_HALF_EVEN ) ) ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellY( cellRowIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * BigDecimal.
     * @param cellID
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellYBigDecimal( CellID cellID, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( dimensions[2].add( dimensions[0].multiply( BigDecimal.valueOf( cellID.cellRowIndex ) ) ) ).add( dimensions[0].divide( BigDecimal.valueOf( 2L ), getDimensionsScale( handleOutOfMemoryError ), BigDecimal.ROUND_HALF_EVEN ) );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellYBigDecimal( cellID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * double.
     * @param cellID
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellYDouble( CellID cellID, boolean handleOutOfMemoryError ) {
        try {
            return getCellYBigDecimal( cellID, handleOutOfMemoryError ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellYDouble( cellID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * @deprecated use getCellYDouble(CellID,boolean)
     * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * double.
     * @param cellID
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public @Deprecated final double getCellY( CellID cellID, boolean handleOutOfMemoryError ) {
        try {
            BigDecimal[] dimensions = getDimensions();
            return ( ( dimensions[2].add( dimensions[0].multiply( BigDecimal.valueOf( cellID.cellRowIndex ) ) ) ).add( dimensions[0].divide( BigDecimal.valueOf( 2L ), getDimensionsScale( handleOutOfMemoryError ), BigDecimal.ROUND_HALF_EVEN ) ) ).doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellY( cellID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns gridBounds (the bounding box of the grid) as a double[] where;
     * gridBounds[0] xmin, left most x-coordinate of this
     * gridBounds[1] ymin, lowest y-coordinate of this
     * gridBounds[2] xmax, right most x-coordinate of this
     * gridBounds[3] ymax, highest y-coordinate of this
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * Are bounds in double range?
     * Is there more than cellsize difference with precision?
     *   Throw appropriate exceptions.
     */
    public final double[] getGridBounds( boolean handleOutOfMemoryError ) {
        try {
            double[] result = new double[ 4 ];
            BigDecimal[] dimensions = getDimensions();
            result[ 0 ] = dimensions[ 1 ].doubleValue();
            result[ 1 ] = dimensions[ 2 ].doubleValue();
            result[ 2 ] = dimensions[ 3 ].doubleValue();
            result[ 3 ] = dimensions[ 4 ].doubleValue();
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getGridBounds( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns double[] where;
     * double[0] xmin, left most x-coordinate of cell at (rowIndex,colIndex)
     * double[1] ymin, lowest y-coordinate of cell at (rowIndex,colIndex)
     * double[2] xmax, right most x-coordinate of cell at (rowIndex,colIndex)
     * double[3] ymax, highest y-coordinate of cell at (rowIndex,colIndex)
     * @param cellRowIndex the row index of the cell for which the bounds are
     *   returned
     * @param cellColIndex the column index of the cell for which the bounds are
     *   returned
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double[] getCellBoundsDoubleArray( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getCellBoundsDoubleArray( getCellX( cellColIndex, handleOutOfMemoryError ), getCellY( cellRowIndex, handleOutOfMemoryError ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellBoundsDoubleArray( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns double[] where;
     * double[0] xmin, left most x-coordinate of cell that intersects point at (x,y)
     * double[1] ymin, lowest y-coordinate of cell that intersects point at (x,y)
     * double[2] xmax, right most x-coordinate of cell that intersects point at (x,y)
     * double[3] ymax, highest y-coordinate of cell that intersects point at (x,y)
     * @param x the x-coordinate in the cell for which the bounds are returned
     * @param y the y-coordinate in the cell for which the bounds are returned
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * Precision may compromise result. More precision is available via
     * BigDecimal arithmetic.
     */
    public final double[] getCellBoundsDoubleArray( double x, double y, boolean handleOutOfMemoryError ) {
        try {
            double[] cellBounds = new double[ 4 ];
            double halfCellsize = getDimensions()[ 0 ].doubleValue() / 2.0d;
            cellBounds[ 0 ] = x - halfCellsize;
            cellBounds[ 1 ] = y - halfCellsize;
            cellBounds[ 2 ] = x + halfCellsize;
            cellBounds[ 3 ] = y + halfCellsize;
            return cellBounds;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getCellBoundsDoubleArray( x, y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the noDataValue of this as a BigDecimal.
     */
    public abstract BigDecimal getNoDataValueBigDecimal( boolean handleOutOfMemoryError );
    
    /**
     * @deprecated This should not be used
     * Returns the noDataValue of this as a double.
     * @param handleOutOfMemoryError Controls whether OutOfMemoryError is
     *   handled via caching by this method. If true then caching initiated in
     *   this method, if false then OutOfMemoryError is thrown to be handled
     *   higher up by caller method.
     * TODO:
     * Test
     * Handling for ArithmeticExceptions caused by rounding.
     */
    public @Deprecated double getNoDataValueDouble( boolean handleOutOfMemoryError ) {
        try {
            BigDecimal noDataValueBigDecimal = getNoDataValueBigDecimal( handleOutOfMemoryError );
            return noDataValueBigDecimal.doubleValue();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getNoDataValueDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the next ChunkID in row major order from chunkID, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * @param handleOutOfMemoryError
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    public ChunkID getNextChunk( ChunkID chunkID, int nChunkRows, int nChunkCols, boolean handleOutOfMemoryError ) {
        try {
            int chunkRowIndex = chunkID.getChunkRowIndex();
            int chunkColIndex = chunkID.getChunkColIndex();
            if ( chunkColIndex < nChunkCols - 1 ) {
                return new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex + 1 );
            } else {
                if ( chunkRowIndex < nChunkRows - 1  ) {
                    return new ChunkID( nChunkCols, chunkRowIndex + 1, 0 );
                }
            }
            return null;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk( false ) );
                initMemoryReserve();
                return  getNextChunk( chunkID, nChunkRows, nChunkCols, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the next ChunkID in row minor order from this, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * @param handleOutOfMemoryError
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    public ChunkID getNextChunkRowMinor( ChunkID chunkID, int nChunkRows, int nChunkCols , boolean handleOutOfMemoryError) {
        try {
            int chunkRowIndex = chunkID.getChunkRowIndex();
            int chunkColIndex = chunkID.getChunkColIndex();
            if ( chunkColIndex < nChunkCols - 1 ) {
                return new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex + 1 );
            } else {
                if ( chunkRowIndex > 0 ) {
                    return new ChunkID( nChunkCols, chunkRowIndex - 1, 0 );
                }
            }
            return null;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk( false ) );
                initMemoryReserve();
                return  getNextChunkRowMinor( chunkID, nChunkRows, nChunkCols, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the next ChunkID in row major order from this, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * @param handleOutOfMemoryError
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    public ChunkID getPreviousChunk( ChunkID chunkID, int nChunkRows, int nChunkCols, boolean handleOutOfMemoryError ) {
        try {
            int chunkRowIndex = chunkID.getChunkRowIndex();
            int chunkColIndex = chunkID.getChunkColIndex();
            if ( chunkColIndex > 0 ) {
                return new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex - 1 );
            } else {
                if ( chunkRowIndex > 0  ) {
                    return new ChunkID( nChunkCols, chunkRowIndex - 1, 0 );
                }
            }
            return null;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk( false ) );
                initMemoryReserve();
                return  getPreviousChunk( chunkID, nChunkRows, nChunkCols, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A simple CellID class for distinguishing cells.
     */
    public static class CellID extends Object implements Serializable, Comparable {
        
        /**
         * For storing the row in which the cell is positioned
         */
        protected long cellRowIndex;
        
        /**
         * For storing the column in which the cell is positioned
         */
        protected long cellColIndex;
        
        /**
         * For storing a hashCode for the cell.
         * NB. When the number of cells in a grid is more than the maximum int
         * value care should be taken that hashCodes are distributed in a
         * relatively even way. Operations will slow down greatly if a large
         * number of cells have a CellID with the same hashCode.
         */
        protected int hashCode;
        
        /**
         * Default constructor.
         */
        public CellID() {}
        
        /**
         * Constructs a new CellID that is a clone of cellID
         * @param cellID The CellID to clone.
         */
        public CellID( CellID cellID ) {
            this.cellColIndex = cellID.cellColIndex;
            this.cellRowIndex = cellID.cellRowIndex;
            this.hashCode = cellID.hashCode;
        }
        
        /**
         * Constructs a new CellID
         * @param cellRowIndex The row of the cell.
         * @param cellColIndex The column of the cell.
         * @param ncols The number of columns in a grid2DSquareCellAbstract.
         */
        public CellID( long ncols, long cellRowIndex, long cellColIndex ) {
            //this( nrows, ncols, cellRowIndex, cellColIndex );
            this.cellRowIndex = cellRowIndex;
            this.cellColIndex = cellColIndex;
            this.hashCode = ( int ) ( ( ncols * cellRowIndex ) + cellColIndex );
        }
        
        //public CellID( long cellRowIndex, long cellColIndex ) {
        //    this( ncols, cellRowIndex, cellColIndex );
        //    this.cellRowIndex = cellRowIndex;
        //    this.cellColIndex = cellColIndex;
        //    this.hashCode = ( int ) ( ( this.ncols * cellRowIndex ) + cellColIndex );
        //}
        
        /**
         * Returns hashCode of this CellID
         */
        public int hashCode() {
            return this.hashCode;
        }
        
        /**
         * Returns cellRowIndex of this CellID
         */
        public long getCellRowIndex() {
            return this.cellRowIndex;
        }
        
        /**
         * Returns cellColIndex of this CellID
         */
        public long getCellColIndex() {
            return this.cellColIndex;
        }
        
        /**
         * Returns a description of this CellID
         */
        public String toString() {
            return "CellID( hashCode( " + hashCode() + "), cellRowIndex( " + getCellRowIndex() + " ), cellColIndex( " + getCellColIndex() + " ) )";
        }
        
        /**
         * Overrides equals in Object
         */
        public boolean equals( Object object ) {
            if ( this == object ) {
                return true;
            }
            if ( ( object == null ) || ( object.getClass() != this.getClass() ) ) {
                return false;
            }
            CellID anotherCellID = ( CellID ) object;
            return ( ( this.cellColIndex == anotherCellID.cellColIndex ) && ( this.cellRowIndex == anotherCellID.cellRowIndex ) );
        }
        
        /**
         * Method required by Comparable
         */
        public int compareTo( Object object ) {
            //try {
            // Cast failure should result in NullPointerException or ClassCastException
            CellID cellID = ( CellID ) object;
            if ( cellID.cellRowIndex > this.cellRowIndex ) {
                return 1;
            }
            if ( cellID.cellRowIndex < this.cellRowIndex ) {
                return -1;
            }
            if ( cellID.cellColIndex > this.cellColIndex ) {
                return 1;
            }
            if ( cellID.cellColIndex < this.cellColIndex ) {
                return -1;
            }
            return 0;
            //} catch ( Exception e0 ) {
            //    return -1;
            //}
        }
    }
    
    /**
     * A simple ChunkID class for distinguishing s in this chunk.
     */
    public static class ChunkID extends Object implements Serializable, Comparable {
        
        /**
         * For storing the chunk row index
         */
        protected int chunkRowIndex;
        
        /**
         * For storing the chunk column index
         */
        protected int chunkColIndex;
        
        /**
         * For storing a hashCode for this ChunkID.
         * NB. When the number of cells in a chunk is more than the maximum int
         * value care should be taken to distributed hashcodes in a relatively
         * even way. Operations may slow down greatly if a large number of
         * chunks have ChunkIDs with the same hashCode.
         */
        protected int hashCode;
        
        /**
         * Default constructor.
         */
        public ChunkID() {}
        
        /**
         * Constructs a new ChunkID that is a clone of chunkID
         * @param chunkID The ChunkID to clone.
         */
        public ChunkID( ChunkID chunkID ) {
            this.chunkColIndex = chunkID.chunkColIndex;
            this.chunkRowIndex = chunkID.chunkRowIndex;
            this.hashCode = chunkID.hashCode;
        }
        
        /**
         * Constructs a new ChunkID
         * @param chunkRowIndex the row of the
         * @param chunkColIndex the column of the
         * @param nChunkCols The number of chunk columns (of the grid2DSquareCellAbstract).
         */
        public ChunkID( int nChunkCols, int chunkRowIndex, int chunkColIndex ) {
            this.chunkRowIndex = chunkRowIndex;
            this.chunkColIndex = chunkColIndex;
            this.hashCode = ( int ) ( ( nChunkCols * chunkRowIndex ) + chunkColIndex );
        }
        
        /**
         * Returns this.hashCode
         */
        public int hashCode() {
            return this.hashCode;
        }
        
        /**
         * Returns this.chunkRowIndex
         */
        public int getChunkRowIndex() {
            return this.chunkRowIndex;
        }
        
        /**
         * Returns this.chunkColIndex
         */
        public int getChunkColIndex() {
            return this.chunkColIndex;
        }
        
        /**
         * Returns a description of this
         */
        public String toString() {
            return "ChunkID( hashCode( " + hashCode() + "), chunkRowIndex( " + getChunkRowIndex() + " ), chunkColIndex( " + getChunkColIndex() + " ) )";
        }
        
        /**
         * Overrides equals in Object
         */
        public boolean equals( Object object ) {
            if ( this == object ) {
                return true;
            }
            if ( ( object == null ) || ( object.getClass() != this.getClass() ) ) {
                return false;
            }
            ChunkID anotherChunkID = ( ChunkID ) object;
            return ( ( this.chunkColIndex == anotherChunkID.chunkColIndex ) && ( this.chunkRowIndex == anotherChunkID.chunkRowIndex ) );
        }
        
        /**
         * Method required by Comparable
         */
        public int compareTo( Object object ) {
            //try {
            // Cast failure should throw NullPointerException or ClassCastException
            ChunkID chunkID = ( ChunkID ) object;
            if ( chunkID.chunkRowIndex > this.chunkRowIndex ) {
                return 1;
            }
            if ( chunkID.chunkRowIndex < this.chunkRowIndex ) {
                return -1;
            }
            if ( chunkID.chunkColIndex > this.chunkColIndex ) {
                return 1;
            }
            if ( chunkID.chunkColIndex < this.chunkColIndex ) {
                return -1;
            }
            return 0;
            //} catch ( Exception e0 ) {
            //    return -1;
            //}
        }
        
    }
    
}