/**
 * Version 1.0 is to handle individual single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
// gnu.trove dependencies
//import gnu.trove.TDoubleHashSet;
// java dependencies
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InvalidClassException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
import uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciiGridImporter;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;

/**
 * A class for representing grids of double precision values.
 * @see Grid2DSquareCellAbstract
 */
public class Grid2DSquareCellDouble extends Grid2DSquareCellAbstract implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    /**
     * For storing the NODATA value of the grid, which by default is
     * Double.NEGATIVE_INFINITY.
     * N.B. Double.NaN should not be used.
     * N.B. Care should be taken so that noDataValue is not a data value.
     */
    protected double noDataValue = Double.NEGATIVE_INFINITY;
    
    /**
     * Creates a new Grid2DSquareCellDouble
     */
    public Grid2DSquareCellDouble() {}
    
    /**
     * Creates a new Grid2DSquareCellDouble.
     * Users should take care of concurrent modification if
     * grid2DSquareCellDouble0Directory is in use. If a completely new
     * instance is wanted then use:
     * Grid2DSquareCellDouble( File,Grid2DSquareCellDoubleAbstract,
     *   Grid2DSquareCellDoubleChunkAbstractFactory,int,int.long,long,long,long,
     *   double)
     * @param directory the File directory to be used for caching information
     *   and data about this
     * @param gridFile the File that ois was constructed from
     * @param ois the ObjectInputStream used in first attempt to construct this
     */
    public Grid2DSquareCellDouble(
            File directory,
            File gridFile,
            ObjectInputStream ois ) {
        try {
            //Grid2DSquareCellDouble grid2DSquareCellDouble = ( Grid2DSquareCellDouble ) ois.readObject(); // For debugging
            //initAll( grid2DSquareCellDouble ); // For debugging
            initAll( ( Grid2DSquareCellDouble ) ois.readObject() );
        } catch ( Exception e0 ) {
            if ( e0 instanceof InvalidClassException || e0 instanceof ClassCastException ) {
                try {
                    ois.close();
                    ois = new ObjectInputStream( new BufferedInputStream( new FileInputStream( new File( gridFile, "Cache" ) ) ) );
                } catch ( Exception e1 ) {
                    System.out.println( e1.toString() );
                    e1.printStackTrace();
                }
                try {
                    // If the object is a Grid2DSquareCellInt
                    Grid2DSquareCellIntFactory grid2DSquareCellIntFactory = new Grid2DSquareCellIntFactory( directory );
                    Grid2DSquareCellInt grid2DSquareCellInt = grid2DSquareCellIntFactory.createGrid2DSquareCellInt( directory, gridFile, ois );
                    grid2DSquareCellInt.initMemoryReserve( 0 );
                    Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory( directory );
                    Grid2DSquareCellDouble grid2DSquareCellDouble = grid2DSquareCellDoubleFactory.createGrid2DSquareCellDouble( grid2DSquareCellInt );
                    initAll( grid2DSquareCellDouble );
                } catch ( Exception e1 ) {
                    System.out.println( e1.toString() );
                    e1.printStackTrace();
                }
            } else {
                System.out.println( e0.toString() );
                e0.printStackTrace();
            }
        }
    }
    
    /**
     * Creates a new Grid2DSquareCellDouble with each cell value equal to
     * noDataValue and all chunks of the same type.
     * N.B. This method concieved for testing purposes to see which type of
     * chunks are more efficient in different situations. If this becomes clear
     * it may become unneccesary as it may be possible for this to automatically
     * optimise chunk storage. This optimisiation is going to be difficult owing
     * a trade off between fast access and low memory.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered for creating chunks.
     * @param chunkNrows The number of rows of cells in any chunk.
     * @param chunkNcols The number of columns of cells in any chunk.
     * @param nrows The number of rows of cells.
     * @param ncols The number of columns of cells.
     * @param dimensions The cellsize, xmin, ymin, xmax and ymax.
     * @param noDataValue The noDataValue.
     * @throws OutOfMemoryError if unable to load a chunk constructed.
     * TODO: make protected and move to a factory class.
     */
    public Grid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            double noDataValue ) {
        // Initialise memory reserve to a default small amount
        initMemoryReserve();
        // Initialise Statistics
        this.gridStatistics = gridStatistics;
        this.gridStatistics.init( this );
        this.directory = directory;
        this.chunkNrows = chunkNrows;
        this.chunkNcols = chunkNcols;
        this.nrows = nrows;
        this.ncols = ncols;
        this.dimensions = dimensions;
        initNoDataValue( noDataValue );
        //this.name = directory.getName();
        initNChunkRows( nrows, chunkNrows );
        initNChunkCols( ncols, chunkNcols );
        long totalChunkCount = ( long ) this.nChunkRows * ( long ) this.nChunkCols;
        this.grid2DSquareCellChunks = new HashMap();
        int chunkRowIndex;
        int chunkColIndex;
        int loadedChunkCount = 0;
        boolean isLoadedChunk = false;
        for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
            for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                do {
                    try {
                        // Try to load chunk.
                        getGrid2DSquareCellChunks().put( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ), grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk( this, new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ) ) );
                        isLoadedChunk = true;
                    } catch ( java.lang.OutOfMemoryError oome0 ) {
                        clearMemoryReserve();
                        if ( loadedChunkCount == 0 ) {
                            // This means the OutOfMemoryError is very difficult to handle here. Either:
                            //  - more memory needs to be allocated to the JVM;
                            //  - memory needs to be freed up from other objects
                            //    (OutOfMemory handling done in a higher class);
                            //  - different dimensions or types of
                            //    Grid2DSquareCellChunkAbstract should be used.
                            System.out.println( "Ran out of memory initialising the first chunk!" );
                            throw oome0;
                        } else {
                            System.out.println( "Ran out of memory initialising chunk " + ( loadedChunkCount + 1 ) + " out of " + totalChunkCount + "!" );
                            if ( loadedChunkCount <= nChunkCols ) {
                                System.out.println( "Warning: A complete row of chunks cannot be initialised and therefore stored in memory. This can have serious implications for computation.");
                            }
                        }
                        boolean sufficientMemoryToContinue = false;
                        do {
                            try {
                                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                                initMemoryReserve();
                                sufficientMemoryToContinue = true;
                            } catch ( java.lang.OutOfMemoryError oome1 ) {
                                clearMemoryReserve();
                                if ( this.grid2DSquareCellChunks.size() == 1 ) {
                                    // This means the OutOfMemoryError is very difficult to handle here. Either:
                                    //  - more memory needs to be allocated to the JVM;
                                    //  - memory needs to be freed up from other objects
                                    //    (OutOfMemory handling done in a higher class);
                                    //  - different dimensions or types of
                                    //    Grid2DSquareCellChunkAbstract should be used.
                                    throw oome1;
                                }
                            }
                        } while ( ! sufficientMemoryToContinue );
                    }
                } while ( ! isLoadedChunk );
                isLoadedChunk = false;
                loadedChunkCount ++;
            }
            System.out.println( "Done chunk row " + chunkRowIndex ); // For debugging
        }
    }
    
    /**
     * Creates a new Grid2DSquareCellDouble based on values in grid2DSquareCell.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which this is
     *   to be constructed.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered to construct chunks
     *   of this.
     * @param chunkNrows The number of rows of cells in any chunk.
     * @param chunkNcols The number of columns of cells in any chunk.
     * @param startRowIndex The grid2DSquareCell row index which is the bottom
     *   most row of this.
     * @param startColIndex The grid2DSquareCell column index which is the left
     *   most column of this.
     * @param endRowIndex The grid2DSquareCell row index which is the top most
     *   row of this.
     * @param endColIndex The grid2DSquareCell column index which is the right
     *   most column of this.
     * @param noDataValue The noDataValue for this.
     */
    public Grid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            double noDataValue ) {
        // Initialise memory reserve to a default small amount
        initMemoryReserve();
        this.directory = directory;
        this.chunkNrows = chunkNrows;
        this.chunkNcols = chunkNcols;
        this.nrows = endRowIndex - startRowIndex + 1L;
        this.ncols = endColIndex - startColIndex + 1L;
        initNoDataValue( noDataValue );
        //this.name = directory.getName();
        BigDecimal[] grid2DSquareCellDimensions = grid2DSquareCell.getDimensions();
        this.dimensions = new BigDecimal[ grid2DSquareCellDimensions.length ];
        initNChunkRows( this.nrows, chunkNrows );
        initNChunkCols( this.ncols, chunkNcols );
        long totalChunkCount = ( long ) this.nChunkRows * ( long ) this.nChunkCols;
        this.dimensions[ 0 ] = new BigDecimal( grid2DSquareCellDimensions[ 0 ].toString() );
        this.dimensions[ 1 ] = grid2DSquareCellDimensions[ 1 ].add( new BigDecimal( ( long ) startColIndex ).multiply( this.dimensions[ 0 ] ) );
        this.dimensions[ 2 ] = grid2DSquareCellDimensions[ 2 ].add( new BigDecimal( ( long ) startColIndex ).multiply( this.dimensions[ 0 ] ) );
        this.dimensions[ 3 ] = this.dimensions[ 1 ].add( new BigDecimal( Long.toString( ncols ) ).multiply( this.dimensions[ 0 ] ) );
        this.dimensions[ 4 ] = this.dimensions[ 2 ].add( new BigDecimal( Long.toString( nrows ) ).multiply( this.dimensions[ 0 ] ) );
        this.grid2DSquareCellChunks = new HashMap( this.nChunkRows * this.nChunkCols );
        int chunkRowIndex;
        int chunkColIndex;
        int loadedChunkCount = 0;
        boolean isLoadedChunk = false;
        int chunkRow;
        int chunkCol;
        long row;
        long col;
        // Initialise Statistics
        this.gridStatistics = gridStatistics;
        this.gridStatistics.init( this );
        if ( gridStatistics.getClass() == GridStatistics0.class ) {
            for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                    do {
                        try {
                            // Try to load chunk.
                            this.grid2DSquareCellChunks.put( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ), grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk( this, new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ) ) );
                            row = ( ( long ) chunkRowIndex * ( long ) this.chunkNrows );
                            for ( chunkRow = 0; chunkRow < this.chunkNrows; chunkRow ++ ) {
                                col = ( ( long ) chunkColIndex * ( long ) this.chunkNcols );
                                for ( chunkCol = 0; chunkCol < this.chunkNcols; chunkCol ++ ) {
                                    initCell( row, col, grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorFalse ), this.handleOutOfMemoryErrorFalse );
                                    col ++;
                                }
                                row ++;
                            }
                            isLoadedChunk = true;
                        } catch ( java.lang.OutOfMemoryError oome0 ) {
                            clearMemoryReserve();
                            if ( loadedChunkCount == 0 ) {
                                // This means the OutOfMemoryError is very difficult to handle here. Either:
                                //  - more memory needs to be allocated to the JVM;
                                //  - memory needs to be freed up from other objects
                                //    (OutOfMemory handling done in a higher class);
                                //  - different dimensions or types of
                                //    Grid2DSquareCellChunkAbstract should be used.
                                System.out.println( "Ran out of memory initialising the first chunk!" );
                                throw oome0;
                            } else {
                                System.out.println( "Ran out of memory initialising chunk " + ( loadedChunkCount + 1 ) + " out of " + totalChunkCount + "!" );
                                if ( loadedChunkCount <= nChunkCols ) {
                                    System.out.println( "Warning: A complete row of chunks cannot be initialised and therefore stored in memory. This can have serious implications for computation.");
                                }
                            }
                            boolean sufficientMemoryToContinue = false;
                            do {
                                try {
                                    clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                                    initMemoryReserve();
                                    sufficientMemoryToContinue = true;
                                } catch ( java.lang.OutOfMemoryError oome1 ) {
                                    clearMemoryReserve();
                                    if ( this.grid2DSquareCellChunks.size() == 1 ) {
                                        // This means the OutOfMemoryError is very difficult to handle here. Either:
                                        //  - more memory needs to be allocated to the JVM;
                                        //  - memory needs to be freed up from other objects
                                        //    (OutOfMemory handling done in a higher class);
                                        //  - different dimensions or types of
                                        //    Grid2DSquareCellChunkAbstract should be used.
                                        throw oome1;
                                    }
                                }
                            } while ( ! sufficientMemoryToContinue );
                        }
                    }  while ( ! isLoadedChunk );
                    isLoadedChunk = false;
                    loadedChunkCount ++;
                }
                System.out.println("Done row " + chunkRowIndex );
            }
        } else {
            // gridStatistics.getClass() == GridStatistics1.class
            // initCellFast(long,long,double) is to be used inplace of
            // initCell(long,long,double).
            for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                    do {
                        try {
                            // Try to load chunk.
                            this.grid2DSquareCellChunks.put( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ), grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk( this, new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ) ) );
                            row = ( ( long ) chunkRowIndex * ( long ) this.chunkNrows );
                            for ( chunkRow = 0; chunkRow < this.chunkNrows; chunkRow ++ ) {
                                col = ( ( long ) chunkColIndex * ( long ) this.chunkNcols );
                                for ( chunkCol = 0; chunkCol < this.chunkNcols; chunkCol ++ ) {
                                    initCellFast( row, col, grid2DSquareCell.getCellDouble( row, col, this.handleOutOfMemoryErrorFalse ), this.handleOutOfMemoryErrorFalse );
                                    col ++;
                                }
                                row ++;
                            }
                            isLoadedChunk = true;
                        } catch ( OutOfMemoryError oome0 ) {
                            clearMemoryReserve();
                            if ( loadedChunkCount == 0 ) {
                                // This means the OutOfMemoryError is very difficult to handle here. Either:
                                //  - more memory needs to be allocated to the JVM;
                                //  - memory needs to be freed up from other objects
                                //    (OutOfMemory handling done in a higher class);
                                //  - different dimensions or types of
                                //    Grid2DSquareCellChunkAbstract should be used.
                                System.out.println( "Ran out of memory initialising the first chunk!" );
                                throw oome0;
                            } else {
                                System.out.println( "Ran out of memory initialising chunk " + ( loadedChunkCount + 1 ) + " out of " + totalChunkCount + "!" );
                                if ( loadedChunkCount <= nChunkCols ) {
                                    System.out.println( "Warning: A complete row of chunks cannot be initialised and therefore stored in memory. This can have serious implications for computation.");
                                }
                            }
                            boolean sufficientMemoryToContinue = false;
                            do {
                                try {
                                    clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                                    initMemoryReserve();
                                    sufficientMemoryToContinue = true;
                                } catch ( java.lang.OutOfMemoryError oome1 ) {
                                    clearMemoryReserve();
                                    if ( this.grid2DSquareCellChunks.size() == 1 ) {
                                        // This means the OutOfMemoryError is very difficult to handle here. Either:
                                        //  - more memory needs to be allocated to the JVM;
                                        //  - memory needs to be freed up from other objects
                                        //    (OutOfMemory handling done in a higher class);
                                        //  - different dimensions or types of
                                        //    Grid2DSquareCellChunkAbstract should be used.
                                        throw oome1;
                                    }
                                }
                            } while ( ! sufficientMemoryToContinue );
                        }
                    } while ( ! isLoadedChunk );
                    isLoadedChunk = false;
                    loadedChunkCount ++;
                }
                System.out.println("Done row " + chunkRowIndex );
            }
        }
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from gridFile.
     * Currently gridFile must be a directory of a Grid2DSquareCellDouble or
     * Grid2DSquareCellInt or a ESRI Asciigrid format Files with a name ending
     * ".asc".
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered to construct chunks
     *   of this.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param noDataValue The noDataValue for this.
     */
    public Grid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            File gridFile,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            double noDataValue ) {
        // Initialise memory reserve to a default small amount
        initMemoryReserve();
        if ( gridFile.isDirectory() ) {
            Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory( directory );
            File cache = new File( gridFile, "Cache"  );
            if ( cache.exists() ) {
                ObjectInputStream ois = null;
                try {
                    ois = new ObjectInputStream( new BufferedInputStream( new FileInputStream( cache ) ) );
                } catch ( Exception e1 ) {
                    System.out.println( e1.toString() );
                    e1.printStackTrace();
                }
                Grid2DSquareCellDouble gridFileGrid2DSquareCellDouble = grid2DSquareCellDoubleFactory.createGrid2DSquareCellDouble( directory, cache, ois );
                initAll( grid2DSquareCellDoubleFactory.createGrid2DSquareCellDouble( directory, gridFileGrid2DSquareCellDouble, grid2DSquareCellDoubleChunkFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex, noDataValue ) );
            }
        } else {
            this.directory = directory;
            this.chunkNrows = chunkNrows;
            this.chunkNcols = chunkNcols;
            this.nrows = endRowIndex - startRowIndex + 1L;
            this.ncols = endColIndex - startColIndex + 1L;
            initNoDataValue( noDataValue );
            //this.name = directory.getName();
            initNChunkRows( nrows, chunkNrows );
            initNChunkCols( ncols, chunkNcols );
            long totalChunkCount = ( long ) this.nChunkRows * ( long ) this.nChunkCols;
            this.grid2DSquareCellChunks = new HashMap( this.chunkNrows * this.chunkNcols );
            this.dimensions = new BigDecimal[ 5 ];
            // Initialise Statistics
            this.gridStatistics = gridStatistics;
            this.gridStatistics.init( this );
            String filename = gridFile.getName();
            int loadedChunkCount = 0;
            boolean isLoadedChunk = false;
            // Value must be set otherwise OutOfMemoryError might cause problems.
            double value = noDataValue;
            if ( filename.endsWith( ".asc" ) ) {
                ESRIAsciiGridImporter eSRIAsciigridImporter = new ESRIAsciiGridImporter( gridFile );
                Object[] header = eSRIAsciigridImporter.readHeaderObject();
                //long inputNcols = ( Long ) header[ 0 ];
                //long inputNrows = ( Long ) header[ 1 ];
                this.dimensions[ 0 ] = ( BigDecimal ) header[ 4 ];
                this.dimensions[ 1 ] = ( ( BigDecimal ) header[ 2 ] ).add( dimensions[ 0 ].multiply( new BigDecimal( startColIndex ) ) );
                this.dimensions[ 2 ] = ( ( BigDecimal ) header[ 3 ] ).add( dimensions[ 0 ].multiply( new BigDecimal( startRowIndex ) ) );
                this.dimensions[ 3 ] = this.dimensions[ 1 ].add( new BigDecimal( Long.toString( this.ncols ) ).multiply( this.dimensions[ 0 ] ) );
                this.dimensions[ 4 ] = this.dimensions[ 2 ].add( new BigDecimal( Long.toString( this.nrows ) ).multiply( this.dimensions[ 0 ] ) );
                // Initialise Chunks
                int chunkRowIndex;
                int chunkColIndex;
                ChunkID nextChunkToCache = new ChunkID( this.nChunkCols, 0, 0 );
                for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
                    for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                        do {
                            try {
                                this.grid2DSquareCellChunks.put( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ), grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk( this, new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ) ) );
                                isLoadedChunk = true;
                            } catch ( java.lang.OutOfMemoryError oome0 ) {
                                clearMemoryReserve();
                                if ( loadedChunkCount == 0 ) {
                                    // This means the OutOfMemoryError is very difficult to handle here. Either:
                                    //  - more memory needs to be allocated to the JVM;
                                    //  - memory needs to be freed up from other objects
                                    //    (OutOfMemory handling done in a higher class);
                                    //  - different dimensions or types of
                                    //    Grid2DSquareCellChunkAbstract should be used.
                                    System.out.println( "Ran out of memory initialising the first chunk!" );
                                    throw oome0;
                                } else {
                                    System.out.println( "Ran out of memory initialising chunk " + ( loadedChunkCount + 1 ) + " out of " + totalChunkCount + "!" );
                                    if ( loadedChunkCount <= nChunkCols ) {
                                        System.out.println( "Warning: A complete row of chunks cannot be initialised and therefore stored in memory. This can have serious implications for computation.");
                                    }
                                }
                                cacheAndClearChunks( true );
                                //boolean sufficientMemoryToContinue = false;
                                //int cachedAndClearedChunkCount = 0;
                                //do {
                                //    try {
                                //        //clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                                //        cacheChunk( nextChunkToCache, false);
                                //        clearChunk( nextChunkToCache );
                                //        nextChunkToCache = getNextChunk( nextChunkToCache, this.nChunkRows, this.nChunkCols , this.handleOutOfMemoryErrorFalse );
                                //        cachedAndClearedChunkCount ++;
                                //        initMemoryReserve();
                                //        sufficientMemoryToContinue = true;
                                //    } catch ( java.lang.OutOfMemoryError oome1 ) {
                                //        clearMemoryReserve();
                                //        if ( cachedAndClearedChunkCount == loadedChunkCount ) {
                                //            // This means the OutOfMemoryError is very difficult to handle here. Either:
                                //            //  - more memory needs to be allocated to the JVM;
                                //            //  - memory needs to be freed up from other objects
                                //            //    (OutOfMemory handling done in a higher class);
                                //            //  - different dimensions or types of
                                //            //    Grid2DSquareCellChunkAbstract should be used.
                                //            throw oome1;
                                //        }
                                //    }
                                //} while ( ! sufficientMemoryToContinue );
                            }
                        } while ( ! isLoadedChunk );
                        isLoadedChunk = false;
                        loadedChunkCount ++;
                    }
                }
                System.out.println("First stage of initialisation complete. Reading data into initialised Chunks");
                // Read Data into Chunks
                // TODO
                // Change so row for loop works the other way and less calculations are needed inside.
                long row;
                long col;
                boolean isInitCellDone = false;
                nextChunkToCache = new ChunkID( this.nChunkCols, this.nChunkRows - 1, 0 );
                if ( gridStatistics.getClass() == GridStatistics0.class ) {
                    for ( row = ( this.nrows - 1 ); row > 0; row -- ) {
                        for ( col = 0; col < this.ncols; col ++ ) {
                            value = eSRIAsciigridImporter.readDouble();
                            do {
                                try {
                                    initCell( row, col, value, this.handleOutOfMemoryErrorFalse );
                                    isInitCellDone = true;
                                } catch ( java.lang.OutOfMemoryError oome0 ) {
                                    clearMemoryReserve();
                                    boolean sufficientMemoryToContinue = false;
                                    do {
                                        try {
                                            cacheChunk( nextChunkToCache, false );
                                            clearChunk( nextChunkToCache );
                                            nextChunkToCache = getNextChunkRowMinor( nextChunkToCache, this.nChunkRows, this.nChunkCols, this.handleOutOfMemoryErrorFalse );
                                            //clearChunk( cacheChunkExceptReverseOrder( getChunkRowIndex( row, this.handleOutOfMemoryErrorFalse ) , getChunkColIndex( col, this.handleOutOfMemoryErrorFalse ), false ) );
                                            initMemoryReserve();
                                            sufficientMemoryToContinue = true;
                                        } catch ( java.lang.OutOfMemoryError oome1 ) {
                                            clearMemoryReserve();
                                            if ( this.grid2DSquareCellChunks.size() == 1 ) {
                                                // This means the OutOfMemoryError is very difficult to handle here. Either:
                                                //  - more memory needs to be allocated to the JVM;
                                                //  - memory needs to be freed up from other objects
                                                //    (OutOfMemory handling done in a higher class);
                                                //  - different dimensions or types of
                                                //    Grid2DSquareCellChunkAbstract should be used.
                                                throw oome1;
                                            }
                                        }
                                    } while ( ! sufficientMemoryToContinue );
                                }
                            } while ( ! isInitCellDone );
                            isInitCellDone = false;
                        }
                        System.out.println("Done row " + row );
                    }
                } else {
                    // gridStatistics.getClass() == GridStatistics1.class
                    for ( row = ( this.nrows - 1 ); row > 0; row -- ) {
                        for ( col = 0; col < this.ncols; col ++ ) {
                            value = eSRIAsciigridImporter.readDouble();
                            do {
                                try {
                                    initCellFast( row, col, value, this.handleOutOfMemoryErrorFalse );
                                    isInitCellDone = true;
                                } catch ( java.lang.OutOfMemoryError oome0 ) {
                                    clearMemoryReserve();
                                    boolean sufficientMemoryToContinue = false;
                                    do {
                                        try {
                                            cacheChunk( nextChunkToCache, false );
                                            clearChunk( nextChunkToCache );
                                            nextChunkToCache = getNextChunkRowMinor( nextChunkToCache, this.nChunkRows, this.nChunkCols, this.handleOutOfMemoryErrorFalse );
                                            //clearChunk( cacheChunkExceptReverseOrder( getChunkRowIndex( row, this.handleOutOfMemoryErrorFalse ) , getChunkColIndex( col, this.handleOutOfMemoryErrorFalse ), false ) );
                                            initMemoryReserve();
                                            sufficientMemoryToContinue = true;
                                        } catch ( java.lang.OutOfMemoryError oome1 ) {
                                            clearMemoryReserve();
                                            if ( this.grid2DSquareCellChunks.size() == 1 ) {
                                                // This means the OutOfMemoryError is very difficult to handle here. Either:
                                                //  - more memory needs to be allocated to the JVM;
                                                //  - memory needs to be freed up from other objects
                                                //    (OutOfMemory handling done in a higher class);
                                                //  - different dimensions or types of
                                                //    Grid2DSquareCellChunkAbstract should be used.
                                                throw oome1;
                                            }
                                        }
                                    } while ( ! sufficientMemoryToContinue );
                                }
                            } while ( ! isInitCellDone );
                            isInitCellDone = false;
                        }
                        System.out.println("Done row " + row );
                    }
                }
            }
        }
    }
    
    /**
     * Returns a string description of the instance. Basically the values of
     * each field.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public String toString( boolean handleOutOfMemoryError ) {
        try {
            return "Grid2DSquareCellDouble( " +
                    //"noDataValue( " + getNoDataValue( handleOutOfMemoryError ) + " ), " +
                    "noDataValue( " + getNoDataValue() + " ), " +
                    super.toString( this.handleOutOfMemoryErrorFalse, true ) + " )";
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return toString( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises all fields from an existing instance except
     * grid2DSquareCellChunks. For use with constructors:
     *   Grid2DSquareCellDouble(File,File,ObjectInputStream).
     *   Grid2DSquareCellInt(File,File,ObjectInputStream).
     * @param grid2DSquareCell the Grid2DSquareCellAbstract from which
     *   the field values of this are set.
     */
    //protected void initMain( Grid2DSquareCellAbstract grid2DSquareCell ) {
    //    try {
    //        this.noDataValue = ( ( Grid2DSquareCellDouble ) grid2DSquareCell ).noDataValue;
    //    } catch ( ClassCastException e ) {
    //        this.noDataValue = Double.NEGATIVE_INFINITY;
    //    }
    //    super.initMain( grid2DSquareCell, true );
    //}
    
    /**
     * Initialises all fields from an existing instance. Main use in
     * constructors:
     *   Grid2DSquareCellDouble(File,File,ObjectInputStream).
     *   Grid2DSquareCellInt(File,File,ObjectInputStream).
     * @param grid2DSquareCell the Grid2DSquareCellAbstract from which
     *   the field values of this are set.
     */
    protected void initAll( Grid2DSquareCellAbstract grid2DSquareCell ) {
        try {
            this.noDataValue = ( ( Grid2DSquareCellDouble ) grid2DSquareCell ).noDataValue;
        } catch ( ClassCastException e ) {
            this.noDataValue = Double.NEGATIVE_INFINITY;
        }
        super.initAll( grid2DSquareCell, true );
    }
    
    /**
     * Returns grid2DSquareCellDoubleChunks.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDoubleChunkAbstract getGrid2DSquareCellDoubleChunk( int chunkRowIndex, int chunkColIndex, boolean handleOutOfMemoryError ) {
        try {
            return getGrid2DSquareCellDoubleChunk( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return getGrid2DSquareCellDoubleChunk( chunkRowIndex, chunkColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns grid2DSquareCellDoubleChunksAbstract for the given ChunkID
     * @param chunkID
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDoubleChunkAbstract getGrid2DSquareCellDoubleChunk( ChunkID chunkID, boolean handleOutOfMemoryError ) {
        try {
            if ( inGrid( chunkID, false ) ) {
                if ( ! getGrid2DSquareCellChunks().containsKey( chunkID ) ) {
                    //if ( ! ( this.grid2DSquareCellChunks.containsKey( chunkID ) ) ) {
                    //        if ( this.grid2DSquareCellDoubleChunks.get( chunkID ) == null ) {
                    //if ( this.grid2DSquareCellDoubleChunks[ chunkRowIndex ][ chunkColIndex ] == null ) {
                    restoreChunk( chunkID, handleOutOfMemoryError );
                }
                return ( Grid2DSquareCellDoubleChunkAbstract ) this.grid2DSquareCellChunks.get( chunkID );
                //return this.grid2DSquareCellDoubleChunks[ chunkRowIndex ][ chunkColIndex ];
            }
            return null;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkID , false) );
                initMemoryReserve();
                return getGrid2DSquareCellDoubleChunk( chunkID, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * TODO:
     * Clears all resources to do with this.
     */
    public void clear() {
        return;
    }
    
    /**
     * If newValue and oldValue are the same then statistics won't change. A test might be
     * appropriate in set cell so that this method is not called. Also want to keep track if
     * underlying data has changed for getting statistics of GridStatistics1 type.
     * @param newValue The value replacing oldValue.
     * @param oldValue The value being replaced.
     * @param noDataValue this.noDataValue
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    private void upDateGridStatistics( double newValue, double oldValue, double noDataValue, boolean handleOutOfMemoryError ) {
        try {
            // The java.lang.OutOfMemoryError removed as it thought better for this to fire outwards
            // and be dealt with higher up.
            //try {
            if ( this.gridStatistics instanceof GridStatistics0 ) {
                if ( oldValue != noDataValue ) {
                    BigDecimal oldValueBigDecimal = new BigDecimal( oldValue );
                    this.gridStatistics.nonNoDataValueCountBigInteger = this.gridStatistics.nonNoDataValueCountBigInteger.subtract( BigInteger.ONE );
                    this.gridStatistics.sumBigDecimal = this.gridStatistics.sumBigDecimal.subtract( oldValueBigDecimal );
                    if ( oldValueBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == 0 ) {
                        this.gridStatistics.minCountBigInteger = this.gridStatistics.minCountBigInteger.subtract( BigInteger.ONE );
                    }
                    if ( oldValueBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == 0 ) {
                        this.gridStatistics.maxCountBigInteger = this.gridStatistics.maxCountBigInteger.subtract( BigInteger.ONE );
                    }
                }
                if ( newValue != noDataValue ) {
                    BigDecimal newValueBigDecimal = new BigDecimal( newValue );
                    this.gridStatistics.nonNoDataValueCountBigInteger = this.gridStatistics.nonNoDataValueCountBigInteger.add( BigInteger.ONE );
                    this.gridStatistics.sumBigDecimal = this.gridStatistics.sumBigDecimal.add( newValueBigDecimal );
                    if ( newValueBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == -1 ) {
                        this.gridStatistics.minBigDecimal = newValueBigDecimal;
                        this.gridStatistics.minCountBigInteger = BigInteger.ONE;
                    } else {
                        if ( newValueBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == 0 ) {
                            this.gridStatistics.minCountBigInteger = this.gridStatistics.minCountBigInteger.add( BigInteger.ONE );
                        } else {
                            if ( this.gridStatistics.minCountBigInteger.compareTo( BigInteger.ONE ) == -1 ) {
                                // The GridStatistics need recalculating
                                this.gridStatistics.update();
                            }
                        }
                    }
                    if ( newValueBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == -1 ) {
                        this.gridStatistics.maxBigDecimal = newValueBigDecimal;
                        this.gridStatistics.maxCountBigInteger = BigInteger.ONE;
                    } else {
                        if ( newValueBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == 0 ) {
                            this.gridStatistics.maxCountBigInteger = this.gridStatistics.maxCountBigInteger.add( BigInteger.ONE );
                        } else {
                            if ( this.gridStatistics.maxCountBigInteger.compareTo( BigInteger.ONE ) == -1 ) {
                                // The GridStatistics need recalculating
                                this.gridStatistics.update();
                            }
                        }
                    }
                }
            } else {
                if ( newValue != oldValue ) {
                    ( ( GridStatistics1 ) this.gridStatistics ).setIsUpToDate( false );
                }
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                upDateGridStatistics( newValue, oldValue, noDataValue, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns noDataValue.
     */
    public final double getNoDataValue() {
        return this.noDataValue;
    }
    
    /**
     * Returns the noDataValue of this as a BigDecimal.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * Handling for Infinite values.
     */
    public BigDecimal getNoDataValueBigDecimal( boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            if ( Double.isInfinite( noDataValue ) ) {
                // Cannot convert to a BigDecimal!
                return null;
            }
            return new BigDecimal( getNoDataValue() );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return getNoDataValueBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises noDataValue.
     * @param noDataValue The value this.noDataValue is initialised to.
     */
    protected final void initNoDataValue( double noDataValue ) {
        // This method is called before this.grid2DSquareCellChunks are
        // constructed so no java.lang.OutOfMemoryError handling is appropriate
        // here. The java.lang.OutOfMemoryError handling should be in the method
        // calling the Grid2DSquareCellDouble constructor.
        if ( Double.isNaN( noDataValue ) ) {
            this.noDataValue = Double.NEGATIVE_INFINITY;
            System.out.println( "noDataValue cannot be set to NaN! Initialised noDataValue as " + this.noDataValue );
            //throw new Exception( "noDataValue cannot be set to NaN" );
        } else {
            this.noDataValue = noDataValue;
        }
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex else returns noDataValue.
     * @param cellRowIndex .
     * @param cellColIndex .
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCell( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            if ( inGrid( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse ) ) {
                int chunkRowIndex = getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse );
                int chunkColIndex = getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse );
                Grid2DSquareCellChunkAbstract grid2DSquareCellChunk = getGrid2DSquareCellChunk( chunkRowIndex, chunkColIndex, this.handleOutOfMemoryErrorFalse );
                if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                    return ( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellChunk ).getCell( ( int ) ( cellRowIndex - ( ( long ) ( chunkRowIndex * getChunkNrows(this.handleOutOfMemoryErrorFalse) ) ) ), ( int ) ( cellColIndex - ( ( long ) ( chunkColIndex * getChunkNcols(this.handleOutOfMemoryErrorFalse) ) ) ), noDataValue );
                } else {
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                        return ( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellChunk ).getCell( ( int ) ( cellRowIndex - ( ( long ) ( chunkRowIndex * getChunkNrows(this.handleOutOfMemoryErrorFalse) ) ) ), ( int ) ( cellColIndex - ( ( long ) ( chunkColIndex * getChunkNcols(this.handleOutOfMemoryErrorFalse) ) ) ), noDataValue );
                    } else {
                        if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                            return ( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellChunk ).getCell( ( int ) ( cellRowIndex - ( ( long ) ( chunkRowIndex * getChunkNrows(this.handleOutOfMemoryErrorFalse) ) ) ), ( int ) ( cellColIndex - ( ( long ) ( chunkColIndex * getChunkNcols(this.handleOutOfMemoryErrorFalse) ) ) ), noDataValue );
                        } else {
                            return ( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellChunk ).getCell( ( int ) ( cellRowIndex - ( ( long ) ( chunkRowIndex * getChunkNrows(this.handleOutOfMemoryErrorFalse) ) ) ), ( int ) ( cellColIndex - ( ( long ) ( chunkColIndex * getChunkNcols(this.handleOutOfMemoryErrorFalse) ) ) ), noDataValue );
                        }
                    }
                }
            }
            return noDataValue;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getCell( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex.
     * @param chunkRowIndex .
     * @param chunkColIndex .
     * @param chunkCellRowIndex .
     * @param chunkCellColIndex .
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCell( Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellChunk, int chunkRowIndex, int chunkColIndex, int chunkCellRowIndex, int chunkCellColIndex, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                return ( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
            } else {
                if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                    return ( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                } else {
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                        return ( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                    } else {
                        if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class  ) {
                            return ( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellChunk ).getCell( chunkCellRowIndex, chunkCellColIndex, noDataValue );
                        } else {
                            //log("Error in " + getClass().getName() + ".setCell(Grid2DSquareCellDoubleChunkAbstract,int,int,int,int,double)" );
                            return noDataValue;
                        }
                    }
                }
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return getCell( grid2DSquareCellChunk, chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value of the cell containing point given by
     * x-coordinate x, y-coordinate y as a int.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCell( double x, double y, boolean handleOutOfMemoryError ) {
        try {
            return getCell( getCellRowIndex( y, this.handleOutOfMemoryErrorFalse ), getCellColIndex( x, this.handleOutOfMemoryErrorFalse ), handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( y, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( x, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getCell( x, y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value of the cell with cell ID cellID as a int.
     * @param cellID the CellID of the cell.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCell( CellID cellID, boolean handleOutOfMemoryError ) {
        // No java.lang.OutOfMemoryError handling needed here?
        return getCell( cellID.cellRowIndex, cellID.cellColIndex, handleOutOfMemoryError );
    }
    
    /**
     * For returning the value at x-coordinate x, y-coordinate y and setting it
     * to newValue.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param newValue .
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double setCell( double x, double y, double newValue, boolean handleOutOfMemoryError ) {
        try {
            return setCell( getCellRowIndex( x, this.handleOutOfMemoryErrorFalse ), getCellColIndex( y, this.handleOutOfMemoryErrorFalse ), newValue, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( y, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( x, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return setCell( x, y, newValue, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value of the cell with cell ID cellID and setting it
     * to newValue.
     * @param cellID the CellID of the cell.
     * @param newValue .
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double setCell( CellID cellID, double newValue, boolean handleOutOfMemoryError ) {
        // No java.lang.OutOfMemoryError handling needed here.
        return setCell( cellID.cellRowIndex, cellID.cellColIndex, newValue, handleOutOfMemoryError );
    }
    
    /**
     * For returning the value at cellRowIndex, cellColIndex and setting it to
     * newValue.
     * @param cellRowIndex .
     * @param cellColIndex .
     * @param newValue .
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double setCell( long cellRowIndex, long cellColIndex, double newValue, boolean handleOutOfMemoryError ) {
        try {
            int chunkRowIndex = getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse );
            int chunkColIndex = getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse );
            int chunkCellRowIndex = getChunkCellRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse );
            int chunkCellColIndex = getChunkCellColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse );
            //if ( inGrid( cellRowIndex, cellColIndex ) ) {
            return setCell( getGrid2DSquareCellDoubleChunk( chunkRowIndex, chunkColIndex, this.handleOutOfMemoryErrorFalse ), chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, newValue, handleOutOfMemoryError );
            //} else {
            //    return getNoDataValue();
            //}
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return setCell( cellRowIndex, cellColIndex, newValue, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value of the cell in chunk given by chunkRowIndex and
     * chunkColIndex and cell in the chunk given by chunkCellColIndex and
     * chunkCellRowIndex and setting it to newValue.
     * @param chunkRowIndex
     * @param chunkColIndex
     * @param chunkCellRowIndex
     * @param chunkCellColIndex
     * @param newValue
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double setCell( int chunkRowIndex, int chunkColIndex, int chunkCellRowIndex, int chunkCellColIndex, double newValue, boolean handleOutOfMemoryError ) {
        try {
            //if ( inGrid( chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex ) ) {
            return setCell( getGrid2DSquareCellDoubleChunk( chunkRowIndex, chunkColIndex, this.handleOutOfMemoryErrorFalse ), chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, newValue, handleOutOfMemoryError );
            //} else {
            //    return getNoDataValue();
            //}
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return setCell( chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, newValue, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex and sets it to newValue.
     * @param chunkRowIndex .
     * @param chunkColIndex .
     * @param newValue .
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double setCell( Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellChunk, int chunkRowIndex, int chunkColIndex, int chunkCellRowIndex, int chunkCellColIndex, double newValue, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            double result = noDataValue;
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                result = ( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellChunk ).setCell( chunkCellRowIndex, chunkCellColIndex, newValue, result );
            } else {
                if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                    result = ( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellChunk ).setCell( chunkCellRowIndex, chunkCellColIndex, newValue, result );
                } else {
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                        result = ( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellChunk ).setCell( chunkCellRowIndex, chunkCellColIndex, newValue, result );
                    } else {
                        if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class  ) {
                            result = ( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellChunk ).setCell( chunkCellRowIndex, chunkCellColIndex, newValue, result );
                        } else {
                            //log("Error in " + getClass().getName() + ".setCell(Grid2DSquareCellDoubleChunkAbstract,int,int,int,int,double)" );
                            return result;
                        }
                    }
                }
            }
            // Update Statistics
            upDateGridStatistics( newValue, result, noDataValue, this.handleOutOfMemoryErrorFalse );
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( chunkRowIndex, chunkColIndex , false) );
                initMemoryReserve();
                return setCell( grid2DSquareCellChunk, chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex, newValue, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initilises the value at cellRowIndex, cellColIndex
     * @param cellRowIndex
     * @param cellColIndex
     * @param valueToInitialise
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected void initCell( long cellRowIndex, long cellColIndex, double valueToInitialise, boolean handleOutOfMemoryError ) {
        try {
            if ( inGrid( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse ) ) {
                int chunkRowIndex = getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse );
                int chunkColIndex = getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse );
                int chunkNrows = getChunkNrows(this.handleOutOfMemoryErrorFalse);
                int chunkNcols = getChunkNcols(this.handleOutOfMemoryErrorFalse);
                //ChunkID chunkID = new ChunkID( chunkRowIndex, chunkColIndex ); // For debugging
                //boolean containsKey = this.grid2DSquareCellChunks.containsKey( chunkID ); // For debugging
                getGrid2DSquareCellDoubleChunk( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ), this.handleOutOfMemoryErrorFalse ).initCell( ( int ) ( cellRowIndex - ( ( long ) chunkRowIndex * ( long ) chunkNrows ) ), ( int ) ( cellColIndex - ( ( long ) chunkColIndex * ( long ) chunkNcols ) ), valueToInitialise );
                // Update Statistics
                double noDataValue = getNoDataValue();
                if ( valueToInitialise != noDataValue ) {
                    BigDecimal cellBigDecimal = new BigDecimal( valueToInitialise );
                    this.gridStatistics.nonNoDataValueCountBigInteger = this.gridStatistics.nonNoDataValueCountBigInteger.add( BigInteger.ONE );
                    this.gridStatistics.sumBigDecimal = this.gridStatistics.sumBigDecimal.add( cellBigDecimal );
                    if ( cellBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == -1 ) {
                        this.gridStatistics.minCountBigInteger = BigInteger.ONE;
                        this.gridStatistics.minBigDecimal = cellBigDecimal;
                    } else {
                        if ( cellBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == 0 ) {
                            this.gridStatistics.minCountBigInteger = this.gridStatistics.minCountBigInteger.add( BigInteger.ONE );
                        }
                    }
                    if ( cellBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == 1 ) {
                        this.gridStatistics.maxCountBigInteger = BigInteger.ONE;
                        this.gridStatistics.maxBigDecimal = cellBigDecimal;
                    } else {
                        if ( cellBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == 0 ) {
                            this.gridStatistics.maxCountBigInteger = this.gridStatistics.maxCountBigInteger.add( BigInteger.ONE );
                        }
                    }
                }
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                initCell( cellRowIndex, cellColIndex, valueToInitialise, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initilises the value at cellRowIndex, cellColIndex and does nothing about
     * this.gridStatistics
     * @param cellRowIndex
     * @param cellColIndex
     * @param valueToInitialise
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected void initCellFast( long cellRowIndex, long cellColIndex, double valueToInitialise, boolean handleOutOfMemoryError ) {
        try {
            if ( inGrid( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse ) ) {
                int chunkRowIndex = getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse );
                int chunkColIndex = getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse );
                int chunkNrows = getChunkNrows(this.handleOutOfMemoryErrorFalse);
                int chunkNcols = getChunkNcols(this.handleOutOfMemoryErrorFalse);
                //ChunkID chunkID = new ChunkID( chunkRowIndex, chunkColIndex ); // Debugging code
                //boolean containsKey = this.grid2DSquareCellChunks.containsKey( chunkID ); // Debugging code
                this.getGrid2DSquareCellDoubleChunk( new ChunkID( this.nChunkCols, chunkRowIndex, chunkColIndex ), this.handleOutOfMemoryErrorFalse ).initCell( ( int ) ( cellRowIndex - ( ( long ) chunkRowIndex * ( long ) chunkNrows ) ), ( int ) ( cellColIndex - ( ( long ) chunkColIndex * ( long ) chunkNcols ) ), valueToInitialise );
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                initCell( cellRowIndex, cellColIndex, valueToInitialise, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a double[] of all cell values for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     *   returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     *   returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double[] getCells( double x, double y, double distance, boolean handleOutOfMemoryError ) {
        try {
            return getCells( x, y, getCellRowIndex( y, this.handleOutOfMemoryErrorFalse ), getCellColIndex( x, this.handleOutOfMemoryErrorFalse ), distance, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( y, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( x, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getCells( x, y, distance, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a int[] of all cell values for cells thats centroids are
     * intersected by circle with centre at centroid of cell given by cell row
     * index cellRowIndex, cell column index cellColIndex, and radius distance.
     * @param cellRowIndex the row index for the cell thats centroid is the
     *   circle centre from which cell values are returned.
     * @param cellColIndex the column index for the cell thats centroid is the
     *   circle centre from which cell values are returned.
     * @param distance the radius of the circle for which intersected cell
     *   values are returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double[] getCells( long cellRowIndex, long cellColIndex, double distance, boolean handleOutOfMemoryError ) {
        try {
            return getCells( getCellX( cellColIndex, this.handleOutOfMemoryErrorFalse ), getCellY( cellRowIndex, this.handleOutOfMemoryErrorFalse ), cellRowIndex, cellColIndex, distance, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getCells( cellRowIndex, cellColIndex, distance, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a int[] of all cell values for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x The x-coordinate of the circle centre from which cell values are
     *   returned.
     * @param y The y-coordinate of the circle centre from which cell values are
     *   returned.
     * @param cellRowIndex The row index at y.
     * @param cellColIndex The col index at x.
     * @param distance The radius of the circle for which intersected cell
     *   values are returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected double[] getCells( double x, double y, long cellRowIndex, long cellColIndex, double distance, boolean handleOutOfMemoryError ) {
        try {
            double[] cells;
            BigDecimal[] dimensions = getDimensions();
            int cellDistance = ( int ) Math.ceil( distance / dimensions[ 0 ].doubleValue() );
            cells = new double[ cellDistance * cellDistance ];
            long row;
            long col;
            double thisX;
            double thisY;
            int count = 0;
            for ( row = cellRowIndex - cellDistance; row <= cellRowIndex + cellDistance; row ++ ) {
                thisY = getCellY( cellRowIndex, this.handleOutOfMemoryErrorFalse );
                for ( col = cellColIndex - cellDistance; col <= cellColIndex + cellDistance; col ++ ) {
                    thisX = getCellX( cellColIndex, this.handleOutOfMemoryErrorFalse );
                    if ( Utilities.distance( x, y, thisX, thisY ) <= distance ) {
                        cells[ count ] = getCell( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
                        count ++;
                    }
                }
            }
            // Trim cells
            System.arraycopy( cells, 0, cells, 0, count );
            return cells;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getCells( x, y, cellRowIndex, cellColIndex, distance, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the average of the nearest data values to point given by
     * x-coordinate x, y-coordinate y as a double.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDouble( double x, double y, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            double result = getCell( x, y, this.handleOutOfMemoryErrorFalse );
            if ( result == noDataValue ) {
                result = getNearestValueDouble( x, y, getCellRowIndex( y, this.handleOutOfMemoryErrorFalse ), getCellColIndex( x, this.handleOutOfMemoryErrorFalse ), noDataValue, handleOutOfMemoryError );
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( y, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( x, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValueDouble( x, y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the average of the nearest data values to position given by row
     * index rowIndex, column index colIndex
     * @param cellRowIndex the row index from which average of the nearest data
     *   values is returned
     * @param cellColIndex the column index from which average of the nearest data
     *   values is returned
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDouble( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            double result = getCell( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            if ( result == noDataValue ) {
                result = getNearestValueDouble( getCellX( cellColIndex, this.handleOutOfMemoryErrorFalse ), getCellY( cellRowIndex, this.handleOutOfMemoryErrorFalse ), cellRowIndex, cellColIndex, noDataValue, handleOutOfMemoryError );
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValueDouble( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the average of the nearest data values to point given by
     * x-coordinate x, y-coordinate y in position given by row index rowIndex,
     * column index colIndex
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param cellRowIndex the row index from which average of the nearest data
     *   values is returned
     * @param cellColIndex the column index from which average of the nearest data
     *   values is returned
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected double getNearestValueDouble( double x, double y, long cellRowIndex, long cellColIndex, double noDataValue, boolean handleOutOfMemoryError ) {
        try {
            CellID nearestCellID = getNearestCellID( x, y, cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            double nearestValue = getCell( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            if ( nearestValue == noDataValue ) {
                long ncols = getNcols();
                // Find a value Seeking outwards from nearestCellID
                // Initialise visitedSet1
                HashSet visitedSet = new HashSet();
                HashSet visitedSet1 = new HashSet();
                visitedSet.add( nearestCellID );
                visitedSet1.add( nearestCellID );
                // Initialise toVisitSet1
                HashSet toVisitSet1 = new HashSet();
                long row;
                long col;
                for ( row = -1; row < 2; row ++ ) {
                    for ( col = -1; col < 2; col ++ ) {
                        if ( ! ( row == 0 && col == 0 ) ) {
                            if ( inGrid( cellRowIndex + row, cellColIndex + col, this.handleOutOfMemoryErrorFalse ) ) {
                                toVisitSet1.add( new CellID( ncols, cellRowIndex + row, cellColIndex + col ) );
                                //toVisitSet1.add( new CellID( cellRowIndex + row, cellColIndex + col ) );
                            }
                        }
                    }
                }
                // Seek
                boolean foundValue = false;
                double value;
                HashSet values = new HashSet();
                HashSet visitedSet2;
                HashSet toVisitSet2;
                Iterator iterator;
                CellID cellID;
                while ( ! foundValue ) {
                    visitedSet2 = new HashSet();
                    toVisitSet2 = new HashSet();
                    iterator = toVisitSet1.iterator();
                    while ( iterator.hasNext() ) {
                        cellID = ( CellID ) iterator.next();
                        visitedSet2.add( cellID );
                        value = getCell( cellID, this.handleOutOfMemoryErrorFalse );
                        if ( value != noDataValue ) {
                            foundValue = true;
                            values.add( cellID );
                        } else {
                            // Add neighbours to toVisitSet2
                            for ( row = -1; row < 2; row ++ ) {
                                for ( col = -1; col < 2; col ++ ) {
                                    if ( ! ( row == 0 && col == 0 ) ) {
                                        if ( inGrid( cellID.cellRowIndex + row, cellID.cellColIndex + col, this.handleOutOfMemoryErrorFalse ) ) {
                                            //toVisitSet2.add( getCellID( cellID.cellRowIndex + row, cellID.cellColIndex + col ) );
                                            toVisitSet2.add( new CellID( ncols, cellID.cellRowIndex + row, cellID.cellColIndex + col ) );
                                            //toVisitSet2.add( new CellID( cellID.cellRowIndex + row, cellID.cellColIndex + col ) );
                                        }
                                    }
                                }
                            }
                        }
                    }
                    toVisitSet2.removeAll( visitedSet1 );
                    toVisitSet2.removeAll( visitedSet2 );
                    visitedSet.addAll( visitedSet2 );
                    visitedSet1 = visitedSet2;
                    toVisitSet1 = toVisitSet2;
                }
                double distance;
                double minDistance = Integer.MAX_VALUE;
                // Go through values and find the closest
                HashSet closest = new HashSet();
                iterator = values.iterator();
                while ( iterator.hasNext() ) {
                    cellID = ( CellID ) iterator.next();
                    distance = Utilities.distance( x, y, getCellX( cellID, this.handleOutOfMemoryErrorFalse ), getCellY( cellID, this.handleOutOfMemoryErrorFalse ) );
                    if ( distance < minDistance ) {
                        closest.clear();
                        closest.add( cellID );
                    } else {
                        if ( distance == minDistance ) {
                            closest.add( cellID );
                        }
                    }
                    minDistance = Math.min( minDistance, distance );
                }
                // Get cellIDs that are within distance of discovered value
                CellID[] cellIDs = getCellIDs( x, y, minDistance, this.handleOutOfMemoryErrorFalse );
                // Go through values and find the closest
                for ( int i = 0; i < cellIDs.length; i ++ ) {
                    if ( ! visitedSet.contains( cellIDs[ i ] ) ) {
                        if ( getCell( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ) != noDataValue ) {
                            distance = Utilities.distance( x, y, getCellX( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ), getCellY( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ) );
                            if ( distance < minDistance ) {
                                closest.clear();
                                closest.add( cellIDs[ i ] );
                            } else {
                                if ( distance == minDistance ) {
                                    closest.add( cellIDs[ i ] );
                                }
                            }
                            minDistance = Math.min( minDistance, distance );
                        }
                    }
                }
                // Go through the closest and calculate the average.
                value = 0;
                iterator = closest.iterator();
                while ( iterator.hasNext() ) {
                    cellID = ( CellID ) iterator.next();
                    value += getCell( cellID, this.handleOutOfMemoryErrorFalse );
                }
                nearestValue = value / ( double ) closest.size();
            }
            return nearestValue;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValueDouble( x, y, cellRowIndex, cellColIndex, noDataValue, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * to point given by x-coordinate x, y-coordinate y.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID[] getNearestValuesCellIDs( double x, double y, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            double value = getCell( x, y, this.handleOutOfMemoryErrorFalse );
            if ( value == noDataValue ) {
                return getNearestValuesCellIDs( x, y, getCellRowIndex( y, this.handleOutOfMemoryErrorFalse ), getCellColIndex( x, this.handleOutOfMemoryErrorFalse ), noDataValue, handleOutOfMemoryError );
            }
            CellID[] cellIDs = new CellID[ 1 ];
            cellIDs[ 0 ] = getCellID( x, y, this.handleOutOfMemoryErrorFalse );
            return cellIDs;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( y, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( x, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValuesCellIDs( x, y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * to position given by row index rowIndex, column index colIndex.
     * @param cellRowIndex The row index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param cellColIndex The column index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID[] getNearestValuesCellIDs( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            double value = getCell( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            if ( value == noDataValue ) {
                return getNearestValuesCellIDs( getCellX( cellColIndex, this.handleOutOfMemoryErrorFalse ), getCellY( cellRowIndex, this.handleOutOfMemoryErrorFalse ), cellRowIndex, cellColIndex, noDataValue, handleOutOfMemoryError );
            }
            CellID[] cellIDs = new CellID[ 1 ];
            cellIDs[ 0 ] = getCellID( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            return cellIDs;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValuesCellIDs( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * nearest to point with position given by: x-coordinate x, y-coordinate y;
     * and, cell row index cellRowIndex, cell column index cellColIndex.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param cellRowIndex The row index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param cellColIndex The column index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param noDataValue the no data value of the grid
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected CellID[] getNearestValuesCellIDs( double x, double y, long cellRowIndex, long cellColIndex, double noDataValue, boolean handleOutOfMemoryError ) {
        try {
            CellID[] nearestCellIDs = new CellID[1];
            nearestCellIDs[0] = getNearestCellID( x, y, cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            double nearestCellValue = getCell( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            if ( nearestCellValue == noDataValue ) {
                long ncols = getNcols();
                // Find a value Seeking outwards from nearestCellID
                // Initialise visitedSet1
                HashSet visitedSet = new HashSet();
                HashSet visitedSet1 = new HashSet();
                visitedSet.add( nearestCellIDs[0] );
                visitedSet1.add( nearestCellIDs[0] );
                // Initialise toVisitSet1
                HashSet toVisitSet1 = new HashSet();
                long row;
                long col;
                for ( row = -1; row < 2; row ++ ) {
                    for ( col = -1; col < 2; col ++ ) {
                        if ( ! ( row == 0 && col == 0 ) ) {
                            if ( inGrid( cellRowIndex + row, cellColIndex + col, this.handleOutOfMemoryErrorFalse ) ) {
                                toVisitSet1.add( getCellID( cellRowIndex + row, cellColIndex + col, this.handleOutOfMemoryErrorFalse ) );
                            }
                        }
                    }
                }
                // Seek
                boolean foundValue = false;
                double value;
                CellID cellID;
                HashSet values = new HashSet();
                HashSet visitedSet2;
                HashSet toVisitSet2;
                Iterator iterator;
                while ( ! foundValue ) {
                    visitedSet2 = new HashSet();
                    toVisitSet2 = new HashSet();
                    iterator = toVisitSet1.iterator();
                    while ( iterator.hasNext() ) {
                        cellID = ( CellID ) iterator.next();
                        visitedSet2.add( cellID );
                        value = getCell( cellID, this.handleOutOfMemoryErrorFalse );
                        if ( value != noDataValue ) {
                            foundValue = true;
                            values.add( cellID );
                        } else {
                            // Add neighbours to toVisitSet2
                            for ( row = -1; row < 2; row ++ ) {
                                for ( col = -1; col < 2; col ++ ) {
                                    if ( ! ( row == 0 && col == 0 ) ) {
                                        if ( inGrid( cellID.cellRowIndex + row, cellID.cellColIndex + col, this.handleOutOfMemoryErrorFalse ) ) {
                                            toVisitSet2.add( getCellID( cellID.cellRowIndex + row, cellID.cellColIndex + col, this.handleOutOfMemoryErrorFalse ) );
                                            //toVisitSet2.add( new CellID( cellID.cellRowIndex + row, cellID.cellColIndex + col ) );
                                        }
                                    }
                                }
                            }
                        }
                    }
                    toVisitSet2.removeAll( visitedSet1 );
                    toVisitSet2.removeAll( visitedSet2 );
                    visitedSet.addAll( visitedSet2 );
                    visitedSet1 = visitedSet2;
                    toVisitSet1 = toVisitSet2;
                }
                double distance;
                double minDistance = Double.MAX_VALUE;
                // Go through values and find the closest
                HashSet closest = new HashSet();
                iterator = values.iterator();
                while ( iterator.hasNext() ) {
                    cellID = ( CellID ) iterator.next();
                    distance = Utilities.distance( x, y, getCellX( cellID, this.handleOutOfMemoryErrorFalse ), getCellY( cellID, this.handleOutOfMemoryErrorFalse ) );
                    if ( distance < minDistance ) {
                        closest.clear();
                        closest.add( cellID );
                    } else {
                        if ( distance == minDistance ) {
                            closest.add( cellID );
                        }
                    }
                    minDistance = Math.min( minDistance, distance );
                }
                // Get cellIDs that are within distance of discovered value
                CellID[] cellIDs = getCellIDs( x, y, minDistance, this.handleOutOfMemoryErrorFalse );
                // Go through values and find the closest
                for ( int i = 0; i < cellIDs.length; i ++ ) {
                    if ( ! visitedSet.contains( cellIDs[ i ] ) ) {
                        if ( getCell( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ) != noDataValue ) {
                            distance = Utilities.distance( x, y, getCellX( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ), getCellY( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ) );
                            if ( distance < minDistance ) {
                                closest.clear();
                                closest.add( cellIDs[ i ] );
                            } else {
                                if ( distance == minDistance ) {
                                    closest.add( cellIDs[ i ] );
                                }
                            }
                            minDistance = Math.min( minDistance, distance );
                        }
                    }
                }
                // Go through the closest and put into an array
                nearestCellIDs = new CellID[ closest.size() ];
                iterator = closest.iterator();
                int counter = 0;
                while ( iterator.hasNext() ) {
                    nearestCellIDs[ counter ] = ( CellID ) iterator.next();
                    counter ++;
                }
            }
            return nearestCellIDs;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValuesCellIDs( x, y, cellRowIndex, cellColIndex, noDataValue, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the distance to the nearest data value from point given by
     * x-coordinate x, y-coordinate y as a double.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDoubleDistance( double x, double y, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            double result = getCell( x, y, this.handleOutOfMemoryErrorFalse );
            if ( result == noDataValue ) {
                result = getNearestValueDoubleDistance( x, y, getCellRowIndex( y, this.handleOutOfMemoryErrorFalse ), getCellColIndex( x, this.handleOutOfMemoryErrorFalse ), noDataValue, this.handleOutOfMemoryErrorFalse );
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( y, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( x, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValueDoubleDistance( x, y, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the distance to the nearest data value from position given by
     * row index rowIndex, column index colIndex as a double.
     * @param cellRowIndex The cell row index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param cellColIndex The cell column index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDoubleDistance( long cellRowIndex, long cellColIndex, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            double result = getCell( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            if ( result == noDataValue ) {
                result = getNearestValueDoubleDistance( getCellX( cellColIndex, this.handleOutOfMemoryErrorFalse ), getCellY( cellRowIndex, this.handleOutOfMemoryErrorFalse ), cellRowIndex, cellColIndex, noDataValue, this.handleOutOfMemoryErrorFalse );
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValueDoubleDistance( cellRowIndex, cellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the distance to the nearest data value from: point given by
     * x-coordinate x, y-coordinate y in position given by row index rowIndex,
     * column index colIndex as a double.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param cellRowIndex The cell row index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param cellColIndex The cell column index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected double getNearestValueDoubleDistance( double x, double y, long cellRowIndex, long cellColIndex, double noDataValue, boolean handleOutOfMemoryError ) {
        try {
            CellID nearestCellID = getNearestCellID( x, y, cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            double result = getCell( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
            if ( result == noDataValue ) {
                // Find a value Seeking outwards from nearestCellID
                // Initialise visitedSet1
                HashSet visitedSet = new HashSet();
                HashSet visitedSet1 = new HashSet();
                visitedSet.add( nearestCellID );
                visitedSet1.add( nearestCellID );
                // Initialise toVisitSet1
                HashSet toVisitSet1 = new HashSet();
                long row;
                long col;
                for ( row = -1; row < 2; row ++ ) {
                    for ( col = -1; col < 2; col ++ ) {
                        if ( ! ( row == 0 && col == 0 ) ) {
                            if ( inGrid( cellRowIndex + row, cellColIndex + col, this.handleOutOfMemoryErrorFalse ) ) {
                                toVisitSet1.add( getCellID( cellRowIndex + row, cellColIndex + col, this.handleOutOfMemoryErrorFalse ) );
                            }
                        }
                    }
                }
                // Seek
                boolean foundValue = false;
                double value;
                CellID cellID;
                HashSet values = new HashSet();
                HashSet visitedSet2;
                HashSet toVisitSet2;
                Iterator iterator;
                while ( ! foundValue ) {
                    visitedSet2 = new HashSet();
                    toVisitSet2 = new HashSet();
                    iterator = toVisitSet1.iterator();
                    while ( iterator.hasNext() ) {
                        cellID = ( CellID ) iterator.next();
                        visitedSet2.add( cellID );
                        value = getCell( cellID, this.handleOutOfMemoryErrorFalse );
                        if ( value != noDataValue ) {
                            foundValue = true;
                            values.add( cellID );
                        } else {
                            // Add neighbours to toVisitSet2
                            for ( row = -1; row < 2; row ++ ) {
                                for ( col = -1; col < 2; col ++ ) {
                                    if ( ! ( row == 0 && col == 0 ) ) {
                                        if ( inGrid( cellID.cellRowIndex + row, cellID.cellColIndex + col, this.handleOutOfMemoryErrorFalse ) ) {
                                            toVisitSet2.add( getCellID( cellID.cellRowIndex + row, cellID.cellColIndex + col, this.handleOutOfMemoryErrorFalse ) );
                                            //toVisitSet2.add( new CellID( cellID.cellRowIndex + row, cellID.cellColIndex + col ) );
                                        }
                                    }
                                }
                            }
                        }
                    }
                    toVisitSet2.removeAll( visitedSet1 );
                    toVisitSet2.removeAll( visitedSet2 );
                    visitedSet.addAll( visitedSet2 );
                    visitedSet1 = visitedSet2;
                    toVisitSet1 = toVisitSet2;
                }
                double distance;
                double minDistance = Double.MAX_VALUE;
                // Go through values and find the closest
                HashSet closest = new HashSet();
                iterator = values.iterator();
                while ( iterator.hasNext() ) {
                    cellID = ( CellID ) iterator.next();
                    distance = Utilities.distance( x, y, getCellX( cellID, this.handleOutOfMemoryErrorFalse ), getCellY( cellID, this.handleOutOfMemoryErrorFalse ) );
                    if ( distance < minDistance ) {
                        closest.clear();
                        closest.add( cellID );
                    } else {
                        if ( distance == minDistance ) {
                            closest.add( cellID );
                        }
                    }
                    minDistance = Math.min( minDistance, distance );
                }
                // Get cellIDs that are within distance of discovered value
                CellID[] cellIDs = getCellIDs( x, y, minDistance, this.handleOutOfMemoryErrorFalse );
                // Go through values and find the closest
                for ( int i = 0; i < cellIDs.length; i ++ ) {
                    if ( ! visitedSet.contains( cellIDs[ i ] ) ) {
                        if ( getCell( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ) != noDataValue ) {
                            distance = Utilities.distance( x, y, getCellX( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ), getCellY( cellIDs[ i ], this.handleOutOfMemoryErrorFalse ) );
                            minDistance = Math.min( minDistance, distance );
                        }
                    }
                }
                result = minDistance;
            } else {
                result = 0.0d;
            }
            return result;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return getNearestValueDoubleDistance( x, y, cellRowIndex, cellColIndex, noDataValue, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns current value of cell containing the point given by x-coordinate
     * x, y-coordinate y, and adds valueToAdd to that cell.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param valueToAdd the value to be added to the cell containing the point
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double addToCell( double x, double y, double valueToAdd, boolean handleOutOfMemoryError ) {
        try {
            return addToCell( getCellRowIndex( y, this.handleOutOfMemoryErrorFalse ), getCellColIndex( x, this.handleOutOfMemoryErrorFalse ), valueToAdd, handleOutOfMemoryError );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( y, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( x, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return addToCell( x, y, valueToAdd, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value of the cell with cell ID cellID and adds valueToAdd to
     * that cell.
     * @param cellID the CellID of the cell.
     * @param valueToAdd the value to be added to the cell containing the point
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double addToCell( CellID cellID, double valueToAdd, boolean handleOutOfMemoryError ) {
        // No need for java.lang.OutOfMemoryError Handling?
        return addToCell( cellID.cellRowIndex, cellID.cellColIndex, valueToAdd, handleOutOfMemoryError );
    }
    
    /**
     * Returns current value of cell with row index rowIndex and column index
     * colIndex and adds valueToAdd to that cell.
     * @param cellRowIndex the row index of the cell.
     * @param cellColIndex the col index of the cell.
     * @param valueToAdd the value to be added to the cell.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * NB1. If cell is not contained in this then then returns noDataValue.
     * NB2. Adding to noDataValue is done as if adding to a cell with value of
     *   0.
     * TODO: Check Arithmetic
     */
    public double addToCell( long cellRowIndex, long cellColIndex, double valueToAdd, boolean handleOutOfMemoryError ) {
        try {
            double noDataValue = getNoDataValue();
            if ( inGrid( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse ) ) {
                double currentValue = getCell( cellRowIndex, cellColIndex, this.handleOutOfMemoryErrorFalse );
                if ( currentValue != noDataValue ) {
                    if ( valueToAdd != noDataValue ) {
                        return setCell( cellRowIndex, cellColIndex, currentValue + valueToAdd, this.handleOutOfMemoryErrorFalse );
                    }
                } else {
                    if ( valueToAdd != noDataValue ) {
                        return setCell( cellRowIndex, cellColIndex, valueToAdd, this.handleOutOfMemoryErrorFalse );
                    }
                }
            }
            return noDataValue;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunkExcept( getChunkRowIndex( cellRowIndex, this.handleOutOfMemoryErrorFalse ), getChunkColIndex( cellColIndex, this.handleOutOfMemoryErrorFalse ) , false) );
                initMemoryReserve();
                return addToCell( cellRowIndex, cellColIndex, valueToAdd, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns an iterator over the cell value in this.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Iterator iterator( boolean handleOutOfMemoryError ) {
        try {
            return new Grid2DSquareCellDoubleIterator( this );
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearChunk( cacheChunk(false) );
                initMemoryReserve();
                return iterator( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
}
