/**
 * Version 1.0 is to handle individual single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
// java dependencies
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.math.BigDecimal;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciiGridImporter;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;

/**
 * TODO:
 * Abstraction and extension as for Grid2DSquareCellIntFactory.
 * Add more control for prefered type of Grid2DSquareCellDoubleChunkFactory 
 */
public class Grid2DSquareCellDoubleFactory {
    
    /**
     * A directory for caching.
     */
    protected File directory;
    
    //    /**
    //     * A store for the number of rows in a chunk.
    //     */
    //    protected int chunkNrows;
    //
    //    /**
    //     * A store for the number of columns in a chunk.
    //     */
    //    protected int chunkNcols;
    
    /**
     * Defaults directory to a new one in
     * System.getProperties( "java.io.tmpdir" );
     */
    public Grid2DSquareCellDoubleFactory() {
        this( FileCreator.createTempFile() );
        //this ( FileCreator.createNewFile() );
    }
    
    /**
     * @param directory A "workspace directory" for storing temporary files and
     * caching Grid2DSquareCellDoubleAbstract data.
     */
    public Grid2DSquareCellDoubleFactory(File directory) {
        this.directory = directory;
        //this( directory, 64, 64 );
    }
    
    /////////////////////////
    // Create from scratch //
    /////////////////////////
    
    /**
     * Returns an Grid2DSquareCellDouble with all values as
     * noDataValues.
     * @param nrows the Grid2DSquareCellDouble nrows.
     * @param ncols the Grid2DSquareCellDouble ncols.
     * Defaults:
     * chunkNrows to 64,
     * chunkNcols to 64.
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( long nrows, long ncols ) {
        return createGrid2DSquareCellDouble( 64, 64, nrows, ncols );
    }
    
    /**
     * Returns an Grid2DSquareCellDouble with all values as
     * noDataValues.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param nrows the Grid2DSquareCellDouble nrows.
     * @param ncols the Grid2DSquareCellDouble ncols.
     * Defaults:
     * dimensions = [ cellsize = 1.0d, xmin = 0.0d, ymin = 0.0d,
     * xmax = ncols * cellsize, ymax = nrows * cellsize ];
     * noDataValue = Double.MIN_VALUE.
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( int chunkNrows, int chunkNcols, long nrows, long ncols ) {
        BigDecimal[] dimensions = new BigDecimal[ 5 ];
        dimensions[ 0 ] = new BigDecimal( "1" );
        dimensions[ 1 ] = new BigDecimal( "0" );
        dimensions[ 2 ] = new BigDecimal( "0" );
        dimensions[ 3 ] = new BigDecimal( ncols );
        dimensions[ 4 ] = new BigDecimal( nrows );
        return createGrid2DSquareCellDouble( chunkNrows, chunkNcols, nrows, ncols, dimensions, Double.MIN_VALUE );
    }
    
    /**
     * Returns an Grid2DSquareCellDouble with all values as
     * noDataValues.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param nrows the Grid2DSquareCellDouble nrows.
     * @param ncols the Grid2DSquareCellDouble ncols.
     * @param noDataValue The grid = Doubleeger.MIN_VALUE;
     * Defaults:
     * dimensions = [ cellsize = 1.0d, xmin = 0.0d, ymin = 0.0d,
     * xmax = ncols * cellsize, ymax = nrows * cellsize ];
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( int chunkNrows, int chunkNcols, long nrows, long ncols, double noDataValue ) {
        BigDecimal[] dimensions = new BigDecimal[ 5 ];
        dimensions[ 0 ] = new BigDecimal( "1" );
        dimensions[ 1 ] = new BigDecimal( "0" );
        dimensions[ 2 ] = new BigDecimal( "0" );
        dimensions[ 3 ] = new BigDecimal( ncols );
        dimensions[ 4 ] = new BigDecimal( nrows );
        return createGrid2DSquareCellDouble( chunkNrows, chunkNcols, nrows, ncols, dimensions, noDataValue );
    }
    
    /**
     * Returns an Grid2DSquareCellDouble with all cell vallue equal to noDataValue.
     * @param nrows the Grid2DSquareCellDouble nrows.
     * @param ncols the Grid2DSquareCellDouble ncols.
     * @param dimensions
     * Defaults:
     * chunkNrows and chunkNcols to 64
     * noDataValue = Double.MIN_VALUE
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( long nrows, long ncols, BigDecimal[] dimensions ) {
        return createGrid2DSquareCellDouble( 64, 64, nrows, ncols, dimensions, Double.MIN_VALUE );
    }
    
    /**
     * Returns an Grid2DSquareCellDouble with all cell vallue equal to noDataValue.
     * @param nrows the Grid2DSquareCellDouble nrows.
     * @param ncols the Grid2DSquareCellDouble ncols.
     * @param dimensions
     * @param noDataValue
     * Defaults:
     * chunkNrows and chunkNcols to 64
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( long nrows, long ncols, BigDecimal[] dimensions, double noDataValue ) {
        return createGrid2DSquareCellDouble( 64, 64, nrows, ncols, dimensions, noDataValue );
    }
    
    /**
     * Returns an Grid2DSquareCellDouble with all values as
     * noDataValues.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param nrows the Grid2DSquareCellDouble nrows.
     * @param ncols the Grid2DSquareCellDouble ncols.
     * Defaults:
     * directory to FileCreator.createNewFile( this.directory )
     * noDataValue = Double.MIN_VALUE
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions ) {
        //        this.chunkNrows = chunkNrows;
        //        this.chunkNcols = chunkNcols;
        return createGrid2DSquareCellDouble( FileCreator.createNewFile( this.directory ), chunkNrows, chunkNcols, nrows, ncols, dimensions, Double.MIN_VALUE );
    }
    
    /**
     * Returns an Grid2DSquareCellDouble with all values as
     * noDataValues.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param nrows the Grid2DSquareCellDouble nrows.
     * @param ncols the Grid2DSquareCellDouble ncols.
     * @param noDataValue
     * Defaults:
     * directory to FileCreator.createNewFile( this.directory )
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions, double noDataValue ) {
        //        this.chunkNrows = chunkNrows;
        //        this.chunkNcols = chunkNcols;
        return createGrid2DSquareCellDouble( FileCreator.createNewFile( this.directory ), chunkNrows, chunkNcols, nrows, ncols, dimensions, noDataValue );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble grid with all values as
     * noDataValues.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param nrows The Grid2DSquareCellDouble nrows.
     * @param ncols The Grid2DSquareCellDouble ncols.
     * @param dimensions
     * @param noDataValue the no data value to be used.
     * Default:
     * Grid2DSquareCellDoubleChunkAbstractFactory to
     *   Grid2DSquareCellDoubleChunkArrayFactory
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File directory, int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions, double noDataValue ) {
        return createGrid2DSquareCellDouble( directory, new Grid2DSquareCellDoubleChunkArrayFactory(), chunkNrows, chunkNcols, nrows, ncols, dimensions, noDataValue );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble grid with all values as
     * noDataValues.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
      * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that 
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param nrows The Grid2DSquareCellDouble nrows.
     * @param ncols The Grid2DSquareCellDouble ncols.
     * @param dimensions
     * @param noDataValue the no data value to be used.
     * Default:
     * GridStatisticsAbstract to GridStatistics1
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File directory, Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory, int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions, double noDataValue ) {
        return createGrid2DSquareCellDouble( new GridStatistics1(), directory, grid2DSquareCellDoubleChunkFactory, chunkNrows, chunkNcols, nrows, ncols, dimensions, noDataValue );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble grid with all values as
     * noDataValues.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that 
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param nrows The Grid2DSquareCellDouble nrows.
     * @param ncols The Grid2DSquareCellDouble ncols.
     * @param dimensions
     * @param noDataValue the no data value to be used.
     * TODO: test
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( GridStatisticsAbstract gridStatistics, File directory, Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory, int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions, double noDataValue ) {
        return new Grid2DSquareCellDouble( gridStatistics, directory, grid2DSquareCellDoubleChunkFactory, chunkNrows, chunkNcols, nrows, ncols, dimensions, noDataValue );
    }
    
    //////////////////////////////////////////////////////
    // Create from an existing Grid2DSquareCellAbstract //
    //////////////////////////////////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellDouble with all values taken from
     * grid2DSquareCell.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which grid
     *   values are used.
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( Grid2DSquareCellAbstract grid2DSquareCell ) {
        //return createGrid2DSquareCellDouble( FileCreator.createNewFile( this.directory ), grid2DSquareCell, 0L, 0L, grid2DSquareCell.getNrows() - 1L, grid2DSquareCell.getNcols() - 1L );
        return createGrid2DSquareCellDouble( this.directory, grid2DSquareCell, 0L, 0L, grid2DSquareCell.getNrows() - 1L, grid2DSquareCell.getNcols() - 1L );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with all values taken from
     * grid2DSquareCell.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which grid
     *   values are used.
     * @param startRowIndex The topmost row index of grid2DSquareCell.
     * @param startColIndex The leftmost column index of grid2DSquareCell.
     * @param endRowIndex The bottom row index of grid2DSquareCell.
     * @param endColIndex The rightmost column index of grid2DSquareCell.
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( Grid2DSquareCellAbstract grid2DSquareCell, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
        return createGrid2DSquareCellDouble( FileCreator.createNewFile( this.directory, "" + startRowIndex + "_" + startColIndex + "_" + endRowIndex + "_" + endColIndex ), grid2DSquareCell, startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with all values taken from
     * grid2DSquareCell.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which grid
     *   values are used.
     * @param startRowIndex The topmost row index of grid2DSquareCell.
     * @param startColIndex The leftmost column index of grid2DSquareCell.
     * @param endRowIndex The bottom row index of grid2DSquareCell.
     * @param endColIndex The rightmost column index of grid2DSquareCell.
     * Default:
     * Grid2DSquareCellDoubleChunkAbstractFactory to Grid2DSquareCellDoubleChunkArrayFactory
     * chunkNrows, chunkNcols to 64
     * noDataValue to grid2DSquareCell.getNoDataValueBigDecimal() ).doubleValue()
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File directory, Grid2DSquareCellAbstract grid2DSquareCell, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
        //return createGrid2DSquareCellDouble( directory, grid2DSquareCell, new Grid2DSquareCellDoubleChunkMapFactory(), 64, 64, startRowIndex, startColIndex, endRowIndex, endColIndex, grid2DSquareCell.getNoDataValueDouble() );
        return createGrid2DSquareCellDouble( directory, grid2DSquareCell, new Grid2DSquareCellDoubleChunkArrayFactory(), 64, 64, startRowIndex, startColIndex, endRowIndex, endColIndex, grid2DSquareCell.getNoDataValueDouble( grid2DSquareCell.handleOutOfMemoryErrorFalse ) );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with all values taken from
     * grid2DSquareCell.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that 
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which grid
     *   values are used.
     * @param startRowIndex The topmost row index of grid2DSquareCell.
     * @param startColIndex The leftmost column index of grid2DSquareCell.
     * @param endRowIndex The bottom row index of grid2DSquareCell.
     * @param endColIndex The rightmost column index of grid2DSquareCell.
     * @param noDataValue the no data value to be used.
     * Default:
     * GridStatisticsAbstract to GridStatistics1
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File directory, Grid2DSquareCellAbstract grid2DSquareCell, Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory, int chunkNrows, int chunkNcols, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, double noDataValue ) {
        return createGrid2DSquareCellDouble( new GridStatistics1(), directory, grid2DSquareCell, grid2DSquareCellDoubleChunkFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex, noDataValue );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with all values taken from
     * grid2DSquareCell.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that 
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which grid
     *   values are used.
     * @param startRowIndex The topmost row index of grid2DSquareCell.
     * @param startColIndex The leftmost column index of grid2DSquareCell.
     * @param endRowIndex The bottom row index of grid2DSquareCell.
     * @param endColIndex The rightmost column index of grid2DSquareCell.
     * @param noDataValue the no data value to be used.
     * TODO: test
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( GridStatisticsAbstract gridStatistics, File directory, Grid2DSquareCellAbstract grid2DSquareCell, Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory, int chunkNrows, int chunkNcols, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, double noDataValue ) {
        return new Grid2DSquareCellDouble( gridStatistics, directory, grid2DSquareCell, grid2DSquareCellDoubleChunkFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex, noDataValue );
    }
    
    
    ////////////////////////
    // Create from a File //
    ////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from
     * gridFile. If gridFile is a directory it is assumed it will contain a
     * file called Cache which can be opened into an object input stream and
     * initailised as an instance of a class extending Grid2DSquareCellAbstract.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File gridFile ) {
        if ( gridFile.isDirectory() ) {
            // Initialise from Cache
            File cache = new File( gridFile, "Cache" );
            /*File cache = new File( gridFile, "Grid2DSquareCellDouble" );
            if ( ! cache.exists() ) {
                cache = new File( gridFile, "Grid2DSquareCellDouble" );
                if ( ! cache.exists() ) {
                    // This error occurs if gridFile is a directory and does not
                    // contain a recognised cache file.
                    System.out.println( "Error0 in createGrid2DSquareCellDouble(File(" + gridFile + ")" );
                    return null;
                }
            }*/
            try {
                return createGrid2DSquareCellDouble( this.directory, gridFile, new ObjectInputStream( new BufferedInputStream( new FileInputStream( cache ) ) ) );
            } catch ( Exception e0 ) {
                System.out.println( e0 );
                e0.printStackTrace();
            }
        }
        // Assume it is ESRI asciigrid
        ESRIAsciiGridImporter eSRIAsciigridImporter = new ESRIAsciiGridImporter( gridFile );
        Object[] header = eSRIAsciigridImporter.readHeaderObject();
        long ncols = ( Long ) header[ 0 ];
        long nrows = ( Long ) header[ 1 ];
        double noDataValue = ( Double ) header[ 5 ];
        eSRIAsciigridImporter.close();
        //return createGrid2DSquareCellDouble( FileCreator.createNewFile( this.directory ), gridFile, 0L, 0L, nrows - 1L, ncols - 1L, noDataValue );
        return createGrid2DSquareCellDouble( this.directory, gridFile, 0L, 0L, nrows - 1L, ncols - 1L, noDataValue );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from gridFile.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param noDataValue the no data value to be used.
     * Default:
     * Grid2DSquareCellDoubleChunkAbstractFactory to Grid2DSquareCellDoubleChunkArrayFactory
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File directory, File gridFile, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, double noDataValue ) {
        return createGrid2DSquareCellDouble( directory, gridFile, new Grid2DSquareCellDoubleChunkArrayFactory(), startRowIndex, startColIndex, endRowIndex, endColIndex, noDataValue );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from gridFile.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that 
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param noDataValue the no data value to be used.
     * Defaults:
     * chunkNrows and chunkNcols to 64
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File directory, File gridFile, Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, double noDataValue ) {
        return createGrid2DSquareCellDouble( directory, gridFile, grid2DSquareCellDoubleChunkFactory, 64, 64, startRowIndex, startColIndex, endRowIndex, endColIndex, noDataValue );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from gridFile.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that 
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param noDataValue the no data value to be used.
     * Default:
     * GridStatisticsAbstract to GridStatistics1
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File directory, File gridFile, Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory, int chunkNrows, int chunkNcols, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, double noDataValue ) {
        return createGrid2DSquareCellDouble( new GridStatistics1(), directory, gridFile, grid2DSquareCellDoubleChunkFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex, noDataValue );
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble with values obtained from gridFile.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellDouble information.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param grid2DSquareCellDoubleChunkFactory The prefered
     *   Grid2DSquareCellDoubleChunkAbstractFactory for creating chunks that 
     *   the constructed Grid2DSquareCellDouble is to be made of.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param noDataValue the no data value to be used.
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( GridStatisticsAbstract gridStatistics, File directory, File gridFile, Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory, int chunkNrows, int chunkNcols, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, double noDataValue ) {
        return new Grid2DSquareCellDouble( gridStatistics, directory, gridFile, grid2DSquareCellDoubleChunkFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex, noDataValue );
    }
    
    /////////////////////////
    // Create from a cache //
    /////////////////////////
    
    /**
     * @param directory
     * @param gridFile
     * @param ois
     */
    public Grid2DSquareCellDouble createGrid2DSquareCellDouble( File directory, File gridFile, ObjectInputStream ois ) {
        return new Grid2DSquareCellDouble( directory, gridFile, ois );
    }
    
    /**
     * Returns directory.
     */
    public File getDirectory() {
        return this.directory;
    }
    
}
