/**
 * Version 1.0 is to handle individual single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
// gnu dependencies
import gnu.trove.TIntHashSet;
import gnu.trove.TIntLongHashMap;
import gnu.trove.TIntLongIterator;
import gnu.trove.TIntLongProcedure;
// java dependencies
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Arrays;
import java.util.ConcurrentModificationException;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
import uk.ac.leeds.ccg.andyt.grids.utilities.UnsignedLong;
import uk.ac.leeds.ccg.andyt.grids.utilities.UnsignedLongPowersOf2;

/**
 * The maximum size of this Grid2DSquareCellIntChunkAbstract is 64 cells.
 */
public class Grid2DSquareCellIntChunk64CellMap extends Grid2DSquareCellIntChunkAbstract {
    
    /**
     * For storing values mapped to a binary encoded long. The long is a key
     * which indicates if the value is that at a given location. Both keys and
     * values are unique.
     */
    private TIntLongHashMap data;
    
    /**
     * Default constructor.
     */
    protected Grid2DSquareCellIntChunk64CellMap() {}
    
    /**
     * TODO: docs
     */
    protected Grid2DSquareCellIntChunk64CellMap( Grid2DSquareCellInt grid2DSquareCellInt, ChunkID chunkID ) {
        //initChunkID( chunkID );
        this.chunkID = chunkID;
        initGrid2DSquareCell( grid2DSquareCellInt );
        if ( grid2DSquareCellInt.getChunkNrows(grid2DSquareCellInt.handleOutOfMemoryErrorFalse) * grid2DSquareCellInt.getChunkNcols(grid2DSquareCellInt.handleOutOfMemoryErrorFalse) <= 64 ) {
            initData();
        } else {
            System.out.println("Warning: grid2DSquareCellIntChunk.chunkNrows * grid2DSquareCellIntChunk.chunkNcols > 64. Cannot create " + this.getClass().getName() + "!");
            // Sample? What to do? Hmmmmmmm!
        }
    }
    
    /**
     * TODO: docs
     */
    protected Grid2DSquareCellIntChunk64CellMap( Grid2DSquareCellIntChunkAbstract grid2DSquareCellIntChunk, ChunkID chunkID ) {
        this.chunkID = chunkID;
        Grid2DSquareCellInt grid2DSquareCellInt = grid2DSquareCellIntChunk.getGrid2DSquareCellInt();
        initGrid2DSquareCell( grid2DSquareCellInt );
        int chunkNrows = grid2DSquareCellInt.getChunkNrows(grid2DSquareCellInt.handleOutOfMemoryErrorFalse);
        int chunkNcols = grid2DSquareCellInt.getChunkNcols(grid2DSquareCellInt.handleOutOfMemoryErrorFalse);
        if ( chunkNrows * chunkNcols <= 64 ) {
            int noDataValue = grid2DSquareCellInt.getNoDataValue( grid2DSquareCell.handleOutOfMemoryErrorFalse );
            initData();
            int value;
            for ( int row = 0; row < chunkNrows; row ++ ) {
                for ( int col = 0; col < chunkNcols; col ++ ) {
                    value = grid2DSquareCellIntChunk.getCell( row, col, noDataValue );
                    if ( value != noDataValue ) {
                        initCell( row, col, value );
                    }
                }
            }
        } else {
            System.out.println("Warning: grid2DSquareCellIntChunk.chunkNrows * grid2DSquareCellIntChunk.chunkNcols > 64. Cannot create " + this.getClass().getName() + "!");
            // Sample? What to do?
        }
    }
    
    /**
     * Returns a description of this.
     */
    public String toString() {
        return "Grid2DSquareCellIntChunk64CellMap( ChunkID( " + this.chunkID.toString() + " ) )";
    }
    
    /**
     * Initialises the data associated with this.
     */
    public void initData() {
        this.data = new TIntLongHashMap();
    }
    
//    /**
//     * TODO: docs
//     */
//    public TDoubleLongHashMap getData(){
//        return this.data;
//    }
    
    /**
     * Clears the data associated with this.
     */
    public void clearData() {
        this.data = null;
        System.gc();
    }
    
    /**
     * Returns this.data
     */
    public TIntLongHashMap getData() {
        return this.data;
    }
    
    /**
     * TODO:
     * docs
     */
    public int[] toArrayIncludingNoDataValues() {
        Grid2DSquareCellInt grid2DSquareCellInt = getGrid2DSquareCellInt();
        int noDataValue = grid2DSquareCellInt.getNoDataValue( grid2DSquareCell.handleOutOfMemoryErrorFalse );
        int chunkNrows = grid2DSquareCellInt.getChunkNrows(grid2DSquareCellInt.handleOutOfMemoryErrorFalse);
        int chunkNcols = grid2DSquareCellInt.getChunkNcols(grid2DSquareCellInt.handleOutOfMemoryErrorFalse);
        int[] array = new int[ chunkNrows * chunkNcols ];
        Arrays.fill( array, noDataValue );
        TIntLongIterator iterator = this.data.iterator();
        UnsignedLong valueMap = new UnsignedLong();
        int ite;
        boolean[] valueMapped;
        int position;
        for ( ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            valueMap.setLong( iterator.value() );
            valueMapped = valueMap.toBooleanArray();
            for ( position = 0; position < valueMapped.length; position ++ ) {
                if ( valueMapped[ position ] ) {
                    array[ position ] = iterator.key();
                }
            }
        }
        return array;
    }
    
    /**
     * TODO:
     * docs
     */
    public int[] toArrayNotIncludingNoDataValues() {
        int[] array = new int[ getNonNoDataValueCountInt() ];
        TIntLongIterator iterator = this.data.iterator();
        UnsignedLong valueMap = new UnsignedLong();
        int ite;
        boolean[] valueMapped;
        int position;
        int index = 0;
        for ( ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            valueMap.setLong( iterator.value() );
            valueMapped = valueMap.toBooleanArray();
            for ( position = 0; position < valueMapped.length; position ++ ) {
                if ( valueMapped[ position ] ) {
                    array[ index ] = iterator.key();
                    index ++;
                }
            }
        }
        return array;
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell row chunkCellColIndex.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param noDataValue the noDataValue of this.grid2DSquareCellInt
     */
    public int getCell( int chunkCellRowIndex, int chunkCellColIndex, int noDataValue ) {
       TIntLongIterator iterator = this.data.iterator();
        UnsignedLong valueMap = new UnsignedLong();
         Grid2DSquareCellInt grid2DSquareCellInt = getGrid2DSquareCellInt();
        int chunkNcols = grid2DSquareCellInt.getChunkNcols(grid2DSquareCellInt.handleOutOfMemoryErrorFalse);
        int ite;
        for ( ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            valueMap.setLong( iterator.value() );
            if ( valueMap.isAtPosition( ( chunkCellRowIndex * chunkNcols ) + chunkCellColIndex ) ) {
                return iterator.key();
            }
        }
        return noDataValue;
    }
    
    /**
     * Initialises the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell column chunkCellColIndex. Utility method for constructor.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToInitialise the value with which the cell is initialised
     */
    protected void initCell(int chunkCellRowIndex, int chunkCellColIndex, int valueToInitialise ) {
         Grid2DSquareCellInt grid2DSquareCellInt = getGrid2DSquareCellInt();
        int chunkNcols = grid2DSquareCellInt.getChunkNcols(grid2DSquareCellInt.handleOutOfMemoryErrorFalse);
       if ( this.data.containsKey( valueToInitialise ) ) {
            this.data.put( valueToInitialise, data.get( valueToInitialise ) + powerOf2( ( ( chunkCellRowIndex * chunkNcols ) + chunkCellColIndex ) ) );
        } else {
            this.data.put( valueToInitialise, powerOf2( ( ( chunkCellRowIndex * chunkNcols ) + chunkCellColIndex ) ) );
        }
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell column chunkCellColIndex and sets it to valueToSet
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToSet the value the cell is to be set to
     * @param noDataValue the noDataValue of this.grid2DSquareCellInt
     */
    public int setCell( int chunkCellRowIndex, int chunkCellColIndex, int valueToSet, int noDataValue ) {
        Grid2DSquareCellInt grid2DSquareCellInt = getGrid2DSquareCellInt();
        int result = noDataValue;
        TIntLongIterator iterator = this.data.iterator();
        UnsignedLong valueMap = new UnsignedLong();
        int value;
        boolean gotValue = false;
        boolean setValue = false;
        if ( valueToSet == result ) {
            setValue = true;
        } else {
            setValue = false;
        }
        int position = ( chunkCellRowIndex * grid2DSquareCellInt.getChunkNcols(grid2DSquareCellInt.handleOutOfMemoryErrorFalse) ) + chunkCellColIndex;
        long powerOf2 = powerOf2( position );
        int ite;
        for ( ite = 0; ite < data.size(); ite ++ ) {
            if ( ! ( gotValue && setValue ) ) {
                try{
                    iterator.advance();
                } catch ( ConcurrentModificationException why ) {}
                valueMap.setLong( iterator.value() );
                value = iterator.key();
                //if ( ! gotValue ) { // Probably slows things down!?
                if ( valueMap.isAtPosition( position ) ) {
                    // May be better to keep the value in data and tidy if
                    // necessary. This would save evaluating the if, but would
                    // mean that the chunk is likely to get heavier. However it
                    // is difficult to know what is best, it probably depends on
                    // the data and what is being done with it!
                    if ( iterator.value() == powerOf2 ) {
                        if ( value == valueToSet ) {
                            return value;
                        } else {
                            //try {
                            this.data.remove( value ); // Removal causes ConcurrentModificationException in iterator.advance()!
                            //} catch ( ConcurrentModificationException e ) {}
                        }
                    } else {
                        iterator.setValue( iterator.value() - powerOf2 );
                    }
                    result = value;
                    gotValue = true;
                }
                //}
                //if ( ! setValue ) { // Probably slows things down!?
                if ( value == valueToSet ) {
                    iterator.setValue( iterator.value() + powerOf2 );
                    setValue = true;
                }
                //}
            } else {
                return result;
            }
        }
        if ( ! setValue ) {
            this.data.put( valueToSet, powerOf2 );
        }
        return result;
    }
    
    /**
     * Returns 2 raised to the power of value as a long.
     * TODO: move this to the chunk master and have as a long[]. The look up
     * should be faster than the creation on the fly.
     */
    public long powerOf2( int value ){
        Grid2DSquareCellInt grid2DSquareCellInt = this.getGrid2DSquareCellInt();
        try {
            //return this.unsignedLongPowersOf2.powersOf2[ 63 - value ];
            return grid2DSquareCellInt.unsignedLongPowersOf2.powersOf2[ value ];
        } catch ( java.lang.NullPointerException e ) {
            grid2DSquareCellInt.unsignedLongPowersOf2 = new UnsignedLongPowersOf2( false );
            //return this.unsignedLongPowersOf2.powersOf2[ 63 - value ];
            return grid2DSquareCellInt.unsignedLongPowersOf2.powersOf2[ value ];
        }
    }
    
    /**
     * Returns the number of cells with noDataValues as a BigInteger
     */
    public BigInteger getNonNoDataValueCountBigInteger() {
        TIntLongIterator iterator = this.data.iterator();
        long cellIDMap = 0L;
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            cellIDMap += iterator.value();
        }
        return new BigInteger( "" + new UnsignedLong( cellIDMap ).getCount() );
    }
    
    /**
     * Returns the number of cells with noDataValues as an int
     */
    public int getNonNoDataValueCountInt() {
        TIntLongIterator iterator = this.data.iterator();
        long cellIDMap = 0L;
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            cellIDMap += iterator.value();
        }
        return new UnsignedLong( cellIDMap ).getCount();
    }
    
    /**
     * For returning the sum of all non noDataValues as a BigDecimal
     */
    public BigDecimal getSumBigDecimal() {
        BigDecimal sum = new BigDecimal( 0.0d );
        BigDecimal thisCount;
        TIntLongIterator iterator = this.data.iterator();
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = BigDecimal.valueOf( ( long ) value.getCount() );
            sum = sum.add( new BigDecimal( iterator.key() ).multiply( thisCount ) );
        }
        return sum;
    }
    
    /**
     * For returning the minimum of all non noDataValues as a int.
     */
    public int getMinInt() {
        int min = Integer.MAX_VALUE;
        TIntLongIterator iterator = this.data.iterator();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            min = Math.min( min, iterator.key() );
        }
        return min;
    }
    
    /**
     * For returning the maximum of all non noDataValues as a int.
     */
    public int getMaxInt() {
        int max = Integer.MIN_VALUE;
        TIntLongIterator iterator = this.data.iterator();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            max = Math.max( max, iterator.key() );
        }
        return max;
    }
    
    /**
     * For returning the Arithmetic Mean of all non noDataValues as a BigDecimal.
     * Using BigDecimal this should be as precise as possible with doubles.
     */
    public BigDecimal getArithmeticMeanBigDecimal() {
        BigDecimal mean = new BigDecimal("0");
        BigDecimal count = new BigDecimal("0");
        BigDecimal thisCount;
        //double mean = 0.0d;
        //double count = 0.0d;
        //byte thisCount = 0;
        TIntLongIterator iterator = this.data.iterator();
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = BigDecimal.valueOf( ( long ) value.getCount() );
            count = count.add( thisCount );
            mean = mean.add( new BigDecimal( iterator.key() ).multiply( thisCount ) );
            //count += ( double ) thisCount;
            //mean += iterator.key() * ( double ) thisCount;
        }
        try {
            return mean.divide( count, mean.scale() + 10, BigDecimal.ROUND_HALF_EVEN );
        } catch ( ArithmeticException e ) {
            System.out.println("Divide by zero in " + this.getClass().getName() + ".getArithmeticMean(). Returning null!" );
            return null;
            //return Double.NaN;
            //return 0.0d;
        }
        //return mean /= count;
    }
    
    /**
     * For returning the Geometric Mean of all non noDataValues as a BigDecimal
     * Warning! This is imprecise and it can happen that Math.pow does not 
     * return what might be expected! (For example, negative powers in the range 
     * (0,1) for negative numbers.) 
     * TODO:
     * Develop a pow function such as com.ibm.icu.math.BigDecimal.pow
     * This resource is not used here due to licensing, but it could be...
     */
    public BigDecimal getGeometricMeanBigDecimal() {
        BigDecimal mean = new BigDecimal(1.0d);
        BigDecimal count = new BigDecimal(0.0d);
        BigDecimal thisCount;
        //double mean = 1.0d;
        //double count = 0.0d;
        //byte thisCount = 0;
        TIntLongIterator iterator = this.data.iterator();
        double key;
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            key = iterator.key();
            value.setLong( iterator.value() );
            thisCount = new BigDecimal( ( double ) value.getCount() );
            count = count.add( thisCount );
            mean = mean.multiply( new BigDecimal( key ).multiply( thisCount ) );
            //thisCount = value.getCount();
            //count += thisCount;
            //mean *= key * ( double ) thisCount;
        }
        //return ( mean.pow( new BigDecimal( 1.0d ).divide( count, BigDecimal.ROUND_HALF_EVEN ) ) ).doubleValue();
        return new BigDecimal( Math.pow( mean.doubleValue(), 1.0d / count.doubleValue() ) );
        //return Math.pow( mean, 1.0d / count );
    }

    /**
     * Returns the Harmonic Mean of all non noDataValues as a BigDecimal. Zero 
     * values are ignored in the calculation.
     */
    public BigDecimal getHarmonicMeanBigDecimal() {
        BigDecimal mean = new BigDecimal(1.0d);
        BigDecimal count = new BigDecimal(0.0d);
        BigDecimal thisCount;
        BigDecimal one = new BigDecimal(1.0d);
        TIntLongIterator iterator = this.data.iterator();
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = BigDecimal.valueOf( ( long ) value.getCount() );
            if ( iterator.key() != 0.0d ) {
                count = count.add( thisCount );
                mean = mean.add( thisCount.multiply( one.divide( new BigDecimal( iterator.key() ), 325, BigDecimal.ROUND_HALF_EVEN ) ) );
            } else {
                System.out.println("Warning: Value of 0.0d encountered in " + this.getClass().getName() + "!" );
            }
        }
        if ( mean.compareTo( new BigDecimal( 0.0d ) ) != 0 ) {
            return count.divide( mean, 325, BigDecimal.ROUND_HALF_EVEN );
        } else {
            return mean;
        }
    }
    
    /**
     * For returning the mode of all non noDataValues as a TDoubleHashSet
     */
    public TIntHashSet getModeTIntHashSet() {
        TIntHashSet mode = new TIntHashSet();
        TIntLongIterator iterator = this.data.iterator();
        byte modeCount = 0;
        byte thisCount;
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = value.getCount();
            if ( thisCount > modeCount ) {
                mode.clear();
                mode.add( iterator.key() );
                modeCount = thisCount;
            } else {
                if ( thisCount == modeCount ) {
                    mode.add( iterator.key() );
                }
            }
        }
        return mode;
    }
    
    /**
     * For returning the median of all non noDataValues as a double
     */
    public double getMedianDouble() {
        int scale = 325;
        int[] keys = this.data.keys();
        sort1( keys, 0, keys.length );
        int nonNoDataValueCount = getNonNoDataValueCountInt();
        UnsignedLong unsignedLong = new UnsignedLong();
        int index = -1;
        if ( nonNoDataValueCount % 2 == 0 ) {
            // Need arithmetic mean of ( ( nonNoDataValueCount / 2 ) - 1 )th
            // and ( nonNoDataValueCount / 2 )th values
            int requiredIndex = ( nonNoDataValueCount / 2 ) - 1;
            boolean got = false;
            //double median = 0.0d;
            BigDecimal medianBigDecimal = null;
            for ( int keyIndex = 0; keyIndex < keys.length; keyIndex ++ ) {
                unsignedLong.setLong( this.data.get( keys[ keyIndex ] ) );
                index += unsignedLong.getCount();
                if ( ! got && index >= requiredIndex ) {
                    //median = keys[ keyIndex ];
                    medianBigDecimal = new BigDecimal( keys[ keyIndex ] );
                    got = true;
                }
                if ( index >= requiredIndex + 1 ) {
                    //median += keys[ keyIndex ];
                    //median /= 2.0d;
                    //return ( median + keys[ keyIndex ] ) / 2.0d;
                    return ( medianBigDecimal.add( new BigDecimal( keys[ keyIndex ] ) ) ).divide( new BigDecimal( 2.0d ), scale, BigDecimal.ROUND_HALF_EVEN ).doubleValue();
                }
            }
        } else {
            // Need ( ( nonNoDataValueCount ) / 2 )th value
            int requiredIndex = nonNoDataValueCount / 2;
            for ( int keyIndex = 0; keyIndex < keys.length; keyIndex ++ ) {
                unsignedLong.setLong( this.data.get( keys[ keyIndex ] ) );
                index += unsignedLong.getCount();
                if ( index >= requiredIndex ) {
                    return keys[ keyIndex ];
                }
            }
        }
        return getGrid2DSquareCell().getNoDataValueDouble( getGrid2DSquareCell().handleOutOfMemoryErrorFalse );
    }
    
    /**
     * For returning the standard deviation of all non noDataValues as a double
     */
    public double getStandardDeviationDouble() {
        double standardDeviation = 0.0d;
        double mean = getArithmeticMeanDouble();
        TIntLongIterator iterator = this.data.iterator();
        UnsignedLong value = new UnsignedLong();
        double count = 0.0d;
        double thisCount;
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = ( double ) value.getCount();
            count += thisCount;
            standardDeviation += ( ( iterator.key() - mean ) * ( iterator.key() - mean ) ) * thisCount;
        }
        if ( ( count - 1.0d ) > 0 ) {
            return Math.sqrt( standardDeviation / ( count - 1.0d ) );
        } else {
            return standardDeviation;
        }
    }
    
    /**
     * For returning the number of different values.
     */
    public BigInteger getDiversity() {
        return new BigInteger( Integer.toString( this.data.size() ) );
    }
    
    /**
     * Returns a Grid2DSquareCellIntChunk64CellMapIterator for iterating over 
     * the cells in this.
     */
    public AbstractIterator iterator() {
        return new Grid2DSquareCellIntChunk64CellMapIterator( this );
    }
    
}