/**
 * Version 1.0 is to handle individual single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
// java dependencies
import java.io.BufferedInputStream;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.StreamTokenizer;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.math.BigDecimal;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciiGridImporter;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;

/**
 * TODO:
 * 1. Abstraction and extension as for Grid2DSquareCellDoubleFactory.
 * 2. Add constructor and setter for prefered type of Grid2DSquareCellIntChunkFactory 
 * 3. docs
 */
public class Grid2DSquareCellIntFactory {
    
    /**
     * A directory for caching.
     */
    protected File directory;

//    /**
//     * A store for the number of rows in a chunk.
//     */
//    protected int chunkNrows;
//
//    /**
//     * A store for the number of columns in a chunk.
//     */
//    protected int chunkNcols;
    
    /**
     * Defaults directory to a new one in
     * System.getProperties( "java.io.tmpdir" );
     */
    public Grid2DSquareCellIntFactory() {
        this( FileCreator.createTempFile() );
        //this ( FileCreator.createNewFile() );
    }
    
    /**
     * @param directory A "workspace directory" for storing temporary files and
     * caching Grid2DSquareCellDoubleAbstract data.
     */
    public Grid2DSquareCellIntFactory(File directory) {
        this.directory = directory;
        //this( directory, 64, 64 );
    }
    
    
    /////////////////////////
    // Create from scratch //
    /////////////////////////
    
    /** 
     * Returns an Grid2DSquareCellInt with all values as 
     * noDataValues.
     * @param nrows the Grid2DSquareCellInt nrows.
     * @param ncols the Grid2DSquareCellInt ncols.
     * Defaults:
     * chunkNrows and chunkNcols to 64
     */
    public Grid2DSquareCellAbstract createGrid2DSquareCellInt( long nrows, long ncols ) {
        return createGrid2DSquareCellInt( 64, 64, nrows, ncols );
    }
    
    /** 
     * Returns an Grid2DSquareCellInt with all values as 
     * noDataValues.
     * @param chunkNrows The Grid2DSquareCellInt chunkNrows.
     * @param chunkNcols The Grid2DSquareCellInt chunkNcols.
     * @param nrows the Grid2DSquareCellInt nrows.
     * @param ncols the Grid2DSquareCellInt ncols.
     * Defaults:
     * dimensions = [ cellsize = 1.0d, xmin = 0.0d, ymin = 0.0d, 
     * xmax = ncols * cellsize, ymax = nrows * cellsize ];
     * noDataValue = Integer.MIN_VALUE
     */
    public Grid2DSquareCellAbstract createGrid2DSquareCellInt( int chunkNrows, int chunkNcols, long nrows, long ncols ) {
        BigDecimal[] dimensions = new BigDecimal[ 5 ];
        dimensions[ 0 ] = new BigDecimal( 1.0d );
        dimensions[ 1 ] = new BigDecimal( 0.0d );
        dimensions[ 2 ] = new BigDecimal( 0.0d );
        dimensions[ 3 ] = new BigDecimal( ncols );
        dimensions[ 4 ] = new BigDecimal( nrows );
        return createGrid2DSquareCellInt( chunkNrows, chunkNcols, nrows, ncols, dimensions );
    }
    
    /** 
     * Returns an Grid2DSquareCellInt with all values as 
     * noDataValues.
     * @param nrows the Grid2DSquareCellInt nrows.
     * @param ncols the Grid2DSquareCellInt ncols.
     * @param dimensions
     * Defaults:
     * chunkNrows and chunkNcols to 64
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( long nrows, long ncols, BigDecimal[] dimensions ) {
        return createGrid2DSquareCellInt( 64, 64, nrows, ncols, dimensions );
    }
    
    /** 
     * Returns an Grid2DSquareCellInt with all values as 
     * noDataValues.
     * @param chunkNrows The Grid2DSquareCellInt chunkNrows.
     * @param chunkNcols The Grid2DSquareCellInt chunkNcols.
     * @param nrows the Grid2DSquareCellInt nrows.
     * @param ncols the Grid2DSquareCellInt ncols.
     * Defaults:
     * directory to FileCreator.createNewFile( this.directory )
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions ) {
//        this.chunkNrows = chunkNrows;
//        this.chunkNcols = chunkNcols;
        return createGrid2DSquareCellInt( FileCreator.createNewFile( this.directory ), chunkNrows, chunkNcols, nrows, ncols, dimensions );
    }
     
    /**
     * Returns a new Grid2DSquareCellInt grid with all values as 
     * noDataValues.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param chunkNrows The Grid2DSquareCellInt chunkNrows.
     * @param chunkNcols The Grid2DSquareCellInt chunkNcols.
     * @param nrows the Grid2DSquareCellInt nrows.
     * @param ncols the Grid2DSquareCellInt ncols.
     * @param dimensions
     */
     public Grid2DSquareCellInt createGrid2DSquareCellInt( File directory, int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions ) {
        return createGrid2DSquareCellInt( directory, new Grid2DSquareCellIntChunkArrayFactory(), chunkNrows, chunkNcols, nrows, ncols, dimensions );
     }

    /**
     * Returns a new Grid2DSquareCellInt grid with all values as 
     * noDataValues.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param chunkFactory The type of chunks this will be made up of.
     * @param nrows The Grid2DSquareCellInt nrows.
     * @param ncols The Grid2DSquareCellInt ncols.
     * @param dimensions
     * defaults GridStatisticsAbstract to GridStatistics1
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( File directory, Grid2DSquareCellIntChunkAbstractFactory chunkFactory, int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions ) {
        return createGrid2DSquareCellInt( new GridStatistics1(), directory, chunkFactory, chunkNrows, chunkNcols, nrows, ncols, dimensions );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt grid with all values as 
     * noDataValues.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param chunkFactory The type of chunks this will be made up of.
     * @param nrows The Grid2DSquareCellInt nrows.
     * @param ncols The Grid2DSquareCellInt ncols.
     * @param dimensions
     * TODO: test
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( GridStatisticsAbstract gridStatistics, File directory, Grid2DSquareCellIntChunkAbstractFactory chunkFactory, int chunkNrows, int chunkNcols, long nrows, long ncols, BigDecimal[] dimensions ) {
        return new Grid2DSquareCellInt( gridStatistics, directory, chunkFactory, chunkNrows, chunkNcols, nrows, ncols, dimensions );
    }
    
    /////////////////////////////////////////////////////////
    // Create from an existing Grid2DSquareCellIntAbstract //
    /////////////////////////////////////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellInt with all values as int values from 
     * grid2DSquareCell.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values 
     * are obtained.
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( Grid2DSquareCellAbstract grid2DSquareCell ) {
        return createGrid2DSquareCellInt( this.directory, grid2DSquareCell, 0L, 0L, grid2DSquareCell.getNrows() - 1L, grid2DSquareCell.getNcols() - 1L );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from 
     * grid2DSquareCell.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values 
     * are obtained.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats 
     * values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats 
     * values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats 
     * values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats 
     * values are used.
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( Grid2DSquareCellAbstract grid2DSquareCell, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
        return createGrid2DSquareCellInt( FileCreator.createNewFile( this.directory, "" + startRowIndex + "_" + startColIndex + "_" + endRowIndex + "_" + endColIndex ), grid2DSquareCell, startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from 
     * grid2DSquareCell.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values 
     * are obtained.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats 
     * values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats 
     * values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats 
     * values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats 
     * values are used.
     * Default:
     * Grid2DSquareCellIntChunkAbstractFactory to Grid2DSquareCellIntChunkMapFactory,
     * chunkNrows to 64,
     * chunkNcols to 64.
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( File directory, Grid2DSquareCellAbstract grid2DSquareCell, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
       return createGrid2DSquareCellInt( directory, grid2DSquareCell, new Grid2DSquareCellIntChunkMapFactory(), 64, 64, startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from 
     * grid2DSquareCell.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values 
     * are obtained.
     * @param grid2DSquareCellIntChunkAbstractFactory The 
     * Grid2DSquareCellIntChunkAbstractFactory used to construct the chunks.
     * @param chunkNrows The (maximum) number of rows in each chunk.
     * @param chunkNcols The (maximum) number of columns in each chunk.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats 
     * values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats 
     * values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats 
     * values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats 
     * values are used.
     * Defaults:
     * GridStatisticsAbstract to GridStatistics1
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( File directory, Grid2DSquareCellAbstract grid2DSquareCell, Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkAbstractFactory, int chunkNrows, int chunkNcols, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
        return createGrid2DSquareCellInt( new GridStatistics1(), directory, grid2DSquareCell, grid2DSquareCellIntChunkAbstractFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from 
     * grid2DSquareCell.
     * @param gridStatistics The GridStatisticsAbstract for the returned 
     * Grid2DSquareCellInt.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values 
     * are obtained.
     * @param grid2DSquareCellIntChunkAbstractFactory The 
     * Grid2DSquareCellIntChunkAbstractFactory used to construct the chunks.
     * @param chunkNrows The (maximum) number of rows in each chunk.
     * @param chunkNcols The (maximum) number of columns in each chunk.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats 
     * values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats 
     * values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats 
     * values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats 
     * values are used.
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( GridStatisticsAbstract gridStatistics, File directory, Grid2DSquareCellAbstract grid2DSquareCell, Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkAbstractFactory, int chunkNrows, int chunkNcols, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
        return new Grid2DSquareCellInt( gridStatistics, directory, grid2DSquareCell, grid2DSquareCellIntChunkAbstractFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
    
    
    ////////////////////////
    // Create from a File //
    ////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from
     * gridFile. If gridFile is a directory then it is assumed to contain a 
     * file called cache which can be opened into an object input stream and 
     * initailised as an instance of a class extending Grid2DSquareCellAbstract.
     * @param gridFile either a directory, or a formatted File with a specific 
     * extension containing the data and information about the 
     * Grid2DSquareCellInt to be returned. 
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( File gridFile ) {
        if ( gridFile.isDirectory() ) {
            // Initialise from Cache
            File cache = new File( gridFile, "Cache" );
            //if ( ! cache.exists() ) {
            //    throw new Exception( "File( " + cache.toString() + " does not exist!" );
            //    return null;
            //}
            try {
                return createGrid2DSquareCellInt( this.directory, gridFile, new ObjectInputStream( new BufferedInputStream( new FileInputStream( cache ) ) ) );
            } catch ( Exception e0 ) {
                System.out.println( e0 );
                e0.printStackTrace();
            }
        }
        // Assume it is ESRI asciigrid
        ESRIAsciiGridImporter eSRIAsciigridImporter = new ESRIAsciiGridImporter( gridFile );
        Object[] header = eSRIAsciigridImporter.readHeaderObject();
        long ncols = ( Long ) header[ 0 ];
        long nrows = ( Long ) header[ 1 ];
        double noDataValue = ( Double ) header[ 5 ];
        eSRIAsciigridImporter.close();
        //return createGrid2DSquareCellInt( FileCreator.createNewFile( this.directory ), gridFile, 0L, 0L, nrows - 1L, ncols - 1L, noDataValue );
        return createGrid2DSquareCellInt( this.directory, gridFile, 0L, 0L, nrows - 1L, ncols - 1L );
        //return null;
    }
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from
     * gridFile.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param gridFile either a directory, or a formatted File with a specific 
     * extension containing the data and information about the 
     * Grid2DSquareCellInt to be returned. 
     * @param startRowIndex The topmost row index of the grid represented in 
     * gridFile thats values are used.
     * @param startColIndex The leftmost column index of the grid represented in 
     * gridFile thats values are used.
     * @param endRowIndex The bottom row index of the grid represented in 
     * gridFile thats values are used.
     * @param endColIndex The rightmost column index of the grid represented in 
     * gridFile thats values are used.
     * Default:
     * Grid2DSquareCellIntChunkAbstractFactory to Grid2DSquareCellIntChunkArrayFactory
     */
    protected Grid2DSquareCellInt createGrid2DSquareCellInt( File directory, File gridFile, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
        return createGrid2DSquareCellInt( directory, gridFile, new Grid2DSquareCellIntChunkArrayFactory(), startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
      
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from
     * gridFile.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param gridFile either a directory, or a formatted File with a specific 
     * extension containing the data and information about the 
     * Grid2DSquareCellInt to be returned. 
     * @param grid2DSquareCellIntChunkFactory The 
     * Grid2DSquareCellIntChunkAbstractFactory used to construct the chunks.
     * @param startRowIndex The topmost row index of the grid represented in 
     * gridFile thats values are used.
     * @param startColIndex The leftmost column index of the grid represented in 
     * gridFile thats values are used.
     * @param endRowIndex The bottom row index of the grid represented in 
     * gridFile thats values are used.
     * @param endColIndex The rightmost column index of the grid represented in 
     * gridFile thats values are used.
     * Defaults:
     * chunkNrows and chunkNcols to 64
     */
    protected Grid2DSquareCellInt createGrid2DSquareCellInt( File directory, File gridFile, Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
        return createGrid2DSquareCellInt( directory, gridFile, grid2DSquareCellIntChunkFactory, 64, 64, startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from
     * gridFile.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param gridFile either a directory, or a formatted File with a specific 
     * extension containing the data and information about the 
     * Grid2DSquareCellInt to be returned. 
     * @param grid2DSquareCellIntChunkFactory The 
     * Grid2DSquareCellIntChunkAbstractFactory used to construct the chunks.
     * @param chunkNrows The (maximum) number of rows in each chunk.
     * @param chunkNcols The (maximum) number of columns in each chunk.
     * @param startRowIndex The topmost row index of the grid represented in 
     * gridFile thats values are used.
     * @param startColIndex The leftmost column index of the grid represented in 
     * gridFile thats values are used.
     * @param endRowIndex The bottom row index of the grid represented in 
     * gridFile thats values are used.
     * @param endColIndex The rightmost column index of the grid represented in 
     * gridFile thats values are used.
     * Defaults:
     * GridStatisticsAbstract to GridStatistics1
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( File directory, File gridFile, Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory, int chunkNrows, int chunkNcols, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
         return createGrid2DSquareCellInt( new GridStatistics1(), directory, gridFile, grid2DSquareCellIntChunkFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from
     * gridFile.
     * @param gridStatistics The GridStatisticsAbstract for the returned 
     * Grid2DSquareCellInt.
     * @param directory The directory to be used for storing cached 
     * Grid2DSquareCellInt information.
     * @param gridFile either a directory, or a formatted File with a specific 
     * extension containing the data and information about the 
     * Grid2DSquareCellInt to be returned. 
     * @param grid2DSquareCellIntChunkFactory The 
     * Grid2DSquareCellIntChunkAbstractFactory used to construct the chunks.
     * @param chunkNrows The (maximum) number of rows in each chunk.
     * @param chunkNcols The (maximum) number of columns in each chunk.
     * @param startRowIndex The topmost row index of the grid represented in 
     * gridFile thats values are used.
     * @param startColIndex The leftmost column index of the grid represented in 
     * gridFile thats values are used.
     * @param endRowIndex The bottom row index of the grid represented in 
     * gridFile thats values are used.
     * @param endColIndex The rightmost column index of the grid represented in 
     * gridFile thats values are used.
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( GridStatisticsAbstract gridStatistics, File directory, File gridFile, Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory, int chunkNrows, int chunkNcols, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex ) {
        return new Grid2DSquareCellInt( gridStatistics, directory, gridFile, grid2DSquareCellIntChunkFactory, chunkNrows, chunkNcols, startRowIndex, startColIndex, endRowIndex, endColIndex );
    }
    
    /////////////////////////
    // Create from a cache //
    /////////////////////////
    
    /**
     * @param directory
     * @param gridFile
     * @param ois
     * Defaults GridStatistics to GridStatistics1
     */
    public Grid2DSquareCellInt createGrid2DSquareCellInt( File directory, File gridFile, ObjectInputStream ois ) {
        return new Grid2DSquareCellInt( directory, gridFile, ois );
    }
    
    /**
     * Returns directory.
     */
    public File getDirectory() {
        return this.directory;
    }

}
