/**
 * Version 1.0 is for IO of 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.exchange;
// java dependencies
import java.io.File;
import java.io.PrintWriter;
import java.io.BufferedWriter;
//import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.math.BigDecimal;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDouble;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellInt;
import uk.ac.leeds.ccg.andyt.grids.utilities.ErrorHandler;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;

/**
 * Class for exporting ESRI Asciigrid.
 */

public class ESRIAsciiGridExporter extends ErrorHandler {
       
    private static final long serialVersionUID = 1L;
    
    /** Creates a new instance of ESRIAsciiGridExporter */
    public ESRIAsciiGridExporter() {
        this.initMemoryReserve();
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * a the File to which it was written.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     */
    public File toAsciiFile( Grid2DSquareCellAbstract grid2DSquareCell ) {
        return toAsciiFile( grid2DSquareCell, FileCreator.createNewFile( grid2DSquareCell.getDirectory(), grid2DSquareCell.getName() + ".asc" ) );
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * file.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     * @param file The File to export to.
     */
    public File toAsciiFile( Grid2DSquareCellAbstract grid2DSquareCell, File file ) {
        return toAsciiFile( grid2DSquareCell, FileCreator.createNewFile( grid2DSquareCell.getDirectory(), grid2DSquareCell.getName() + ".asc" ), grid2DSquareCell.getNoDataValueBigDecimal( true ) );
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * file.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     * @param file The File to export to.
     * @param noDataValueBigDecimal The value to be used or substituted as a 
     *   noDataValue for grid2DSquareCell.
     */
    public File toAsciiFile( Grid2DSquareCellAbstract grid2DSquareCell, File file, BigDecimal noDataValueBigDecimal ) {
        try {
            boolean handleOutOfMemoryError = true;
            String noDataValueBigDecimalString = noDataValueBigDecimal.toString();
            PrintWriter pw = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( file ) ) ) );
            BigDecimal[] dimensions = grid2DSquareCell.getDimensions();
            long nrows = grid2DSquareCell.getNrows();
            long ncols = grid2DSquareCell.getNcols();
            pw.println( "ncols " + ncols );
            pw.println( "nrows " + nrows );
            pw.println( "xllcorner " + dimensions[1].toString() );
            pw.println( "yllcorner " + dimensions[2].toString() );
            pw.println( "cellsize " + dimensions[0].toString() );
            
            
            if ( grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
                Grid2DSquareCellInt grid2DSquareCellInt = ( Grid2DSquareCellInt ) grid2DSquareCell;
                int noDataValue = grid2DSquareCellInt.getNoDataValue( this.handleOutOfMemoryErrorFalse );
                pw.println( "NODATA_Value " + noDataValueBigDecimalString );
                long row;
                long col;
                int value;
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0 ; col < ncols; col ++ ) {
                        value = grid2DSquareCellInt.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                        if ( value == noDataValue ) {
                            pw.print( noDataValueBigDecimalString + " " );
                        } else {
                            pw.print( value + " " );
                        }
                    }
                    pw.println( "" );
                }
            } else {
                //grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class
                Grid2DSquareCellDouble grid2DSquareCellDouble = ( Grid2DSquareCellDouble ) grid2DSquareCell;
                double noDataValue = grid2DSquareCellDouble.getNoDataValue();
                if (  Double.isInfinite( noDataValue ) ) {
                    System.out.println( "Warning!!! noDataValue = " + noDataValue + " in ESRIAsciigridExporter.toAsciiFile( Grid2DSquareCellAbstract( " + grid2DSquareCellDouble.toString( handleOutOfMemoryError ) + " ), File( " + file.toString() + " ) )" );
                }
                pw.println( "NODATA_Value " + noDataValueBigDecimalString );
                long row;
                long col;
                double value;
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0 ; col < ncols; col ++ ) {
                        //pw.print( grid.getCell( row, col ) + " " );
                        value = grid2DSquareCellDouble.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                        if ( Double.isNaN( value ) ) {
                            pw.print( noDataValueBigDecimalString + " " );
                            System.out.println( "Warning!!! NaN encountered at row " + row + ", column " + col + " set to noDataValue " + noDataValueBigDecimalString + "." );
                        } else {
                            if ( value == noDataValue ) {
                                pw.print( noDataValueBigDecimalString + " " );
                            } else {
                                pw.print( value + " " );
                            }
                        }
                    }
                    pw.println( "" );
                }
            }
            // Close output
                pw.flush();
                pw.close();
        } catch ( java.io.FileNotFoundException e ) {
            System.out.println( e.toString() );
            System.out.println( " in uk.ac.leeds.ccg.grids.Grid2DSquareCellDoubleExchange.toAsciiFile( File(" + file + ") )" );
            e.printStackTrace();
        }
        return file;
    }
    
}
