/**
 * Version 1.0 is for IO of 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.exchange;
// java dependencies
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.awt.image.MemoryImageSource;
import java.io.File;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDouble;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellInt;
import uk.ac.leeds.ccg.andyt.grids.utilities.ErrorHandler;

/**
 * Class for exporting to images.
 */
public class ImageExporter extends ErrorHandler {
    
    private static final long serialVersionUID = 1L;
    
    /** Creates a new instance of ImageExporter */
    public ImageExporter() {
        this.initMemoryReserve();
    }
    
    /**
     * Writes this grid as a Grey scale image
     * @param file The File exported to.
     * @param type The name of the type of image to be written e.g. "png", "jpeg"
     * TODO: need different toImage methods based on grid (see end of method)
     */
    public void toGreyScaleImage( Grid2DSquareCellAbstract grid2DSquareCell, File file, String type ) {
        long nrows = grid2DSquareCell.getNrows();
        long ncols = grid2DSquareCell.getNcols();
        // TODO:
        // Check int precision OK here.
        int size = ( int ) ( ncols * nrows );
        long row;
        long col;
        int iValue = 0;
        int[] gridImageArray = new int[ size ];
        int pos;
        // Test what writers are available as this may vary on different systems!
        boolean writerAvailable = io.isImageWriterAvailable( type );
        if ( ! writerAvailable ) {
            System.out.println( "Warning!!! Failed to write grid as " + type + " to File(" + file.toString() + ") as writer is unavailable." );
        } else {
            if ( grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
                Grid2DSquareCellInt grid2DSquareCellInt = ( Grid2DSquareCellInt ) grid2DSquareCell;
                int noDataValue = grid2DSquareCellInt.getNoDataValue( this.handleOutOfMemoryErrorFalse );
                int max = grid2DSquareCellInt.getGridStatistics().getMaxInt();
                int min = grid2DSquareCellInt.getGridStatistics().getMinInt();
                int value;
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        value = grid2DSquareCellInt.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                        //pos = ( int ) ( ( row * ncols ) + col );
                        pos = ( int ) ( ( ( ( nrows - 1 ) - row ) * ncols ) + col );
                        //pos = ( int ) ( ( ( ( nrows - row ) - 1 ) * ncols ) + col - 1 );
                        // Construct an RGB integer by byte operation : 32 bytes, first 8 bytes is transparency value
                        // second, third, forth 8 byte is Red, green, blue value, which will be used by MemoryImageSource
                        // class.
                        iValue = 255;
                        // Set noDataValue as blue
                        if (  value == noDataValue ) {
                            gridImageArray[ pos ] = ( 255 << 24 ) | ( 0 << 16 ) | ( 0 << 8 ) | iValue;
                        } else {
                            if ( max != min ) {
                                iValue = 255 * ( ( value - min ) / ( max - min ) );
                                gridImageArray[ pos ] = ( 255 << 24 ) | ( iValue << 16 ) | ( iValue << 8 ) | iValue;
                            } else {
                                // Set white
                                gridImageArray[ pos ] = ( 255 << 24 ) | ( iValue << 16 ) | ( iValue << 8 ) | iValue;
                            }
                        }
                    }
                }
            } else {
                //grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class
                Grid2DSquareCellDouble grid2DSquareCellDouble = ( Grid2DSquareCellDouble ) grid2DSquareCell;
                double noDataValue = grid2DSquareCellDouble.getNoDataValue();
                if (  Double.isInfinite( noDataValue ) ) {
                    System.out.println( "Warning!!! noDataValue = " + noDataValue + " in ESRIAsciigridExporter.toAsciiFile( Grid2DSquareCellAbstract( " + grid2DSquareCellDouble.toString() + " ), File( " + file.toString() + " ) )" );
                }
                double max = grid2DSquareCell.getGridStatistics().getMaxDouble();
                double min = grid2DSquareCell.getGridStatistics().getMinDouble();
                // Read all data into an array and scale all values into the range [ 0, 255 ]
                double value;
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        value = grid2DSquareCellDouble.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                        //pos = ( int ) ( ( row * ncols ) + col );
                        pos = ( int ) ( ( ( ( nrows - 1 ) - row ) * ncols ) + col );
                        //pos = ( int ) ( ( ( ( nrows - row ) - 1 ) * ncols ) + col - 1 );
                        // Construct an RGB integer by byte operation : 32 bytes, first 8 bytes is transparency value
                        // second, third, forth 8 byte is Red, green, blue value, which will be used by MemoryImageSource
                        // class.
                        iValue = 255;
                        // Set noDataValue as blue
                        if (  value == noDataValue ) {
                            gridImageArray[ pos ] = ( 255 << 24 ) | ( 0 << 16 ) | ( 0 << 8 ) | iValue;
                        } else {
                            if ( max != min ) {
                                iValue = new Double( 255.0d * ( value - min ) / ( max - min ) ).intValue();
                                gridImageArray[ pos ] = ( 255 << 24 ) | ( iValue << 16 ) | ( iValue << 8 ) | iValue;
                            } else {
                                // Set white
                                gridImageArray[ pos ] = ( 255 << 24 ) | ( iValue << 16 ) | ( iValue << 8 ) | iValue;
                            }
                        }
                    }
                }
            }
        }
        // Use gridImageArray to create a MemoryImageSource
        // Construct a BufferedImage using Toolkit
        // because Image class does not implement the RenderedImage
        // interface and imageio needs it.
        MemoryImageSource gridImageSource = new MemoryImageSource( ( int ) ncols, ( int ) nrows, gridImageArray, 0, ( int ) ncols );
        Image tempImage = Toolkit.getDefaultToolkit().createImage( gridImageSource );
        BufferedImage gridImage = new BufferedImage( ( int ) ncols, ( int ) nrows, BufferedImage.TYPE_INT_RGB );
        Graphics2D g = ( Graphics2D ) gridImage.getGraphics();
        g.drawImage( tempImage, 0, 0, new java.awt.Panel() );
        try {
            javax.imageio.ImageIO.write( gridImage, type, file );
        } catch( java.io.IOException e1 ) {
            e1.printStackTrace();
            System.out.println( "Warning!!! Failed to write grid as " + type + " to File(" + file.toString() + ")" );
        }
        g.dispose();
        gridImage.flush();
        tempImage.flush();
    }
}
