/**
 * Version 1.0 is for processing of 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.process;
// java dependencies
import java.awt.geom.Point2D;
import java.io.File;
import java.io.BufferedReader;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StreamTokenizer;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.Calendar;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDouble;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.CellID;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunk64CellMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkArrayFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkJAIFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunk64CellMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkArrayFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkJAIFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntFactory;
import uk.ac.leeds.ccg.andyt.grids.utilities.ErrorHandler;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;

/**
 * A class holding methods for processing an individual Grid2DSquareCellAbstract or
 * multiple Grid2DSquareCellAbstracts.
 * TODO:
 * Implement a general replace method ( depricate/replace mask methods ).
 */

public class Grid2DSquareCellProcessor extends ErrorHandler {
    
    /**
     * For storing the start time of the processing.
     */
    protected long startTime;
    
    /**
     * The log for recording progress and information about the processing.
     */
    protected PrintWriter log;
    
    /**
     * The log indentation (how many spaces before a log message line is output).
     */
    protected int logIndentation;
    
    /**
     * The workspace directory for the processing.
     */
    protected File workspace;
    
    /**
     * A collection of the Grid2dSquareCellAbstracts being processed
     */
    protected HashSet grid2DSquareCells;
    
    
    /** Creates a new instance of Grid2DSquareCellDoubleProcessor */
    public Grid2DSquareCellProcessor() {
        //this( FileCreator.createNewFile() );
        //this( FileCreator.createNewFile( new File( System.getProperty( "tmpdir" ) ) ) );
        this( FileCreator.createNewFile( new File( System.getProperty( "java.io.tmpdir" ) ) ) );
    }
    
    /**
     * Creates a new instance of Grid2DSquareCellDoubleProcessor.
     * By default the logs are appended to the end of the log file if it exists.
     * To overwrite the log file use:
     * Grid2DSquareCellDoubleProcessor( workspace, false );
     * @param workspace
     */
    public Grid2DSquareCellProcessor( File workspace ) {
        this( workspace, true );
    }
    
    /**
     * Creates a new instance of Grid2DSquareCellDoubleProcessor. The log file
     * in workspace will be overwritten if append is false.
     * @param workspace
     * @param append
     */
    public Grid2DSquareCellProcessor( File workspace, boolean append ) {
        initMemoryReserve();
        initWorkspace( workspace );
        this.workspace = workspace;
        try {
            this.log = new PrintWriter( new FileOutputStream( new File( workspace, "log.txt" ), append ), true );
        } catch ( FileNotFoundException fnfe0 ) {
            fnfe0.printStackTrace();
        }
        this.startTime = System.currentTimeMillis();
        this.logIndentation = 0;
        log( this.logIndentation, "log file " + workspace.toString() + File.separator + "log.txt set up " + Calendar.getInstance().getTime().toString(), this.handleOutOfMemoryErrorFalse );
        // Initialise Factories
        //        initFactories();
        this.grid2DSquareCells = new HashSet();
    }
    
    /**
     * Returns a copy of this.startTime.
     */
    public long getTime0() { 
        long result = this.startTime;
        return result;
    }
    
    /**
     * Returns a copy of this.workspace
     */
    public File getWorkspace() {
        return new File( this.workspace.toString() );
    }
    
    /**
     * Returns this.grid2DSquareCells
     */
    protected HashSet getGrid2DSquareCells() {
        return this.grid2DSquareCells;
    }
    
    /**
     * Initialises workspace. If it already exists then the log file is appended
     * to. Otherwise it is created and the log is set up. If the directory into
     * which workspace is being created is not writable this will fail.
     * @param workspace The workspace to initialise.
     */
    protected void initWorkspace( File workspace ) {
        try {
            File logFile;
            if ( workspace.exists() ) {
                logFile = new File( workspace, "log.txt" );
                if ( ! logFile.exists() ) {
                    logFile.createNewFile();
                }
            } else {
                workspace.mkdir();
                logFile = new File( workspace, "log.txt" );
                logFile.createNewFile();
            }
            this.log = new PrintWriter( new FileOutputStream( logFile, true ) );
        } catch ( IOException ioe0 ) {
            ioe0.printStackTrace();
        }
    }
    
    //    /**
    //     * Attempts to cache a seriailized version of a Chunk. Returns as an Object[]
    //     * a reference to the Grid2DSquareCellAbstract from which a Chunk was cached and
    //     * the ChunkID of that Chunk.
    //     */
    //    public final Object[] cacheChunk() {
    //        Object[] result = new Object[ 2 ];
    //        Iterator iterator = getGrid2DSquareCells().iterator();
    //        Grid2DSquareCellAbstract grid2DSquareCell;
    //        ChunkID cachedChunkID = null;
    //        while ( iterator.hasNext() ) {
    //            grid2DSquareCell = ( Grid2DSquareCellAbstract ) iterator.next();
    //            result[ 0 ] = grid2DSquareCell;
    //            cachedChunkID = grid2DSquareCell.cacheChunk();
    //            if ( cachedChunkID != null ) {
    //                result[ 1 ] = cachedChunkID;
    //                return result;
    //            }
    //        }
    //        return null;
    //    }
    
    /**
     * Attempts to: clear the memory reserve, cache a seriailized version of a Chunk
     * of a Grid2DSquareCellAbstract in this.grid2DSquareCells then clear it. Then
     * reinitialise the memory reserve. It either succeeds or throws an 
     * OutOfMemoryError.
     * TODO:
     * May want more sophistication as to what chunks are cached (what order etc...)
     */
    public final void freeMemoryOrThrowError() {
        clearMemoryReserve();
        Iterator iterator = getGrid2DSquareCells().iterator();
        Grid2DSquareCellAbstract grid2DSquareCell;
        while ( iterator.hasNext() ) {
            grid2DSquareCell = ( Grid2DSquareCellAbstract ) iterator.next();
            grid2DSquareCell.clearChunk( grid2DSquareCell.cacheChunk(false) );
            try {
                initMemoryReserve();
                return;
            } catch ( OutOfMemoryError oome0 ) {
            }
        }
        throw new OutOfMemoryError( "Unable to free up enough memory." );
    }
    
    /**
     * Changes workspace to that passed in if it can be created. By default this
     * does not copy the logfile from the existing workspace when it sets up a
     * log in the new location. To do this use:
     * setWorkspace( workspace, true )
     * @param workspace The workspace to change to.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setWorkspace( File workspace, boolean handleOutOfMemoryError ) {
        try {
            this.workspace = workspace;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                freeMemoryOrThrowError();
                setWorkspace( workspace, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Changes workspace to that passed in if it can be created. If copyLogFile
     * is true, this copies the logfile from the existing workspace and sets up
     * the log to append in the new location.
     * @param workspace The workspace to change to.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setWorkspace( File workspace, boolean copyLogFile, boolean handleOutOfMemoryError ) {
        try{
            boolean mkdirSuccess = false;
            File newLog = new File( workspace, "log.txt" );
            try {
                if ( ! workspace.exists() ) {
                    mkdirSuccess = workspace.mkdir();
                    if ( copyLogFile ) {
                        copyAndSetUpNewLog( newLog, this.handleOutOfMemoryErrorFalse );
                    } else {
                        newLog.createNewFile();
                        this.log = new PrintWriter( new FileOutputStream( newLog, true ) );
                    }
                } else {
                    if ( ! newLog.exists() ) {
                        newLog.createNewFile();
                        this.log = new PrintWriter( new FileOutputStream( newLog, true ) );
                    } else {
                        if ( copyLogFile ) {
                            this.log = new PrintWriter( new FileOutputStream( newLog, true ) );
                        } else {
                            this.log = new PrintWriter( new FileOutputStream( newLog, false ) );
                        }
                    }
                }
            } catch ( IOException ioe0 ) {
                ioe0.printStackTrace();
            }
            this.workspace = workspace;
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                freeMemoryOrThrowError();
                setWorkspace( workspace, copyLogFile, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Copies the log
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void copyAndSetUpNewLog( File newLog, boolean handleOutOfMemoryError ) {
        try {
            this.log.flush();
            this.log.close();
            File workspace = getWorkspace();
            File oldLog = new File( workspace, "log.txt" );
            BufferedInputStream bis = new BufferedInputStream( new FileInputStream( oldLog ) );
            BufferedOutputStream bos = new BufferedOutputStream( new FileOutputStream( newLog ) );
            for ( int i = 0; i < oldLog.length(); i ++ ) {
                bos.write( bis.read() );
            }
            bos.flush();
            bos.close();
            bis.close();
            this.log = new PrintWriter( new FileOutputStream( newLog, true ) );
            log( "log file copied from " + oldLog.toString() + " " + Calendar.getInstance().toString(), this.handleOutOfMemoryErrorFalse );
        } catch ( IOException ioe0 ) {
            ioe0.printStackTrace();
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                freeMemoryOrThrowError();
                copyAndSetUpNewLog( newLog, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Writes string to log file and the console (standard output)
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void log( String string, boolean handleOutOfMemoryError ) {
        // No OutOfMemoryError handling needed?
        log( this.logIndentation, string, handleOutOfMemoryError );
    }
    
    /**
     * Writes string to log file and the console (standard output) indenting
     * string by logIndentation amount of white-space.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void log( int logIndentation, String string, boolean handleOutOfMemoryError ) {
        try {
            if ( string.endsWith( "}" ) ) {
                logIndentation --;
                this.logIndentation --;
            }
            for ( int i = 0; i < logIndentation; i ++ ) {
                System.out.print( " " );
                this.log.write( " " );
            }
            System.out.println( string );
            this.log.write( string );
            this.log.write( System.getProperty( "line.separator" ) );
            this.log.flush();
            if ( string.endsWith( "{" ) ) {
                this.logIndentation ++;
            }
        } catch ( java.lang.OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                freeMemoryOrThrowError();
                log( logIndentation, string, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a Grid2DSquareCellDouble identical to grid2DSquareCell but with 
     * cells with values in range given by (minValueInRangeToBeReplaced, 
     * maxValueInRangeToBeReplaced) replaced with valueToInsert, boolean handleOutOfMemoryError ) {
     * coincident with grid2DSquareCellMask if their
     * centroids intersect a cell of mask that has a noDataValue.
     * @param grid2DSquareCellToReplaceValues The Grid2DSquareCellDouble to replace
     *   cells with value replaceValue.
     * @param mask The Grid2DSquareCellDouble mask.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     */
    public Grid2DSquareCellDouble replace( Grid2DSquareCellDouble grid2DSquareCell, long[] cellConstraints, BigDecimal[] dimensionsConstraints, double minValueInRangeToBeReplaced, double maxValueInRangeToBeReplaced, double valueToInsert, boolean handleOutOfMemoryError ) {
//                Grid2DSquareCellAbstract result;
//                int chunkNrows = grid2DSquareCellToReplaceFrom.getChunkNrows();
//                int chunkNcols = grid2DSquareCellToReplaceFrom.getChunkNcols();
    return null;
    }
    
   /**
     * Returns a Grid2DSquareCellAbstract identical to grid2DSquareCellToMask but
     * with cells coincident with grid2DSquareCellMask if their
     * centroids intersect a cell of mask that has a noDataValue.
     * @param grid2DSquareCellToReplaceValues The Grid2DSquareCellDouble to replace
     *   cells with value replaceValue.
     * @param mask The Grid2DSquareCellDouble mask.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     */
    public Grid2DSquareCellDouble mask( Grid2DSquareCellAbstract grid2DSquareCellToMask, Grid2DSquareCellAbstract grid2DSquareCellMask, boolean handleOutOfMemoryError ) {
                Grid2DSquareCellAbstract result;
                int chunkNrows = grid2DSquareCellToMask.getChunkNrows(this.handleOutOfMemoryErrorFalse);
                int chunkNcols = grid2DSquareCellToMask.getChunkNcols(this.handleOutOfMemoryErrorFalse);
//                try {
//                    if ( grid2DSquareCellToMask.getClass() == Grid2DSquareCellInt ) {
//                        Grid2DSquareCellIntFactory grid2DSquareCellIntFactory = new Grid2DSquareCellIntFactory( directory )
//                        result =
//                                long gridNrows = grid.getNrows();
//                        long gridNcols = grid.getNcols();
//                        int nChunkRows = grid.getNChunkRows();
//                        int nChunkCols = grid.getNChunkCols();
//                        double gridNoDataValue = grid.getNoDataValue();
//                        double maskNoDataValue = mask.getNoDataValue();
//                        BigDecimal[] gridDimensions = grid.getDimensions();
//                        BigDecimal[] maskDimensions = grid.getDimensions();
//                        boolean sameDimensions = true;
//                        for ( int i = 0; i < 3; i ++ ) {
//                            if ( maskDimensions[ i ].compareTo( gridDimensions[ i ] ) != 0 ) {
//                                //System.out.println( "maskBounds[" + i + "] " + maskBounds[ i ] );
//                                //System.out.println( "bounds[" + i + "] " + bounds[ i ] );
//                                sameDimensions = false;
//                            }
//                        }
//                        int chunkRowIndex;
//                        int chunkColIndex;
//                        int row;
//                        int col;
//                        double value;
//                        if ( ( gridNrows == mask.getNrows() ) && ( gridNcols == mask.getNcols() ) && sameDimensions ) {
//                            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
//                                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
//                                    Grid2DSquareCellDoubleChunkAbstract gridChunk = grid.getGrid2DSquareCellDoubleChunk( chunkRowIndex, chunkColIndex, this.handleOutOfMemoryErrorTrue );
//                                    Grid2DSquareCellDoubleChunkAbstract maskChunk = mask.getGrid2DSquareCellDoubleChunk( chunkRowIndex, chunkColIndex, this.handleOutOfMemoryErrorFalse );
//                                    for ( row = 0; row < chunkNrows; row ++ ) {
//                                        for ( col = 0; col < chunkNcols; col ++ ) {
//                                            value = mask.getCell( maskChunk, chunkRowIndex, chunkColIndex, row, col, this.handleOutOfMemoryErrorFalse );
//                                            if ( value == maskNoDataValue ) {
//                                                grid.setCell( gridChunk, chunkRowIndex, chunkColIndex, row, col, gridNoDataValue, this.handleOutOfMemoryErrorFalse );
//                                            }
//                                        }
//                                    }
//                                }
//                            }
//                        } else {
//                            //System.out.println( "maskCellsize " + maskCellsize );
//                            //System.out.println( "cellsize " + cellsize );
//                            double x;
//                            double y;
//                            for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
//                                for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
//                                    Grid2DSquareCellDoubleChunkAbstract gridChunk = grid.getGrid2DSquareCellDoubleChunk( chunkRowIndex, chunkColIndex, this.handleOutOfMemoryErrorFalse );
//                                    Grid2DSquareCellDoubleChunkAbstract maskChunk = mask.getGrid2DSquareCellDoubleChunk( chunkRowIndex, chunkColIndex, this.handleOutOfMemoryErrorFalse );
//                                    for ( row = 0; row < chunkNrows; row ++ ) {
//                                        for ( col = 0; col < chunkNcols; col ++ ) {
//                                            value = mask.getCell( maskChunk, chunkRowIndex, chunkColIndex, row, col, this.handleOutOfMemoryErrorFalse );
//                                            if ( value == maskNoDataValue ) {
//                                                grid.setCell( gridChunk, chunkRowIndex, chunkColIndex, row, col, gridNoDataValue, this.handleOutOfMemoryErrorFalse );
//                                            }
//                                        }
//                                    }
//                                }
//                            }
//                            System.out.println( "Warning: Spatial frame of grid is different to mask in Grid2DSquareCellDoubleProcessor.mask( grid( " + grid.toString() + " ), mask( " + mask.toString() + " ) )" );
//                        }
//                        //grid.setName( grid.getName() + "_mask" );
//                        return grid;
//                    } catch ( java.lang.OutOfMemoryError oome0 ) {
//                        if ( handleOutOfMemoryError ) {
//                            freeMemoryOrThrowError();
//                            log( logIndentation, string, handleOutOfMemoryError );
//                        } else {
//                            throw oome0;
//                        }
        return null;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * masked using mask.
     * @param grid the Grid2DSquareCellDouble to be masked
     * @param mask the Grid2DSquareCellDouble mask
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble mask( Grid2DSquareCellDouble grid, Grid2DSquareCellDouble mask, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        //result.setName( grid.getName() + "_mask" );
        return mask( result, mask, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with the values of cells in the range [min,max] set to its
     * noDataValue. (Existing noDataValue cells in grid remain as noDataValue.)
     * @param grid the Grid2DSquareCellDouble to be masked
     * @param min the minimum value in the range
     * @param max the maximum value in the range
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble mask( Grid2DSquareCellDouble grid, double min, double max, boolean handleOutOfMemoryError ) {
        double noDataValue = grid.getNoDataValue();
        double value;
        long ncols = grid.getNcols();
        long nrows = grid.getNrows();
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                value = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                if ( value >= min && value <= max ) {
                    grid.setCell( row, col, noDataValue, this.handleOutOfMemoryErrorTrue );
                }
            }
        }
        //grid.setName( grid.getName() + "_mask" );
        return grid;
    }
    
    /**
     * Returns grid with the value of cells that's centroids intersect the
     * rectangle given by (xmin,ymin,xmax,ymax) set to its noDataValue.
     * @param grid the Grid2DSquareCellDouble to be masked
     * @param xmin the minimum x-coordinate of the masking rectangle
     * @param ymin the minimum y-coordinate of the masking rectangle
     * @param xmax the maximum x-coordinate of the masking rectangle
     * @param ymax the maximum y-coordinate of the masking rectangle
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble mask( Grid2DSquareCellDouble grid, double xmin, double ymin, double xmax, double ymax, boolean handleOutOfMemoryError ) {
        return mask( grid, grid.getCellRowIndex( ymax, this.handleOutOfMemoryErrorTrue ), grid.getCellColIndex( xmin, this.handleOutOfMemoryErrorTrue ), grid.getCellRowIndex( ymin, this.handleOutOfMemoryErrorTrue ), grid.getCellColIndex( xmax, this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with the value of cells that intersect rectangle given by
     * (startRowIndex,startColIndex,endRowIndex,endColIndex) set to its
     * noDataValue.
     * @param grid the Grid2DSquareCellDouble to be masked
     * @param startRowIndex the index of the first row to be masked
     * @param startColIndex the index of the first column to be masked
     * @param endRowIndex the index of the final row to be masked
     * @param endColIndex the index of the final column to be masked
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble mask( Grid2DSquareCellDouble grid, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, boolean handleOutOfMemoryError ) {
        double noDataValue = grid.getNoDataValue();
        long rowIndex;
        long colIndex;
        for ( rowIndex = startRowIndex; rowIndex <= endRowIndex; rowIndex ++ ) {
            for ( colIndex = startColIndex; colIndex <= endColIndex; colIndex ++ ) {
                grid.setCell( rowIndex, colIndex, noDataValue, this.handleOutOfMemoryErrorTrue );
            }
        }
        //grid.setName( grid.getName() + "_mask" );
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with the value of cells in grid that intersect rectangle given by
     * (startRowIndex,startColIndex,endRowIndex,endColIndex) set to its
     * noDataValue.
     * @param grid The Grid2DSquareCellDouble to be masked.
     * @param startRowIndex The index of the first row to be masked.
     * @param startColIndex The index of the first column to be masked.
     * @param endRowIndex The index of the final row to be masked.
     * @param endColIndex The index of the final column to be masked.
     * @param gridFactory The Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble mask( Grid2DSquareCellDouble grid, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        //result.setName( grid.getName() + "_mask" );
        return mask( result, startRowIndex, startColIndex, endRowIndex, endColIndex, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with values linearly rescaled into the range [min,max].
     * @param grid the Grid2DSquareCellDouble to be processed.
     * @param min the minimum value in the rescaled range.
     * @param max the maximum value in the rescaled range.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble linearRescale( Grid2DSquareCellDouble grid, double min, double max, boolean handleOutOfMemoryError ) {
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        double noDataValue = grid.getNoDataValue();
        double thisMin = grid.getGridStatistics().getMinDouble();
        double thisMax = grid.getGridStatistics().getMaxDouble();
        double thisValue;
        long row;
        long col;
        // if range of either input or output is zero
        if ( ( thisMax - thisMin ) == 0.0d || ( max - min ) == 0.0d ) {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    thisValue = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( thisValue != noDataValue ) {
                        grid.setCell( row, col, min, this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        } else {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    thisValue = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( thisValue != noDataValue ) {
                        grid.setCell( row, col, ( ( ( ( thisValue - thisMin ) / ( thisMax - thisMin ) ) ) * ( max - min ) ) + min, this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        }
        //grid.setName( grid.getName() + "_linearRescale" );
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with values linearly rescaled into the range [min,max].
     * @param grid the Grid2DSquareCellDouble to be processed.
     * @param min the minimum value in the rescaled range.
     * @param max the maximum value in the rescaled range.
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble linearRescale( Grid2DSquareCellDouble grid, double min, double max, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        //result.setName( grid.getName() + "_linearRescale" );
        return linearRescale( result, min, max, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with values linearly scaled to range [1.0d,1000000.0d] then
     * logged using Math.log(double) and then linearly scaled into the range
     * [min,max]
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param min the minimum value in the rescaled range
     * @param max the maximum value in the rescaled range
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble logRescale( Grid2DSquareCellDouble grid, double min, double max, boolean handleOutOfMemoryError ) {
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        double noDataValue = grid.getNoDataValue();
        double thisMin = grid.getGridStatistics().getMinDouble();
        double thisMax = grid.getGridStatistics().getMaxDouble();
        linearRescale( grid, 1.0d, 1000000.0d, this.handleOutOfMemoryErrorTrue );
        double thisValue;
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                thisValue = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                if ( thisValue != noDataValue ) {
                    grid.setCell( row, col, Math.log( thisValue ), this.handleOutOfMemoryErrorTrue );
                }
            }
        }
        linearRescale( grid, min, max, this.handleOutOfMemoryErrorTrue );
        //grid.setName( grid.getName() + "_logRescale" );
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with values linearly scaled to range [1.0d,1000000.0d] then logged using
     * Math.log(double) and then linearly scaled into the range [min,max]
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param min the minimum value in the rescaled range
     * @param max the maximum value in the rescaled range
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble logRescale( Grid2DSquareCellDouble grid, double min, double max, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        //result.setName( grid.getName() + "_logRescale" );
        return logRescale( result, min, max, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with value of cells with keys in cellKeys set to a value a
     * little bit larger.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble setValueALittleBitLarger( Grid2DSquareCellDouble grid, HashSet cellIDs, boolean handleOutOfMemoryError ) {
        CellID cellID;
        double noDataValue = grid.getNoDataValue();
        Iterator iterator1 = cellIDs.iterator();
        double thisValue;
        int counter = 0;
        while ( iterator1.hasNext() ) {
            cellID = ( ( CellID ) iterator1.next() );
            thisValue = grid.getCell( cellID.getCellRowIndex(), cellID.getCellColIndex(), this.handleOutOfMemoryErrorTrue );
            if ( thisValue != noDataValue ) {
                grid.setCell( cellID, Utilities.getValueALittleBitLarger( thisValue ), this.handleOutOfMemoryErrorTrue );
            }
        }
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with value of cells with keys in cellKeys set to a value a little bit
     * larger.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble setValueALittleBitLarger( Grid2DSquareCellDouble grid, HashSet cellIDs, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        return setValueALittleBitLarger( result, cellIDs, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with value of cells with keys in cellKeys set to a value a
     * little bit smaller.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDKeys
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble setValueALittleBitSmaller( Grid2DSquareCellDouble grid, HashSet cellIDKeys, boolean handleOutOfMemoryError ) {
        CellID cellID;
        double noDataValue = grid.getNoDataValue();
        Iterator iterator1 = cellIDKeys.iterator();
        double thisValue;
        while ( iterator1.hasNext() ) {
            cellID = ( CellID ) iterator1.next();
            thisValue = grid.getCell( cellID.getCellRowIndex(), cellID.getCellColIndex(), this.handleOutOfMemoryErrorTrue );
            if ( thisValue != noDataValue ) {
                grid.setCell( cellID, Utilities.getValueALittleBitSmaller( thisValue ), this.handleOutOfMemoryErrorTrue );
            }
        }
        return grid;
    }
    
    /**
     * Returns a new Grid2DSquareCellDouble that is a copy of grid, but
     * with value of cells with keys in cellKeys set to a value a little bit
     * smaller.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble setValueALittleBitSmaller( Grid2DSquareCellDouble grid, HashSet cellIDs, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        return setValueALittleBitSmaller( result, cellIDs, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with value added to all cells with IDs as Integers in
     * cellKeys.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param value the value to be added
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, HashSet cellIDs, double value, boolean handleOutOfMemoryError ) {
        Iterator iterator1 = cellIDs.iterator();
        while ( iterator1.hasNext() ) {
            //grid.addToCell( ( CellID ) iterator1.next(), value );
            CellID cellID = ( CellID ) iterator1.next();
            if ( cellID != null ) {
                grid.addToCell( cellID, value, this.handleOutOfMemoryErrorTrue );
            }
        }
        return grid;
    }
    
    /**
     * Returns a copy of grid with value added to all cells with IDs as Integer
     * keys in cellIDKeys
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs
     * @param value the value to be added
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, HashSet cellIDs, double value, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        return addToGrid( result, cellIDs, value, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with value added to every cell via
     * AbstractGrid2DSquareCellDouble.addToCell(value)
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param value the value to be added
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, double value, boolean handleOutOfMemoryError ) {
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                grid.addToCell( row, col, value, this.handleOutOfMemoryErrorTrue );
            }
        }
        return grid;
    }
    
    /**
     * Returns grid with value added to every cell via
     * Grid2DSquareCellDouble.addToCell(value)
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param value the value to be added
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, CellID[] cellIDs, double value, boolean handleOutOfMemoryError ) {
        for ( int cellIDIndex = 0; cellIDIndex < cellIDs.length; cellIDIndex ++ ) {
            grid.addToCell( cellIDs[ cellIDIndex ].getCellRowIndex(), cellIDs[ cellIDIndex ].getCellColIndex(), value, this.handleOutOfMemoryErrorTrue );
        }
        return grid;
    }
    
    /**
     * Returns a copy of grid with value added to every cell via
     * AbstractGrid2DSquareCellDouble.addToCell(value)
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param value the value to be added
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to
     *   generate result.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, double value, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        return addToGrid( result, value, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns grid with values added from gridToAdd
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param gridToAdd the Grid2DSquareCellDouble from which values are
     *   added
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, Grid2DSquareCellDouble gridToAdd, boolean handleOutOfMemoryError ) {
        return addToGrid( grid, gridToAdd, 1.0d, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns a Grid2DSquareCellDouble with values of grid2DSquareCellDouble
     * added with values from grid2DSquareCellDoubleToAdd multiplied by weight.
     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to be processed.
     * @param grid2DSquareCellDoubleToAdd The Grid2DSquareCellDouble from which
     *   values are added.
     * @param weight The value grid2DSquareCellDoubleToAdd values are multiplied
     *   by.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid2DSquareCellDouble, Grid2DSquareCellDouble grid2DSquareCellDoubleToAdd, double weight, boolean handleOutOfMemoryError ) {
        return addToGrid( grid2DSquareCellDouble, grid2DSquareCellDoubleToAdd, 0L, 0L, grid2DSquareCellDoubleToAdd.getNrows() - 1L, grid2DSquareCellDoubleToAdd.getNcols() - 1L, weight, this.handleOutOfMemoryErrorTrue );
    }
    
    /**
     * Returns a Grid2DSquareCellDouble with values of grid2DSquareCellDouble
     * added with values from grid2DSquareCellDoubleToAdd (with row index between
     * startRowIndex, endRowIndex and column index between startColIndex,
     * endColIndex) multiplied by weight.
     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to be processed.
     * @param grid2DSquareCellDoubleToAdd The Grid2DSquareCellDouble from which
     *   values are added.
     * @param startRowIndex The index of the first row from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param startColIndex the index of the first column from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param endRowIndex the index of the final row from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param endColIndex the index of the final column from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param weight The value grid2DSquareCellDoubleToAdd values are multiplied
     *   by.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid2DSquareCellDouble, Grid2DSquareCellDouble grid2DSquareCellDoubleToAdd, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, double weight, boolean handleOutOfMemoryError ) {
        BigDecimal[] dimensions = grid2DSquareCellDoubleToAdd.getDimensions();
        BigDecimal[] dimensionConstraints = new BigDecimal[ 5 ];
        dimensionConstraints[ 1 ] = dimensions[ 1 ].add( new BigDecimal( startColIndex ).multiply( dimensions[ 0 ] ) );
        dimensionConstraints[ 2 ] = dimensions[ 2 ].add( new BigDecimal( startRowIndex ).multiply( dimensions[ 0 ] ) );
        dimensionConstraints[ 3 ] = dimensions[ 1 ].add( new BigDecimal( endColIndex - startColIndex + 1L ).multiply( dimensions[ 0 ] ) );
        dimensionConstraints[ 4 ] = dimensions[ 2 ].add( new BigDecimal( endRowIndex - startRowIndex + 1L ).multiply( dimensions[ 0 ] ) );
        return addToGrid( grid2DSquareCellDouble, grid2DSquareCellDoubleToAdd, startRowIndex, startColIndex, endRowIndex, endColIndex, dimensionConstraints, weight, this.handleOutOfMemoryErrorTrue );
    }
    
    //    /**
    //     * Returns grid with values added from gridToAdd for cells within bounds
    //     * given by (xmin,ymin,xmax,ymax) after they are multiplied by weight
    //     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to be processed.
    //     * @param grid2DSquareCellDoubleToAdd The Grid2DSquareCellDouble from which
    //     *   values are added.
    //     * @param dimensionContraints
    //     * @param weight The value grid2DSquareCellDoubleToAdd values are multiplied
    //     *   by.
    //     */
    //    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid2DSquareCellDouble, Grid2DSquareCellDouble grid2DSquareCellDoubleToAdd, BigDecimal[] dimensionConstraints, double weight ) {
    //        //        long startRowIndex = gridToAdd.getCellRowIndex( ymin );
    //        //        long startColIndex = gridToAdd.getCellColIndex( xmin );
    //        //        long endRowIndex = gridToAdd.getCellRowIndex( ymax );
    //        //        long endColIndex = gridToAdd.getCellColIndex( xmax );
    //        BigDecimal cellsizeDividedBy2BigDecimal =
    //        long startRowIndex = grid2DSquareCellDoubleToAdd.getCellRowIndex( dimensionConstraints[ 2 ] );
    //        long startColIndex = grid2DSquareCellDoubleToAdd.getCellColIndex( dimensionConstraints[ 1 ] );
    //        long endRowIndex = grid2DSquareCellDoubleToAdd.getCellRowIndex( dimensionConstraints[ 4 ] );
    //        long endColIndex = grid2DSquareCellDoubleToAdd.getCellColIndex( dimensionConstraints[ 3 ] );
    //        return addToGrid( grid2DSquareCellDouble, grid2DSquareCellDoubleToAdd, startRowIndex, startColIndex, endRowIndex, endColIndex, dimensionConstraints, weight );
    //    }
    
    /**
     * Returns a Grid2DSquareCellDouble with values of grid2DSquareCellDouble
     * added with values from grid2DSquareCellDoubleToAdd (with row index between
     * startRowIndex, endRowIndex and column index between startColIndex,
     * endColIndex) multiplied by weight.
     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to be processed.
     * @param grid2DSquareCellDoubleToAdd The Grid2DSquareCellDouble from which
     *   values are added.
     * @param startRowIndex The index of the first row from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param startColIndex the index of the first column from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param endRowIndex the index of the final row from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param endColIndex the index of the final column from which
     *   grid2DSquareCellDoubleToAdd values are added.
     * @param dimensionContraints
     * @param weight The value grid2DSquareCellDoubleToAdd values are multiplied
     *   by.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO: Check that reasonable answers are returned for intersections and aggregations.
     */
    protected Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid2DSquareCellDouble, Grid2DSquareCellDouble grid2DSquareCellDoubleToAdd, long startRowIndex, long startColIndex, long endRowIndex, long endColIndex, BigDecimal[] dimensionConstraints, double weight, boolean handleOutOfMemoryError ) {
        long nrows = grid2DSquareCellDouble.getNrows();
        long ncols = grid2DSquareCellDouble.getNcols();
        double noDataValue = grid2DSquareCellDouble.getNoDataValue();
        BigDecimal[] grid2DSquareCellDoubleDimensions = grid2DSquareCellDouble.getDimensions();
        
        long grid2DSquareCellDoubleToAddNrows = grid2DSquareCellDoubleToAdd.getNrows();
        long grid2DSquareCellDoubleToAddNcols = grid2DSquareCellDoubleToAdd.getNcols();
        double grid2DSquareCellDoubleToAddNoDataValue = grid2DSquareCellDoubleToAdd.getNoDataValue();
        BigDecimal[] grid2DSquareCellDoubleToAddDimensions = grid2DSquareCellDoubleToAdd.getDimensions();
        
        //Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory();
        Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory( grid2DSquareCellDouble.getDirectory() );
        // TODO:
        // Implement higher method to pass in a Grid2DSquareCellDoubleChunkFactory.
        //Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory( grid.getDirectory() );
        
        if ( ( dimensionConstraints[ 1 ].compareTo( grid2DSquareCellDoubleDimensions[ 3 ] ) == 1 ) ||
                ( dimensionConstraints[ 3 ].compareTo( grid2DSquareCellDoubleDimensions[ 1 ] ) == -1 ) ||
                ( dimensionConstraints[ 2 ].compareTo( grid2DSquareCellDoubleDimensions[ 4 ] ) == 1 ) ||
                ( dimensionConstraints[ 4 ].compareTo( grid2DSquareCellDoubleDimensions[ 2 ] ) == -1 ) ) {
            return grid2DSquareCellDouble;
        }
        if ( grid2DSquareCellDoubleToAddDimensions[ 0 ].compareTo( grid2DSquareCellDoubleDimensions[ 0 ] ) == -1 ) {
            // TODO:
            //            if ( startRowIndex != 0 || startColIndex != 0 || endRowIndex != grid2DSquareCellDoubleToAdd.getNrows() - 1 || endColIndex != grid2DSquareCellDoubleToAdd.getNcols() - 1 ) {
            //                System.out.println("Might be aggregating more than necessary!!!");
            //            }
            //            Grid2DSquareCellDouble aggregateGrid2DSquareCellDoubleToAddToAdd = null;
            //            try {
            //                aggregateGrid2DSquareCellDoubleToAddToAdd = aggregate( grid2DSquareCellDoubleToAdd, "sum", grid2DSquareCellDoubleDimensions, grid2DSquareCellDoubleFactory );
            //            } catch ( java.lang.OutOfMemoryError e ) {
            //                // TODO:
            //                // Clear some memory, cache, use chankRAF etc... try again. Perhaps fail and throw error to deal with higher up...
            //                System.out.println( "Warning!!! OutOfMemoryError in Grid2DSquareCellProcessor.addToGrid( grid ( ))" );
            //                int outOfMemoryErrorDebug0 = 1;
            //            }
            //            return addToGrid( grid2DSquareCellDouble, aggregateGrid2DSquareCellDoubleToAddToAdd, dimensionConstraints, weight );
            return null;
        } else {
            // If gridToAddCellsize is the same and the grids align
            if ( ( grid2DSquareCellDoubleToAddDimensions[ 0 ].compareTo( grid2DSquareCellDoubleDimensions[ 0 ] ) == 0 ) &&
                    ( ( grid2DSquareCellDoubleToAddDimensions[ 1 ].remainder( grid2DSquareCellDoubleDimensions[ 0 ] ) ).compareTo( ( grid2DSquareCellDoubleDimensions[ 1 ].remainder( grid2DSquareCellDoubleDimensions[ 0 ] ) ) ) == 0 ) &&
                    ( ( grid2DSquareCellDoubleToAddDimensions[ 2 ].remainder( grid2DSquareCellDoubleDimensions[ 0 ] ) ).compareTo( ( grid2DSquareCellDoubleDimensions[ 2 ].remainder( grid2DSquareCellDoubleDimensions[ 0 ] ) ) ) == 0 ) ) {
                //System.out.println( "Grids Align!" );
                double x;
                double y;
                double value;
                long row;
                long col;
                // TODO:
                // Make more robust if necessary using xBigDecimal and yBigDecimal
                // rather than using x and y?
                // The necessity can be calculated given the precision and size of
                // the double and the grid2DSquareCellDouble cellsize.
                for ( row = startRowIndex; row <= endRowIndex; row ++ ) {
                    for ( col = startColIndex; col <= endColIndex; col ++ ) {
                        x = grid2DSquareCellDoubleToAdd.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCellDoubleToAdd.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        value = grid2DSquareCellDoubleToAdd.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value != grid2DSquareCellDoubleToAddNoDataValue ) {
                            grid2DSquareCellDouble.addToCell( x, y, value * weight, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
                return grid2DSquareCellDouble;
            } else {
                // System.out.println("Intersection!!!!");
                // Need to intersect
                // TODO:
                // Clip grid2DSquareCellDoubleToAdd might improve matters here.
                // Check
                Grid2DSquareCellDouble tempGrid1 = null;
                Grid2DSquareCellDouble tempGrid2 = null;
                try {
                    tempGrid1 = grid2DSquareCellDoubleFactory.createGrid2DSquareCellDouble( nrows, ncols, grid2DSquareCellDoubleDimensions );
                    tempGrid2 = grid2DSquareCellDoubleFactory.createGrid2DSquareCellDouble( nrows, ncols, grid2DSquareCellDoubleDimensions );
                } catch ( java.lang.OutOfMemoryError e ) {
                    System.out.println( "Warning!!! OutOfMemoryError: Unable to create temprorary grids trying another way but may run out of filespace!" );
                    // TODO:
                    // Clear some memory, cache, use chankRAF etc... try again. Perhaps fail and throw error to deal with higher up...
                    System.out.println( "Warning!!! OutOfMemoryError in Grid2DSquareCellProcessor.addToGrid( grid ( ))" );
                    int outOfMemoryErrorDebug1 = 1;
                }
                // TODO:
                // Check scale and rounding appropriate
                int scale = 324;
                int roundingMode = BigDecimal.ROUND_HALF_EVEN;
                BigDecimal grid2DSquareCellDoubleToAddCellsizeDividedBy2 = grid2DSquareCellDoubleToAddDimensions[ 0 ].divide( new BigDecimal( "2" ), scale, roundingMode );
                BigDecimal grid2DSquareCellDoubleToAddCellsizeSquared = grid2DSquareCellDoubleToAddDimensions[ 0 ].multiply( grid2DSquareCellDoubleToAddDimensions[ 0 ] );
                double[] bounds = new double[ 4 ];
                CellID cellID1;
                CellID cellID2;
                CellID cellID3;
                CellID cellID4;
                double d1;
                double d2;
                double d3;
                double d4;
                double x;
                double y;
                long row;
                long col;
                double areaProportion = 0.0d;
                // TODO:
                // precision checking and use of BigDecimal?
                for ( row = 0; row < nrows; row ++ ) {
                    for ( col = 0; col < ncols; col ++ ) {
                        bounds = grid2DSquareCellDouble.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                        x = grid2DSquareCellDouble.getCellXDouble( col, this.handleOutOfMemoryErrorTrue );
                        y = grid2DSquareCellDouble.getCellYDouble( row, this.handleOutOfMemoryErrorTrue );
                        cellID1 = grid2DSquareCellDoubleToAdd.getCellID( bounds[ 0 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                        cellID2 = grid2DSquareCellDoubleToAdd.getCellID( bounds[ 2 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                        cellID3 = grid2DSquareCellDoubleToAdd.getCellID( bounds[ 0 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                        cellID4 = grid2DSquareCellDoubleToAdd.getCellID( bounds[ 2 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                        if ( cellID1.equals( cellID2 ) && cellID2.equals( cellID3 ) ) {
                            d1 = grid2DSquareCellDoubleToAdd.getCell( cellID1, this.handleOutOfMemoryErrorTrue );
                            if ( d1 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                areaProportion = ( grid2DSquareCellDoubleDimensions[ 0 ].multiply( grid2DSquareCellDoubleDimensions[ 0 ] ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue();
                                tempGrid1.addToCell( row, col, d1 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                            }
                        } else {
                            d1 = grid2DSquareCellDoubleToAdd.getCell( cellID1, this.handleOutOfMemoryErrorTrue );
                            d2 = grid2DSquareCellDoubleToAdd.getCell( cellID2, this.handleOutOfMemoryErrorTrue );
                            d3 = grid2DSquareCellDoubleToAdd.getCell( cellID3, this.handleOutOfMemoryErrorTrue );
                            d4 = grid2DSquareCellDoubleToAdd.getCell( cellID4, this.handleOutOfMemoryErrorTrue );
                            
                            if ( ! grid2DSquareCellDoubleToAdd.inGrid( cellID1, this.handleOutOfMemoryErrorTrue ) && d1 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                if ( cellID1.equals( cellID2 ) || cellID1.equals( cellID3 ) ) {
                                    if ( cellID1.equals( cellID2 ) ) {
                                        //areaProportion = ( Math.abs( bounds[3] - ( grid2DSquareCellDoubleToAdd.getCellY( cellID1 ) - grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) * cellsize ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize );
                                        areaProportion = Math.abs( ( ( BigDecimal.valueOf( bounds[3] ).subtract( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID1, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).multiply( grid2DSquareCellDoubleDimensions[ 0 ] ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    } else {
                                        //areaProportion = ( Math.abs( ( grid2DSquareCellDoubleToAdd.getCellX( cellID1 ) + grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) - bounds[0] ) * cellsize ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize );
                                        areaProportion = Math.abs( ( ( ( ( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID1, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[0] ) ) ).multiply( grid2DSquareCellDoubleDimensions[ 0 ] ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    }
                                } else {
                                    //areaProportion = ( ( Math.abs( bounds[3] - ( grid2DSquareCellDoubleToAdd.getCellY( cellID1 ) - grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) * Math.abs( ( grid2DSquareCellDoubleToAdd.getCellX( cellID1 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[0] ) ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize ) );
                                    areaProportion = Math.abs( ( ( BigDecimal.valueOf( bounds[3] ).subtract( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID1, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ).multiply( ( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID1, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2.subtract( BigDecimal.valueOf( bounds[ 0 ] ) ) ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ) ).doubleValue() );
                                }
                                tempGrid1.addToCell( row, col, d1 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                            }
                            if ( ! grid2DSquareCellDoubleToAdd.inGrid( cellID2, this.handleOutOfMemoryErrorTrue ) && d2 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                if ( cellID2.equals( cellID1 ) ) {
                                    if ( cellID2.equals( cellID4 ) ) {
                                        //areaProportion = ( Math.abs( bounds[2] - ( grid2DSquareCellDoubleToAdd.getCellX( cellID2 ) - grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) * cellsize ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize );
                                        areaProportion = Math.abs( ( ( ( BigDecimal.valueOf( bounds[ 2 ] ).subtract( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID2, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ).multiply( grid2DSquareCellDoubleDimensions[ 0 ] ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    } else {
                                        //areaProportion = ( ( Math.abs( bounds[3] - ( grid2DSquareCellDoubleToAdd.getCellY( cellID2 ) - grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) * Math.abs( bounds[2] - ( grid2DSquareCellDoubleToAdd.getCellX( cellID2 ) - ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) ) ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize ) );
                                        areaProportion = Math.abs( ( ( BigDecimal.valueOf( bounds[ 3 ] ).subtract( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID2, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ).multiply( BigDecimal.valueOf( bounds[ 2 ] ).subtract( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID2, this.handleOutOfMemoryErrorTrue ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    }
                                    tempGrid1.addToCell( row, col, d2 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                    tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                            if ( ! grid2DSquareCellDoubleToAdd.inGrid( cellID3, this.handleOutOfMemoryErrorTrue ) && d3 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                if ( ! cellID3.equals( cellID1 ) ) {
                                    if ( cellID3.equals( cellID4 ) ) {
                                        //areaProportion = ( Math.abs( ( grid2DSquareCellDoubleToAdd.getCellY( cellID3 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[1] ) * cellsize ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize );
                                        areaProportion = Math.abs( ( ( ( ( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID3, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[ 1 ] ) ) ).multiply( grid2DSquareCellDoubleDimensions[ 0 ] ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    } else {
                                        //areaProportion = ( ( Math.abs( ( grid2DSquareCellDoubleToAdd.getCellY( cellID3 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( grid2DSquareCellDoubleToAdd.getCellX( cellID3 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[0] ) ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize ) );
                                        areaProportion = Math.abs( ( ( ( ( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID3, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[ 1 ] ) ) ).multiply( ( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID3, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[ 0 ] ) ) ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    }
                                    tempGrid1.addToCell( row, col, d3 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                    tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                            if ( ! grid2DSquareCellDoubleToAdd.inGrid( cellID4, this.handleOutOfMemoryErrorTrue ) && d4 != grid2DSquareCellDoubleToAddNoDataValue ) {
                                if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                    //areaProportion = ( ( Math.abs( ( grid2DSquareCellDoubleToAdd.getCellY( cellID4 ) + ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( grid2DSquareCellDoubleToAdd.getCellX( cellID4 ) - ( grid2DSquareCellDoubleToAddCellsize / 2.0d ) ) ) ) / ( grid2DSquareCellDoubleToAddCellsize * grid2DSquareCellDoubleToAddCellsize ) );
                                    areaProportion = Math.abs( ( ( ( ( grid2DSquareCellDoubleToAdd.getCellYBigDecimal( cellID4, this.handleOutOfMemoryErrorTrue ).add( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ).subtract( BigDecimal.valueOf( bounds[ 1 ] ) ) ).multiply( BigDecimal.valueOf( bounds[ 2 ] ).subtract( ( grid2DSquareCellDoubleToAdd.getCellXBigDecimal( cellID4, this.handleOutOfMemoryErrorTrue ) ).subtract( grid2DSquareCellDoubleToAddCellsizeDividedBy2 ) ) ) ).divide( grid2DSquareCellDoubleToAddCellsizeSquared, scale, roundingMode ) ).doubleValue() );
                                    tempGrid1.addToCell( row, col, d4 * areaProportion, this.handleOutOfMemoryErrorTrue );
                                    tempGrid2.addToCell( row, col, areaProportion, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                            //// Check
                            // Check fails due to rounding errors!
                            //if ( cellID1 != Integer.MIN_VALUE && cellID2 != Integer.MIN_VALUE && cellID3 != Integer.MIN_VALUE && cellID4 != Integer.MIN_VALUE && totalArea != 1.0 ) { System.out.println( "id = " + i + " : totalArea = " + totalArea + " (cellID1,cellID2,cellID3,cellID4) = (" + cellID1 + "," + cellID2 + "," + cellID3 + "," + cellID4 + ")" ); System.exit(0);}
                        }
                    }
                }
                System.out.println( "tempGrid1" );
                System.out.println( tempGrid1.toString() );
                System.out.println( "tempGrid2" );
                System.out.println( tempGrid2.toString() );
                // The values are normalised by dividing the aggregate Grid sum by the proportion of cells with grid values.
                for ( row = 0; row <= nrows; row ++ ) {
                    for ( col = 0; col <= ncols; col ++ ) {
                        d1 = tempGrid2.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( ! ( d1 == 0.0d || d1 == noDataValue ) ) {
                            //setCell( i, tempGrid2.getCell( i ) );
                            grid2DSquareCellDouble.addToCell( row, col, weight * tempGrid1.getCell( row, col, this.handleOutOfMemoryErrorTrue ) / d1, this.handleOutOfMemoryErrorTrue );
                            //addToCell( i, tempGrid1.getCell( i ) / ( Math.pow( ( gridToAddCellsize / cellsize ), 2.0d ) / d1 ) );
                        }
                    }
                }
                tempGrid1.clear();
                tempGrid2.clear();
            }
        }
        return grid2DSquareCellDouble;
    }
    
    /**
     * Returns grid with values added from a file.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param file the file contining values to be added.
     * @param type the type of file.
     *   Supported types include "xyv", "xy", "idxy" )
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble addToGrid( Grid2DSquareCellDouble grid, File file, String type, boolean handleOutOfMemoryError ) {
        if ( type.equalsIgnoreCase( "xyv" ) ) {
            try {
                StreamTokenizer st =
                        new StreamTokenizer(
                        new BufferedReader(
                        new InputStreamReader(
                        new FileInputStream( file ) ) ) );
                st.eolIsSignificant( false );
                st.parseNumbers();
                st.whitespaceChars( ',', ',' );
                st.wordChars( '"', '"' );
                int tokenType = st.nextToken();
                String alternator = new String( "x" );
                double x = 0.0d;
                double y = 0.0d;
                double value = 0.0d;
                while ( tokenType != StreamTokenizer.TT_EOF ) {
                    switch ( tokenType ) {
                        case StreamTokenizer.TT_NUMBER :
                            if ( alternator.equals( "x" ) ) {
                                x = st.nval;
                                alternator = "y";
                            } else if ( alternator.equals( "y" ) ) {
                                y = st.nval;
                                alternator = "value";
                            } else {
                                grid.addToCell( x, y, st.nval, this.handleOutOfMemoryErrorTrue );
                                alternator = "x";
                            }
                            break;
                        default :
                            break;
                    }
                    tokenType = st.nextToken();
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() );
                System.out.println( "in uk.ac.leeds.ccg.grids.AbstractGrid2DSquareCellDouble.addToGrid( file( " + file.toString() + "), type( " + type + " ) )");
                e.printStackTrace();
            }
        }
        if ( type.equalsIgnoreCase( "xy" ) ) {
            try {
                StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( file ) ) ) );
                st.eolIsSignificant( false );
                st.parseNumbers();
                st.whitespaceChars( ',', ',' );
                st.wordChars( '"', '"' );
                int tokenType = st.nextToken();
                String alternator = new String( "x" );
                double x = 0.0d;
                double y = 0.0d;
                double value = 0.0d;
                while ( tokenType != StreamTokenizer.TT_EOF ) {
                    switch ( tokenType ) {
                        case StreamTokenizer.TT_NUMBER :
                            if ( alternator.equals( "x" ) ) {
                                x = st.nval;
                                alternator = "y";
                            } else {
                                y = st.nval;
                                grid.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                                alternator = "x";
                            }
                            break;
                        default :
                            break;
                    }
                    tokenType = st.nextToken();
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() );
                System.out.println( "in uk.ac.leeds.ccg.grids.AbstractGrid2DSquareCellDouble.addToGrid( file( " + file.toString() + "), type( " + type + " ) )");
                e.printStackTrace();
            }
        }
        if ( type.equalsIgnoreCase( "idxy" ) ) {
            try {
                StreamTokenizer st = new StreamTokenizer( new BufferedReader( new InputStreamReader( new FileInputStream( file ) ) ) );
                st.eolIsSignificant( false );
                st.parseNumbers();
                st.ordinaryChar( 'e' );
                st.ordinaryChar( 'd' );
                st.ordinaryChar( 'E' );
                st.ordinaryChar( 'D' );
                int tokenType = st.nextToken();
                int nextTokenType;
                String alternator = new String( "id" );
                int id;
                double x = 0.0d;
                double y = 0.0d;
                double value = 0.0d;
                while ( tokenType != StreamTokenizer.TT_EOF ) {
                    switch ( tokenType ) {
                        case StreamTokenizer.TT_NUMBER :
                            if ( alternator.equals( "id" ) ) {
                                id = ( int ) st.nval;
                                alternator = "x";
                            } else {
                                if ( alternator.equals( "x" ) ) {
                                    x = st.nval;
                                    nextTokenType = st.nextToken();
                                    if ( nextTokenType != StreamTokenizer.TT_NUMBER ) {
                                        nextTokenType = st.nextToken();
                                        nextTokenType = st.nextToken();
                                        x = x * Math.pow( 10.0, st.nval );
                                    } else {
                                        st.pushBack();
                                    }
                                    alternator = "y";
                                } else {
                                    y = st.nval;
                                    nextTokenType = st.nextToken();
                                    if ( nextTokenType != StreamTokenizer.TT_NUMBER ) {
                                        nextTokenType = st.nextToken();
                                        nextTokenType = st.nextToken();
                                        y = y * Math.pow( 10.0, st.nval );
                                    } else {
                                        st.pushBack();
                                    }
                                    alternator = "id";
                                    //System.out.println( " x, y = " + x + ", " + y );
                                    grid.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                                }
                            }
                            break;
                        default :
                            break;
                    }
                    tokenType = st.nextToken();
                }
            } catch ( java.io.IOException e ) {
                System.out.println( e.toString() );
                System.out.println( "in uk.ac.leeds.ccg.grids.AbstractGrid2DSquareCellDouble.addToGrid( file( " + file.toString() + "), type( " + type + " ) )");
                e.printStackTrace();
            }
        }
        return grid;
    }
    
    /**
     * TODO:
     * 1. Documentiation
     * 2. Change so Grid2DSquareCellDouble are Grid2DSquareCellAbstract
     * Returns a Grid2DSquareCellDouble that's values are those of grid0 minus
     * grid1.
     * @param data controls if result is produced even if one of the inputs is a
     *   NoData value
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble minus( Grid2DSquareCellDouble grid0, Grid2DSquareCellDouble grid1, boolean data, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        double grid0NoDataValue = grid0.getNoDataValue();
        double grid1NoDataValue = grid1.getNoDataValue();
        long nrows = grid0.getNrows();
        long ncols = grid0.getNcols();
        BigDecimal[] dimensions0 = grid0.getDimensions();
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, grid0NoDataValue );
        double value0;
        double value1;
        long row;
        long col;
        if ( data ) {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value0 == grid0NoDataValue ) {
                        value0 = 0.0d;
                        value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value1 != grid1NoDataValue ) {
                            result.setCell( row, col, value0 - value1, this.handleOutOfMemoryErrorTrue );
                        }
                    } else {
                        value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value1 == grid1NoDataValue ) {
                            value1 = 0.0d;
                        }
                        result.setCell( row, col, value0 - value1, this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        } else {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value0 != grid0NoDataValue && value1 != grid1NoDataValue ) {
                        result.setCell( row, col, value0 - value1, this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        }
        //result.setName( "minus_" + grid0.getName( this.handleOutOfMemoryErrorTrue ) + "_" + grid1.getName( this.handleOutOfMemoryErrorTrue ) );
        return result;
    }
    
    /**
     * TODO:
     * 1. Documentiation
     * 2. Change so Grid2DSquareCellDouble are Grid2DSquareCellAbstract
     * @param data controls if result is produced even if one of the inputs is a
     *   NoData value
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @see Grid2DSquareCellProcessor#addToGrid(Grid2DSquareCellDouble,Grid2DSquareCellDouble,boolean)
     */
    public Grid2DSquareCellDouble add( Grid2DSquareCellDouble grid0, Grid2DSquareCellDouble grid1, boolean data, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        double grid0NoDataValue = grid0.getNoDataValue();
        double grid1NoDataValue = grid1.getNoDataValue();
        long nrows = grid0.getNrows();
        long ncols = grid0.getNcols();
        BigDecimal[] dimensions0 = grid0.getDimensions();
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, grid0NoDataValue );
        long row;
        long col;
        double value0;
        double value1;
        if ( data ) {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value0 == grid0NoDataValue ) {
                        value0 = 0.0d;
                        value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value1 != grid1NoDataValue ) {
                            result.setCell( row, col, value0 + value1, this.handleOutOfMemoryErrorTrue );
                        }
                    } else {
                        value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value1 == grid1NoDataValue ) {
                            value1 = 0.0d;
                        }
                        result.setCell( row, col, value0 + value1, this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        } else {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value0 != grid0NoDataValue && value1 != grid1NoDataValue ) {
                        result.setCell( row, col, value0 + value1, this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        }
        //result.setName( "add_" + grid0.getName( this.handleOutOfMemoryErrorTrue ) + "_" + grid1.getName( this.handleOutOfMemoryErrorTrue ) );
        return result;
    }
    
    
    /**
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * 1. Documentiation
     * 2. Change so Grid2DSquareCellDouble are Grid2DSquareCellAbstract
     */
    public Grid2DSquareCellDouble multiply( Grid2DSquareCellDouble grid0, Grid2DSquareCellDouble grid1, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        double grid0NoDataValue = grid0.getNoDataValue();
        double grid1NoDataValue = grid1.getNoDataValue();
        long nrows = grid0.getNrows();
        long ncols = grid0.getNcols();
        BigDecimal[] dimensions0 = grid0.getDimensions();
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, grid0NoDataValue );
        double value0;
        double value1;
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                result.setCell( row, col, value0 * value1, this.handleOutOfMemoryErrorTrue );
            }
        }
        //result.setName( "multiply_" + grid0.getName( this.handleOutOfMemoryErrorTrue ) + "_" + grid1.getName( this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
        return result;
    }
    
    /**
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * 1. Documentiation
     * 2. Change so Grid2DSquareCellDouble are Grid2DSquareCellAbstract
     */
    public Grid2DSquareCellDouble divide( Grid2DSquareCellDouble grid0, Grid2DSquareCellDouble grid1, Grid2DSquareCellDoubleFactory gridFactory, double defaultValue, boolean handleOutOfMemoryError ) {
        double grid0NoDataValue = grid0.getNoDataValue();
        double grid1NoDataValue = grid1.getNoDataValue();
        long nrows = grid0.getNrows();
        long ncols = grid0.getNcols();
        BigDecimal[] dimensions0 = grid0.getDimensions();
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, grid0NoDataValue );
        double value0;
        double value1;
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                value0 = grid0.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                value1 = grid1.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                result.setCell( row, col, value0 / value1, this.handleOutOfMemoryErrorTrue );
            }
        }
        //result.setName( "divide_" + grid0.getName() + "_" + grid1.getName() );
        return result;
    }
    
    //    /**
    //     * Returns an Grid2DSquareCellDouble at a lower level of resolution than grid.  The result values
    //     * are either the sum, mean, max or min of values in grid depending on statistic.
    //     * xllcorner and yllcorner are the same as grid.
    //     * @param grid - the Grid2DSquareCellDouble to be processed
    //     * @param cellfactor - the number of times wider/higher the aggregated grid cells are
    //     * @param statistic - "sum", "mean", "max", or "min" depending on what aggregate of values are wanted
    //     * NB. In the calculation of the sum and the mean if there is a cell in grid which has a data value then
    //     *     the result which incorporates that cell has a data value.  For this result cell, any of the cells in
    //     *     grid which have noDataValues their value is taken as that of the average of its nearest cells with
    //     *     a value.
    //     *     In the calculation of the max and the min noDataValues are simply ignored.
    //     *     Formerly noDataValues were treated as the average of values within a result cell.
    //     * TODO:
    //     * 1. Implement median, mode and variance aggregations.
    //     */
    //    public Grid2DSquareCellDouble aggregate( Grid2DSquareCellDouble grid, int cellFactor, String statistic ) {
    //        try {
    //            return aggregate( grid, cellFactor, statistic, 0, 0, new Grid2DSquareCellDoubleJAIFactory() );
    //        } catch ( java.lang.OutOfMemoryError e ) {
    //            return aggregate( grid, cellFactor, statistic, 0, 0, new Grid2DSquareCellDoubleFileFactory() );
    //        }
    //    }
    
    /**
     * Returns an Grid2DSquareCellDouble at a lower level of resolution
     * than grid.  The result values are either the sum, mean, max or min of
     * values in grid depending on statistic.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellFactor the number of times wider/higher the aggregated grid
     *                   cells are to be
     * @param statistic "sum", "mean", "max", or "min" depending on what
     *                   aggregate of values are wanted
     * @param rowOffset the number of rows above or below the origin of grid
     *                  where the aggregation is to start.
     *                  > 0 result yllcorner will be above grid yllcorner
     *                  < 0 result yllcorner will be below grid yllcorner
     * @param colOffset the number of columns above or below the origin of grid
     *                  where the aggregation is to start.
     *                  > 0 result xllcorner will be right of grid xllcorner
     *                  < 0 result xllcorner will be left of grid xllcorner
     * @param gridFactory the Abstract2DSquareCellDoubleFactory used to create
     *                    result and temporary AbstractGrid2DSquareCellDoubles.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * NB. In the calculation of the sum and the mean if there is a cell in grid which has a data value then
     *     the result which incorporates that cell has a data value.  For this result cell, any of the cells in
     *     grid which have noDataValues their value is taken as that of the average of its nearest cells with
     *     a value.
     *     In the calculation of the max and the min noDataValues are simply ignored.
     *     Formerly noDataValues were treated as the average of values within a result cell.
     * TODO: implement median, mode and variance aggregations.
     */
    public Grid2DSquareCellDouble aggregate( Grid2DSquareCellDouble grid, int cellFactor, String statistic, int rowOffset, int colOffset, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        // Initial tests
        if ( cellFactor <= 0 ) {
            System.out.println( "Warning!!! cellFactor <= 0 : Returning null." );
            return null;
        }
        // Initialisation
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        BigDecimal[] dimensions = grid.getDimensions();
        double noDataValue = grid.getNoDataValue();
        BigDecimal[] resultDimensions = new BigDecimal[ 5 ];
        resultDimensions[0] = dimensions[0].multiply( new BigDecimal( Integer.toString( cellFactor ) ) );
        resultDimensions[1] = dimensions[1].add( dimensions[0].multiply( new BigDecimal( Integer.toString( colOffset ) ) ) );
        resultDimensions[2] = dimensions[2].add( dimensions[0].multiply( new BigDecimal( Integer.toString( rowOffset ) ) ) );
        //double resultCellsize = cellsize * ( double ) cellFactor;
        //double width = cellsize * ncols;
        //double height = cellsize * nrows;
        //double resultXllcorner = xllcorner + ( colOffset * cellsize );
        //double resultYllcorner = yllcorner + ( rowOffset * cellsize );
        // Calculate resultNrows and resultHeight
        long resultNrows = 1L;
        BigDecimal resultHeight = new BigDecimal( resultDimensions[0].toString() );
        //double resultHeight = resultCellsize;
        while ( resultDimensions[ 2 ].add( resultHeight ).compareTo( dimensions[ 4 ] ) == -1 ) {
            resultNrows ++;
            resultHeight = resultHeight.add( resultDimensions[0] );
        }
        //while ( ( resultYllcorner + resultHeight ) < ( yllcorner + height ) ) {
        //    resultNrows ++;
        //    resultHeight += resultCellsize;
        //}
        // Calculate resultNcols and resultWidth
        long resultNcols = 1L;
        BigDecimal resultWidth = new BigDecimal( resultDimensions[0].toString() );
        //double resultWidth = resultCellsize;
        while ( resultDimensions[ 1 ].add( resultWidth ).compareTo( dimensions[ 3 ] ) == -1 ) {
            resultNrows ++;
            resultWidth = resultWidth.add( resultDimensions[0] );
        }
        //while ( ( resultXllcorner + resultWidth ) < ( xllcorner + width ) ) {
        //    resultNcols ++;
        //    resultWidth += resultCellsize;
        //}
        resultDimensions[3] = dimensions[1].add( resultWidth );
        resultDimensions[4] = dimensions[2].add( resultHeight );
        // Initialise result
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultDimensions, noDataValue );
        
        long row;
        long col;
        double x;
        double y;
        double value;
        
        // sum
        if ( statistic.equalsIgnoreCase( "sum" ) ) {
            Grid2DSquareCellDouble dataCount = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultDimensions, noDataValue );
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    x = grid.getCellX( col, this.handleOutOfMemoryErrorTrue );
                    y = grid.getCellY( row, this.handleOutOfMemoryErrorTrue );
                    if ( result.inGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                        value = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            dataCount.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
            // Add the nearest values for the noDataValues so long as there is a value
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    x = grid.getCellX( col, this.handleOutOfMemoryErrorTrue );
                    y = grid.getCellY( row, this.handleOutOfMemoryErrorTrue );
                    if ( result.inGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                        if ( dataCount.getCell( x, y, this.handleOutOfMemoryErrorTrue ) != noDataValue ) {
                            result.addToCell( x, y, grid.getNearestValueDouble( row, col, this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
        }
        
        // mean
        if ( statistic.equalsIgnoreCase( "mean" ) ) {
            Grid2DSquareCellDouble numerator = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultDimensions, noDataValue );
            Grid2DSquareCellDouble denominator = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultDimensions, noDataValue );
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    x = grid.getCellX( col, this.handleOutOfMemoryErrorTrue );
                    y = grid.getCellY( row, this.handleOutOfMemoryErrorTrue );
                    if ( result.inGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                        value = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            numerator.addToCell( x, y, value, this.handleOutOfMemoryErrorTrue );
                            denominator.addToCell( x, y, 1.0d, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
            for ( row = 0; row < resultNrows; row ++ ) {
                for ( col = 0; col < resultNcols; col ++ ) {
                    value = numerator.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        result.setCell( row, col, value / denominator.getCell( row, col, this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
                    }
                }
            }
        }
        
        // min
        if ( statistic.equalsIgnoreCase( "min" ) ) {
            double min;
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    x = grid.getCellX( col, this.handleOutOfMemoryErrorTrue );
                    y = grid.getCellY( row, this.handleOutOfMemoryErrorTrue );
                    if ( result.inGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                        value = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            min = result.getCell( x, y, this.handleOutOfMemoryErrorTrue );
                            if ( min != noDataValue ) {
                                result.setCell( x, y, Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                            } else {
                                result.setCell( x, y, value, this.handleOutOfMemoryErrorTrue );
                            }
                        }
                    }
                }
            }
        }
        
        // max
        if ( statistic.equalsIgnoreCase( "max" ) ) {
            double max;
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    x = grid.getCellX( col, this.handleOutOfMemoryErrorTrue );
                    y = grid.getCellY( row, this.handleOutOfMemoryErrorTrue );
                    if ( result.inGrid( x, y, this.handleOutOfMemoryErrorTrue ) ) {
                        value = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            max = result.getCell( x, y, this.handleOutOfMemoryErrorTrue );
                            if ( max != noDataValue ) {
                                result.setCell( x, y, Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                            } else {
                                result.setCell( x, y, value, this.handleOutOfMemoryErrorTrue );
                            }
                        }
                    }
                }
            }
        }
        
        return result;
    }
    
    //    /**
    //     * Returns an Grid2DSquareCellDouble at a lower level of resolution than grid.  The result values
    //     * are either the sum, mean, max or min of values in grid depending on statistic.
    //     * @param grid - the Grid2DSquareCellDouble to be processed
    //     * @param resultCellsize - output grid cellsize
    //     * @param statistic - "sum", "mean", "max", or "min" depending on what aggregate of values are wanted
    //     * @param resultXllcorner - the x-coordinate of the aggregate grid lower left corner
    //     * @param resultYllcorner - the y-coordinate of the aggregate grid lower left corner
    //     * Use this aggregate method to force origin of the result to be ( resultXllcorner, resultYllcorner ) and
    //     * if resultCellsize is not an integer multiple of this.cellsize.
    //     * NB. In the calculation of the sum and the mean if there is a cell in grid which has a data value then
    //     *     the result which incorporates that cell has a data value.  For this result cell, any of the cells in
    //     *     grid which have noDataValues their value is taken as that of the average of its nearest cells with
    //     *     a value.
    //     *     In the calculation of the max and the min noDataValues are simply ignored.
    //     *     Formerly noDataValues were treated as the average of values within a result cell.
    //     * TODO: implement median, mode and variance aggregations.
    //     */
    //    public Grid2DSquareCellDouble aggregate( Grid2DSquareCellDouble grid, double resultCellsize, String statistic, double resultXllcorner, double resultYllcorner ) {
    //        try {
    //            return aggregate( grid, resultCellsize, statistic, resultXllcorner, resultYllcorner, new Grid2DSquareCellDoubleFactory() );
    //        } catch ( java.lang.OutOfMemoryError e ) {
    //            return aggregate( grid, resultCellsize, statistic, resultXllcorner, resultYllcorner, new Grid2DSquareCellDoubleFileFactory() );
    //        }
    //    }
    
    /**
     * Returns an Grid2DSquareCellDouble at a lower level of resolution than grid.  The result values
     * are either the sum, mean, max or min of values in grid depending on statistic.
     * @param grid The Grid2DSquareCellDouble to be processed
     * @param statistic "sum", "mean", "max", or "min" depending on what aggregate of values are wanted
     * @param resultDimensions
     * @param gridFactory The Abstract2DSquareCellDoubleFactory used to create grids
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * Use this aggregate method if result is to have a new spatial frame.
     * NB. In the calculation of the sum and the mean if there is a cell in grid which has a data value then
     *     the result which incorporates that cell has a data value.  For this result cell, any of the cells in
     *     grid which have noDataValues their value is taken as that of the average of its nearest cells with
     *     a value.
     *     In the calculation of the max and the min noDataValues are simply ignored.
     *     Formerly noDataValues were treated as the average of values within a result cell.
     * TODO: implement median, mode and variance aggregations.
     */
    public Grid2DSquareCellDouble aggregate( Grid2DSquareCellDouble grid, String statistic, BigDecimal[] resultDimensions, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        int scale = 325;
        // Initialistaion
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        BigDecimal[] dimensions = grid.getDimensions();
        double noDataValue = grid.getNoDataValue();
        double cellsize = resultDimensions[0].doubleValue();
        
        //double width = cellsize * ncols;
        //double height = cellsize * nrows;
        
        // Test this is an aggregation
        if ( resultDimensions[0].compareTo( dimensions[ 0 ] ) != 1 ) {
            System.out.println( "!!!Warning: Not an aggregation as resultCellsize < cellsize. Returning null!" );
            return null;
        }
        // Test for intersection
        if ( ( resultDimensions[1].compareTo( dimensions[1].add( dimensions[ 0 ].multiply( new BigDecimal( Long.toString( ncols ) ) ) ) ) == 1 )
        || ( resultDimensions[2].compareTo( dimensions[2].add( dimensions[ 0 ].multiply( new BigDecimal( Long.toString( nrows ) ) ) ) ) == 1 ) ) {
            System.out.println( "!!!Warning: No intersection for aggregation. Returning null!" );
            return null;
        }
        // If resultCellsize is an integer multiple of cellsize and grid aligns with result then use
        // a cellFactor aggregation as it should be faster.
        //System.out.println("resultCellsize % cellsize == " + ( resultCellsize % cellsize ) );
        //System.out.println("resultXllcorner % cellsize = " + ( resultXllcorner % cellsize ) + ", xllcorner % cellsize = " + ( xllcorner % cellsize ) );
        //System.out.println("resultYllcorner % cellsize = " + ( resultYllcorner % cellsize ) + ", yllcorner % cellsize = " + ( yllcorner % cellsize ) );
        if ( true ) {
            BigDecimal t0 = resultDimensions[0].divide( dimensions[0], Math.max( resultDimensions[0].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            BigDecimal t1 = resultDimensions[1].divide( dimensions[0], Math.max( resultDimensions[1].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            BigDecimal t2 = dimensions[1].divide( dimensions[0], Math.max( dimensions[1].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            BigDecimal t3 = resultDimensions[2].divide( dimensions[0], Math.max( resultDimensions[2].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            BigDecimal t4 = dimensions[2].divide( dimensions[0], Math.max( dimensions[2].scale(), dimensions[0].scale() ) + 2, BigDecimal.ROUND_HALF_EVEN );
            if (
                    ( t0.compareTo( new BigDecimal( t0.toBigInteger().toString() ) ) == 0 ) &&
                    ( t1.compareTo( new BigDecimal( t1.toBigInteger().toString() ) ) == t2.compareTo( new BigDecimal( t2.toBigInteger().toString() ) ) ) &&
                    ( t3.compareTo( new BigDecimal( t3.toBigInteger().toString() ) ) == t4.compareTo( new BigDecimal( t4.toBigInteger().toString() ) ) ) ) {
                int cellFactor = resultDimensions[0].divide( dimensions[0], 2, BigDecimal.ROUND_UNNECESSARY ).intValue();
                int rowOffset = dimensions[2].subtract( resultDimensions[2].divide( dimensions[ 0 ], scale, BigDecimal.ROUND_HALF_EVEN ) ).intValue();
                int colOffset = dimensions[1].subtract( resultDimensions[1].divide( dimensions[ 0 ], scale, BigDecimal.ROUND_HALF_EVEN ) ).intValue();
                return aggregate( grid, cellFactor, statistic, rowOffset, colOffset, gridFactory, this.handleOutOfMemoryErrorTrue );
            }
        }
        // Calculate resultNrows and resultHeight
        long resultNrows = 1L;
        BigDecimal resultHeight = new BigDecimal( resultDimensions[0].toString() );
        //double resultHeight = resultCellsize;
        while ( resultDimensions[ 2 ].add( resultHeight ).compareTo( dimensions[ 4 ] ) == -1 ) {
            resultNrows ++;
            resultHeight = resultHeight.add( resultDimensions[0] );
        }
        //while ( ( resultYllcorner + resultHeight ) < ( yllcorner + height ) ) {
        //    resultNrows ++;
        //    resultHeight += resultCellsize;
        //}
        // Calculate resultNcols and resultWidth
        long resultNcols = 1L;
        BigDecimal resultWidth = new BigDecimal( resultDimensions[0].toString() );
        //double resultWidth = resultCellsize;
        while ( resultDimensions[ 1 ].add( resultWidth ).compareTo( dimensions[ 3 ] ) == -1 ) {
            resultNrows ++;
            resultWidth = resultWidth.add( resultDimensions[0] );
        }
        //while ( ( resultXllcorner + resultWidth ) < ( xllcorner + width ) ) {
        //    resultNcols ++;
        //    resultWidth += resultCellsize;
        //}
        resultDimensions[3] = dimensions[1].add( resultWidth );
        resultDimensions[4] = dimensions[2].add( resultHeight );
        
        // Initialise result
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultDimensions, noDataValue );
        
        long row;
        long col;
        double x;
        double y;
        double value;
        
        // sum
        if ( statistic.equalsIgnoreCase( "sum" ) ) {
            Grid2DSquareCellDouble dataCount = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultDimensions, noDataValue );
            //Grid2DSquareCellDouble tempGrid1 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            //Grid2DSquareCellDouble tempGrid2 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            double areaProportion;
            double[] bounds = new double[ 4 ];
            CellID[] cellIDs = new CellID[ 4 ];
            //double totalArea = 0;
            
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    bounds = grid.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                    
                    if ( result.inGrid( bounds[ 0 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue ) ) {
                        value = grid.getCell( bounds[ 0 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            dataCount.addToCell( bounds[ 0 ], bounds[ 1 ], 1.0d, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                    if ( result.inGrid( bounds[ 0 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue ) ) {
                        value = grid.getCell( bounds[ 0 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            dataCount.addToCell( bounds[ 0 ], bounds[ 3 ], 1.0d, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                    if ( result.inGrid( bounds[ 2 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue ) ) {
                        value = grid.getCell( bounds[ 2 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            dataCount.addToCell( bounds[ 2 ], bounds[ 1 ], 1.0d, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                    if ( result.inGrid( bounds[ 2 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue ) ) {
                        value = grid.getCell( bounds[ 2 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                        if ( value != noDataValue ) {
                            dataCount.addToCell( bounds[ 2 ], bounds[ 3 ], 1.0d, this.handleOutOfMemoryErrorTrue);
                        }
                    }
                }
            }
            // Add the nearest values for the noDataValues so long as there is a value
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    bounds = grid.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                    cellIDs[ 0 ] = result.getCellID( bounds[ 0 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                    cellIDs[ 1 ] = result.getCellID( bounds[ 2 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue );
                    cellIDs[ 2 ] = result.getCellID( bounds[ 0 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                    cellIDs[ 3 ] = result.getCellID( bounds[ 2 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue );
                    if ( dataCount.getCell( bounds[ 0 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue ) != noDataValue ||
                            dataCount.getCell( bounds[ 2 ], bounds[ 3 ], this.handleOutOfMemoryErrorTrue ) != noDataValue ||
                            dataCount.getCell( bounds[ 0 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue ) != noDataValue ||
                            dataCount.getCell( bounds[ 2 ], bounds[ 1 ], this.handleOutOfMemoryErrorTrue ) != noDataValue ) {
                        value = grid.getNearestValueDouble( row, col, this.handleOutOfMemoryErrorTrue );
                        if ( cellIDs[ 0 ].equals( cellIDs[ 1 ] ) && cellIDs[ 1 ].equals( cellIDs[2] ) ) {
                            result.addToCell( cellIDs[ 0 ], value, this.handleOutOfMemoryErrorTrue );
                        } else {
                            if ( cellIDs[ 0 ].equals( cellIDs[ 1 ] ) || cellIDs[ 0 ].equals( cellIDs[ 2 ] ) ) {
                                if ( cellIDs[ 0 ].equals( cellIDs[ 1 ] ) ) {
                                    areaProportion = ( Math.abs( bounds[3] - ( result.getCellY( cellIDs[ 0 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( Math.abs( ( result.getCellX( cellIDs[ 0 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[0] ) * cellsize ) / ( cellsize * cellsize );
                                }
                            } else {
                                areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellY( cellIDs[ 0 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) * Math.abs( ( result.getCellX( cellIDs[ 0 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                            }
                            result.addToCell( cellIDs[0], value * areaProportion, this.handleOutOfMemoryErrorTrue );
                        }
                        if ( ! cellIDs[ 1 ].equals( cellIDs[ 0 ] ) ) {
                            if ( cellIDs[ 1 ].equals( cellIDs[ 3 ] ) ) {
                                areaProportion = ( Math.abs( bounds[2] - ( result.getCellX( cellIDs[ 1 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                            } else {
                                areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellY( cellIDs[ 1 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellX( cellIDs[ 1 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                            }
                            result.addToCell( cellIDs[ 1 ], value * areaProportion, this.handleOutOfMemoryErrorTrue );
                        }
                        if ( ! cellIDs[ 2 ].equals( cellIDs[ 0 ] ) ) {
                            if ( ! cellIDs[ 2 ].equals( cellIDs[ 3 ] ) ) {
                                areaProportion = ( Math.abs( ( result.getCellY( cellIDs[ 2 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[1] ) * cellsize ) / ( cellsize * cellsize );
                            } else {
                                areaProportion = ( ( Math.abs( ( result.getCellY( cellIDs[ 2 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellX( cellIDs[ 2 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                            }
                            result.addToCell( cellIDs[ 2 ], value * areaProportion, this.handleOutOfMemoryErrorTrue );
                        }
                        if ( ! cellIDs[ 3 ].equals( cellIDs[ 1 ] ) && ! cellIDs[ 3 ].equals( cellIDs[ 2 ] ) ) {
                            areaProportion = ( ( Math.abs( ( result.getCellY( cellIDs[ 3 ], this.handleOutOfMemoryErrorTrue ) + ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellX( cellIDs[ 3 ], this.handleOutOfMemoryErrorTrue ) - ( resultDimensions[ 0 ].doubleValue() / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                            result.addToCell( cellIDs[ 3 ], value * areaProportion, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
        }
        //        // Add the nearest values for the noDataValues so long as there is a value
        //            for ( int i = 0; i < nrows; i ++ ) {
        //                for ( int j = 0; j < ncols; j ++ ) {
        //                    bounds = grid.getCellBounds( i, j );
        //                    cellID1 = result.getCellID( bounds[ 0 ], bounds[ 3 ] );
        //                    cellID2 = result.getCellID( bounds[ 2 ], bounds[ 3 ] );
        //                    cellID3 = result.getCellID( bounds[ 0 ], bounds[ 1 ] );
        //                    cellID4 = result.getCellID( bounds[ 2 ], bounds[ 1 ] );
        //                    if ( dataCount.getCell( bounds[ 0 ], bounds[ 3 ] ) != noDataValue ||
        //                    dataCount.getCell( bounds[ 2 ], bounds[ 3 ] ) != noDataValue ||
        //                    dataCount.getCell( bounds[ 0 ], bounds[ 1 ] ) != noDataValue ||
        //                    dataCount.getCell( bounds[ 2 ], bounds[ 1 ] ) != noDataValue ) {
        //                        value = grid.getNearestValueDouble( i, j );
        //                        if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
        //                            if ( cellID1 != Integer.MIN_VALUE ) {
        //                                result.addToCell( cellID1, value );
        //                            }
        //                        } else {
        //                            if ( cellID1 != Integer.MIN_VALUE ) {
        //                                if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
        //                                    if ( cellID1 == cellID2 ) {
        //                                        areaProportion = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
        //                                    } else {
        //                                        areaProportion = ( Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize ) / ( cellsize * cellsize );
        //                                    }
        //                                } else {
        //                                    areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
        //                                }
        //                                result.addToCell( cellID1, value * areaProportion );
        //                            }
        //                            if ( cellID2 != Integer.MIN_VALUE ) {
        //                                if ( cellID2 != cellID1 ) {
        //                                    if ( cellID2 == cellID4 ) {
        //                                        areaProportion = ( Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
        //                                    } else {
        //                                        areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellY( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
        //                                    }
        //                                    result.addToCell( cellID2, value * areaProportion );
        //                                }
        //                            }
        //                            if ( cellID3 != Integer.MIN_VALUE ) {
        //                                if ( cellID3 != cellID1 ) {
        //                                    if ( cellID3 == cellID4 ) {
        //                                        areaProportion = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize ) / ( cellsize * cellsize );
        //                                    } else {
        //                                        areaProportion = ( ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellX( cellID3 ) + ( resultCellsize / 2.0d) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
        //                                    }
        //                                    result.addToCell( cellID3, value * areaProportion );
        //                                }
        //                            }
        //                            if ( cellID4 != Integer.MIN_VALUE ) {
        //                                if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
        //                                    areaProportion = ( ( Math.abs( ( result.getCellY( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellX( cellID4 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
        //                                    result.addToCell( cellID4, value * areaProportion );
        //                                }
        //                            }
        //                        }
        //                    }
        //                }
        //            }
        //        }
        
        // mean
        if ( statistic.equalsIgnoreCase( "mean" ) ) {
            double denominator = ( resultDimensions[ 0 ].doubleValue() * resultDimensions[ 0 ].doubleValue() ) / ( cellsize * cellsize );
            Grid2DSquareCellDouble sum = aggregate( grid, "sum", resultDimensions, gridFactory, this.handleOutOfMemoryErrorTrue);
            addToGrid( result, sum, 1.0d / denominator, this.handleOutOfMemoryErrorTrue );
        }
        
        // max
        if ( statistic.equalsIgnoreCase( "max" ) ) {
            double max;
            double[] bounds = new double[4];
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    value = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        x = grid.getCellX( col, this.handleOutOfMemoryErrorTrue );
                        y = grid.getCellY( row, this.handleOutOfMemoryErrorTrue );
                        bounds = grid.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                        max = result.getCell( bounds[0], bounds[3], this.handleOutOfMemoryErrorTrue );
                        if ( max != noDataValue ) {
                            result.setCell( bounds[0], bounds[3], Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[0], bounds[3], value, this.handleOutOfMemoryErrorTrue );
                        }
                        max = result.getCell( bounds[2], bounds[3], this.handleOutOfMemoryErrorTrue );
                        if ( max != noDataValue ) {
                            result.setCell( bounds[2], bounds[3], Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[2], bounds[3], value, this.handleOutOfMemoryErrorTrue );
                        }
                        max = result.getCell( bounds[0], bounds[1], this.handleOutOfMemoryErrorTrue );
                        if ( max != noDataValue ) {
                            result.setCell( bounds[0], bounds[1], Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[0], bounds[1], value, this.handleOutOfMemoryErrorTrue );
                        }
                        max = result.getCell( bounds[2], bounds[1], this.handleOutOfMemoryErrorTrue );
                        if ( max != noDataValue ) {
                            result.setCell( bounds[2], bounds[1], Math.max( max, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[2], bounds[1], value, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
        }
        
        // min
        if ( statistic.equalsIgnoreCase( "min" ) ) {
            double min;
            double[] bounds = new double[4];
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    value = grid.getCell( row, col, this.handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        x = grid.getCellX( col, this.handleOutOfMemoryErrorTrue );
                        y = grid.getCellY( row, this.handleOutOfMemoryErrorTrue );
                        bounds = grid.getCellBoundsDoubleArray( row, col, this.handleOutOfMemoryErrorTrue );
                        min = result.getCell( bounds[0], bounds[3], this.handleOutOfMemoryErrorTrue );
                        if ( min != noDataValue ) {
                            result.setCell( bounds[0], bounds[3], Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[0], bounds[3], value, this.handleOutOfMemoryErrorTrue );
                        }
                        min = result.getCell( bounds[2], bounds[3], this.handleOutOfMemoryErrorTrue );
                        if ( min != noDataValue ) {
                            result.setCell( bounds[2], bounds[3], Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[2], bounds[3], value, this.handleOutOfMemoryErrorTrue );
                        }
                        min = result.getCell( bounds[0], bounds[1], this.handleOutOfMemoryErrorTrue );
                        if ( min != noDataValue ) {
                            result.setCell( bounds[0], bounds[1], Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[0], bounds[1], value, this.handleOutOfMemoryErrorTrue );
                        }
                        min = result.getCell( bounds[2], bounds[1], this.handleOutOfMemoryErrorTrue );
                        if ( min != noDataValue ) {
                            result.setCell( bounds[2], bounds[1], Math.min( min, value ), this.handleOutOfMemoryErrorTrue );
                        } else {
                            result.setCell( bounds[2], bounds[1], value, this.handleOutOfMemoryErrorTrue );
                        }
                    }
                }
            }
        }
        
        /*
        // Initialistaion
        int nrows = grid.getNrows();
        int ncols = grid.getNcols();
        double xllcorner = grid.getXllcorner();
        double yllcorner = grid.getYllcorner();
        double cellsize = grid.getCellsize();
        double noDataValue = grid.getNoDataValue();
        double width = cellsize * ncols;
        double height = cellsize * nrows;
        // Test this is an aggregation
        if ( resultCellsize <= cellsize ) {
            System.out.println( "!!!Warning: resultCellsize <= cellsize in aggregate( cellsize( " + resultCellsize + " ), statistic( " + statistic + " ), resultXllcorner( " + resultXllcorner + " ), resultYllcorner( " + resultYllcorner + " ), noDataValue( " + noDataValue + " ), gridFactory( " + gridFactory.toString() + " ) ). Returning null!" );
            return null;
        }
        // Test for intersection
        if ( ( resultXllcorner > xllcorner + ( ( double ) ncols * cellsize ) ) || ( resultYllcorner > yllcorner + ( ( double ) nrows * cellsize ) ) ) {
            System.out.println( "!!!Warning: No intersection in aggregate( cellsize( " + resultCellsize + " ), statistic( " + statistic + " ), resultXllcorner( " + resultXllcorner + " ), resultYllcorner( " + resultYllcorner + " ), noDataValue( " + noDataValue + " ), gridFactory( " + gridFactory.toString() + " ) ). Returning null!" );
            return null;
        }
        // If resultCellsize is an integer multiple of cellsize and grid aligns with result then use
        // a cellFactor aggregation as it should be faster.
        //System.out.println("resultCellsize % cellsize == " + ( resultCellsize % cellsize ) );
        //System.out.println("resultXllcorner % cellsize = " + ( resultXllcorner % cellsize ) + ", xllcorner % cellsize = " + ( xllcorner % cellsize ) );
        //System.out.println("resultYllcorner % cellsize = " + ( resultYllcorner % cellsize ) + ", yllcorner % cellsize = " + ( yllcorner % cellsize ) );
        if ( ( resultCellsize % cellsize == 0.0d ) && ( ( resultXllcorner % cellsize ) == ( xllcorner % cellsize ) ) && ( ( resultYllcorner % cellsize ) == ( yllcorner % cellsize ) ) ) {
            int cellFactor = ( int ) ( resultCellsize / cellsize );
            int rowOffset = ( int ) ( yllcorner - resultYllcorner / cellsize );
            int colOffset = ( int ) ( xllcorner - resultXllcorner / cellsize );
            return aggregate( grid, cellFactor, statistic, rowOffset, colOffset, gridFactory );
        }
        // Calculate resultNrows and resultHeight
        int resultNrows = 1;
        double resultHeight = resultCellsize;
        while ( ( resultYllcorner + resultHeight ) < ( yllcorner + height ) ) {
            resultNrows ++;
            resultHeight += resultCellsize;
        }
        // Calculate resultNcols and resultWidth
        int resultNcols = 1;
        double resultWidth = resultCellsize;
        while ( ( resultXllcorner + resultWidth ) < ( xllcorner + width ) ) {
            resultNcols ++;
            resultWidth += resultCellsize;
        }
        //System.out.println( "resultNcols " + resultNcols + ", resultNrows " + resultNrows );
        //System.out.println( "gridToAddNcols " + ncols + ", gridToAddNrows " + nrows );
        // Initialise result
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue );
         
        // sum
        if ( statistic.equalsIgnoreCase( "sum" ) ) {
            Grid2DSquareCellDouble tempGrid1 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            Grid2DSquareCellDouble tempGrid2 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            double x = 0.0d;
            double y = 0.0d;
            double d1 = noDataValue;
            double areaProportion = 0.0d;
            double[] bounds = new double[ 4 ];
            int cellID1 = -1;
            int cellID2 = -1;
            int cellID3 = -1;
            int cellID4 = -1;
            //double totalArea = 0;
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellX( i );
                    y = grid.getCellY( i );
                    bounds = grid.getCellBounds( i );
                    cellID1 = result.getCellID( bounds[ 0 ], bounds[ 3 ] );
                    cellID2 = result.getCellID( bounds[ 2 ], bounds[ 3 ] );
                    cellID3 = result.getCellID( bounds[ 0 ], bounds[ 1 ] );
                    cellID4 = result.getCellID( bounds[ 2 ], bounds[ 1 ] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            areaProportion = 1.0d;
                            tempGrid1.addToCell( x, y, d1 * areaProportion );
                            tempGrid2.addToCell( x, y, areaProportion );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    areaProportion = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize ) / ( cellsize * cellsize );
                                }
                            } else {
                                areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                            }
                            tempGrid1.addToCell( cellID1, d1 * areaProportion );
                            tempGrid2.addToCell( cellID1, areaProportion );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    areaProportion = ( Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( ( Math.abs( bounds[3] - ( result.getCellY( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                                }
                                tempGrid1.addToCell( cellID2, d1 * areaProportion );
                                tempGrid2.addToCell( cellID2, areaProportion );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    areaProportion = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize ) / ( cellsize * cellsize );
                                } else {
                                    areaProportion = ( ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellX( cellID3 ) + ( resultCellsize / 2.0d) ) - bounds[0] ) ) / ( cellsize * cellsize ) );
                                }
                                tempGrid1.addToCell( cellID3, d1 * areaProportion );
                                tempGrid2.addToCell( cellID3, areaProportion );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                areaProportion = ( ( Math.abs( ( result.getCellY( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellX( cellID4 ) - ( resultCellsize / 2.0d ) ) ) ) / ( cellsize * cellsize ) );
                                tempGrid1.addToCell( cellID4, d1 * areaProportion );
                                tempGrid2.addToCell( cellID4, areaProportion );
                            }
                        }
                        // Check fails due to rounding errors!
                        //if ( cellID1 != Integer.MIN_VALUE && cellID2 != Integer.MIN_VALUE && cellID3 != Integer.MIN_VALUE && cellID4 != Integer.MIN_VALUE && totalArea != 1.0 ) { System.out.println( "id = " + i + " : totalArea = " + totalArea + " (cellID1,cellID2,cellID3,cellID4) = (" + cellID1 + "," + cellID2 + "," + cellID3 + "," + cellID4 + ")" );
                        //    throw an exception!!!
                        //}
                    }
                }
            }
            // The values are normalised by dividing the aggregate Grid sum by the proportion of cells with grid values.
            for ( int i = 0; i < resultNrows * resultNcols; i ++ ) {
                d1 = tempGrid2.getCell( i );
                if ( d1 != 0.0d && d1 != noDataValue ) {
                    result.setCell( i, tempGrid1.getCell( i ) / ( Math.pow( ( resultCellsize / cellsize ), 2.0d ) / d1 ) );
                }
            }
            tempGrid1.clear();
            tempGrid2.clear();
        }
         
        // mean
        if ( statistic.equalsIgnoreCase( "mean" ) ) {
            // To calculate the mean and cope with NODATA it is necessary to pass
            // through the data twice or for each aggregated cell get all
            // intersecting cells. This is because each cells area as a proportion
            // of the non noDataValue area of an aggregated cell is needed. This
            // cannot be simply done as it all depends on NODATA. In the
            // implementation below the data is read through twice. First read
            // involves calculating NODATA in each aggregated cell.
            Grid2DSquareCellDouble tempGrid1 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            Grid2DSquareCellDouble tempGrid2 = gridFactory.createGrid2DSquareCellDouble( resultNrows, resultNcols, resultXllcorner, resultYllcorner, resultCellsize, noDataValue, 1 );
            double x;
            double y;
            double d1;
            double area;
            double[] bounds = new double[4];
            int cellID1;
            int cellID2;
            int cellID3;
            int cellID4;
            //double totalArea = 0;
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellX( i );
                    y = grid.getCellY( i );
                    bounds = grid.getCellBounds( i );
                    cellID1 = result.getCellID( bounds[0], bounds[3] );
                    cellID2 = result.getCellID( bounds[2], bounds[3] );
                    cellID3 = result.getCellID( bounds[0], bounds[1] );
                    cellID4 = result.getCellID( bounds[2], bounds[1] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            area = cellsize * cellsize;
                            tempGrid1.addToCell( x, y, area );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    area = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    area = ( Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize );
                                }
                            } else {
                                area = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                            }
                            tempGrid1.addToCell( cellID1, area );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    area = ( Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    area = ( Math.abs( bounds[3] - ( result.getCellY( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) );
                                }
                                tempGrid1.addToCell( cellID2, area );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    area = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize );
                                } else {
                                    area = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellX( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                                }
                                tempGrid1.addToCell( cellID3, area );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                area = ( Math.abs( ( result.getCellY( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellX( cellID4 ) - ( resultCellsize / 2.0d ) ) ) );
                                tempGrid1.addToCell( cellID4, area );
                            }
                        }
                    }
                }
            }
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                double areaIntersect;
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellX( i );
                    y = grid.getCellY( i );
                    bounds = grid.getCellBounds( i );
                    cellID1 = result.getCellID( bounds[0], bounds[3] );
                    cellID2 = result.getCellID( bounds[2], bounds[3] );
                    cellID3 = result.getCellID( bounds[0], bounds[1] );
                    cellID4 = result.getCellID( bounds[2], bounds[1] );
                    if ( cellID1 == cellID2 && cellID2 == cellID3 ) {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            area = tempGrid1.getCell( x, y );
                            result.addToCell( x, y, d1 * ( cellsize * cellsize ) / area );
                        }
                    } else {
                        if ( cellID1 != Integer.MIN_VALUE ) {
                            if ( cellID1 == cellID2 || cellID1 == cellID3 ) {
                                if ( cellID1 == cellID2 ) {
                                    areaIntersect = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) * cellsize );
                                }
                            } else {
                                areaIntersect = ( Math.abs( bounds[3] - ( result.getCellY( cellID1 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( ( result.getCellX( cellID1 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                            }
                            area = tempGrid1.getCell( cellID1 );
                            result.addToCell( cellID1, d1 * areaIntersect / area );
                        }
                        if ( cellID2 != Integer.MIN_VALUE ) {
                            if ( cellID2 != cellID1 ) {
                                if ( cellID2 == cellID4 ) {
                                    areaIntersect = ( Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( bounds[3] - ( result.getCellY( cellID2 ) - ( resultCellsize / 2.0d ) ) ) * Math.abs( bounds[2] - ( result.getCellX( cellID2 ) - ( resultCellsize / 2.0d ) ) ) );
                                }
                                area = tempGrid1.getCell( cellID2 );
                                result.addToCell( cellID2, d1 * areaIntersect / area );
                            }
                        }
                        if ( cellID3 != Integer.MIN_VALUE ) {
                            if ( cellID3 != cellID1 ) {
                                if ( cellID3 == cellID4 ) {
                                    areaIntersect = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * cellsize );
                                } else {
                                    areaIntersect = ( Math.abs( ( result.getCellY( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( ( result.getCellX( cellID3 ) + ( resultCellsize / 2.0d ) ) - bounds[0] ) );
                                }
                                area = tempGrid1.getCell( cellID3 );
                                result.addToCell( cellID3, d1 * areaIntersect / area );
                            }
                        }
                        if ( cellID4 != Integer.MIN_VALUE ) {
                            if ( cellID4 != cellID2 && cellID4 != cellID3 ) {
                                areaIntersect = ( Math.abs( ( result.getCellY( cellID4 ) + ( resultCellsize / 2.0d ) ) - bounds[1] ) * Math.abs( bounds[2] - ( result.getCellX( cellID4 ) - ( resultCellsize / 2.0d ) ) ) );
                                area = tempGrid1.getCell( cellID4 );
                                result.addToCell( cellID4, d1 * areaIntersect / area );
                            }
                        }
                    }
                }
            }
            tempGrid1.clear();
        }
         
        // max
        if ( statistic.equalsIgnoreCase( "max" ) ) {
            double x;
            double y;
            double d1;
            double d2;
            double[] bounds = new double[4];
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellX( i );
                    y = grid.getCellY( i );
                    bounds = grid.getCellBounds( i );
                    d2 = result.getCell( bounds[0], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[3], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[3], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[0], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[1], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[1], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[1], Math.max( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[1], d1 );
                    }
                }
            }
        }
         
        // min
        if ( statistic.equalsIgnoreCase( "min" ) ) {
            double x;
            double y;
            double d1;
            double d2;
            double[] bounds = new double[4];
            for ( int i = 0; i < nrows * ncols; i ++ ) {
                d1 = grid.getCell( i );
                if ( d1 != noDataValue ) {
                    x = grid.getCellX( i );
                    y = grid.getCellY( i );
                    bounds = grid.getCellBounds( i );
                    d2 = result.getCell( bounds[0], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[3], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[3] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[3], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[3], d1 );
                    }
                    d2 = result.getCell( bounds[0], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[0], bounds[1], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[0], bounds[1], d1 );
                    }
                    d2 = result.getCell( bounds[2], bounds[1] );
                    if ( d2 != noDataValue ) {
                        result.setCell( bounds[2], bounds[1], Math.min( d2, d1 ) );
                    } else {
                        result.setCell( bounds[2], bounds[1], d1 );
                    }
                }
            }
        }
         */
        return result;
    }
    
    //    TODO: Move to a Extended Statistics class
    //    /**
    //     * Returns a double global statistic for an Grid2DSquareCellDouble grid
    //     * NB. Only for grids in the same spatial frame.
    //     */
    //    public double globalBivariateStatistics( Grid2DSquareCellDouble grid0, Grid2DSquareCellDouble grid1, String comparator  ) {
    //        // Initialisation
    //        int grid0Nrows = grid0.getNrows();
    //        int grid0Ncols = grid0.getNcols();
    //        double grid0Xllcorner = grid0.getXllcorner();
    //        double grid0Yllcorner = grid0.getYllcorner();
    //        double grid0Cellsize = grid0.getCellsize();
    //        double grid0NoDataValue = grid0.getNoDataValue();
    //        int grid1Nrows = grid1.getNrows();
    //        int grid1Ncols = grid1.getNcols();
    //        double grid1Xllcorner = grid1.getXllcorner();
    //        double grid1Yllcorner = grid1.getYllcorner();
    //        double grid1Cellsize = grid1.getCellsize();
    //        double grid1NoDataValue = grid1.getNoDataValue();
    //        AbstractGridStatistics grid0Statistics = grid0.getGridStatistics();
    //        AbstractGridStatistics grid1Statistics = grid1.getGridStatistics();
    //        // TODO: Check spatial frame
    //
    //        // Calculation
    //        double thisDistance;
    //        double x0;
    //        double x1;
    //        double y0;
    //        double y1;
    //        double value0;
    //        double value1;
    //        // diff: The sum of all the differences between the grid cells
    //        if ( comparator.equalsIgnoreCase( "diff" ) ) {
    //            double diff = 0.0d;
    //            for ( int i = 0; i < grid0Nrows; i ++ ) {
    //                for ( int j = 0; j < grid0Ncols; j ++ ) {
    //                    value0 = grid0.getCell( i, j );
    //                    if ( value0 != grid0NoDataValue ) {
    //                        value1 = grid0.getCell( i, j );
    //                        if ( value1 != grid1NoDataValue ) {
    //                            diff += ( value0 - value1 );
    //                        }
    //                    }
    //                }
    //            }
    //            return diff;
    //        }
    //
    //        // abs: The sum of all the absolute differences between the grid cells
    //        if ( comparator.equalsIgnoreCase( "abs" ) ) {
    //            double abs = 0.0d;
    //            for ( int i = 0; i < grid0Nrows; i ++ ) {
    //                for ( int j = 0; j < grid0Ncols; j ++ ) {
    //                    value0 = grid0.getCell( i, j );
    //                    if ( value0 != grid0NoDataValue ) {
    //                        value1 = grid0.getCell( i, j );
    //                        if ( value1 != grid1NoDataValue ) {
    //                            abs += Math.abs( value0 - value1 );
    //                        }
    //                    }
    //                }
    //            }
    //            return abs;
    //        }
    //
    //        // pearsons: The persons correlation coefficient
    //        if ( comparator.equalsIgnoreCase( "pearsons" ) ) {
    //            double pearsons = grid0NoDataValue;
    //            double sum0 = 0.0d;
    //            double sum1 = 0.0d;
    //            double sum0Squared = 0.0d;
    //            double sum1Squared = 0.0d;
    //            double sum01 = 0.0d;
    //            double n = 0.0d;
    //            for ( int i = 0; i < grid0Nrows; i ++ ) {
    //                for ( int j = 0; j < grid0Ncols; j ++ ) {
    //                    value0 = grid0.getCell( i, j );
    //                    value1 = grid1.getCell( i, j );
    //                    if ( value0 != grid0NoDataValue && value1 != grid0NoDataValue ) {
    //                        sum0 += value0;
    //                        sum1 += value1;
    //                        sum0Squared += Math.pow( value0, 2.0d );
    //                        sum1Squared += Math.pow( value1, 2.0d );
    //                        sum01 += value0 * value1;
    //                        n += 1.0d;
    //                    }
    //                }
    //            }
    //            double numerator = ( n * sum01 ) - ( sum0 * sum1 );
    //            double denominator = Math.sqrt( Math.abs( (  ( n * sum0Squared ) - Math.pow( sum0, 2.0d )  ) * ( ( n * sum1Squared ) - Math.pow( sum1, 2.0d ) ) ) );
    //            if ( denominator != 0.0d ) {
    //                pearsons = numerator / denominator;
    //            }
    //            return pearsons;
    //        }
    //
    //        // momentcorrelation: The moment correlation coefficient
    //        if ( comparator.equalsIgnoreCase( "momentCorrelation" ) ) {
    //            double momentCorrelation = grid0NoDataValue;
    //            double mean0 = grid0Statistics.getMean();
    //            double mean1 = grid1Statistics.getMean();
    //            double sd0 = 0.0d;
    //            double sd1 = 0.0d;
    //            double m01 = 0.0d;
    //            double n = 0.0d;
    //            double denominator;
    //            for ( int i = 0; i < grid0Nrows; i ++ ) {
    //                for ( int j = 0; j < grid0Ncols; j ++ ) {
    //                    value0 = grid0.getCell( i, j );
    //                    value1 = grid1.getCell( i, j );
    //                    if ( value0 != grid0NoDataValue && value1 != grid1NoDataValue ) {
    //                        n += 1.0d;
    //                        sd0 += Math.pow( value0 - mean0, 2.0d );
    //                        sd1 += Math.pow( value1 - mean1, 2.0d );
    //                        m01 += ( value0 - mean0 ) * ( value1 - mean1 );
    //                    }
    //                }
    //            }
    //            denominator = Math.sqrt( sd0 ) * Math.sqrt( sd1 );
    //            if ( denominator != 0.0d ) {
    //                momentCorrelation = m01 / denominator;
    //            }
    //            return momentCorrelation;
    //        }
    //        return grid0NoDataValue;
    //    }
    
    /**
     * Returns a new Grid2DSquareCellDouble the values of which are the distance to the nearest data value
     * TODO: Optimise as it is currently very slow and inefficient!!!
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDouble distanceToDataValue( Grid2DSquareCellDouble grid0, Grid2DSquareCellDoubleFactory gridFactory, boolean handleOutOfMemoryError ) {
        long nrows = grid0.getNrows();
        long ncols = grid0.getNcols();
        BigDecimal[] dimensions0 = grid0.getDimensions();
        double noDataValue = grid0.getNoDataValue();
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid0.getChunkNrows(this.handleOutOfMemoryErrorFalse), grid0.getChunkNcols(this.handleOutOfMemoryErrorFalse), nrows, ncols, dimensions0, noDataValue );
        // Calculate distances
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                result.setCell( row, col, grid0.getNearestValueDoubleDistance( row, col, this.handleOutOfMemoryErrorTrue ), this.handleOutOfMemoryErrorTrue );
            }
        }
        return result;
    }
    
    //    TODO:
    //    /**
    //     * Returns a new Grid2DSquareCellDouble which is a copy of grid0 but
    //     * with all noDataValues replaced as follows:
    //     * At each iteration we deal with the nearest band of noDataValues.
    //     * (NB. If we dealt only with the nearest noDataValues it may take a very long time to compute!)
    //     * We replace these nearest noDataValues with the average values within distance.
    //     * We then repeat until all cells have a value.
    //     */
    //    public Grid2DSquareCellDouble replaceNoDataValues( Grid2DSquareCellDouble grid0, double distance, Grid2DSquareCellDoubleFactory gridFactory ) {
    //        long nrows = grid0.getNrows();
    //        long ncols = grid0.getNcols();
    //        BigDecimal[] dimensions = grid0.getDimensions();
    //        double noDataValue = grid0.getNoDataValue();
    //        Grid2DSquareCellDouble temp1 = gridFactory.createGrid2DSquareCellDouble( grid0 );
    //        Grid2DSquareCellDouble temp2 = gridFactory.createGrid2DSquareCellDouble( grid0 );
    //        // Get distances
    //        Grid2DSquareCellDouble distanceGrid = distanceToDataValue( grid0, gridFactory );
    //        AbstractGridStatistics distanceGridStatistics = distanceGrid.getGridStatistics();
    //        double maxDistance = distanceGridStatistics.getMax();
    //
    //        AbstractGrid2DSquareCellDouble[] geometricDensity = Grid2DSquareCellDoubleProcessorGWS.geometricDensity( grid0, maxDistance, gridFactory );
    //        //return geometricDensity[ geometricDensity.length - 1 ];
    //        return geometricDensity[ geometricDensity.length / 2 ];
    //        /*
    //        int maxCellDistance = ( int ) Math.ceil( maxDistance / cellsize );
    //        boolean alternator = true;
    //        Grid2DSquareCellDouble thisGrid = null;
    //        Grid2DSquareCellDouble thatGrid = null;
    //        double thisValue;
    //        double thatValue;
    //        double thisDistance;
    //        for ( int iterations = 1; iterations <= maxCellDistance; iterations ++ ) {
    //            for ( int i = 0; i < nrows; i ++ ) {
    //                for ( int j = 0; j < ncols; j ++ ) {
    //                    thisValue = thisGrid.getCell( i, j );
    //                    thatValue = thatGrid.getCell( i, j );
    //                    thisDistance = distanceGrid.getCell( i, j );
    //                    if ( thatValue != noDataValue ) {
    //                        thisGrid.setCell( i, j, thatValue );
    //                    } else {
    //                        if ( thatValue == noDataValue && thisDistance < ( iterations * cellsize ) ) {
    //                            thisGrid.setCell( i, j, Grid2DSquareCellDoubleProcessor.regionUnivariateStatistics( thatGrid, i, j, "mean", distance, 1.0d, 1.0d, gridFactory ) );
    //                        }
    //                    }
    //                }
    //            }
    //        }
    //        thatGrid.clear();
    //        return thisGrid;
    //         */
    //
    //    }
    
    
    /**
     * Shuffles dummyDiff numberOfShuffles places
     */
    private void shuffle( double[] dummyDiff, double[] diff, int numberOfShuffles ) {
        for ( int i = 0; i < dummyDiff.length - numberOfShuffles; i ++ ) {
            dummyDiff[ i ] = diff[ i + numberOfShuffles ];
        }
        for ( int i = dummyDiff.length - 1; i > dummyDiff.length - 1 + numberOfShuffles; i -- ) {
            dummyDiff[ i ] = diff[ i - dummyDiff.length + numberOfShuffles ];
        }
    }
    
    /**
     * Returns a double[][] of grid values
     *
     */
    protected double[][] getRowProcessInitialData( Grid2DSquareCellDouble grid, int cellDistance, long row ) {
        double[][] result = new double[ ( cellDistance * 2 ) + 1 ][ ( cellDistance * 2 ) + 1 ];
        long col;
        long r;
        for ( r = - cellDistance; r <= cellDistance; r ++ ) {
            for ( col = - cellDistance; col <= cellDistance; col ++ ) {
                result[ ( int ) row + cellDistance ][ ( int ) col + cellDistance ] = grid.getCell( r + row, col, this.handleOutOfMemoryErrorTrue );
            }
        }
        return result;
    }
    
    /**
     * Returns a double[][] based on previous which has been shuffled
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then caching operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected double[][] getRowProcessData( Grid2DSquareCellDouble grid, double[][] previous, int cellDistance, long row, long col ) {
        double[][] result = previous;
        if ( col == 0 ) {
            return getRowProcessInitialData( grid, cellDistance, row );
        } else {
            // shift columns one left
            for ( int i = 0; i <= cellDistance * 2; i ++ ) {
                for ( int j = 0; j <= ( cellDistance * 2 ) - 1; j ++ ) {
                    result[ i ][ j ] = previous[ i ][ j + 1 ];
                }
            }
            // getLastColumn
            for ( int i = -cellDistance; i <= cellDistance; i ++ ) {
                result[ i + cellDistance ][ cellDistance * 2 ] = grid.getCell( ( long ) i + row, ( long ) col + cellDistance, this.handleOutOfMemoryErrorTrue );
            }
        }
        return result;
    }
    
    
    //    // TODO:
    //    // tests
    //    private double[][] getChunkProcessInitialData( Grid2DSquareCellDouble grid, int chunkCells, int rowChunk ) {
    //        double[][] result = new double[ chunkCells * 3 ][ chunkCells * 3 ];
    //        for ( int i = -chunkCells; i <= ( chunkCells * 2 ) - 1; i ++ ) {
    //            for ( int j = -chunkCells; j <= ( chunkCells * 2 ) - 1; j ++ ) {
    //                result[ i + chunkCells ][ j + chunkCells ] = grid.getCell( i + ( rowChunk * chunkCells ), j );
    //            }
    //        }
    //        return result;
    //    }
    //
    //    // Needs testing!!
    //    private double[][] getChunkProcessData( Grid2DSquareCellDouble grid, double[][] previous, int chunkCells, int rowChunk, int colChunk ) {
    //        double[][] result = previous;
    //        if ( colChunk == 0 ) {
    //            return getRowProcessInitialData( grid, chunkCells, rowChunk );
    //        } else {
    //            // shift end columns to start columns
    //            for ( int i = -chunkCells; i <= ( chunkCells * 2 ) - 1; i ++ ) {
    //                for ( int j = 0; j <= chunkCells; j ++ ) {
    //                    result[ i + chunkCells ][ j ] = previous[ i + ( chunkCells * 2 ) ][ j + ( chunkCells * 2 ) ];
    //                }
    //            }
    //            // getOtherData
    //            for ( int i = -chunkCells; i <= ( chunkCells * 2 ) - 1; i ++ ) {
    //                for ( int j = 0; i <= ( chunkCells * 2 ) - 1; i ++ ) {
    //                    result[ i + chunkCells ][ j + chunkCells ] = grid.getCell( i + ( rowChunk * chunkCells ), j + ( colChunk * chunkCells ) );
    //                }
    //            }
    //        }
    //        return result;
    //    }
    
}