/**
 * Version 1.0 is for processing of 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.process;
// java dependencies
import java.util.HashSet;
import java.awt.geom.Point2D;
import java.math.BigDecimal;
import java.util.Iterator;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.utilities.Kernel;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDouble;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.CellID;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleFactory;

/**
 * A class of methods relevant to the processing of Digital Elevation Model Data.
 */
public class Grid2DSquareCellProcessorDEM extends Grid2DSquareCellProcessor {
    
    private HashSet cellIDs;
    
    /** Creates a new Grid2DSquareCellDoubleProcessorDEM */
    public Grid2DSquareCellProcessorDEM() {
        super();
    }
    
    /**
     * Returns an Grid2DSquareCellDouble[] slopeAndAspect where:
     * slopeAndAspect[0] is the aggregate slope over the region weighted by
     *   distance, weightIntersect and weightFactor;
     * slopeAndAspect[1] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the clockwize angle
     *   from north.
     * slopeAndAspect[2] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the sine of the
     *   clockwize angle from north.
     * slopeAndAspect[3] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the cosine of the
     *   clockwize angle from north.
     * (N.B. Effectively no weighting is applied.)
     * @param grid the Grid2DSquareCellDouble to be processed.
     * @param gridFactory the factory used for constructing slopeAndAspect.
     */
    public Grid2DSquareCellDouble[] getSlopeAndAspect(
            Grid2DSquareCellDouble grid,
            Grid2DSquareCellDoubleFactory gridFactory ) {
        // Default distance to contain centroids of immediate neighbours
        double distance = ( grid.getDimensions()[ 0 ].doubleValue() ) * ( 3.0d / 2.0d ); // The ( ( square root of 2 ) * cellsize ) is just enough, this is OK as it is larger but less than ( 2 * cellsize ).
        double weightIntersect = 1.0d;
        double weightFactor = 0.0d;
        return getSlopeAndAspect( grid, distance, weightIntersect, weightFactor, grid, gridFactory );
    }
    
    /**
     * Returns an Grid2DSquareCellDouble[] slopeAndAspect where:
     * slopeAndAspect[0] is the aggregate slope over the region weighted by
     *   distance, weightIntersect and weightFactor;
     * slopeAndAspect[1] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the clockwize angle
     *   from north.
     * slopeAndAspect[2] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the sine of the
     *   clockwize angle from north.
     * slopeAndAspect[3] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the cosine of the
     *   clockwize angle from north.
     * @param grid the Grid2DSquareCellDouble to be processed.
     * @param distance the distance which defines the aggregate region.
     * @param weightIntersect the kernel weighting weight at centre.
     * @param weightFactor the kernel weighting distance decay.
     * @param slopeAndAspectDimensions is an AbstractGrid2DSquareCellDouble
     *   used for setting the resolutions, dimensions and noDataValue of
     *   slopeAndAspect.
     * @param gridFactory the factory used for generating slopeAndAspect.
     * (NB. There are various strategies to reduce bias caused by noDataValues.
     *   Here:
     *     If the cell in grid for which slopeAndAspect is being calculated is a
     *       noDataValue then the cells in slopeAndAspect are assigned their
     *       noDataValue.
     *     If one of the cells in the calculation of slope and aspect is a
     *       noDataValue then its height is taken as the nearest cell value.
     *       (Formerly the difference in its height was taken as the average
     *       difference in height for those cells with values.)
     * )
     */
    public Grid2DSquareCellDouble[] getSlopeAndAspect(
            Grid2DSquareCellDouble grid,
            double distance,
            double weightIntersect,
            double weightFactor,
            Grid2DSquareCellDouble slopeAndAspectDimensions,
            Grid2DSquareCellDoubleFactory gridFactory ) {
        long gridNcols = grid.getNcols();
        long gridNrows = grid.getNrows();
        BigDecimal[] gridDimensions = grid.getDimensions();
        double gridNoDataValue = grid.getNoDataValue();
        long ncols = slopeAndAspectDimensions.getNcols();
        long nrows = slopeAndAspectDimensions.getNrows();
        BigDecimal[] dimensions = slopeAndAspectDimensions.getDimensions();
        double noDataValue = slopeAndAspectDimensions.getNoDataValue();
        int cellDistance = ( int ) Math.ceil( distance / gridDimensions[ 0 ].doubleValue() );
        double thisDistance;
        double x;
        double y;
        double height;
        double thisX;
        double thisY;
        double thisHeight;
        double diffX;
        double diffY;
        double diffHeight;
        double slope;
        double aspect;
        Grid2DSquareCellDouble[] slopeAndAspect = new Grid2DSquareCellDouble[ 4 ];
        slopeAndAspect[ 0 ] = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, dimensions, noDataValue );
        slopeAndAspect[ 1 ] = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, dimensions, noDataValue );
        slopeAndAspect[ 2 ] = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, dimensions, noDataValue );
        slopeAndAspect[ 3 ] = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, dimensions, noDataValue );
        slopeAndAspect[ 0 ].setName( grid.getName() + "_slope", this.handleOutOfMemoryErrorFalse );
        slopeAndAspect[ 1 ].setName( grid.getName() + "_aspect", this.handleOutOfMemoryErrorFalse );
        slopeAndAspect[ 2 ].setName( grid.getName() + "_sin_aspect", this.handleOutOfMemoryErrorFalse );
        slopeAndAspect[ 3 ].setName( grid.getName() + "_cos_aspect", this.handleOutOfMemoryErrorFalse );
        Grid2DSquareCellDouble dummyGrid = null;
        if ( dimensions[0].compareTo( gridDimensions[0] ) == 1 ) {
            dummyGrid = aggregate( grid, "mean", dimensions, gridFactory, this.handleOutOfMemoryErrorFalse );
        } else {
            dummyGrid = grid;
        }
        long rowIndex = 0;
        long colIndex = 0;
        double[][] weights = Kernel.getKernelWeights( dummyGrid, distance, weightIntersect, weightFactor );
        double weight;
        long row;
        long col;
        int p;
        int q;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                x = dummyGrid.getCellX( col, this.handleOutOfMemoryErrorFalse );
                y = dummyGrid.getCellY( row, this.handleOutOfMemoryErrorFalse );
                height = dummyGrid.getCell( x, y, this.handleOutOfMemoryErrorFalse );
                if ( height != noDataValue ) {
                    rowIndex = grid.getCellRowIndex( y, this.handleOutOfMemoryErrorFalse );
                    colIndex = grid.getCellColIndex( x, this.handleOutOfMemoryErrorFalse );
                    diffX = 0.0d;
                    diffY = 0.0d;
                    slope = 0.0d;
                    for ( p = - cellDistance; p <= cellDistance; p ++ ) {
                        for ( q = - cellDistance; q <= cellDistance; q ++ ) {
                            thisX = grid.getCellX( ( colIndex + q ), this.handleOutOfMemoryErrorFalse );
                            thisY = grid.getCellY( ( rowIndex + p ), this.handleOutOfMemoryErrorFalse );
                            thisDistance = Utilities.distance( x, y, thisX, thisY );
                            if ( thisDistance <= distance ) {
                                //weight = Utilities.getKernelWeight( distance, weightIntersect, weightFactor, thisDistance );
                                weight = weights[ p + cellDistance ][ q + cellDistance ];
                                thisHeight = grid.getCell( thisX, thisY, this.handleOutOfMemoryErrorFalse );
                                if ( thisHeight != noDataValue ) {
                                    diffHeight = ( height - thisHeight ) * weight;
                                    diffX += ( x - thisX ) * diffHeight;
                                    diffY += ( y - thisY ) * diffHeight;
                                    slope += diffHeight;
                                }
                            }
                        }
                    }
                    slopeAndAspect[ 0 ].setCell( row, col, slope, this.handleOutOfMemoryErrorFalse );
                    slopeAndAspect[ 1 ].setCell( row, col, Utilities.angle( x, y, ( x + diffX ), ( y + diffY ) ), this.handleOutOfMemoryErrorFalse );
                    slopeAndAspect[ 2 ].setCell( row, col, Math.sin( Utilities.angle( x, y, ( x + diffX ), ( y + diffY ) ) ), this.handleOutOfMemoryErrorFalse );
                    slopeAndAspect[ 3 ].setCell( row, col, Math.cos( Utilities.angle( x, y, ( x + diffX ), ( y + diffY ) ) ), this.handleOutOfMemoryErrorFalse );
                }
            }
        }
        return slopeAndAspect;
    }
    
    /**
     * Returns a double[] slopeAndAspect where:
     * slopeAndAspect[0] is the aggregate slope over the region weighted by
     *   distance, weightIntersect and weightFactor;
     * slopeAndAspect[1] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the clockwize angle
     *   from north.
     * slopeAndAspect[2] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the sine of the
     *   clockwize angle from north.
     * slopeAndAspect[3] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the cosine of the
     *   clockwize angle from north.
     * @param grid the Grid2DSquareCellDouble to be processed.
     * @param rowIndex the rowIndex where slopeAndAspect is calculated.
     * @param colIndex the colIndex where slopeAndAspect is calculated.
     * @param distance the distance which defines the aggregate region.
     * @param weightIntersect the kernel weighting weight at centre.
     * @param weightFactor the kernel weighting distance decay.
     */
    protected double[] getSlopeAndAspect(
            Grid2DSquareCellDouble grid,
            long rowIndex,
            long colIndex,
            double distance,
            double weightIntersect,
            double weightFactor ) {
        return getSlopeAndAspect( grid, rowIndex, colIndex, grid.getCellX( colIndex, this.handleOutOfMemoryErrorFalse ), grid.getCellY( rowIndex, this.handleOutOfMemoryErrorFalse ), distance, weightIntersect, weightFactor );
    }
    
    /**
     * Returns a double[] slopeAndAspect where:
     * slopeAndAspect[0] is the aggregate slope over the region weighted by
     *   distance, weightIntersect and weightFactor;
     * slopeAndAspect[1] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the clockwize angle
     *   from north.
     * slopeAndAspect[2] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the sine of the
     *   clockwize angle from north.
     * slopeAndAspect[3] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the cosine of the
     *   clockwize angle from north.
     * @param grid the Grid2DSquareCellDouble to be processed.
     * @param x the x coordinate from where the aspect is calculated
     * @param y the y coordinate from where the aspect is calculated
     * @param distance the distance which defines the aggregate region.
     * @param weightIntersect the kernel weighting weight at centre.
     * @param weightFactor the kernel weighting distance decay.
     */
    protected double[] getSlopeAndAspect(
            Grid2DSquareCellDouble grid,
            double x,
            double y,
            double distance,
            double weightIntersect,
            double weightFactor ) {
        return getSlopeAndAspect( grid, grid.getCellRowIndex( y, this.handleOutOfMemoryErrorFalse ), grid.getCellColIndex( x, this.handleOutOfMemoryErrorFalse ), x, y, distance, weightIntersect, weightFactor );
    }
    
    /**
     * Returns a double[] slopeAndAspect where:
     * slopeAndAspect[0] is the aggregate slope over the region weighted by
     *   distance, weightIntersect and weightFactor;
     * slopeAndAspect[1] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the clockwize angle
     *   from north.
     * slopeAndAspect[2] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the sine of the
     *   clockwize angle from north.
     * slopeAndAspect[3] is the aggregate aspect over the region weighted by
     *   distance, weightIntersect and weightFactor. This is the cosine of the
     *   clockwize angle from north.
     * @param grid - the Grid2DSquareCellDouble to be processed
     * @param rowIndex the rowIndex where the result is calculated
     * @param colIndex the colIndex where the result is calculated
     * @param x the x coordinate from where the aspect is calculated
     * @param y the y coordinate from where the aspect is calculated
     * @param distance the distance which defines the region
     * @param weightIntersect
     * @param weightFactor
     * NB. If grid.getCell( x, y ) == grid.getNoDataValue() then;
     *     result[ 0 ] = grid.getNoDataValue()
     *     result[ 1 ] = grid.getNoDataValue()
     * TODO:
     * x and y can be offset from a cell centroid so consider interpolation
     */
    protected double[] getSlopeAndAspect(
            Grid2DSquareCellDouble grid,
            long rowIndex,
            long colIndex,
            double x,
            double y,
            double distance,
            double weightIntersect,
            double weightFactor ) {
        double noDataValue = grid.getNoDataValue();
        double[] slopeAndAspect = new double[ 2 ];
        slopeAndAspect[ 0 ] = noDataValue;
        slopeAndAspect[ 1 ] = noDataValue;
        slopeAndAspect[ 2 ] = noDataValue;
        slopeAndAspect[ 3 ] = noDataValue;
        double height = grid.getCell( x, y, this.handleOutOfMemoryErrorFalse );
        if ( height != noDataValue ) {
            double cellsize = grid.getDimensions()[0].doubleValue();
            int cellDistance = ( int ) Math.ceil( ( distance + cellsize ) / cellsize );
            double thisDistance;
            double weight;
            double thisX;
            double thisY;
            double thisHeight;
            double diffX = 0.0d;
            double diffY = 0.0d;
            double diffHeight;
            double slope = 0.0d;
            double aspect;
            // Calculate slope and aspect
            for ( int p = - cellDistance; p <= cellDistance; p ++ ) {
                for ( int q = - cellDistance; q <= cellDistance; q ++ ) {
                    thisY = y + ( p * distance );
                    thisX = x + ( q * distance );
                    thisDistance = Utilities.distance( x, y, thisX, thisY );
                    if ( thisDistance <= distance ) {
                        weight = Kernel.getKernelWeight( distance, weightIntersect, weightFactor, thisDistance );
                        thisHeight = grid.getNearestValueDouble( thisX, thisY, this.handleOutOfMemoryErrorFalse );
                        diffHeight = ( height - thisHeight ) * weight;
                        diffX += ( x - thisX ) * diffHeight;
                        diffY += ( y - thisY ) * diffHeight;
                        slope += diffHeight;
                    }
                }
            }
            slopeAndAspect[ 0 ] = slope;
            slopeAndAspect[ 1 ] = Utilities.angle( x, y, ( x + diffX ), ( y + diffY ) );
            slopeAndAspect[ 2 ] = Math.sin( Utilities.angle( x, y, ( x + diffX ), ( y + diffY ) ) );
            slopeAndAspect[ 3 ] = Math.cos( Utilities.angle( x, y, ( x + diffX ), ( y + diffY ) ) );
        }
        return slopeAndAspect;
    }
    
    /**
     * Returns an Grid2DSquareCellDouble hollowFilledDEM which has cell
     * values as in grid except with any hollows raised so that they are not
     * hollows. All cells with height < outflowFilling is done by finding all
     * hollows DEM
     * TODO: docs
     * @param grid the Grid2DSquareCellDouble to be processed
     */
    public Grid2DSquareCellDouble getHollowFilledDEM(
            Grid2DSquareCellDouble grid,
            double outflowHeight,
            int maxIterations,
            HashSet outflowCellIDsSet,
            Grid2DSquareCellDoubleFactory gridFactory ) {
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( grid );
        long nrows = result.getNrows();
        long ncols = result.getNcols();
        double noDataValue = result.getNoDataValue();
        double minHeight = result.getGridStatistics().getMinDouble();
        double height;
        // Initialise outflowCellIDs
        HashSet outflowCellIDs = getHollowFilledDEMOutflowCellIDs( outflowCellIDsSet, outflowHeight, minHeight, grid, nrows, ncols, noDataValue );
        // Initialise hollowsHashSet
        HashSet hollowsHashSet = getHollowFilledDEMInitialHollowsHashSet( grid, nrows, ncols, noDataValue );
        // Remove outflowCellIDs from hollowsHashSet
        hollowsHashSet.removeAll( outflowCellIDs );
        HashSet hollows2 = hollowsHashSet;
        int numberOfHollows = hollowsHashSet.size();
        boolean calculated1 = false;
        boolean calculated2 = false;
        if ( numberOfHollows == 0 ) {
            calculated1 = true;
        }
        int iteration1 = 0;
        int iteration2;
        int counter1 = 0;
        long row;
        long col;
        long p;
        long q;
        long r;
        long s;
        
        //int cellID1;
        //int cellID2;
        //int cellID3;
        Iterator iterator1;
        Iterator iterator2;
        
        //Integer cellID1;
        //Integer cellID2;
        //Integer cellID3;
        CellID[] cellIDs = new CellID[3];
        
        HashSet toVisitSet1 = null;
        HashSet toVisitSet2 = null;
        HashSet toVisitSet3 = null;
        
        HashSet visitedSet1 = null;
        HashSet visitedSet2 = null;
        
        HashSet hollows1 = null;
        HashSet hollowsVisited = null;
        HashSet hollowSet = null;
        double height0;
        int noDataCount;
        int outflowCellCount;
        
        // Fill in hollows
        while ( ! calculated1 ) {
            if ( iteration1 < maxIterations ) {
                iteration1 ++;
                numberOfHollows = hollows2.size();
                System.out.println( "Iteration " + iteration1 + " out of a maximum " + maxIterations + ": Number of hollows " + numberOfHollows );
                if ( numberOfHollows > 0 ) {
                    visitedSet1 = new HashSet();
                    hollowsVisited = new HashSet();
                    
                    //hollowsVisited.addAll( outflowCellIDs );
                    
                    // Raise all hollows by a small amount
                    setValueALittleBitLarger( result, hollows2, this.handleOutOfMemoryErrorFalse );
                    
                    // Recalculate hollows in hollows2 neighbourhood
                    toVisitSet1 = new HashSet();
                    iterator1 = hollows2.iterator();
                    while ( iterator1.hasNext() ) {
                        cellIDs[ 0 ] = ( CellID ) iterator1.next();
                        row = cellIDs[ 0 ].getCellRowIndex();
                        col = cellIDs[ 0 ].getCellColIndex();
                        for ( p = -1; p < 2; p ++ ) {
                            for ( q = -1; q < 2; q ++ ) {
                                //if ( ! ( p == 0 && q == 0 ) ) {
                                if ( grid.inGrid( row + p, col + q, this.handleOutOfMemoryErrorFalse ) ) {
                                    toVisitSet1.add( grid.getCellID( row + p, col + q, this.handleOutOfMemoryErrorFalse ) );
                                }
                                //}
                            }
                        }
                    }
                    hollows1 = hollowFilledDEMCalculateHollowsInNeighbourhood( result, toVisitSet1 );
                    hollows1.removeAll( outflowCellIDs );
                    hollows2.clear();
                    toVisitSet1.clear();
                    /*
                    hollows1 = getHollowFilledDEMCalculateHollowsInNeighbourhood( result, hollows2 );
                    hollows1.removeAll( outflowCellIDs );
                    hollows2.clear();
                     */
                    // Trace bottom of each hollow and raise to the height of the lowest cell around it.
                    
                    
                    iterator1 = hollows1.iterator();
                    while ( iterator1.hasNext() ) {
                        cellIDs[ 0 ] = ( CellID ) iterator1.next();
                        if ( ! hollowsVisited.contains( cellIDs[ 0 ] ) ) {
                            hollowSet = new HashSet();
                            hollowSet.add( cellIDs[ 0 ] );
                            row = cellIDs[ 0 ].getCellRowIndex();
                            col = cellIDs[ 0 ].getCellColIndex();
                            toVisitSet1 = new HashSet();
                            
                            // Step 1: Add all cells in adjoining hollows to hollowSet
                            for ( p = -1; p < 2; p ++ ) {
                                for ( q = -1; q < 2; q ++ ) {
                                    if ( ! ( p == 0 && q == 0 ) ) {
                                        if ( grid.inGrid( row + p, col + q, this.handleOutOfMemoryErrorFalse ) ) {
                                            cellIDs[ 1 ] = grid.getCellID( row + p, col + q, this.handleOutOfMemoryErrorFalse );
                                            toVisitSet1.add( cellIDs[ 1 ] );
                                        }
                                    }
                                }
                            }
                            toVisitSet1.removeAll( outflowCellIDs );
                            visitedSet2 = new HashSet();
                            visitedSet2.add( cellIDs[ 0 ] );
                            toVisitSet3 = new HashSet();
                            toVisitSet3.addAll( toVisitSet1 );
                            calculated2 = false;
                            while ( ! calculated2 ) {
                                toVisitSet2 = new HashSet();
                                iterator2 = toVisitSet1.iterator();
                                while ( iterator2.hasNext() ) {
                                    cellIDs[ 1 ] = ( CellID ) iterator2.next();
                                    visitedSet2.add( cellIDs[ 1 ] );
                                    row = cellIDs[ 1 ].getCellRowIndex();
                                    col = cellIDs[ 1 ].getCellColIndex();
                                    for ( p = -1; p < 2; p ++ ) {
                                        for ( q = -1; q < 2; q ++ ) {
                                            if ( ! ( p == 0 && q == 0 ) ) {
                                                if ( grid.inGrid( row + p, col + q, this.handleOutOfMemoryErrorFalse ) ) {
                                                    cellIDs[ 2 ] = grid.getCellID( row + p, col + q, this.handleOutOfMemoryErrorFalse );
                                                    visitedSet1.add( cellIDs[ 2 ] );
                                                    // If a hollow then add to hollow set and visit neighbours if not done already
                                                    if ( hollows1.contains( cellIDs[ 2 ] ) ) {
                                                        hollowSet.add( cellIDs[ 2 ] );
                                                        for ( r = -1; r < 2; r ++ ) {
                                                            for ( s = -1; s < 2; s ++ ) {
                                                                if ( ! ( r == 0 && s == 0 ) ) { // Is this correct?
                                                                    if ( grid.inGrid( row + p + r, col + q + s, this.handleOutOfMemoryErrorFalse ) ) {
                                                                        toVisitSet2.add( grid.getCellID( row + p + r, col + q + s, this.handleOutOfMemoryErrorFalse ) );
                                                                    }
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                toVisitSet2.removeAll( outflowCellIDs );
                                toVisitSet3.addAll( toVisitSet2 );
                                toVisitSet1 = toVisitSet2;
                                toVisitSet1.removeAll( visitedSet2 );
                                if ( toVisitSet1.isEmpty() ) {
                                    calculated2 = true;
                                }
                            }
                            
                            // Step 2 Examine neighbours of each hollow
                            toVisitSet3.removeAll( hollowSet );
                            // NB. toVisitSet3 contains all cells which neighbour the traced hollow
                            calculated2 = false;
                            minHeight = Double.MAX_VALUE;
                            height0 = result.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                            while ( ! calculated2 ) {
                                toVisitSet2 = new HashSet();
                                //toVisitSet2.addAll( toVisitSet3 );
                                iterator2 = toVisitSet3.iterator();
                                noDataCount = 0;
                                outflowCellCount = 0;
                                
                                // Step 2.1 Calculate height of the lowest neighbour minHeight // (that is not an outflow cell???)
                                while ( iterator2.hasNext() ) {
                                    cellIDs[ 1 ] = ( CellID ) iterator2.next();
                                    row = cellIDs[ 1 ].getCellRowIndex();
                                    col = cellIDs[ 1 ].getCellColIndex();
                                    height = result.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                                    if ( height == noDataValue ) {
                                        noDataCount ++;
                                    } else {
                                        if ( outflowCellIDs.contains( cellIDs[ 1 ] ) ) {
                                            outflowCellCount ++;
                                        } else {
                                            minHeight = Math.min( minHeight, height );
                                        }
                                        // Is this correct?
                                        //minHeight = Math.min( minHeight, height );
                                    }
                                }
                                if ( noDataCount + outflowCellCount == toVisitSet3.size() ) {
                                    // System.out.println("Hollow surrounded by noDataValue or outflow cells!!!");
                                    // Add cellIDs of this hollow to outflowCellIDs so that it is not revisited.
                                    outflowCellIDs.addAll( hollowSet );
                                    calculated2 = true;
                                } else {
                                    // Step 2.2 Treat cells:
                                    // If minHeight is higher then add cells with this height to the
                                    // hollow set and their neighbours to toVisitSet2
                                    if ( minHeight > height0 ) {
                                        iterator2 = toVisitSet3.iterator();
                                        while ( iterator2.hasNext() ) {
                                            cellIDs[ 1 ] = ( CellID ) iterator2.next();
                                            row = cellIDs[ 1 ].getCellRowIndex();
                                            col = cellIDs[ 1 ].getCellColIndex();
                                            height = result.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                                            if ( height == minHeight ) {
                                                hollowSet.add( cellIDs[ 1 ] );
                                                toVisitSet2.remove( cellIDs[ 1 ] );
                                                for ( r = -1; r < 2; r ++ ) {
                                                    for ( s = -1; s < 2; s ++ ) {
                                                        if ( ! ( r == 0L && s == 0L ) ) {
                                                            if ( grid.inGrid( row + r, col + s, this.handleOutOfMemoryErrorFalse ) ) {
                                                                toVisitSet2.add( grid.getCellID( row + r, col + s, this.handleOutOfMemoryErrorFalse ) );
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        height0 = minHeight;
                                        toVisitSet2.removeAll( hollowSet );
                                        //toVisitSet2.removeAll( outflowCellIDs );
                                        toVisitSet3 = toVisitSet2;
                                    } else {
                                        calculated2 = true;
                                    }
                                }
                            }
                            // Step 3 Raise all cells in hollowSet
                            hollowSet.removeAll( outflowCellIDs );
                            iterator2 = hollowSet.iterator();
                            while ( iterator2.hasNext() ) {
                                cellIDs[ 1 ] = ( CellID ) iterator2.next();
                                row = cellIDs[ 1 ].getCellRowIndex();
                                col = cellIDs[ 1 ].getCellColIndex();
                                result.setCell( row, col, Utilities.getValueALittleBitLarger( height0 ), this.handleOutOfMemoryErrorFalse );
                            }
                            hollowsVisited.addAll( hollowSet );
                            visitedSet1.addAll( hollowSet );
                        }
                    }
                    hollows2 = hollowFilledDEMCalculateHollowsInNeighbourhood( result, visitedSet1 );
                } else {
                    calculated1 = true;
                }
            } else {
                calculated1 = true;
            }
        }
        return result;
    }
    
    /**
     * Returns outflowCellIDs a HashSet containing Integer cellIDs of cell IDs
     * in grid. These are cellIDs in outflowCellIDsSet or those cell IDs thats values
     *
     * @param outflowCellIDsSet
     * @param grid
     * @param minHeight the minimum value in grid
     */
    private HashSet getHollowFilledDEMOutflowCellIDs( HashSet outflowCellIDsSet, double outflowHeight, double minHeight, Grid2DSquareCellDouble grid, long nrows, long ncols, double noDataValue ) {
        HashSet outflowCellIDs = new HashSet();
        if ( !( outflowCellIDsSet == null ) ) {
            outflowCellIDs.addAll( outflowCellIDsSet );
        }
        double height;
        long row;
        long col;
        if ( outflowHeight < minHeight ) {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    height = grid.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                    if ( ( height != noDataValue ) && ( height == minHeight ) ) {
                        outflowCellIDs.add( grid.getCellID( row, col, this.handleOutOfMemoryErrorFalse ) );
                    }
                }
            }
        } else {
            for ( row = 0; row < nrows; row ++ ) {
                for ( col = 0; col < ncols; col ++ ) {
                    height = grid.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                    if ( ( height != noDataValue ) && ( height <= outflowHeight ) ) {
                        outflowCellIDs.add( grid.getCellID( row, col, this.handleOutOfMemoryErrorFalse ) );
                    }
                }
            }
        }
        return outflowCellIDs;
    }
    
    /**
     * Returns a HashSet containing cellIDs which identifies cells for which
     * neighbouring cells in the immediate 8 cell neighbourhood that are either
     * the same value, higher or noDataValues
     * @param grid - the Grid2DSquareCellDouble to be processed
     */
    private HashSet getHollowFilledDEMInitialHollowsHashSet( Grid2DSquareCellDouble grid, long nrows, long ncols, double noDataValue ) {
        HashSet initialHollowsHashSet = new HashSet();
        double[ ] heights = new double[ 9 ];
        int k;
        // Initialise hollows
        long row;
        long col;
        long p;
        long q;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                heights[ 0 ] = grid.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                if ( heights[ 0 ] != noDataValue ) {
                    k = 0;
                    for ( p = -1; p < 2; p ++ ) {
                        for ( q = -1; q < 2; q ++ ) {
                            if ( ! ( p == 0 && q == 0 ) ) {
                                k ++;
                                heights[ k ] = grid.getCell( row + p, col + q, this.handleOutOfMemoryErrorFalse );
                            }
                        }
                    }
                    // This deals with single isolated cells surrounded by noDataValues
                    if ( ( heights[ 1 ] >= heights[ 0 ] || heights[ 1 ] == noDataValue ) &&
                            ( heights[ 2 ] >= heights[ 0 ] || heights[ 2 ] == noDataValue ) &&
                            ( heights[ 3 ] >= heights[ 0 ] || heights[ 3 ] == noDataValue ) &&
                            ( heights[ 4 ] >= heights[ 0 ] || heights[ 4 ] == noDataValue ) &&
                            ( heights[ 5 ] >= heights[ 0 ] || heights[ 5 ] == noDataValue ) &&
                            ( heights[ 6 ] >= heights[ 0 ] || heights[ 6 ] == noDataValue ) &&
                            ( heights[ 7 ] >= heights[ 0 ] || heights[ 7 ] == noDataValue ) &&
                            ( heights[ 8 ] >= heights[ 0 ] || heights[ 8 ] == noDataValue ) ) {
                        initialHollowsHashSet.add( grid.getCellID( row, col, this.handleOutOfMemoryErrorFalse ) );
                    }
                }
            }
        }
        return initialHollowsHashSet;
    }
    
    /**
     * Returns a HashSet of cellIDs for identifying any cells that might be
     * hollows in grid, only those cells with IDs in the neighbourhood of those
     * cells with IDs in cellIDs need be checked.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param cellIDs the HashSet storing cellIDs that must be examined.
     */
    private HashSet hollowFilledDEMCalculateHollowsInNeighbourhood( Grid2DSquareCellDouble grid, HashSet cellIDs ) {
        double noDataValue = grid.getNoDataValue();
        HashSet result = new HashSet();
        HashSet visited1 = new HashSet();
        CellID cellID;
        long row;
        long col;
        long a;
        long b;
        long p;
        long q;
        double[ ] heights = new double[ 9 ];
        int k;
        Iterator iterator1 = cellIDs.iterator();
        while ( iterator1.hasNext() ) {
            cellID = ( CellID ) iterator1.next();
            if ( ! visited1.contains( cellID ) ) {
                row = cellID.getCellRowIndex();
                col = cellID.getCellColIndex();
                // Examine neighbourhood
                for ( a = -1; a < 2; a ++ ) {
                    for ( b = -1; b < 2; b ++ ) {
                        visited1.add( grid.getCellID( row + a, col + b, this.handleOutOfMemoryErrorFalse ) );
                        heights[ 0 ] = grid.getCell( row + a, col + b, this.handleOutOfMemoryErrorFalse );
                        if ( heights[ 0 ] != noDataValue ) {
                            k = 0;
                            for ( p = -1; p < 2; p ++ ) {
                                for ( q = -1; q < 2; q ++ ) {
                                    if ( ! ( p == 0 && q == 0 ) ) {
                                        k ++;
                                        heights[ k ] = grid.getCell( row + a + q, col + b + p, this.handleOutOfMemoryErrorFalse );
                                    }
                                }
                            }
                            if ( ( heights[ 1 ] >= heights[ 0 ] || heights[ 1 ] == noDataValue ) &&
                                    ( heights[ 2 ] >= heights[ 0 ] || heights[ 2 ] == noDataValue ) &&
                                    ( heights[ 3 ] >= heights[ 0 ] || heights[ 3 ] == noDataValue ) &&
                                    ( heights[ 4 ] >= heights[ 0 ] || heights[ 4 ] == noDataValue ) &&
                                    ( heights[ 5 ] >= heights[ 0 ] || heights[ 5 ] == noDataValue ) &&
                                    ( heights[ 6 ] >= heights[ 0 ] || heights[ 6 ] == noDataValue ) &&
                                    ( heights[ 7 ] >= heights[ 0 ] || heights[ 7 ] == noDataValue ) &&
                                    ( heights[ 8 ] >= heights[ 0 ] || heights[ 8 ] == noDataValue ) ) {
                                result.add( grid.getCellID( row + a, col + b, this.handleOutOfMemoryErrorFalse ) );
                            }
                        }
                    }
                }
            }
        }
        return result;
    }
    
    
    private HashSet hollowFilledDEMCalculateHollows( Grid2DSquareCellDouble grid, HashSet cellIDs ) {
        if ( ( grid.getNcols() * grid.getNrows() ) / 4 < cellIDs.size() ) {
            // return getInitialHollowsHashSet( grid );
        }
        double noDataValue = grid.getNoDataValue();
        HashSet result = new HashSet();
        CellID cellID;
        long row;
        long col;
        long p;
        long q;
        double[ ] heights = new double[ 9 ];
        int k;
        //int noDataCount;
        Iterator iterator1 = cellIDs.iterator();
        while ( iterator1.hasNext() ) {
            cellID = ( CellID ) iterator1.next();
            row = cellID.getCellRowIndex();
            col = cellID.getCellColIndex();
            heights[ 0 ] = grid.getCell( row, col, this.handleOutOfMemoryErrorFalse );
            if ( heights[ 0 ] != noDataValue ) {
                //noDataCount = 0;
                k = 0;
                for ( p = -1; p < 2; p ++ ) {
                    for ( q = -1; q < 2; q ++ ) {
                        if ( ! ( p == 0 && q == 0 ) ) {
                            k ++;
                            heights[ k ] = grid.getCell( row + p, col + q, this.handleOutOfMemoryErrorFalse );
                            //if ( heights[ k ] == noDataValue ) {
                            //    noDataCount ++;
                            //}
                        }
                    }
                }
                // This deals with single isolated cells surrounded by noDataValues
                //if ( noDataCount < 8 ) {
                if ( ( heights[ 1 ] >= heights[ 0 ] || heights[ 1 ] == noDataValue ) &&
                        ( heights[ 2 ] >= heights[ 0 ] || heights[ 2 ] == noDataValue ) &&
                        ( heights[ 3 ] >= heights[ 0 ] || heights[ 3 ] == noDataValue ) &&
                        ( heights[ 4 ] >= heights[ 0 ] || heights[ 4 ] == noDataValue ) &&
                        ( heights[ 5 ] >= heights[ 0 ] || heights[ 5 ] == noDataValue ) &&
                        ( heights[ 6 ] >= heights[ 0 ] || heights[ 6 ] == noDataValue ) &&
                        ( heights[ 7 ] >= heights[ 0 ] || heights[ 7 ] == noDataValue ) &&
                        ( heights[ 8 ] >= heights[ 0 ] || heights[ 8 ] == noDataValue ) ) {
                    result.add( cellID );
                }
                //}
            }
        }
        return result;
    }
    
    /**
     * Returns an Grid2DSquareCellDouble[] metrics1 where:
     * metrics1[0] = no data count;
     * metrics1[1] = flatness;
     * metrics1[2] = roughness;
     * metrics1[3] = slopyness;
     * metrics1[4] = levelness;
     * metrics1[5] = totalDownness;
     * metrics1[6] = averageDownness;
     * metrics1[7] = totalUpness;
     * metrics1[8] = averageUpness;
     * metrics1[9] = maxd_hhhh [ sum of distance weighted maximum height differences ];
     * metrics1[10] = mind_hhhh [ sum of distance weighted minimum height differences ];
     * metrics1[11] = sumd_hhhh [ sum of distance weighted height differences ];
     * metrics1[12] = aved_hhhh [ sum of distance weighted average height difference ];
     * metrics1[13] = count_hhhh [ count ];
     * metrics1[14] = w_hhhh [ sum of distance weights ];
     * metrics1[15] = mind_hxhx_ai_hhhl [ sum of distance weighted ( minimum difference of cells adjacent to lower cell ) ];
     * metrics1[16] = maxd_hxhx_ai_hhhl [ sum of distance weighted ( maximum difference of cells adjacent to lower cell ) ];
     * metrics1[17] = sumd_hxhx_ai_hhhl [ sum of distance weighted ( sum of differences of cells adjacent to lower cell ) ];
     * metrics1[18] = d_xhxx_ai_hhhl [ sum of distance weighted ( difference of cell opposite lower cell ) ];
     * metrics1[19] = d_xxxl_ai_hhhl [ sum of distance weighted ( difference of lower cell ) ];
     * metrics1[20] = sumd_xhxl_ai_hhhl [ sum of distance weighted ( sum of differences of lower cell and cell opposite ) ];
     * metrics1[21] = mind_abs_xhxl_ai_hhhl [ sum of distance weighted ( minimum difference magnitude of lower cell and cell opposite ) ];
     * metrics1[22] = maxd_abs_xhxl_ai_hhhl [ sum of distance weighted ( maximum difference magnitude of lower cell and cell opposite ) ];
     * metrics1[23] = sumd_abs_xhxl_ai_hhhl [ sum of distance weighted ( sum of difference magnitudes of lower cell and cell opposite ) ];
     * metrics1[24] = count_hhhl [ count ];
     * metrics1[25] = w_hhhl [ sum of distance weights ];
     * metrics1[26] = mind_hxhx_ai_hlhl [ sum of distance weighted ( minimum difference of higher cells ) ];
     * metrics1[27] = maxd_hxhx_ai_hlhl [ sum of distance weighted ( maximum difference of higher cells ) ];
     * metrics1[28] = sumd_hxhx_ai_hlhl [ sum of distance weighted ( sum differences of higher cells ) ];
     * metrics1[29] = mind_xlxl_ai_hlhl [ sum of distance weighted ( minimum difference of lower cells ) ];
     * metrics1[30] = maxd_xlxl_ai_hlhl [ sum of distance weighted ( maximum difference of lower cells ) ];
     * metrics1[31] = sumd_xlxl_ai_hlhl [ sum of distance weighted ( sum of differences of lower cells ) ];
     * metrics1[32] = mind_abs_hlhl [ sum of distance weighted ( minimum difference magnitude of cells ) ];
     * metrics1[33] = maxd_abs_hlhl [ sum of distance weighted ( maximum difference magnitude of cells ) ];
     * metrics1[34] = sumd_abs_hlhl [ sum of distance weighted ( sum of difference magnitudes of cells ) ];
     * metrics1[35] = count_hlhl [ count ];
     * metrics1[36] = w_hlhl [ sum of distance weights ];
     * metrics1[37] = mind_hhxx_ai_hhll [ sum of distance weighted ( minimum difference of higher cells ) ];
     * metrics1[38] = maxd_hhxx_ai_hhll [ sum of distance weighted ( maximum difference of higher cells ) ];
     * metrics1[39] = sumd_hhxx_ai_hhll [ sum of distance weighted ( sum of differences of higher cells ) ];
     * metrics1[40] = mind_xxll_ai_hhll [ sum of distance weighted ( minimum difference of lower cells ) ];
     * metrics1[41] = maxd_xxll_ai_hhll [ sum of distance weighted ( maximum difference of lower cells ) ];
     * metrics1[42] = sumd_xxll_ai_hhll [ sum of distance weighted ( sum of differences of lower cells ) ];
     * metrics1[43] = mind_abs_hhll [ sum of distance weighted ( minimum difference magnitude of cells ) ];
     * metrics1[44] = maxd_abs_hhll [ sum of distance weighted ( maximum difference magnitude of cells ) ];
     * metrics1[45] = sumd_abs_hhll [ sum of distance weighted ( sum of difference magnitudes of cells ) ];
     * metrics1[46] = count_hhll [ count ];
     * metrics1[47] = w_hhll [ sum of distance weights ];
     * metrics1[48] = mind_lxlx_ai_lllh [ sum of distance weighted ( minimum difference of cells adjacent to higher cell ) ];
     * metrics1[49] = maxd_lxlx_ai_lllh [ sum of distance weighted ( maximum difference of cells adjacent to higher cell ) ];
     * metrics1[50] = sumd_lxlx_ai_lllh [ sum of distance weighted ( sum of differences of cells adjacent to higher cell ) ];
     * metrics1[51] = d_xlxx_ai_lllh [ sum of distance weighted ( difference of cell opposite higher cell ) ];
     * metrics1[52] = d_xxxh_ai_lllh [ sum of distance weighted ( difference of higher cell ) ];
     * metrics1[53] = sumd_xlxh_ai_lllh [ sum of distance weighted ( sum of differences of higher cell and cell opposite ) ];
     * metrics1[54] = mind_abs_xlxh_ai_lllh [ sum of distance weighted ( minimum difference magnitude of higher cell and cell opposite ) ];
     * metrics1[55] = maxd_abs_xlxh_ai_lllh [ sum of distance weighted ( maximum difference magnitude of higher cell and cell opposite ) ];
     * metrics1[56] = sumd_abs_xlxh_ai_lllh [ sum of distance weighted ( sum of difference magnitudes of higher cell and cell opposite ) ];
     * metrics1[57] = count_lllh [ count ];
     * metrics1[58] = w_lllh [ sum of distance weights ];
     * metrics1[59] = maxd_llll [ sum of distance weighted maximum height differences ];
     * metrics1[60] = mind_llll [ sum of distance weighted minimum height differences ];
     * metrics1[61] = sumd_llll [ sum of distance weighted height differences ];
     * metrics1[62] = aved_llll [ sum of distance weighted average height difference ];
     * metrics1[63] = count_llll [ count ];
     * metrics1[64] = w_llll [ sum of distance weights ];
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param distance the distance within which metrics will be calculated
     * @param weightIntersect kernel parameter ( weight at the centre )
     * @param weightFactor kernel parameter ( distance decay )
     * @param gridFactory the preferred factory for creating grids
     */
    public Grid2DSquareCellDouble[] getMetrics1(
            Grid2DSquareCellDouble grid,
            double distance,
            double weightIntersect,
            double weightFactor,
            Grid2DSquareCellDoubleFactory gridFactory ) {
        Grid2DSquareCellDouble[] metrics1 = new Grid2DSquareCellDouble[ 65 ];
        long ncols = grid.getNcols();
        long nrows = grid.getNrows();
        BigDecimal[] dimensions = grid.getDimensions();
        double noDataValue = Double.NEGATIVE_INFINITY;
        int cachedIndex = 0;
        boolean isInitialised = false;
        for ( int i = 0; i < metrics1.length; i ++ ) {
            do {
                try {
                    isInitialised = false;
                    metrics1[ i ] = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, dimensions, noDataValue );
                    System.out.println( "Initialised result[" + i + "]" );
                    isInitialised = true;
                } catch ( OutOfMemoryError oome0 ) {
                    clearMemoryReserve();
                    for ( int ii = cachedIndex; ii < i; ii ++ ) {
                        metrics1[ ii ].cacheChunks( false );
                        metrics1[ ii ].clearChunks( false );
                    }
                    initMemoryReserve();
                }
            } while ( ! isInitialised );
        }
        return getMetrics1( metrics1, grid, dimensions, distance, weightIntersect, weightFactor );
    }
    
    /**
     * Returns an Grid2DSquareCellDouble[] metrics1 where:
     * metrics1[0] = no data count;
     * metrics1[1] = flatness;
     * metrics1[2] = roughness;
     * metrics1[3] = slopyness;
     * metrics1[4] = levelness;
     * metrics1[5] = totalDownness;
     * metrics1[6] = averageDownness;
     * metrics1[7] = totalUpness;
     * metrics1[8] = averageUpness;
     * metrics1[9] = maxd_hhhh [ sum of distance weighted maximum height differences ];
     * metrics1[10] = mind_hhhh [ sum of distance weighted minimum height differences ];
     * metrics1[11] = sumd_hhhh [ sum of distance weighted height differences ];
     * metrics1[12] = aved_hhhh [ sum of distance weighted average height difference ];
     * metrics1[13] = count_hhhh [ count ];
     * metrics1[14] = w_hhhh [ sum of distance weights ];
     * metrics1[15] = mind_hxhx_ai_hhhl [ sum of distance weighted ( minimum difference of cells adjacent to lower cell ) ];
     * metrics1[16] = maxd_hxhx_ai_hhhl [ sum of distance weighted ( maximum difference of cells adjacent to lower cell ) ];
     * metrics1[17] = sumd_hxhx_ai_hhhl [ sum of distance weighted ( sum of differences of cells adjacent to lower cell ) ];
     * metrics1[18] = d_xhxx_ai_hhhl [ sum of distance weighted ( difference of cell opposite lower cell ) ];
     * metrics1[19] = d_xxxl_ai_hhhl [ sum of distance weighted ( difference of lower cell ) ];
     * metrics1[20] = sumd_xhxl_ai_hhhl [ sum of distance weighted ( sum of differences of lower cell and cell opposite ) ];
     * metrics1[21] = mind_abs_xhxl_ai_hhhl [ sum of distance weighted ( minimum difference magnitude of lower cell and cell opposite ) ];
     * metrics1[22] = maxd_abs_xhxl_ai_hhhl [ sum of distance weighted ( maximum difference magnitude of lower cell and cell opposite ) ];
     * metrics1[23] = sumd_abs_xhxl_ai_hhhl [ sum of distance weighted ( sum of difference magnitudes of lower cell and cell opposite ) ];
     * metrics1[24] = count_hhhl [ count ];
     * metrics1[25] = w_hhhl [ sum of distance weights ];
     * metrics1[26] = mind_hxhx_ai_hlhl [ sum of distance weighted ( minimum difference of higher cells ) ];
     * metrics1[27] = maxd_hxhx_ai_hlhl [ sum of distance weighted ( maximum difference of higher cells ) ];
     * metrics1[28] = sumd_hxhx_ai_hlhl [ sum of distance weighted ( sum differences of higher cells ) ];
     * metrics1[29] = mind_xlxl_ai_hlhl [ sum of distance weighted ( minimum difference of lower cells ) ];
     * metrics1[30] = maxd_xlxl_ai_hlhl [ sum of distance weighted ( maximum difference of lower cells ) ];
     * metrics1[31] = sumd_xlxl_ai_hlhl [ sum of distance weighted ( sum of differences of lower cells ) ];
     * metrics1[32] = mind_abs_hlhl [ sum of distance weighted ( minimum difference magnitude of cells ) ];
     * metrics1[33] = maxd_abs_hlhl [ sum of distance weighted ( maximum difference magnitude of cells ) ];
     * metrics1[34] = sumd_abs_hlhl [ sum of distance weighted ( sum of difference magnitudes of cells ) ];
     * metrics1[35] = count_hlhl [ count ];
     * metrics1[36] = w_hlhl [ sum of distance weights ];
     * metrics1[37] = mind_hhxx_ai_hhll [ sum of distance weighted ( minimum difference of higher cells ) ];
     * metrics1[38] = maxd_hhxx_ai_hhll [ sum of distance weighted ( maximum difference of higher cells ) ];
     * metrics1[39] = sumd_hhxx_ai_hhll [ sum of distance weighted ( sum of differences of higher cells ) ];
     * metrics1[40] = mind_xxll_ai_hhll [ sum of distance weighted ( minimum difference of lower cells ) ];
     * metrics1[41] = maxd_xxll_ai_hhll [ sum of distance weighted ( maximum difference of lower cells ) ];
     * metrics1[42] = sumd_xxll_ai_hhll [ sum of distance weighted ( sum of differences of lower cells ) ];
     * metrics1[43] = mind_abs_hhll [ sum of distance weighted ( minimum difference magnitude of cells ) ];
     * metrics1[44] = maxd_abs_hhll [ sum of distance weighted ( maximum difference magnitude of cells ) ];
     * metrics1[45] = sumd_abs_hhll [ sum of distance weighted ( sum of difference magnitudes of cells ) ];
     * metrics1[46] = count_hhll [ count ];
     * metrics1[47] = w_hhll [ sum of distance weights ];
     * metrics1[48] = mind_lxlx_ai_lllh [ sum of distance weighted ( minimum difference of cells adjacent to higher cell ) ];
     * metrics1[49] = maxd_lxlx_ai_lllh [ sum of distance weighted ( maximum difference of cells adjacent to higher cell ) ];
     * metrics1[50] = sumd_lxlx_ai_lllh [ sum of distance weighted ( sum of differences of cells adjacent to higher cell ) ];
     * metrics1[51] = d_xlxx_ai_lllh [ sum of distance weighted ( difference of cell opposite higher cell ) ];
     * metrics1[52] = d_xxxh_ai_lllh [ sum of distance weighted ( difference of higher cell ) ];
     * metrics1[53] = sumd_xlxh_ai_lllh [ sum of distance weighted ( sum of differences of higher cell and cell opposite ) ];
     * metrics1[54] = mind_abs_xlxh_ai_lllh [ sum of distance weighted ( minimum difference magnitude of higher cell and cell opposite ) ];
     * metrics1[55] = maxd_abs_xlxh_ai_lllh [ sum of distance weighted ( maximum difference magnitude of higher cell and cell opposite ) ];
     * metrics1[56] = sumd_abs_xlxh_ai_lllh [ sum of distance weighted ( sum of difference magnitudes of higher cell and cell opposite ) ];
     * metrics1[57] = count_lllh [ count ];
     * metrics1[58] = w_lllh [ sum of distance weights ];
     * metrics1[59] = maxd_llll [ sum of distance weighted maximum height differences ];
     * metrics1[60] = mind_llll [ sum of distance weighted minimum height differences ];
     * metrics1[61] = sumd_llll [ sum of distance weighted height differences ];
     * metrics1[62] = aved_llll [ sum of distance weighted average height difference ];
     * metrics1[63] = count_llll [ count ];
     * metrics1[64] = w_llll [ sum of distance weights ];
     * @param metrics1 an Grid2DSquareCellDouble[] for storing result
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param distance the distance within which metrics will be calculated
     * @param weightIntersect kernel parameter ( weight at the centre )
     * @param weightFactor kernel parameter ( distance decay )
     * Going directly to this method is useful if the initialisation of the
     * metrics1 is slow and has already been done.
     */
    public Grid2DSquareCellDouble[] getMetrics1(
            Grid2DSquareCellDouble[] metrics1,
            Grid2DSquareCellDouble grid,
            BigDecimal[] dimensions,
            double distance,
            double weightIntersect,
            double weightFactor ) {
        long ncols = grid.getNcols();
        long nrows = grid.getNrows();
        double cellsize = dimensions[0].doubleValue();
        double gridNoDataValue = grid.getNoDataValue();
        double noDataValue = Double.MIN_VALUE;
        double[] heights = new double[ 4 ];
        double[] diff = new double[ 4 ];
        double[] dummyDiff = new double[ 4 ];
        double[][] weights = Kernel.getKernelWeights( grid, distance, weightIntersect, weightFactor );
        double[] metrics1ForCell = new double[ metrics1.length ];
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                if ( grid.getCell( row, col, this.handleOutOfMemoryErrorFalse ) != gridNoDataValue ) {
                    metrics1Calculate_All( grid, cellsize, row, col, distance, weights, metrics1ForCell, heights, diff, dummyDiff );
                    for ( int i = 0; i < metrics1.length; i ++ ) {
                        metrics1[ i ].setCell( row, col, metrics1ForCell[ i ], this.handleOutOfMemoryErrorFalse );
                    }
                    //} else {
                    //    for ( int i = 0; i < metrics1.length; i ++ ) {
                    //        metrics1[ i ].setCell( row, col, noDataValue );
                    //    }
                }
            }
            System.out.println("Done row " + row );
        }
        metrics1[ 0 ].setName( "noDataCount", this.handleOutOfMemoryErrorFalse );
        metrics1[ 1 ].setName( "flatness", this.handleOutOfMemoryErrorFalse );
        metrics1[ 2 ].setName( "roughness", this.handleOutOfMemoryErrorFalse );
        metrics1[ 3 ].setName( "slopyness", this.handleOutOfMemoryErrorFalse );
        metrics1[ 4 ].setName( "levelness", this.handleOutOfMemoryErrorFalse );
        metrics1[ 5 ].setName( "totalDownness", this.handleOutOfMemoryErrorFalse );
        metrics1[ 6 ].setName( "averageDownness", this.handleOutOfMemoryErrorFalse );
        metrics1[ 7 ].setName( "totalUpness", this.handleOutOfMemoryErrorFalse );
        metrics1[ 8 ].setName( "averageUpness", this.handleOutOfMemoryErrorFalse );
        metrics1[ 9 ].setName( "maxd_hhhh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 10 ].setName( "mind_hhhh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 11 ].setName( "sumd_hhhh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 12 ].setName( "aved_hhhh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 13 ].setName( "count_hhhh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 14 ].setName( "w_hhhh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 15 ].setName( "mind_hxhx_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 16 ].setName( "maxd_hxhx_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 17 ].setName( "sumd_hxhx_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 18 ].setName( "d_xhxx_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 19 ].setName( "d_xxxl_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 20 ].setName( "sumd_xhxl_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 21 ].setName( "mind_abs_xhxl_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 22 ].setName( "maxd_abs_xhxl_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 23 ].setName( "sumd_abs_xhxl_ai_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 24 ].setName( "count_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 25 ].setName( "w_hhhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 26 ].setName( "mind_hxhx_ai_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 27 ].setName( "maxd_hxhx_ai_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 28 ].setName( "sumd_hxhx_ai_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 29 ].setName( "mind_xlxl_ai_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 30 ].setName( "maxd_xlxl_ai_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 31 ].setName( "sumd_xlxl_ai_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 32 ].setName( "mind_abs_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 33 ].setName( "maxd_abs_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 34 ].setName( "sumd_abs_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 35 ].setName( "count_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 36 ].setName( "w_hlhl", this.handleOutOfMemoryErrorFalse );
        metrics1[ 37 ].setName( "mind_hhxx_ai_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 38 ].setName( "maxd_hhxx_ai_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 39 ].setName( "sumd_hhxx_ai_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 40 ].setName( "mind_xxll_ai_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 41 ].setName( "maxd_xxll_ai_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 42 ].setName( "sumd_xxll_ai_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 43 ].setName( "mind_abs_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 44 ].setName( "maxd_abs_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 45 ].setName( "sumd_abs_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 46 ].setName( "count_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 47 ].setName( "w_hhll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 48 ].setName( "mind_lxlx_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 49 ].setName( "maxd_lxlx_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 50 ].setName( "sumd_lxlx_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 51 ].setName( "d_xlxx_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 52 ].setName( "d_xxxh_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 53 ].setName( "sumd_xlxh_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 54 ].setName( "mind_abs_xlxh_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 55 ].setName( "maxd_abs_xlxh_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 56 ].setName( "sumd_abs_xlxh_ai_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 57 ].setName( "count_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 58 ].setName( "w_lllh", this.handleOutOfMemoryErrorFalse );
        metrics1[ 59 ].setName( "maxd_llll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 60 ].setName( "mind_llll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 61 ].setName( "sumd_llll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 62 ].setName( "aved_llll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 63 ].setName( "count_llll", this.handleOutOfMemoryErrorFalse );
        metrics1[ 64 ].setName( "w_llll", this.handleOutOfMemoryErrorFalse );
        return metrics1;
    }
    
    /**
     * Returns a double[] metrics1 of the cells in grid upto distance from a
     * cell given by rowIndex and colIndex. The elements of metrics1 do not
     * explicitly take into account any axis such as that which can be defined
     * from a metric of slope (general slope direction). Distance weighting is
     * done via a kernel precalculated as weights. Some elements of metrics1 are
     * weighted based on the difference in value (height) of the cell at
     * (rowIndex,colIndex) and other cell values within distance. Within
     * distance equidistant cells in 4 orthoganol directions are accounted for
     * in the metrics1.
     * NB. Every cell is either higher, lower or the same height as the cell
     * at (rowIndex,colIndex). Some DEMs will have few cells in distance with
     * the same value.
     * 9 basic metrics:
     * metrics1[0] = no data count;
     * metrics1[1] = flatness;
     * metrics1[2] = roughness;
     * metrics1[3] = slopyness;
     * metrics1[4] = levelness;
     * metrics1[5] = totalDownness;
     * metrics1[6] = averageDownness;
     * metrics1[7] = totalUpness;
     * metrics1[8] = averageUpness;
     * 6 metrics with all cells higher or same:
     * metrics1[9] = maxd_hhhh [ sum of distance weighted maximum height differences ];
     * metrics1[10] = mind_hhhh [ sum of distance weighted minimum height differences ];
     * metrics1[11] = sumd_hhhh [ sum of distance weighted height differences ];
     * metrics1[12] = aved_hhhh [ sum of distance weighted average height difference ];
     * metrics1[13] = count_hhhh [ count ];
     * 11 metrics with one cell lower or same:
     * metrics1[14] = w_hhhh [ sum of distance weights ];
     * metrics1[15] = mind_hxhx_ai_hhhl [ sum of distance weighted ( minimum difference of cells adjacent to lower cell ) ];
     * metrics1[16] = maxd_hxhx_ai_hhhl [ sum of distance weighted ( maximum difference of cells adjacent to lower cell ) ];
     * metrics1[17] = sumd_hxhx_ai_hhhl [ sum of distance weighted ( sum of differences of cells adjacent to lower cell ) ];
     * metrics1[18] = d_xhxx_ai_hhhl [ sum of distance weighted ( difference of cell opposite lower cell ) ];
     * metrics1[19] = d_xxxl_ai_hhhl [ sum of distance weighted ( difference of lower cell ) ];
     * metrics1[20] = sumd_xhxl_ai_hhhl [ sum of distance weighted ( sum of differences of lower cell and cell opposite ) ];
     * metrics1[21] = mind_abs_xhxl_ai_hhhl [ sum of distance weighted ( minimum difference magnitude of lower cell and cell opposite ) ];
     * metrics1[22] = maxd_abs_xhxl_ai_hhhl [ sum of distance weighted ( maximum difference magnitude of lower cell and cell opposite ) ];
     * metrics1[23] = sumd_abs_xhxl_ai_hhhl [ sum of distance weighted ( sum of difference magnitudes of lower cell and cell opposite ) ];
     * metrics1[24] = count_hhhl [ count ];
     * 22 metrics with two cells lower:
     * (N.B. Could have more metrics e.g. minimum magnitude of minimum of higher
     * and maximum of lower cells.)
     * Metrics with opposite cells lower/higher:
     * metrics1[25] = w_hhhl [ sum of distance weights ];
     * metrics1[26] = mind_hxhx_ai_hlhl [ sum of distance weighted ( minimum difference of higher cells ) ];
     * metrics1[27] = maxd_hxhx_ai_hlhl [ sum of distance weighted ( maximum difference of higher cells ) ];
     * metrics1[28] = sumd_hxhx_ai_hlhl [ sum of distance weighted ( sum differences of higher cells ) ];
     * metrics1[29] = mind_xlxl_ai_hlhl [ sum of distance weighted ( minimum difference of lower cells ) ];
     * metrics1[30] = maxd_xlxl_ai_hlhl [ sum of distance weighted ( maximum difference of lower cells ) ];
     * metrics1[31] = sumd_xlxl_ai_hlhl [ sum of distance weighted ( sum of differences of lower cells ) ];
     * metrics1[32] = mind_abs_hlhl [ sum of distance weighted ( minimum difference magnitude of cells ) ];
     * metrics1[33] = maxd_abs_hlhl [ sum of distance weighted ( maximum difference magnitude of cells ) ];
     * metrics1[34] = sumd_abs_hlhl [ sum of distance weighted ( sum of difference magnitudes of cells ) ];
     * metrics1[35] = count_hlhl [ count ];
     * metrics1[36] = w_hlhl [ sum of distance weights ];
     * Metrics with adjacent cells lower/higher:
     * metrics1[37] = mind_hhxx_ai_hhll [ sum of distance weighted ( minimum difference of higher cells ) ];
     * metrics1[38] = maxd_hhxx_ai_hhll [ sum of distance weighted ( maximum difference of higher cells ) ];
     * metrics1[39] = sumd_hhxx_ai_hhll [ sum of distance weighted ( sum of differences of higher cells ) ];
     * metrics1[40] = mind_xxll_ai_hhll [ sum of distance weighted ( minimum difference of lower cells ) ];
     * metrics1[41] = maxd_xxll_ai_hhll [ sum of distance weighted ( maximum difference of lower cells ) ];
     * metrics1[42] = sumd_xxll_ai_hhll [ sum of distance weighted ( sum of differences of lower cells ) ];
     * metrics1[43] = mind_abs_hhll [ sum of distance weighted ( minimum difference magnitude of cells ) ];
     * metrics1[44] = maxd_abs_hhll [ sum of distance weighted ( maximum difference magnitude of cells ) ];
     * metrics1[45] = sumd_abs_hhll [ sum of distance weighted ( sum of difference magnitudes of cells ) ];
     * metrics1[46] = count_hhll [ count ];
     * metrics1[47] = w_hhll [ sum of distance weights ];
     * 11 metrics with one cell higher:
     * metrics1[48] = mind_lxlx_ai_lllh [ sum of distance weighted ( minimum difference of cells adjacent to higher cell ) ];
     * metrics1[49] = maxd_lxlx_ai_lllh [ sum of distance weighted ( maximum difference of cells adjacent to higher cell ) ];
     * metrics1[50] = sumd_lxlx_ai_lllh [ sum of distance weighted ( sum of differences of cells adjacent to higher cell ) ];
     * metrics1[51] = d_xlxx_ai_lllh [ sum of distance weighted ( difference of cell opposite higher cell ) ];
     * metrics1[52] = d_xxxh_ai_lllh [ sum of distance weighted ( difference of higher cell ) ];
     * metrics1[53] = sumd_xlxh_ai_lllh [ sum of distance weighted ( sum of differences of higher cell and cell opposite ) ];
     * metrics1[54] = mind_abs_xlxh_ai_lllh [ sum of distance weighted ( minimum difference magnitude of higher cell and cell opposite ) ];
     * metrics1[55] = maxd_abs_xlxh_ai_lllh [ sum of distance weighted ( maximum difference magnitude of higher cell and cell opposite ) ];
     * metrics1[56] = sumd_abs_xlxh_ai_lllh [ sum of distance weighted ( sum of difference magnitudes of higher cell and cell opposite ) ];
     * metrics1[57] = count_lllh [ count ];
     * metrics1[58] = w_lllh [ sum of distance weights ];
     * 6 metrics with all cells higher:
     * metrics1[59] = maxd_llll [ sum of distance weighted maximum height differences ];
     * metrics1[60] = mind_llll [ sum of distance weighted minimum height differences ];
     * metrics1[61] = sumd_llll [ sum of distance weighted height differences ];
     * metrics1[62] = aved_llll [ sum of distance weighted average height difference ];
     * metrics1[63] = count_llll [ count ];
     * metrics1[64] = w_llll [ sum of distance weights ];
     * @param grid the Grid2DSquareCellDouble being processed
     * @param rowIndex the row index of the cell being classified
     * @param colIndex the column index of the cell being classified
     * @param distance the distance within which metrics1 will be calculated
     * @param weights an array of kernel weights for weighting metrics1
     */
    private void metrics1Calculate_All(
            Grid2DSquareCellDouble grid,
            double cellsize,
            long rowIndex,
            long colIndex,
            double distance,
            double[][] weights,
            double[] metrics1,
            double[] heights,
            double[] diff,
            double[] dummyDiff ) {
        for ( int i = 0; i < metrics1.length; i ++ ) {
            metrics1[ i ] = 0.0d;
        }
        double noDataValue = grid.getNoDataValue();
        double cellHeight = grid.getCell( rowIndex, colIndex, this.handleOutOfMemoryErrorFalse );
        int cellDistance = ( int ) Math.ceil( distance / cellsize );
        double cellX = grid.getCellX( colIndex, this.handleOutOfMemoryErrorFalse );
        double cellY = grid.getCellY( rowIndex, this.handleOutOfMemoryErrorFalse );
        double thisCellX;
        double thisCellY;
        double weight;
        double upCount;
        double downCount;
        double upness;
        double downness;
        double averageDiff;
        double averageHeight;
        double noDataCount;
        double xDiff;
        double yDiff;
        double sumWeight;
        int p;
        int q;
        for ( p = 0; p <= cellDistance; p ++ ) {
            for ( q = 1; q <= cellDistance; q ++ ) {
                //if ( p == 2 && q == 2 ) { int debug2 = 0; }
                noDataCount = 0.0d;
                thisCellX = grid.getCellX( colIndex + q, this.handleOutOfMemoryErrorFalse );
                thisCellY = grid.getCellY( rowIndex + p, this.handleOutOfMemoryErrorFalse );
                weight = weights[ p ][ q ];
                xDiff = thisCellX - cellX;
                yDiff = thisCellY - cellY;
                heights[ 0 ] = grid.getCell( thisCellX, thisCellY, this.handleOutOfMemoryErrorFalse );
                if ( heights[ 0 ] == noDataValue ) {
                    heights[ 0 ] = cellHeight;
                    noDataCount += 1.0d;
                }
                heights[ 1 ] = grid.getCell( cellX + yDiff, cellY - xDiff, this.handleOutOfMemoryErrorFalse );
                if ( heights[ 1 ] == noDataValue ) {
                    heights[ 1 ] = cellHeight;
                    noDataCount += 1.0d;
                }
                heights[ 2 ] = grid.getCell( cellX - xDiff, cellY - yDiff, this.handleOutOfMemoryErrorFalse );
                if ( heights[ 2 ] == noDataValue ) {
                    heights[ 2 ] = cellHeight;
                    noDataCount += 1.0d;
                }
                heights[ 3 ] = grid.getCell( cellX - yDiff, cellY + xDiff, this.handleOutOfMemoryErrorFalse );
                if ( heights[ 3 ] == noDataValue ) {
                    heights[ 3 ] = cellHeight;
                    noDataCount += 1.0d;
                }
                metrics1[ 0 ] += noDataCount;
                
                if ( noDataCount < 4.0d ) {
                    // height[1]   height[0]
                    //      cellHeight
                    // height[2]   height[3]
                    
                    // Calculate basic metrics
                    averageHeight = 0.0d;
                    averageDiff = 0.0d;
                    downCount = 0.0d;
                    upCount = 0.0d;
                    upness = 0.0d;
                    downness = 0.0d;
                    for ( int r = 0; r < 4; r ++ ) {
                        averageHeight += heights[ r ];
                        diff[ r ] = heights[ r ] - cellHeight;
                        averageDiff += diff[ r ];
                        if ( diff[ r ] > 0.0d ) {
                            downness += diff[ r ];
                            downCount += 1.0d;
                        } else {
                            if ( diff[ r ] < 0.0d ) {
                                upness += diff[ r ];
                                upCount += 1.0d;
                            } else {
                                metrics1[ 1 ] += weight; // flatness
                            }
                        }
                        metrics1[ 2 ] += weight * Math.abs( diff [ r ] ); // roughness
                    }
                    averageHeight /= ( 4.0d - noDataCount );
                    averageDiff /= ( 4.0d - noDataCount );
                    metrics1[ 5 ] += weight * downness; // totalDownness
                    if ( downCount > 0.0d ) {
                        metrics1[ 6 ] += metrics1[ 5 ] / downCount; // averageDownness
                    }
                    metrics1[ 7 ] += weight * upness; // totalUpness
                    if ( upCount > 0.0d ) {
                        metrics1[ 8 ] += metrics1[ 7 ] / upCount; // averageUpness
                    }
                    // Slopyness and levelness similar to slope in getSlopeAndAspect
                    // slopyness
                    metrics1[ 3 ] += weight * Math.sqrt( ( ( diff[ 0 ] - diff [ 2 ] ) * ( diff[ 0 ] - diff [ 2 ] ) ) + ( ( diff[ 1 ] - diff [ 3 ] ) * ( diff[ 1 ] - diff [ 3 ] ) ) );
                    //levelness
                    metrics1[ 4 ] += weight * averageDiff;
                    //levelness += weight * Math.abs( averageHeight - cellsize );
                    // diff[1]   diff[0]
                    //    cellHeight
                    // diff[2]   diff[3]
                    metrics1Calculate_Complex( metrics1, diff, dummyDiff, weight, averageDiff );
                }
            }
        }
        return;
    }
    
    
    private void metrics1Calculate_Complex( double[] metrics1, double[] diff, double[] dummyDiff, double weight, double averageDiff ) {
        int caseSwitch = metrics1Calculate_CaseSwitch( diff );
        // 81 cases
        // Each orthoganal equidistant cell is either heigher, lower, or
        // the same height as the cell at centre.
        switch ( caseSwitch ) {
            case 0:
                // hhhh
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                break;
            case 1:
                // hhhl
                metrics1Calculate_hhhl( metrics1, diff, weight );
                break;
            case 2:
                // hhhs
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                metrics1Calculate_hhhl( metrics1, diff, weight );
                //count_hhhs += 1.0d;
                //w_hhhs += weight;
                break;
            case 3:
                // hhlh
                // Shuffle diff once for hhhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                break;
            case 4:
                // hhll
                metrics1Calculate_hhll( metrics1, diff, weight );
                break;
            case 5:
                // hhls
                // Shuffle diff once for hhhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, diff, weight );
                //count_hhsl += 1.0d;
                //w_hhsl += weight;
                break;
            case 6:
                // hhsh
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff once for hhhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                //count_hhhs += 1.0d;
                //w_hhhs += weight;
                break;
            case 7:
                // hhsl
                metrics1Calculate_hhhl( metrics1, diff, weight );
                metrics1Calculate_hhll( metrics1, diff, weight );
                //count_hhsl += 1.0d;
                //w_hhsl += weight;
                break;
            case 8:
                // hhss
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                metrics1Calculate_hhhl( metrics1, diff, weight );
                metrics1Calculate_hhll( metrics1, diff, weight );
                //count_hhss += 1.0d;
                //w_hhss += weight;
                break;
            case 9:
                // hlhh
                // Shuffle diff twice for hhhl
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                break;
            case 10:
                // metrics1Calculate_hlhl
                metrics1Calculate_hlhl( metrics1, diff, weight );
                break;
            case 11:
                // hlhs
                // Shuffle diff twice for hhhl
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, diff, weight );
                //count_hshl += 1.0d;
                //w_hshl += weight;
                break;
            case 12:
                // hllh
                // Shuffle diff once for hhll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 13:
                // hlll
                metrics1Calculate_hlll( metrics1, diff, weight );
                break;
            case 14:
                // hlls
                // Shuffle diff once for hhll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                //count_hsll += 1.0d;
                //w_hsll += weight;
                break;
            case 15:
                // hlsh
                // Shuffle diff twice for hhll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                // Shuffle diff once for hhll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                //count_hhsl += 1.0d;
                //w_hhsl += weight;
                break;
            case 16:
                // hlsl
                metrics1Calculate_hlhl( metrics1, diff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                //count_hlsl += 1.0d;
                //w_hlsl += weight;
                break;
            case 17:
                // hlss
                // Shuffle diff twice for hhhl
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, diff, weight );
                // Shuffle diff once for hhll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                break;
            case 18:
                // hshh
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff twice for hhhl
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                break;
            case 19:
                // hshl
                metrics1Calculate_hhhl( metrics1, diff, weight );
                metrics1Calculate_hlhl( metrics1, diff, weight );
                break;
            case 20:
                // hshs
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                metrics1Calculate_hhhl( metrics1, diff, weight );
                metrics1Calculate_hlhl( metrics1, diff, weight );
                break;
            case 21:
                // hslh
                // Shuffle diff once for hhhl and hhll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 22:
                // hsll
                metrics1Calculate_hhll( metrics1, diff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                break;
            case 23:
                // hsls
                // Shuffle diff once for hhhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, diff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                break;
            case 24:
                // hssh
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff once for hhhl and hhll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 25:
                // hssl
                metrics1Calculate_hhhl( metrics1, diff, weight );
                metrics1Calculate_hhll( metrics1, diff, weight );
                metrics1Calculate_hlhl( metrics1, diff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                break;
            case 26:
                // metrics1Calculate_hsss
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                metrics1Calculate_hhhl( metrics1, diff, weight );
                metrics1Calculate_hhll( metrics1, diff, weight );
                metrics1Calculate_hlhl( metrics1, diff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                break;
            case 27:
                // lhhh
                // Shuffle diff thrice for hhhl
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                break;
            case 28:
                // lhhl
                // Shuffle diff thrice for hhll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 29:
                // lhhs
                // Shuffle diff thrice for hhhl and hhll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 30:
                // lhlh
                // Shuffle once for hlhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                break;
            case 31:
                // lhll
                // Shuffle diff thrice for hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 32:
                // lhls
                // Shuffle diff once for hlhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                // Shuffle diff thrice for hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 33:
                // lhsh
                // Shuffle diff thrice for hhhl
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                // Shuffle diff once for hlhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                break;
            case 34:
                // lhsl
                // Shuffle diff thrice for hhll and hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 35:
                // lhss
                // Shuffle diff thrice for hhhl, hhll, hlhl, hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                break;
            case 36:
                // llhh
                // Shuffle diff twice for hhll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 37:
                // llhl
                // Shuffle diff twice for hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 38:
                // llhs
                // Shuffle diff twice for hhll and hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 39:
                // lllh
                // Shuffle diff once for hlll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 40:
                // llll
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 41:
                // llls
                // Shuffle diff once for hlll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 42:
                // llsh
                // Shuffle diff twice for hhll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                // Shuffle diff once for hlll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 43:
                // llsl
                // Shuffle diff twice for hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 44:
                // llss
                // Shuffle diff twice for hhll hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 45:
                // lshh
                // Shuffle diff thrice for hhhl
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                // Shuffle diff twice for hhll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 46:
                // lshl
                // Shuffle diff thrice for hhll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                // Shuffle diff twice for hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 47:
                // lshs
                // Shuffle diff thrice for hhhl
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                // Shuffle diff twice for hhll hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 48:
                // lslh
                // Shuffle diff once for hlhl and hlll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 49:
                // lsll
                // Shuffle diff thrice for hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 50:
                // lsls
                // Shuffle diff once for hlhl hlll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 51:
                // lssh
                // Shuffle diff thrice for hhhl
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                // Shuffle diff twice for hhll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                // Shuffle diff once for hlll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 52:
                // lssl
                // Shuffle diff thrice for hhll hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 53:
                // lsss
                // Shuffle diff thrice for hhhl hhll hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 54:
                // shhh
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff thrice for hhhl
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                break;
            case 55:
                // shhl
                metrics1Calculate_hhhl( metrics1, diff, weight );
                // Shuffle diff thrice for hhll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 56:
                // shhs
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                metrics1Calculate_hhhl( metrics1, diff, weight );
                // Shuffle diff twice for hhll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 57:
                // shlh
                // Shuffle diff once for hhhl hlhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                break;
            case 58:
                // shll
                metrics1Calculate_hhll( metrics1, diff, weight );
                // Shuffle diff thrice for hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 59:
                // shls
                // Shuffle diff once for hhhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, diff, weight );
                // Shuffle diff thrice for hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 60:
                // shsh
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff once for hhhl hlhl
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                break;
            case 61:
                // shsl
                metrics1Calculate_hhhl( metrics1, diff, weight );
                metrics1Calculate_hhll( metrics1, diff, weight );
                // Shuffle diff thrice for hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 62:
                // shss
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff thrice for hhhl hhll hlhl hlll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 63:
                // slhh
                // Shuffle diff twice for hhhl hhll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 64:
                // slhl
                metrics1Calculate_hlhl( metrics1, diff, weight );
                // Shuffle diff twice for hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 65:
                // slhs
                // Shuffle diff twice for hhhl hhll hlhl hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 66:
                // sllh
                // Shuffle diff twice for hhll hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 67:
                // slll
                metrics1Calculate_hlll( metrics1, diff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 68:
                // slls
                // Shuffle diff once for hhll
                metrics1Shuffle1( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 69:
                // slsh
                // Shuffle diff twice for hhhl hhll hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 70:
                // slsl
                metrics1Calculate_hlhl( metrics1, diff, weight );
                metrics1Calculate_hlll( metrics1, diff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 71:
                // slss
                // Shuffle diff twice for hhhl hhll hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 72:
                // sshh
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff twice for hhhl hhll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                break;
            case 73:
                // sshl
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                // Shuffle diff thrice for hhll
                metrics1Shuffle3( dummyDiff, diff );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                // Shuffle diff twice for hlhl hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 74:
                // sshs
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff twice for hhhl hhll hlhl hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 75:
                // sslh
                // Shuffle diff once for hhhl hhll hlhl hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 76:
                // ssll
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 77:
                // ssls
                // Shuffle diff once for hhhl hhll hlhl hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 78:
                // sssh
                metrics1Calculate_hhhh( metrics1, diff, weight, averageDiff );
                // Shuffle diff once for hhhl hhll hlhl hlll
                metrics1Shuffle2( dummyDiff, diff );
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                break;
            case 79:
                // sssl
                metrics1Calculate_hhhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hhll( metrics1, dummyDiff, weight );
                metrics1Calculate_hlhl( metrics1, dummyDiff, weight );
                metrics1Calculate_hlll( metrics1, dummyDiff, weight );
                metrics1Calculate_llll( metrics1, diff, weight, averageDiff );
                break;
            case 80:
                // ssss
                // This case should not happen!
                break;
        }
        return;
    }
    
    /**
     * Returns caseSwitch for the 81 different cases of higher, lower or same
     * height orthoganol equidistant cells in:
     * metrics1(AbstractGrid2DSquareCellDouble,int,int,double,double[][])
     * @param diff the array of height differences
     */
    private int metrics1Calculate_CaseSwitch( double[] diff ) {
        if ( diff[ 0 ] > 0.0d ) {
            if ( diff[ 1 ] > 0.0d ) {
                if ( diff [ 2 ] > 0.0d ) {
                    if ( diff[ 3 ] > 0.0d ) {
                        return 0; // metrics1Calculate_hhhh
                    } else {
                        if ( diff[ 3 ] < 0.0d ) {
                            return 1; // metrics1Calculate_hhhl
                        } else {
                            return 2; // metrics1Calculate_hhhs
                        }
                    }
                } else {
                    if ( diff [ 2 ] < 0.0d ) {
                        if ( diff[ 3 ] > 0.0d ) {
                            return 3; // metrics1Calculate_hhlh
                        } else {
                            if ( diff[ 3 ] < 0.0d ) {
                                return 4; // metrics1Calculate_hhll
                            } else {
                                return 5; // metrics1Calculate_hhls
                            }
                        }
                    } else {
                        if ( diff[ 3 ] > 0.0d ) {
                            return 6; // metrics1Calculate_hhsh
                        } else {
                            if ( diff[ 3 ] < 0.0d ) {
                                return 7; // metrics1Calculate_hhsl
                            } else {
                                return 8; // metrics1Calculate_hhss
                            }
                        }
                    }
                }
            } else {
                if ( diff[ 1 ] < 0.0d ) {
                    if ( diff[ 2 ] > 0.0d ) {
                        if ( diff[ 3 ] > 0.0d ) {
                            return 9; // metrics1Calculate_hlhh
                        } else {
                            if ( diff[ 3 ] < 0.0d ) {
                                return 10; // metrics1Calculate_hlhl
                            } else {
                                return 11; // metrics1Calculate_hlhs
                            }
                        }
                    } else {
                        if ( diff [ 2 ] < 0.0d ) {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 12; // metrics1Calculate_hllh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 13; // metrics1Calculate_hlll
                                } else {
                                    return 14; // metrics1Calculate_hlls
                                }
                            }
                        } else {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 15; // metrics1Calculate_hlsh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 16; // metrics1Calculate_hlsl
                                } else {
                                    return 17; // metrics1Calculate_hlss
                                }
                            }
                        }
                    }
                } else {
                    if ( diff[ 2 ] > 0.0d ) {
                        if ( diff[ 3 ] > 0.0d ) {
                            return 18; // metrics1Calculate_hshh
                        } else {
                            if ( diff[ 3 ] < 0.0d ) {
                                return 19; // metrics1Calculate_hshl
                            } else {
                                return 20; // metrics1Calculate_hshs
                            }
                        }
                    } else {
                        if ( diff [ 2 ] < 0.0d ) {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 21; // metrics1Calculate_hslh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 22; // metrics1Calculate_hsll
                                } else {
                                    return 23; // metrics1Calculate_hsls
                                }
                            }
                        } else {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 24; // metrics1Calculate_hssh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 25; // metrics1Calculate_hssl
                                } else {
                                    return 26; // metrics1Calculate_hsss
                                }
                            }
                        }
                    }
                }
            }
        } else {
            if ( diff[ 0 ] < 0.0d ) {
                if ( diff[ 1 ] > 0.0d ) {
                    if ( diff [ 2 ] > 0.0d ) {
                        if ( diff[ 3 ] > 0.0d ) {
                            return 27; // metrics1Calculate_lhhh
                        } else {
                            if ( diff[ 3 ] < 0.0d ) {
                                return 28; // metrics1Calculate_lhhl
                            } else {
                                return 29; // metrics1Calculate_lhhs
                            }
                        }
                    } else {
                        if ( diff [ 2 ] < 0.0d ) {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 30; // metrics1Calculate_lhlh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 31; // metrics1Calculate_lhll
                                } else {
                                    return 32; // metrics1Calculate_lhls
                                }
                            }
                        } else {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 33; // metrics1Calculate_lhsh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 34; // metrics1Calculate_lhsl
                                } else {
                                    return 35; // metrics1Calculate_lhss
                                }
                            }
                        }
                    }
                } else {
                    if ( diff[ 1 ] < 0.0d ) {
                        if ( diff[ 2 ] > 0.0d ) {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 36; // metrics1Calculate_llhh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 37; // metrics1Calculate_llhl
                                } else {
                                    return 38; // metrics1Calculate_llhs
                                }
                            }
                        } else {
                            if ( diff [ 2 ] < 0.0d ) {
                                if ( diff[ 3 ] > 0.0d ) {
                                    return 39; // metrics1Calculate_lllh
                                } else {
                                    if ( diff[ 3 ] < 0.0d ) {
                                        return 40; // metrics1Calculate_llll
                                    } else {
                                        return 41; // metrics1Calculate_llls
                                    }
                                }
                            } else {
                                if ( diff[ 3 ] > 0.0d ) {
                                    return 42; // metrics1Calculate_llsh
                                } else {
                                    if ( diff[ 3 ] < 0.0d ) {
                                        return 43; // metrics1Calculate_llsl
                                    } else {
                                        return 44; // metrics1Calculate_llss
                                    }
                                }
                            }
                        }
                    } else {
                        if ( diff[ 2 ] > 0.0d ) {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 45; // metrics1Calculate_lshh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 46; // metrics1Calculate_lshl
                                } else {
                                    return 47; // metrics1Calculate_lshs
                                }
                            }
                        } else {
                            if ( diff [ 2 ] < 0.0d ) {
                                if ( diff[ 3 ] > 0.0d ) {
                                    return 48; // metrics1Calculate_lslh
                                } else {
                                    if ( diff[ 3 ] < 0.0d ) {
                                        return 49; // metrics1Calculate_lsll
                                    } else {
                                        return 50; // metrics1Calculate_lsls
                                    }
                                }
                            } else {
                                if ( diff[ 3 ] > 0.0d ) {
                                    return 51; // metrics1Calculate_lssh
                                } else {
                                    if ( diff[ 3 ] < 0.0d ) {
                                        return 52; // metrics1Calculate_lssl
                                    } else {
                                        return 53; // metrics1Calculate_lsss
                                    }
                                }
                            }
                        }
                    }
                }
            } else {
                if ( diff[ 1 ] > 0.0d ) {
                    if ( diff [ 2 ] > 0.0d ) {
                        if ( diff[ 3 ] > 0.0d ) {
                            return 54; // metrics1Calculate_shhh
                        } else {
                            if ( diff[ 3 ] < 0.0d ) {
                                return 55; // metrics1Calculate_shhl
                            } else {
                                return 56; // metrics1Calculate_shhs
                            }
                        }
                    } else {
                        if ( diff [ 2 ] < 0.0d ) {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 57; // metrics1Calculate_shlh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 58; // metrics1Calculate_shll
                                } else {
                                    return 59; // metrics1Calculate_shls
                                }
                            }
                        } else {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 60; // metrics1Calculate_shsh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 61; // metrics1Calculate_shsl
                                } else {
                                    return 62; // metrics1Calculate_shss
                                }
                            }
                        }
                    }
                } else {
                    if ( diff[ 1 ] < 0.0d ) {
                        if ( diff[ 2 ] > 0.0d ) {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 63; // metrics1Calculate_slhh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 64; // metrics1Calculate_slhl
                                } else {
                                    return 65; // metrics1Calculate_slhs
                                }
                            }
                        } else {
                            if ( diff [ 2 ] < 0.0d ) {
                                if ( diff[ 3 ] > 0.0d ) {
                                    return 66; // metrics1Calculate_sllh
                                } else {
                                    if ( diff[ 3 ] < 0.0d ) {
                                        return 67; // metrics1Calculate_slll
                                    } else {
                                        return 68; // metrics1Calculate_slls
                                    }
                                }
                            } else {
                                if ( diff[ 3 ] > 0.0d ) {
                                    return 69; // metrics1Calculate_slsh
                                } else {
                                    if ( diff[ 3 ] < 0.0d ) {
                                        return 70; // metrics1Calculate_slsl
                                    } else {
                                        return 71; // metrics1Calculate_slss
                                    }
                                }
                            }
                        }
                    } else {
                        if ( diff[ 2 ] > 0.0d ) {
                            if ( diff[ 3 ] > 0.0d ) {
                                return 72; // metrics1Calculate_sshh
                            } else {
                                if ( diff[ 3 ] < 0.0d ) {
                                    return 73; // metrics1Calculate_sshl
                                } else {
                                    return 74; // metrics1Calculate_sshs
                                }
                            }
                        } else {
                            if ( diff [ 2 ] < 0.0d ) {
                                if ( diff[ 3 ] > 0.0d ) {
                                    return 75; // metrics1Calculate_sslh
                                } else {
                                    if ( diff[ 3 ] < 0.0d ) {
                                        return 76; // metrics1Calculate_ssll
                                    } else {
                                        return 77; // metrics1Calculate_ssls
                                    }
                                }
                            } else {
                                if ( diff[ 3 ] > 0.0d ) {
                                    return 78; // metrics1Calculate_sssh
                                } else {
                                    if ( diff[ 3 ] < 0.0d ) {
                                        return 79; // metrics1Calculate_sssl
                                    } else {
                                        return 80; // metrics1Calculate_ssss
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    /**
     * Shuffles dummyDiff such that:
     * dummyDiff[0] = diff[3]
     * dummyDiff[1] = diff[0]
     * dummyDiff[2] = diff[1]
     * dummyDiff[3] = diff[2]
     */
    private void metrics1Shuffle1( double[] dummyDiff, double[] diff ) {
        dummyDiff[0] = diff[3];
        dummyDiff[1] = diff[0];
        dummyDiff[2] = diff[1];
        dummyDiff[3] = diff[2];
    }
    
    /**
     * Shuffles dummyDiff such that:
     * dummyDiff[0] = diff[2]
     * dummyDiff[1] = diff[3]
     * dummyDiff[2] = diff[0]
     * dummyDiff[3] = diff[1]
     */
    private void metrics1Shuffle2( double[] dummyDiff, double[] diff ) {
        dummyDiff[0] = diff[2];
        dummyDiff[1] = diff[3];
        dummyDiff[2] = diff[0];
        dummyDiff[3] = diff[1];
    }
    
    /**
     * Shuffles dummyDiff such that:
     * dummyDiff[0] = diff[2]
     * dummyDiff[1] = diff[3]
     * dummyDiff[2] = diff[0]
     * dummyDiff[3] = diff[1]
     */
    private void metrics1Shuffle3( double[] dummyDiff, double[] diff ) {
        dummyDiff[0] = diff[1];
        dummyDiff[1] = diff[2];
        dummyDiff[2] = diff[3];
        dummyDiff[3] = diff[0];
    }
    
    /**
     * For processing 6 metrics with all cells higher or same in:
     * metrics1(AbstractGrid2DSquareCellDouble,int,int,double,double[][])
     * metrics1[9] - maxd_hhhh [sum of distance weighted maximum height differences]
     * metrics1[10] - mind_hhhh [sum of distance weighted minimum height differences]
     * metrics1[11] - sumd_hhhh [sum of distance weighted height differences]
     * metrics1[12] - aved_hhhh [sum of distance weighted average height difference]
     * metrics1[13] - count_hhhh [count]
     * metrics1[14] - w_hhhh [sum of distance weights]
     * @param metrics1 the array of metrics to be processed
     * @param diff the array of differences of cell values
     * @param weight the weight to be applied to weighted metrics
     * @param averageDiff the average difference in height for diff (N.B This is
     *   passed in rather than calculated here because of cell values that were
     *   noDataValue in the Grid2DSquareCellDouble for which metrics1
     *   are being processed.
     */
    private void metrics1Calculate_hhhh( double[] metrics1, double[] diff, double weight, double averageDiff ) {
        metrics1[ 9 ] += weight * Math.max( Math.max( diff[ 0 ], diff[ 1 ] ), Math.max( diff[ 2 ], diff[ 3 ] ) );
        metrics1[ 10 ] += weight * Math.min( Math.min( diff[ 0 ], diff[ 1 ] ), Math.min( diff[ 2 ], diff[ 3 ] ) );
        metrics1[ 11 ] += weight * ( diff[ 0 ] + diff[ 1 ] + diff[ 2 ] + diff[ 3 ] );
        metrics1[ 12 ] += weight * averageDiff;
        metrics1[ 13 ] += 1.0d;
        metrics1[ 14 ] += weight;
    }
    
    /**
     * For processing 11 metrics with one cell lower or same in:
     * metrics1(AbstractGrid2DSquareCellDouble,int,int,double,double[][])
     * metrics1[15] - mind_hxhx_ai_hhhl [sum of distance weighted ( minimum difference of cells adjacent to lower cell )]
     * metrics1[16] - maxd_hxhx_ai_hhhl [sum of distance weighted ( maximum difference of cells adjacent to lower cell )]
     * metrics1[17] - sumd_hxhx_ai_hhhl [sum of distance weighted ( sum of differences of cells adjacent to lower cell )]
     * metrics1[18] - d_xhxx_ai_hhhl [sum of distance weighted ( difference of cell opposite lower cell )]
     * metrics1[19] - d_xxxl_ai_hhhl [sum of distance weighted ( difference of lower cell )]
     * metrics1[20] - sumd_xhxl_ai_hhhl [sum of distance weighted ( sum of differences of lower cell and cell opposite )]
     * metrics1[21] - mind_abs_xhxl_ai_hhhl [sum of distance weighted ( minimum difference magnitude of lower cell and cell opposite )]
     * metrics1[22] - maxd_abs_xhxl_ai_hhhl [sum of distance weighted ( maximum difference magnitude of lower cell and cell opposite )]
     * metrics1[23] - sumd_abs_xhxl_ai_hhhl [sum of distance weighted ( sum of difference magnitudes of lower cell and cell opposite )]
     * metrics1[24] - count_hhhl [count]
     * metrics1[25] - w_hhhl [sum of distance weights]
     * @param metrics1 the array of metrics to be processed
     * @param diff the array of differences of cell values
     * @param weight the weight to be applied to weighted metrics
     **/
    private void metrics1Calculate_hhhl( double[] metrics1, double[] diff, double weight ) {
        metrics1[ 15 ] += weight * Math.min( diff[ 0 ], diff[ 2 ] );
        metrics1[ 16 ] += weight * Math.max( diff[ 0 ], diff[ 2 ] );
        metrics1[ 17 ] += weight * ( diff[ 0 ] + diff[ 2 ] );
        metrics1[ 18 ] += weight * diff[ 1 ];
        metrics1[ 19 ] += weight * diff[ 3 ];
        metrics1[ 20 ] += weight * ( diff[ 1 ] + diff[ 3 ] );
        metrics1[ 21 ] += weight * Math.min( diff[ 1 ], Math.abs( diff[ 3 ] ) );
        metrics1[ 22 ] += weight * Math.max( diff[ 1 ], Math.abs( diff[ 3 ] ) );
        metrics1[ 23 ] += weight * ( diff[ 1 ] + Math.abs( diff[ 3 ] ) );
        metrics1[ 24 ] += 1.0d;
        metrics1[ 25 ] += weight;
    }
    
    /**
     * For processing 11 metrics with opposite cells lower/higher or same in:
     * metrics1(AbstractGrid2DSquareCellDouble,int,int,double,double[][])
     * metrics1[26] mind_hxhx_ai_hlhl [sum of distance weighted ( minimum difference of higher cells )]
     * metrics1[27] maxd_hxhx_ai_hlhl [sum of distance weighted ( maximum difference of higher cells )]
     * metrics1[28] sumd_hxhx_ai_hlhl [sum of distance weighted ( sum differences of higher cells )]
     * metrics1[29] mind_xlxl_ai_hlhl [sum of distance weighted ( minimum difference of lower cells )]
     * metrics1[30] maxd_xlxl_ai_hlhl [sum of distance weighted ( maximum difference of lower cells )]
     * metrics1[31] sumd_xlxl_ai_hlhl [sum of distance weighted ( sum of differences of lower cells )]
     * metrics1[32] mind_abs_hlhl [sum of distance weighted ( minimum difference magnitude of cells )]
     * metrics1[33] maxd_abs_hlhl [sum of distance weighted ( maximum difference magnitude of cells )]
     * metrics1[34] sumd_abs_hlhl [sum of distance weighted ( sum of difference magnitudes of cells )]
     * metrics1[35] count_hlhl [count]
     * metrics1[36] w_hlhl [sum of distance weights]
     * @param metrics1 the array of metrics to be processed
     * @param diff the array of differences of cell values
     * @param weight the weight to be applied to weighted metrics
     **/
    private void metrics1Calculate_hlhl( double[] metrics1, double[] diff, double weight ) {
        metrics1[ 26 ] += weight * Math.min( diff[ 0 ], diff[ 2 ] );
        metrics1[ 27 ] += weight * Math.max( diff[ 0 ], diff[ 2 ] );
        metrics1[ 28 ] += weight * ( diff[ 0 ] + diff[ 2 ] );
        metrics1[ 29 ] += weight * Math.min( diff[ 1 ], diff[ 3 ] );
        metrics1[ 30 ] += weight * Math.max( diff[ 1 ], diff[ 3 ] );
        metrics1[ 31 ] += weight * ( diff[ 1 ] + diff[ 3 ] );
        metrics1[ 32 ] += weight * ( Math.min( Math.abs( Math.max( diff[ 1 ], diff[ 3 ] ) ), Math.min( diff[ 0 ], diff[ 2 ] ) ) );
        metrics1[ 33 ] += weight * ( Math.max( Math.abs( Math.min( diff[ 1 ], diff[ 3 ] ) ), Math.max( diff[ 0 ], diff[ 2 ] ) ) );
        metrics1[ 34 ] += weight * ( diff[ 0 ] + Math.abs( diff[ 1 ] ) + diff[ 2 ] + Math.abs( diff[ 3 ] ) );
        metrics1[ 35 ] += 1.0d;
        metrics1[ 36 ] += weight;
    }
    
    /**
     * For processing 11 metrics with adjacent cells lower/higher or same in:
     * metrics1(AbstractGrid2DSquareCellDouble,int,int,double,double[][])
     * metrics1[37] - mind_hhxx_ai_hhll [sum of distance weighted ( minimum difference of higher cells )]
     * metrics1[38] - maxd_hhxx_ai_hhll [sum of distance weighted ( maximum difference of higher cells )]
     * metrics1[39] - sumd_hhxx_ai_hhll [sum of distance weighted ( sum of differences of higher cells )]
     * metrics1[40] - mind_xxll_ai_hhll [sum of distance weighted ( minimum difference of lower cells )]
     * metrics1[41] - maxd_xxll_ai_hhll [sum of distance weighted ( maximum difference of lower cells )]
     * metrics1[42] - sumd_xxll_ai_hhll [sum of distance weighted ( sum of differences of lower cells )]
     * metrics1[43] - mind_abs_hhll [sum of distance weighted ( minimum difference magnitude of cells )]
     * metrics1[44] - maxd_abs_hhll [sum of distance weighted ( maximum difference magnitude of cells )]
     * metrics1[45] - sumd_abs_hhll [sum of distance weighted ( sum of difference magnitudes of cells )]
     * metrics1[46] - count_hhll [count]
     * metrics1[47] - w_hhll [sum of distance weights]
     * @param metrics1 the array of metrics to be processed
     * @param diff the array of differences of cell values
     * @param weight the weight to be applied to weighted metrics
     **/
    private void metrics1Calculate_hhll( double[] metrics1, double[] diff, double weight ) {
        metrics1[ 37 ] += weight * Math.min( diff[ 0 ], diff[ 1 ] );
        metrics1[ 38 ] += weight * Math.max( diff[ 0 ], diff[ 1 ] );
        metrics1[ 39 ] += weight * ( diff[ 0 ] + diff[ 1 ] );
        metrics1[ 40 ] += weight * Math.min( diff[ 2 ], diff[ 3 ] );
        metrics1[ 41 ] += weight * Math.max( diff[ 2 ], diff[ 3 ] );
        metrics1[ 42 ] += weight * ( diff[ 2 ] + diff[ 3 ] );
        metrics1[ 43 ] += weight * ( Math.min( Math.abs( Math.max( diff[ 2 ], diff[ 3 ] ) ), Math.min( diff[ 1 ], diff[ 0 ] ) ) );
        metrics1[ 44 ] += weight * ( Math.max( Math.abs( Math.min( diff[ 2 ], diff[ 3 ] ) ), Math.max( diff[ 1 ], diff[ 0 ] ) ) );
        metrics1[ 45 ] += weight * ( diff[ 1 ] + Math.abs( diff[ 2 ] ) + diff[ 0 ] + Math.abs( diff[ 3 ] ) );
        metrics1[ 46 ] += 1.0d;
        metrics1[ 47 ] += weight;
    }
    
    /**
     * For processing 11 metrics with one cell higher or same in:
     * metrics1(AbstractGrid2DSquareCellDouble,int,int,double,double[][])
     * metrics1[ 48 ] - mind_lxlx_ai_lllh [sum of distance weighted ( minimum difference of cells adjacent to higher cell )]
     * metrics1[ 49 ] - maxd_lxlx_ai_lllh [sum of distance weighted ( maximum difference of cells adjacent to higher cell )]
     * metrics1[ 50 ] - sumd_lxlx_ai_lllh [sum of distance weighted ( sum of differences of cells adjacent to higher cell )]
     * metrics1[ 51 ] - d_xlxx_ai_lllh [sum of distance weighted ( difference of cell opposite higher cell )]
     * metrics1[ 52 ] - d_xxxh_ai_lllh [sum of distance weighted ( difference of higher cell )]
     * metrics1[ 53 ] - sumd_xlxh_ai_lllh [sum of distance weighted ( sum of differences of higher cell and cell opposite )]
     * metrics1[ 54 ] - mind_abs_xlxh_ai_lllh [sum of distance weighted ( minimum difference magnitude of higher cell and cell opposite )]
     * metrics1[ 55 ] - maxd_abs_xlxh_ai_lllh [sum of distance weighted ( maximum difference magnitude of higher cell and cell opposite )]
     * metrics1[ 56 ] - sumd_abs_xlxh_ai_lllh [sum of distance weighted ( sum of difference magnitudes of higher cell and cell opposite )]
     * metrics1[ 57 ] - count_lllh [count]
     * metrics1[ 58 ] - w_lllh [sum of distance weights]
     * @param metrics1 the array of metrics to be processed
     * @param diff the array of differences of cell values
     * @param weight the weight to be applied to weighted metrics
     * @param averageDiff the average difference in height for diff (N.B This is
     *   passed in rather than calculated here because of cell values that were
     *   noDataValue in the Grid2DSquareCellDouble for which metrics1
     *   are being processed.
     **/
    private void metrics1Calculate_hlll( double[] metrics1, double[] diff, double weight ) {
        metrics1[ 48 ] += weight * Math.min( diff[ 1 ], diff[ 3 ] );
        metrics1[ 49 ] += weight * Math.max( diff[ 1 ], diff[ 3 ] );
        metrics1[ 50 ] += weight * ( diff[ 1 ] + diff[ 3 ] );
        metrics1[ 51 ] += weight * diff[ 2 ];
        metrics1[ 52 ] += weight * diff[ 0 ];
        metrics1[ 53 ] += weight * ( diff[ 2 ] + diff[ 0 ] );
        metrics1[ 54 ] = weight * Math.min( diff[ 0 ], Math.abs( diff[ 2 ] ) );
        metrics1[ 55 ] = weight * Math.max( diff[ 0 ], Math.abs( diff[ 2 ] ) );
        metrics1[ 56 ] = weight * ( diff[ 0 ] + Math.abs( diff[ 2 ] ) );
        metrics1[ 57 ] += 1.0d;
        metrics1[ 58 ] += weight;
    }
    
    /**
     * For processing 6 metrics with all cells lower or same in:
     * metrics1(AbstractGrid2DSquareCellDouble,int,int,double,double[][])
     * metrics1[59] - maxd_llll [sum of distance weighted maximum height differences]
     * metrics1[60] - mind_llll [sum of distance weighted minimum height differences]
     * metrics1[61] - sumd_llll [sum of distance weighted height differences]
     * metrics1[62] - aved_llll [sum of distance weighted average height difference]
     * metrics1[63] - count_llll [count]
     * metrics1[64] - w_llll [sum of distance weights]
     */
    private void metrics1Calculate_llll( double[] metrics1, double[] diff, double weight, double averageDiff ) {
        metrics1[ 59 ] += weight * Math.max( Math.max( diff[ 0 ], diff[ 1 ] ), Math.max( diff[ 2 ], diff[ 3 ] ) );
        metrics1[ 60 ] += weight * Math.min( Math.min( diff[ 0 ], diff[ 1 ] ), Math.min( diff[ 2 ], diff[ 3 ] ) );
        metrics1[ 61 ] += weight * ( diff[ 0 ] + diff[ 1 ] + diff[ 2 ] + diff[ 3 ] );
        metrics1[ 62 ] += weight * averageDiff;
        metrics1[ 63 ] += 1.0d;
        metrics1[ 64 ] += weight;
    }
    
    /**
     * Returns an Grid2DSquareCellDouble[] metrics2 where:
     * TODO: metrics2 is a mess. Need to decide what to do with regard to
     * contour tracing and profile trace for axes and comparisons.
     * metrics2[0] = slope;
     * metrics2[1] = aspect;
     * metrics2[2] = no data count;
     * metrics2[3] = contourConcavity;
     * metrics2[4] = contourConvexity;
     * metrics2[5] = profileConcavity;
     * metrics2[6] = profileConvexity;
     */
    public Grid2DSquareCellDouble[] getMetrics2( Grid2DSquareCellDouble grid, double distance, double weightIntersect, double weightFactor, int samplingDensity, Grid2DSquareCellDoubleFactory gridFactory ) {
        // Initialisation
        Grid2DSquareCellDouble[] result = new Grid2DSquareCellDouble[ 7 ];
        long ncols = grid.getNcols();
        long nrows = grid.getNrows();
        BigDecimal[] dimensions = grid.getDimensions();
        double gridNoDataValue = grid.getNoDataValue();
        double noDataValue = Double.MIN_VALUE;
        //double noDataValue = gridNoDataValue;
        //double noDataValue = 0.0d;
        //double noDataValue = Double.NaN;
        Grid2DSquareCellDouble[] slopeAndAspect = getSlopeAndAspect( grid, distance, weightIntersect, weightFactor, grid, gridFactory );
        result[ 0 ] = slopeAndAspect[ 0 ];
        result[ 1 ] = slopeAndAspect[ 1 ];
        for ( int i = 0; i < result.length; i ++ ) {
            result[ i ] = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, dimensions, noDataValue );
        }
        double[] metrics2 = null;
        double slope;
        double aspect;
        Point2D.Double[] metrics2Points;
        double[] weights = null;
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                if ( grid.getCell( row, col, this.handleOutOfMemoryErrorFalse ) != gridNoDataValue ) {
                    slope = result[ 0 ].getCell( row, col, this.handleOutOfMemoryErrorFalse );
                    aspect = result[ 1 ].getCell( row, col, this.handleOutOfMemoryErrorFalse );
                    metrics2Points = getMetrics2Points( slopeAndAspect, distance, samplingDensity );
                    weights = Kernel.getKernelWeights( grid, row, col, distance, weightIntersect, weightFactor, metrics2Points );
                    metrics2 = getMetrics2( grid, row, col, slopeAndAspect, distance, weights );
                    for ( int i = 0; i < result.length; i ++ ) {
                        result[ i ].setCell( row, col, metrics2[ i ], this.handleOutOfMemoryErrorFalse );
                    }
                }
            }
            System.out.println("Done row " + row );
        }
        result[ 2 ].setName( "", this.handleOutOfMemoryErrorFalse );
        result[ 3 ].setName( "", this.handleOutOfMemoryErrorFalse );
        result[ 4 ].setName( "", this.handleOutOfMemoryErrorFalse );
        result[ 5 ].setName( "", this.handleOutOfMemoryErrorFalse );
        result[ 6 ].setName( "", this.handleOutOfMemoryErrorFalse );
        return result;
    }
    
    /**
     * Returns a Point2D.Double[] points that are sample points based on a
     * regular sampling around slope
     * If samplingDensity
     *
     *
     */
    private Point2D.Double[] getMetrics2Points( Grid2DSquareCellDouble[] slopeAndAspect, double distance, int samplingDensity ) {
        Point2D.Double[] metrics2Points = null;
        return metrics2Points;
    }
    
    private double[] getMetrics2( Grid2DSquareCellDouble grid, long row, long col, Grid2DSquareCellDouble[] slopeAndAspect, double distance, double[] weights ) {
        double[] metrics2 = null;
        return metrics2;
    }
    
    /**
     * Returns an Grid2DSquareCellDouble result containing values which
     * indicate the direction of the maximum down slope for the immediate 8 cell
     * neighbourhood.
     * 1 2 3
     * 4 0 5
     * 6 7 8
     * If there is no downhill slope then the flow direction is 0.
     * @param grid the Grid2DSquareCellDouble to be processed
     * @param gridFactory the Grid2DSquareCellDoubleFactory used to create result
     */
    public Grid2DSquareCellDouble getMaxFlowDirection( Grid2DSquareCellDouble grid, Grid2DSquareCellDoubleFactory gridFactory ) {
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        double noDataValue = grid.getNoDataValue();
        Grid2DSquareCellDouble result = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, grid.getDimensions(), noDataValue );
        CellID cellID;
        long row;
        long col;
        int k;
        int[] flowDirections = new int[ 9 ];
        int flowDirection = 0;
        double[ ] z = new double[ 9 ];
        double minz;
        int minzCount;
        int minzCountNoDataValue;
        long p;
        long q;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                z[ 0 ] = grid.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                if ( z[ 0 ] != noDataValue ) {
                    minz = Double.MAX_VALUE;
                    minzCount = 0;
                    minzCountNoDataValue = 0;
                    flowDirection = 0;
                    k = 0;
                    for ( p = -1; p < 2; p ++ ) {
                        for ( q = -1; q < 2; q ++ ) {
                            if ( ! ( p == 0 && q == 0 ) ) {
                                k ++;
                                z[ k ] = grid.getCell( row + p, col + q, this.handleOutOfMemoryErrorFalse );
                                if ( z[ k ] != noDataValue ) {
                                    if ( z[ k ] <= minz && z[ k ] < z[ 0 ] ) {
                                        if ( z[ k ] == minz ) {
                                            minzCount ++;
                                        } else {
                                            minz = z[ k ];
                                            minzCount = 1;
                                            flowDirection = k;
                                        }
                                    }
                                } else {
                                    minzCountNoDataValue ++;
                                }
                            }
                        }
                    }
                    // If more than one flowDirection randomly assign one
                    if ( minzCount + minzCountNoDataValue > 1 ) {
                        int[] min = new int[ minzCount + minzCountNoDataValue ];
                        int minID = 0;
                        double random = Math.random();
                        for ( int k2 = 1; k2 < z.length; k2 ++ ) {
                            if ( z[ k2 ] == minz || z[ k2 ] == noDataValue ) {
                                min[ minID ] = k2;
                                minID ++;
                            }
                        }
                        flowDirection = min[ ( int ) Math.floor( random * ( minzCount +  minzCountNoDataValue ) ) ];
                    }
                    result.setCell( row, col, ( double ) flowDirection, this.handleOutOfMemoryErrorFalse );
                }
            }
        }
        return result;
    }
    
    /**
     * Returns an Grid2DSquareCellDouble[] each element of which
     * corresponds to a metrics of up slope cells of grid - a DEM
     * The steeper the slope the higher the runoff?
     *
     */
    public Grid2DSquareCellDouble getUpSlopeAreaMetrics( Grid2DSquareCellDouble grid, double distance, double weightFactor, double weightIntersect, Grid2DSquareCellDoubleFactory gridFactory ) {
        Grid2DSquareCellDouble upSlopeAreaMetrics = gridFactory.createGrid2DSquareCellDouble( grid.getNrows(), grid.getNcols(), grid.getDimensions(), grid.getNoDataValue() );
        // Get Peaks and set their value to 1.0d
        HashSet initialPeaksHashSet = getInitialPeaksHashSetAndSetTheirValue( grid, upSlopeAreaMetrics );
        // For each Peak find its neighbours and add a proportional value to
        // them based on slope. If the slope is zero then the neighbour is still
        // passed a proportion. This can be configured based on infiltration
        // rates or slope dependent distance decay stuff.
        //        HashSet neighboursOfInitialPeaksHashSet = getNeighboursOfInitialPeaksHashSetAndSetTheirValue( initialPeaksHashSet, grid, upSlopeAreaMetrics );
        // Add to neighbouring cells a value based on the amount of slope
        //        upSlopeMetricsAddToNeighbours( grid, peaks );
        return upSlopeAreaMetrics;
    }
    
    
    /**
     * Returns a HashSet containing cellIDs which identifies cells for which
     * neighbouring cells in the immediate 8 cell neighbourhood that are either
     * the same value, lower or noDataValues
     * @param grid - the Grid2DSquareCellDouble to be processed
     */
    public HashSet getInitialPeaksHashSetAndSetTheirValue( Grid2DSquareCellDouble grid, Grid2DSquareCellDouble upSlopeAreaMetrics ) {
        HashSet initialPeaksHashSet = new HashSet();
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        double noDataValue = grid.getNoDataValue();
        double[ ] heights = new double[ 9 ];
        int k;
        for ( int row = 0; row < nrows; row ++ ) {
            for ( int col = 0; col < ncols; col ++ ) {
                heights[ 0 ] = grid.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                if ( heights[ 0 ] != noDataValue ) {
                    k = 0;
                    for ( int p = -1; p < 2; p ++ ) {
                        for ( int q = -1; q < 2; q ++ ) {
                            if ( ! ( p == 0 && q == 0 ) ) {
                                k ++;
                                heights[ k ] = grid.getCell( row + p, col + q, this.handleOutOfMemoryErrorFalse );
                            }
                        }
                    }
                    // This deals with single isolated cells surrounded by noDataValues
                    if ( ( heights[ 1 ] <= heights[ 0 ] || heights[ 1 ] == noDataValue ) &&
                            ( heights[ 2 ] <= heights[ 0 ] || heights[ 2 ] == noDataValue ) &&
                            ( heights[ 3 ] <= heights[ 0 ] || heights[ 3 ] == noDataValue ) &&
                            ( heights[ 4 ] <= heights[ 0 ] || heights[ 4 ] == noDataValue ) &&
                            ( heights[ 5 ] <= heights[ 0 ] || heights[ 5 ] == noDataValue ) &&
                            ( heights[ 6 ] <= heights[ 0 ] || heights[ 6 ] == noDataValue ) &&
                            ( heights[ 7 ] <= heights[ 0 ] || heights[ 7 ] == noDataValue ) &&
                            ( heights[ 8 ] <= heights[ 0 ] || heights[ 8 ] == noDataValue ) ) {
                        initialPeaksHashSet.add( grid.getCellID( row, col, this.handleOutOfMemoryErrorFalse ) );
                        upSlopeAreaMetrics.addToCell( row, col, 1.0d, this.handleOutOfMemoryErrorFalse );
                    }
                }
            }
        }
        return initialPeaksHashSet;
    }
    
    /**
     * @param grid the Grid2DSquareCellDouble to be processed
     */
    /*protected HashSet getNeighboursOfInitialPeaksHashSetAndSetTheirValue( HashSet initialPeaksHashSet, Grid2DSquareCellDouble grid, Grid2DSquareCellDouble upSlopeAreaMetrics ) {
        double noDataValue = grid.getNoDataValue();
        double[ ] heights = new double[ 9 ];
        double[ ] diff = new double[ 9 ];
        HashSet neighboursOfInitialPeaksHashSet = Utilities.
        Iterator ite = hashSet.iterator();
        Integer cellID;
        int cellID;
        int row;
        int col;
        int k;
        int lowerCount = 0;
        double lowerHeight = 0.0d;
        while ( ite.hasNext() ) {
            cellID = ( Integer ) ite.next();
            cellID = cellID.intValue();
            row = grid.getRowIndex( cellID );
            col = grid.getColIndex( cellID );
            heights[ 0 ] = grid.getCell( row, col );
            if ( heights[ 0 ] != noDataValue ) {
                k = 0;
                for ( int p = -1; p < 2; p ++ ) {
                    for ( int q = -1; q < 2; q ++ ) {
                        if ( ! ( p == 0 && q == 0 ) ) {
                            k ++;
                            heights[ k ] = grid.getCell( row + p, col + q );
                            if ( heights[ k ] != noDataValue ) {
                                diff[ k ] = heights[ k ] - heights[ 0 ];
                                if ( diff[ k ] >= 0.0d ) {
                                    lowerCount ++;
                                    lowerHeight += diff[ k ];
                                }
                            }
                        }
                    }
                    // This deals with single isolated cells surrounded by noDataValues
                    if ( ( heights[ 1 ] <= heights[ 0 ] || heights[ 1 ] == noDataValue ) &&
                    ( heights[ 2 ] <= heights[ 0 ] || heights[ 2 ] == noDataValue ) &&
                    ( heights[ 3 ] <= heights[ 0 ] || heights[ 3 ] == noDataValue ) &&
                    ( heights[ 4 ] <= heights[ 0 ] || heights[ 4 ] == noDataValue ) &&
                    ( heights[ 5 ] <= heights[ 0 ] || heights[ 5 ] == noDataValue ) &&
                    ( heights[ 6 ] <= heights[ 0 ] || heights[ 6 ] == noDataValue ) &&
                    ( heights[ 7 ] <= heights[ 0 ] || heights[ 7 ] == noDataValue ) &&
                    ( heights[ 8 ] <= heights[ 0 ] || heights[ 8 ] == noDataValue ) ) {
                    }
                }
            }
        }
        return;
    }*/
    
    /**
     * There are many estimates of flow that can be generated and many models
     * developed in hydrology. These methods are simplistic but are based on
     * the work of others. The basics are that discharge from any cell is a
     * simple mutliple of velocity and depth. A measure of velocity can be
     * obtained by measuring slope and the depth of discharge itself where the
     * slope is given by the change in height divided by the distance.
     * The algorithm is this:
     * An Grid2DSquareCellDouble height is initialised using grid
     * A coincident Grid2DSquareCellDouble accumulation is initialised
     * Step 1: A value of rainfall is added to all cells in accumulation.
     * Step 2: A proportion of this rainfall is then distributed to neighbouring
     *         cells based on Mannings discharge equations.
     *
     * proportionally based on the difference in height of
     *         neighbouring cells which are down slope. If no immediate
     *         neighbours are downslope then the height cell is raised by value.
     * Step 3: Repeat Steps 2 and 3 iterations number of times.
     * Step 4: Return height and accumulation.
     * NB Care needs to be taken to specify outflow cells
     * TODO:
     * 1. Change precipitation to be a grid
     * 2. Variable frictionFactor
     */
    public Grid2DSquareCellDouble getFlowAccumulation( Grid2DSquareCellDouble grid, int iterations, double precipitation, HashSet outflowCellIDs, Grid2DSquareCellDoubleFactory gridFactory ) {
        Grid2DSquareCellDouble flowAccumulation = getInitialFlowAccumulation( grid, precipitation, outflowCellIDs, gridFactory );
        System.out.println( "intitialFlowAccumulation" );
        System.out.println( flowAccumulation.toString() );
        for ( int iteration = 0; iteration < iterations; iteration ++ ) {
            flowAccumulation = doFlowAccumulation( flowAccumulation, grid, precipitation, outflowCellIDs, gridFactory );
            System.out.println( "flowAccumulation iteration " + ( iteration + 1 ) );
            System.out.println( flowAccumulation.toString() );
        }
        return flowAccumulation;
    }
    
    /**
     * TODO: docs
     * frictionFactor = 75.0d;
     * constant = 8.0d * 9.81d / frictionFactor;
     * velocity = Math.sqrt( constant * waterDepth * changeInDepth / ChangeInLength );
     * discharge = velocity * waterDepth
     */
    public Grid2DSquareCellDouble getInitialFlowAccumulation( Grid2DSquareCellDouble grid, double precipitation,  HashSet outflowCellIDs, Grid2DSquareCellDoubleFactory gridFactory ) {
        //double constant = 8.0d * 9.81d / 75.0d ;
        double constant = 1.0d;
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        BigDecimal[] dimensions = grid.getDimensions();
        double noDataValue = grid.getNoDataValue();
        // Precipitate
        Grid2DSquareCellDouble flowAccumulation = gridFactory.createGrid2DSquareCellDouble( nrows, ncols, dimensions, noDataValue );
        flowAccumulation = addToGrid( flowAccumulation, precipitation, this.handleOutOfMemoryErrorFalse );
        flowAccumulation = mask( flowAccumulation, grid, this.handleOutOfMemoryErrorFalse );
        Grid2DSquareCellDouble tempFlowAccumulation = gridFactory.createGrid2DSquareCellDouble( flowAccumulation );
        double[][] surfaceHeights = new double[3][3];
        double[][] discharge = new double[3][3];
        double slope;
        double velocity;
        double waterDepth;
        double movingWaterDepth;
        double numberOfDownSlopes;
        double totalDischarge;
        double sumDischarge;
        long row;
        long col;
        int p;
        int q;
        // Deal with outflowCellIDs
        Iterator ite = outflowCellIDs.iterator();
        CellID cellID;
        while ( ite.hasNext() ) {
            cellID = ( CellID ) ite.next();
            row = cellID.getCellRowIndex();
            col = cellID.getCellColIndex();
            waterDepth = tempFlowAccumulation.getCell( row, col, this.handleOutOfMemoryErrorFalse );
            flowAccumulation.addToCell( row, col, - waterDepth / 2.0d, this.handleOutOfMemoryErrorFalse );
        }
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                surfaceHeights[1][1] = grid.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                if ( surfaceHeights[1][1] != noDataValue ) {
                    waterDepth = tempFlowAccumulation.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                    surfaceHeights[1][1] += waterDepth;
                    numberOfDownSlopes = 0.0d;
                    sumDischarge = 0.0d;
                    totalDischarge = 0.0d;
                    for ( p = 0; p < 3; p ++ ) {
                        for ( q = 0; q < 3; q ++ ) {
                            if ( ! ( p == 1 && q == 1 ) ) {
                                surfaceHeights[p][q] = grid.getCell( row + p - 1, col + q - 1, this.handleOutOfMemoryErrorFalse );
                                movingWaterDepth = Math.min( waterDepth, surfaceHeights[1][1] - surfaceHeights[p][q] );
                                if ( ( surfaceHeights[p][q] != noDataValue ) && ( surfaceHeights[p][q] < surfaceHeights[1][1] ) ) {
                                    numberOfDownSlopes += 1.0d;
                                    if ( p == q || ( p == 0 && q == 2 ) || ( p == 2 && q == 0 ) ) {
                                        slope = surfaceHeights[1][1] - surfaceHeights[p][q] / ( Math.sqrt( 2.0d ) );
                                    } else {
                                        slope = surfaceHeights[1][1] - surfaceHeights[p][q];
                                    }
                                    velocity = Math.sqrt( constant * movingWaterDepth * slope );
                                    discharge[p][q] = velocity * movingWaterDepth;
                                    sumDischarge += discharge[p][q];
                                }
                            }
                        }
                    }
                    if ( numberOfDownSlopes > 0.0d ) {
                        for ( p = 0; p < 3; p ++ ) {
                            for ( q = 0; q < 3; q ++ ) {
                                if ( ! ( p == 1 && q == 1 ) ) {
                                    if ( surfaceHeights[p][q] != noDataValue && surfaceHeights[p][q] < surfaceHeights[1][1] ) {
                                        movingWaterDepth = Math.min( waterDepth, surfaceHeights[1][1] - surfaceHeights[p][q] );
                                        discharge[p][q] = ( discharge[p][q] / sumDischarge ) * ( movingWaterDepth / 2.0d ); // 50%
                                        totalDischarge += discharge[p][q];
                                        flowAccumulation.addToCell( row + p - 1, col + q - 1, discharge[p][q], this.handleOutOfMemoryErrorFalse );
                                    }
                                }
                            }
                        }
                        flowAccumulation.addToCell( row, col, - totalDischarge, this.handleOutOfMemoryErrorFalse );
                    }
                }
            }
        }
        return flowAccumulation;
    }
    
    /**
     * TODO: docs
     * frictionFactor = 75.0d;
     * constant = 8.0d * 9.81d / frictionFactor;
     * velocity = Math.sqrt( constant * waterDepth * changeInDepth / ChangeInLength );
     * discharge = velocity * waterDepth
     */
    public Grid2DSquareCellDouble doFlowAccumulation( Grid2DSquareCellDouble flowAccumulation, Grid2DSquareCellDouble grid, double precipitation,  HashSet outflowCellIDs, Grid2DSquareCellDoubleFactory gridFactory ) {
        //double constant = 8.0d * 9.81d / 75.0d ;
        double constant = 1.0d;
        long nrows = grid.getNrows();
        long ncols = grid.getNcols();
        BigDecimal[] dimensions = grid.getDimensions();
        double noDataValue = grid.getNoDataValue();
        int gridStatisticsType = 1;
        // Precipitate
        flowAccumulation = addToGrid( flowAccumulation, precipitation, this.handleOutOfMemoryErrorFalse );
        flowAccumulation = mask( flowAccumulation, grid, this.handleOutOfMemoryErrorFalse );
        Grid2DSquareCellDouble tempFlowAccumulation = gridFactory.createGrid2DSquareCellDouble( flowAccumulation );
        double waterDepth;
        double movingWaterDepth;
        double[][] surfaceHeights = new double[3][3];
        double[][] discharge = new double[3][3];
        double slope;
        double velocity;
        double numberOfDownSlopes;
        double totalDischarge;
        double sumDischarge;
        long row;
        long col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                surfaceHeights[1][1] = grid.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                if ( surfaceHeights[1][1] != noDataValue ) {
                    waterDepth = tempFlowAccumulation.getCell( row, col, this.handleOutOfMemoryErrorFalse );
                    surfaceHeights[1][1] += waterDepth;
                    numberOfDownSlopes = 0.0d;
                    sumDischarge = 0.0d;
                    totalDischarge = 0.0d;
                    if ( outflowCellIDs.contains( grid.getCellID( row, col, this.handleOutOfMemoryErrorFalse ) ) ) {
                        // Simply lose a proportion of waterDepth (consider a friction factor)
                        flowAccumulation.addToCell( row, col, - waterDepth / 2.0d, this.handleOutOfMemoryErrorFalse );
                        /*for ( int p = 0; p < 3; p ++ ) {
                            for ( int q = 0; q < 3; q ++ ) {
                                if ( ! ( p == 1 && q == 1 ) ) {
                                    if ( grid.getCell( row + p - 1, col + q - 1 ) == noDataValue ) {
                                        numberOfDownSlopes += 1.0d;
                                        if ( p == q || ( p == 0 && q == 2 ) || ( p == 2 && q == 0 ) ) {
                                            slope = waterDepth / ( Math.sqrt( 2.0d ) );
                                        } else {
                                            slope = waterDepth;
                                        }
                                        velocity = Math.sqrt( constant * waterDepth * slope );
                                        discharge[p][q] = velocity * waterDepth;
                                        sumDischarge += discharge[p][q];
                                    }
                                }
                            }
                        }
                        if ( numberOfDownSlopes > 0.0d ) {
                            for ( int p = 0; p < 3; p ++ ) {
                                for ( int q = 0; q < 3; q ++ ) {
                                    if ( ! ( p == 1 && q == 1 ) ) {
                                        if ( grid.getCell( row + p - 1, col + q - 1 ) == noDataValue ) {
                                            discharge[p][q] = ( discharge[p][q] / sumDischarge ) * ( waterDepth / 2.0d ); // 50%
                                            totalDischarge += discharge[p][q];
                                        }
                                    }
                                }
                            }
                            flowAccumulation.addToCell( row, col, - totalDischarge );
                        }*/
                    } else {
                        for ( int p = 0; p < 3; p ++ ) {
                            for ( int q = 0; q < 3; q ++ ) {
                                if ( ! ( p == 1 && q == 1 ) ) {
                                    surfaceHeights[p][q] = grid.getCell( row + p - 1, col + q - 1, this.handleOutOfMemoryErrorFalse );
                                    if ( surfaceHeights[p][q] != noDataValue ) {
                                        surfaceHeights[p][q] += tempFlowAccumulation.getCell( row + p - 1, col + q - 1, this.handleOutOfMemoryErrorFalse );
                                        if ( surfaceHeights[p][q] < surfaceHeights[1][1] ) {
                                            movingWaterDepth = Math.min( waterDepth, ( surfaceHeights[1][1] - surfaceHeights[p][q] ) );
                                            numberOfDownSlopes += 1.0d;
                                            if ( p == q || ( p == 0 && q == 2 ) || ( p == 2 && q == 0 ) ) {
                                                slope = ( surfaceHeights[1][1] - surfaceHeights[p][q] ) / ( Math.sqrt( 2.0d ) );
                                            } else {
                                                slope = ( surfaceHeights[1][1] - surfaceHeights[p][q] );
                                            }
                                            velocity = Math.sqrt( constant * movingWaterDepth * slope );
                                            discharge[p][q] = velocity * movingWaterDepth;
                                            sumDischarge += discharge[p][q];
                                        }
                                    }
                                }
                            }
                        }
                        if ( numberOfDownSlopes > 0.0d ) {
                            for ( int p = 0; p < 3; p ++ ) {
                                for ( int q = 0; q < 3; q ++ ) {
                                    if ( ! ( p == 1 && q == 1 ) ) {
                                        if ( surfaceHeights[p][q] != noDataValue && surfaceHeights[p][q] < surfaceHeights[1][1] ) {
                                            movingWaterDepth = Math.min( waterDepth, ( surfaceHeights[1][1] - surfaceHeights[p][q] ) );
                                            discharge[p][q] = ( discharge[p][q] / sumDischarge ) * ( movingWaterDepth / 2.0d ); // 50%
                                            totalDischarge += discharge[p][q];
                                            flowAccumulation.addToCell( row + p - 1, col + q - 1, discharge[p][q], this.handleOutOfMemoryErrorFalse );
                                        }
                                    }
                                }
                            }
                            flowAccumulation.addToCell( row, col, - totalDischarge, this.handleOutOfMemoryErrorFalse );
                        }
                    }
                }
            }
        }
        return flowAccumulation;
    }
}